/*===========================================================================*
 * FILE:                     C D E V _ I N T . C
 *===========================================================================*
 *
 * COPYRIGHT (C) 2005-2010 BY
 *          GE INTELLIGENT PLATFORMS, INC. SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:    Application Programmer's Interface for a CEI-x30 ARINC device.
 *              Integrity Operating System Specific Functionality
 *
 *    Internal CEI-x30 routines implemented herein:
 *
 *    cei_acquire_device          - opens a session with the specified x30 device
 *    cei_acquire_device_identity - supplies the PCI configuration space ID values
 *    cei_acquire_thread_lock     - acquires sempahore/mutex lock
 *    cei_delay_milliseconds      - o/s specific delay utility in milliseconds
 *    cei_delay_seconds           - o/s specific delay utility in fractional seconds
 *    cei_get_host_system_time_x  - acquires host system time-of-year 64-bit uSec     
 *    cei_get_host_system_time    - acquires host system time-of-year 32-bit mSec     
 *    cei_get_irig_initial_time   - acquires host system time-of-year IRIG format
 *    cei_map_shared_memory       - maps memory region for sharing globals across processes
 *    cei_open_shared_lock        - creates a shared semaphore across processes
 *    cei_release_device          - closes a session with the specified x30 device
 *    cei_release_thread_lock     - releases sempahore/mutex lock     
 *    cei_read_32bit              - read a 32-bit value from an x30 device
 *    cei_write_32bit             - write a 32-bit value to an x30 device
 *    cei_init_interrupt          - configures an interrupt and ISR
 *    cei_terminate_interrupt     - terminate an interrupt
 *          
 *---------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ------------------------------------------------------------
 * 11/28/05  BH   Created.
 * 04/11/06  BDW  Removed utility routines cei_read_plx_32bit & cei_write_plx_32bit
 * 05/04/06  BDW  Added basicConfiguration[] declaration
 * 08/24/06  BDW  Incorporated CEI_TYPES.H data type definitions in 
 *                  support of 64-bit Linux operating systems - this 
 *                  modified every data declaration and prototype.
 *                Swapped the BAR reference order by the device structure 
 *                   members baseAddr and devicePtr
 * 02/18/08  BDW  Removed the boardInitialized array, relocated as a structure
 *                   member in the data structure array "device"
 * 03/14/08  BDW  Removed the declaration of the following global arrays, 
 *                relocated to CDEV_GLB.H:
 *                   bypassWrapTest
 *                   irigDacThreshold
 *                   a717syncWords
 *                Removed declaration of basicConfiguration, it was unused.
 * 04/15/08  BDW  Added support for initiation and termination of h/w interrupts
 *                via routines cei_init_interrupt and cei_terminate_interrupt.
 * 06/16/09  BDW  Added the function cei_open_preload_lock for multiprocess support.
 * 07/01/09  BDW  Added new function cei_delay_milliseconds (o/s thread suspend).
 * 09/14/10  BDW  Replaced the fixed user-mode isr invocation in cei_init_interrupt
 *                (API-defined func) with the function defined in x30InterruptFunctionPtr
 * 09/16/10  BDW  Renamed interruptFunctionPtr to x30InterruptFunctionPtr
 *                Renamed interruptServiceRoutine to x30InterruptFunctionPtr
 *                to eliminate duplicate names with the PIO product line API
 * 10/28/10  BDW  Modified the parameter list for cei_init_interrupt to account for      
 *                   interrupt setup requirements in the Windows-only CEI_Install library
 *                
 *===========================================================================*/
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <util/error_string.h>
#include "cei_types.h"
#include "lowlevel.h"
#include "cdev_api.h"

#define MAX_DEV_OPEN_ERROR_STR_LEN 512

/* this array contains a reference to the active user-mode ISR */
extern void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32, pCEI_UINT32);

/*========================================================================* 
 * Board-based globals (Board-based items only supporting 2 devices)
 *========================================================================*/
static DEVMAP_T devObj[MAX_BOARDS];
static Task iTask[MAX_BOARDS];

/* hardware interrupts are only supported on a single CEI-830 device */
static CEI_INT32 hwInterruptDeviceAssignment = 99;

/*========================================================================* 
 * Operating System specific functionality
 *========================================================================*/
/*===========================================================================*
 * INTERNAL ROUTINE:      L I T T L E _ E N D I A N _ 3 2
 *===========================================================================*
 *
 * FUNCTION:    Return the little-endian equivalent of the 4-byte integral
 *              argument (endian swap if host is big-endian).
 *
 * RETURN VAL:  The little-endian representation of the argument.
 *
 *===========================================================================*/
#define little_endian_32(longword) ( ((longword & 0x000000ff) << 24) | \
                                     ((longword & 0x0000ff00) << 8)  | \
                                     ((longword & 0xff0000)   >> 8)  | \
                                     ((longword & 0xff000000) >> 24) )

pthread_mutex_t ceiAccessMutex[MAX_BOARDS] = { /* Device Access mutex  */
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
pthread_mutexattr_t m_attr;

/* storage for Windows driver low-level board mapping error messages      */
static CEI_CHAR deviceOpenErrorDesc[MAX_DEV_OPEN_ERROR_STR_LEN] = "Specified PCI Device Not Found";


/*========================================================================* 
 * Non-Integrity Operating System specific functionality - here to resolve
 *========================================================================*/
CEI_INT32 cei_open_shared_lock (CEI_INT16 sharedLockId, pCEI_INT32 sharedLockState)
{
   return ARS_NORMAL;
}
CEI_INT32 cei_open_preload_lock (CEI_INT16 device)
{
   return ARS_NORMAL;
}

CEI_INT32 cei_map_shared_memory ( CEI_UCHAR    blockId,
                                  CEI_UINT32   blockSize,
                                  CEI_VOID  ** blockPtr,
                                  CEI_VOID  ** blockHandle,
                                  CEI_UINT32 * alreadyExists,
                                  CEI_UINT32 * isExclusive )
{
   return 1;
}

/*===========================================================================*
 * ENTRY POINT:            G E T   O P E N   E R R O R
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  openErrorString -- (output) reference to assign to error string
 * 
 * RETURN VAL:  ARS_NORMAL  -- Mutex was acquired successfully
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved
 *              when a device open error was encountered.
 *
 *===========================================================================*/
CEI_INT32 cei_get_open_error (pCEI_CHAR openErrorString)
{
  openErrorString = (pCEI_CHAR)&deviceOpenErrorDesc[0];
  return ARS_NORMAL;
}

/*------------------------------------------------------------------------* 
 * Mutex Protection for Device Access
 *------------------------------------------------------------------------*/
CEI_INT32 cei_acquire_thread_lock (CEI_INT16 deviceIndex)
{
   CEI_INT32 status=0;
   
   if((status = pthread_mutex_lock(&ceiAccessMutex[deviceIndex])) != Success) {
     printf("Fail: cei_acquire_thread_lock. pthread_mutex_lock - %d\n", status);
     return ARS_FAILURE; 
   }
   
   return ARS_NORMAL;
}

CEI_INT32 cei_release_thread_lock (CEI_INT16 deviceIndex)
{
   CEI_INT32 status=0;
   
   if((status = pthread_mutex_unlock(&ceiAccessMutex[deviceIndex])) != Success) {
     printf("Fail: cei_acquire_thread_lock. pthread_mutex_lock - %d\n", status);
     return ARS_FAILURE; 
   }

   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:              D E L A Y   M I L L I S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  numberOfMillieconds -- (input) the number of milliseconds
 *                                             to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the BSP-respective delay invocation
 *              as the basis to suspend thread execution for a specified
 *              number of milliseconds (resolution is platform dependent).
 *
 *===========================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds)
{
   struct timespec ts_d;
   struct timespec ts_rm;
   CEI_UINT32 numberOfSeconds;
   CEI_UINT32 remainingMilliSeconds;

   /* first handle parameters for a delay greater than one second */
   if (numberOfMilliseconds > 1000) {
      numberOfSeconds = numberOfMilliseconds/1000;
      remainingMilliSeconds = numberOfMilliseconds - (numberOfSeconds * 1000);
   }
   else {
      numberOfSeconds = 0;
      remainingMilliSeconds = numberOfMilliseconds;
   }
   ts_d.tv_sec = numberOfSeconds;
   ts_d.tv_nsec = remainingMilliSeconds * 1000000; 

   /* use nanosleep to suspend this thread/task for the specified delay duration */
   if(nanosleep(&ts_d, &ts_rm) != 0) {
     printf("Fail: cei_delay_seconds. nanosleep - %d\n", errno);
     return ARS_FAILURE; 
   }

   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:              D E L A Y   S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  numberOfSeconds -- (input) a floating point representation
 *                                         of the number (or fraction) of
 *                                         seconds to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the target-respective delay invocation
 *              as the basis for a delay defined in fractional seconds.
 *
 *===========================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds)
{
   struct timespec ts_d;
   struct timespec ts_rm;

   /* first handle parameters for a delay greater than one second */
   ts_d.tv_sec = (CEI_UINT32)numberOfSeconds;
   ts_d.tv_nsec = (CEI_UINT32)(numberOfSeconds - (CEI_FLOAT)ts_d.tv_sec) * 1000000000; 
   
   /* use nanosleep to suspend this thread/task for the specified delay duration */
   if(nanosleep(&ts_d, &ts_rm) != 0) {
     printf("Fail: cei_delay_seconds. nanosleep - %d\n", errno);
     return ARS_FAILURE; 
   }

   return ARS_NORMAL;
}

/*------------------------------------------------------------------------* 
 * Retrieve the host 64-bit 1us resolution reference time value
 * Windows System Time in "seconds since midnight Jan 1 of this year"
 *------------------------------------------------------------------------*/
CEI_INT32 cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime)
{
   time_t now;               /* Current time, modify to get Time Reference Point. */
   struct tm start;        /* Formatted "now" current time, modified for TRP.   */
   time_t then;            /* Time Reference Point, Midnight or Jan 1.          */
   struct timespec ts;
   TIME_TAG_TYPE diff;       /* This is the difference time we return in microseconds, */

   time(&now);               /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now); /* Local                                          */
   start.tm_sec  = 0;        /*   time                                         */
   start.tm_min  = 0;        /*     at                                         */
   start.tm_hour = 0;        /*       Jan 1                                    */
   start.tm_mday = 1;        /*         midnight (This is day 0)               */
   start.tm_mon  = 0;        /*           this                                 */
   start.tm_yday = 0;        /*             year                               */
   then = mktime(&start);    /* local Jan 1 to time in seconds since Jan 1     */
                             /* We do IRIG time, which makes Jan 1 "day 0"     */ 

   if(clock_gettime(CLOCK_REALTIME, &ts) != 0) { 
     printf("Fail: cei_get_host_system_time_x. clock_gettime - %d\n", errno);
     return ARS_FAILURE;
   }
   
   diff = ts.tv_sec - then;
   diff *= 1000000;           /* Number of Seconds converted to microseconds    */
   diff += (ts.tv_nsec / 1000);
   *hostTime = diff;          /* microseconds.  Passed back to the caller.      */

   return ARS_NORMAL;
}

/*------------------------------------------------------------------------* 
 * Retrieve the host 32-bit 1ms reference time value
 *------------------------------------------------------------------------*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime)
{
   struct timespec ts;
	   
   if(clock_gettime(CLOCK_REALTIME, &ts) != 0) {
     printf("Fail: cei_get_host_system_time. clock_gettime - %d\n", errno);
     return ARS_FAILURE;
   }
   
   *hostTime = (ts.tv_sec * 1000) + (ts.tv_nsec / 1000000); 

   return ARS_NORMAL;
}

/*------------------------------------------------------------------------* 
 * Retrieve the host reference calendar time/date value - used for
 * CEI-830 IRIG timer initialization
 *------------------------------------------------------------------------*/
CEI_INT32 cei_get_irig_initial_time (pCEI_UINT32 time)
{
   TIME_TAG_TYPE system_time;
   CEI_NATIVE_UINT seconds;
   CEI_NATIVE_UINT tens_of_seconds;
   CEI_NATIVE_UINT minutes;
   CEI_NATIVE_UINT tens_of_minutes;
   CEI_NATIVE_UINT hours;
   CEI_NATIVE_UINT tens_of_hours;
   CEI_NATIVE_UINT days;
   CEI_NATIVE_UINT tens_of_days;
   CEI_NATIVE_UINT hundreds_of_days;

   union
   {
      struct irigTimeFormat
           {
              CEI_NATIVE_UINT seconds          :4;
              CEI_NATIVE_UINT tens_of_seconds  :3;
              CEI_NATIVE_UINT minutes          :4;
              CEI_NATIVE_UINT tens_of_minutes  :3;
              CEI_NATIVE_UINT hours            :4;
              CEI_NATIVE_UINT tens_of_hours    :2;
              CEI_NATIVE_UINT days             :4;
              CEI_NATIVE_UINT tens_of_days     :4;
              CEI_NATIVE_UINT hundreds_of_days :2;
           } IRIG_TIME_FORMAT;
      CEI_UINT32 timeValue;
   } timeConvert;

   if(cei_get_host_system_time_x (&system_time) != ARS_NORMAL)
     return ARS_FAILURE;

   seconds = (CEI_UINT32)(system_time / (TIME_TAG_TYPE)1000000);
   if ( (system_time % (TIME_TAG_TYPE)1000000) > (TIME_TAG_TYPE)500000) {
      seconds++;
   }

   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;
   tens_of_seconds = seconds/10;
   seconds = seconds%10;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;
   tens_of_minutes = minutes/10;
   minutes = minutes%10;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;     
   tens_of_hours = hours/10;    
   hours = hours%10;    
   days++;                       /* account for no such thing as day 0 */
   hundreds_of_days = days/100;
   days = days%100;
   tens_of_days = days/10;
   days = days%10;

   timeConvert.IRIG_TIME_FORMAT.seconds          = seconds;
   timeConvert.IRIG_TIME_FORMAT.tens_of_seconds  = tens_of_seconds;
   timeConvert.IRIG_TIME_FORMAT.minutes          = minutes;
   timeConvert.IRIG_TIME_FORMAT.tens_of_minutes  = tens_of_minutes;
   timeConvert.IRIG_TIME_FORMAT.hours            = hours;
   timeConvert.IRIG_TIME_FORMAT.tens_of_hours    = tens_of_hours;
   timeConvert.IRIG_TIME_FORMAT.days             = days;
   timeConvert.IRIG_TIME_FORMAT.tens_of_days     = tens_of_days;
   timeConvert.IRIG_TIME_FORMAT.hundreds_of_days = hundreds_of_days;
   
   *time = timeConvert.timeValue;

   return ARS_NORMAL;
}

/*========================================================================* 
 * Device Access Routines for the Integrity Operating System
 *========================================================================*/

/*===========================================================================*
 * ENTRY POINT:   A C Q U I R E   D E V I C E   I D E N T I T Y
 *===========================================================================*
 *
 * FUNCTION:    Read the specified device PCI Configuration Space Identifiers
 *
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 7)
 *              vendorId    -- (output) ptr to storage of vendor PCI ID
 *              deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL  -- successful acquisition of the device ID's
 *              ARS_FAILURE -- low-level failed due to invalid device ID
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine acquires the specified device PCI Configuration  
 *              Space Vendor ID and Device ID fields via low-level function.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device_identity (
   CEI_INT32  deviceIndex,                /* device number to test (0-7)     */
   pCEI_INT32 vendorId,                   /* ptr to storage of vendor PCI ID */
   pCEI_INT32 deviceId                    /* ptr to storage of device PCI ID */
)
{
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   CEI_INT32 pciDeviceId;

   /* read the PCI manufacturer and device ID fields from PCI Configuration Space */
   status = vbtGetPCIConfigRegister ((unsigned int)deviceIndex, 0, 2, (unsigned int *)&pciDeviceId);
   *vendorId = (CEI_INT32)pciDeviceId;
   status = vbtGetPCIConfigRegister ((unsigned int)deviceIndex, 1, 2, (unsigned int *)&pciDeviceId);
   *deviceId = (CEI_INT32)pciDeviceId;

   status = ARS_NORMAL;
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:          A C Q U I R E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  deviceIndex   -- (input) device to reference (0 to 7)
 *              device        -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL    -- Session was opened successfully
 *              ARS_FAILURE   -- Driver Open Device failed
 *
 * GLOBALS:     pci_state[8]
 *              deviceMapped[8][3]
 *              ceiAccessLock[8][3]
 *
 * DESCRIPTION: This routine searches for up to eight CEI-x30 devices, 
 *              mapping each CEI-x30 device found with the first execution 
 *              of this routine.  Subsequent executions will simply retrieve
 *              the mapped addresses for the specified device index.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INT32 status = 0;
	 
   memset (&devObj[deviceIndex], 0, sizeof(DEVMAP_T));

   /* request the addresses mapped to both BAR0 and BAR1 for the CEI-830 */
   devObj[deviceIndex].flagMapToHost[0] = TRUE;
   devObj[deviceIndex].flagMapToHost[1] = TRUE;

   if (vbtMapBoardAddresses(deviceIndex, &devObj[deviceIndex]) != BTD_OK)
     return ARS_WINRTFAIL;
   
   device->baseAddr = (pCEI_UINT32)devObj[deviceIndex].memHostBase[0];
   device->devicePtr = (pCEI_UINT32)devObj[deviceIndex].memHostBase[1];

   /* initialize multi-threading mutex (using PTRHEADS) */
   if (pthread_mutexattr_init(&m_attr) == Success) {
      /* PTHREAD_PRIO_NONE, PTHREAD_PRIO_INHERIT, PTHREAD_PRIO_PROTECT */
      pthread_mutexattr_setprotocol(&m_attr, PTHREAD_PRIO_NONE);
      /* PTHREAD_MUTEX_DEFAULT, PTHREAD_MUTEX_NORMAL, PTHREAD_MUTEX_ERRORCHECK, PTHREAD_MUTEX_RECURSIVE */
      pthread_mutexattr_settype(&m_attr, PTHREAD_MUTEX_ERRORCHECK);
      status = pthread_mutex_init(&ceiAccessMutex[deviceIndex], &m_attr);
   }
   else {
      /* use the defaults (PTHREAD_PRIO_INHERIT and PTHREAD_MUTEX_DEFAULT) for the mutex */
      status = pthread_mutex_init(&ceiAccessMutex[deviceIndex], NULL);
   }

   /* convert return status to API type */
   status = (status == Success) ? ARS_NORMAL : ARS_FAILURE; 

   return status;
}

/*===========================================================================*
 * ENTRY POINT:             R E L E A S E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 7)
 *              device      -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     ceiAccessLock[8][3]
 *
 * DESCRIPTION: This routine releases a driver handle and unmaps the two PCI 
 *              memory regions previously mapped for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_release_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INT32 status=0;

   vbtFreeBoardAddresses(&devObj[deviceIndex]);

   /* release multi-thread protection mutex */
   if ((status = pthread_mutex_destroy(&ceiAccessMutex[deviceIndex])) != Success) {
     printf("Fail: cei_release_device. pthread_mutex_destroy - %d\n", status);
     return ARS_FAILURE; 
   }

   pthread_mutexattr_destroy(&m_attr);

   return ARS_NORMAL;
}

/*------------------------------------------------------------------------* 
 * Perform a 32-bit read access
 *------------------------------------------------------------------------*/
CEI_INT32 cei_read_32bit (CEI_INT16 board, pCEI_UINT32 address, pCEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   CEI_UINT32 localData, modifiedData;
   localData = *address;
   modifiedData = little_endian_32(localData);
   *data = modifiedData;
#else
   *data = *address;
#endif          
   return ARS_NORMAL;
}

/*------------------------------------------------------------------------* 
 * Perform a 32-bit write access
 *------------------------------------------------------------------------*/
CEI_INT32 cei_write_32bit (CEI_INT16 board, pCEI_UINT32 address, CEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   *address = little_endian_32(data);
#else
   *address = data;
#endif          
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:   i n t e r r u p t S e r v i c e R o u t i n e
 *===========================================================================*
 *
 * DESCRIPTION: Invoked whenever an interrupt occurs, this routine clears 
 *              the hardware Interrupt Pending bit and invokes the ISR via 
 *              semaphore to launch "x30InterruptFunctionPtr".
 *
 *===========================================================================*/
void x30InterruptServiceRoutine (void)
{
   Error err;
   char dbuf[40];
   IODevice cboard;
   int device;
   short int headTail;

   /* retrieve the device index the ISR is assigned to */
   device = hwInterruptDeviceAssignment;

   /* redundant check we're dealing with a valid device index */
   if (hwInterruptDeviceAssignment < MAX_BOARDS) {
      
      /* connect to this device's driver resource id */
      sprintf (dbuf, "CEIboard%d\n", device);  
      err = RequestResource ((Object *)&cboard, dbuf, "!systempassword");
      if (err != Success) {
         printf("Error from resource acquisition\n");
         return;
      }
      
      /* this "isr" is really a task that blocks on any interrupt event for the resource id */
      while (TRUE) {
         err = SynchronousReceive ((Connection)cboard, NULL);
         if (err != Success) {
            printf("SynchronousReceive error:  %d - %s\n",err, ErrorString(err));
         }
   
         err = ReadIODeviceStatus (cboard, 3, (void *)&headTail, sizeof(headTail));
         if (err != Success) {
            printf("Error Calling 3 ReadIODeviceStatus %d - %s\n", err, ErrorString(err));
            continue;
         }

         /* if an event was encountered, invoke the specified ISR */
         (x30InterruptFunctionPtr[device])(device, 0);
      }	
   }
   return;
}


/*===========================================================================*
 * ENTRY POINT:         C E I _ I N I T _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Configures an interrupt and associates an ISR with it
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 15)
 *              pCEI_VOID   isr               -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 *              CEI_INT32   memRegion         -- (input)  not used
 *              CEI_UINT32  disOffset         -- (input)  not used
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 isrEnabled        -- (output) thread handle if successfully defined
 * 
 * RETURN VAL:  -- Interrupt was initialized successfully
 *              -- Interrupt was not initialized successfully
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device.
 *
 *===========================================================================*/
CEI_INT32 cei_init_interrupt (
   CEI_INT32   deviceIndex,
   pCEI_VOID   isr, 
   CEI_INT32   priority,
   pCEI_UINT32 deviceBaseAddress,
   CEI_INT32   memRegion,
   CEI_UINT32  disOffset,
   CEI_UINT32  interruptBit,
   pCEI_UINT32 isrHandle
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the low-level global virtual base address array     */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   char sbuf[40];
   Error err;

   /* allow only one device ISR assignment per API instance */
   if (deviceIndex < MAX_BOARDS) {
      if (hwInterruptDeviceAssignment == 99) {

         /* create a task that monitors for events for the specified device/resource */
         sprintf (sbuf,"intProcess%d",deviceIndex);
         err = CommonCreateTask (1, (Address)x30InterruptServiceRoutine, 0x1000, sbuf, &iTask[deviceIndex]);
         if (err != Success) {
            printf("Bad Status from CommonCreateTask: %d\n", (int)err);
            return ARS_FAILURE;
         }
      
         /* save the device index for the ISR assignment */
         hwInterruptDeviceAssignment = deviceIndex;
   
         /* launch the task we just created */
         err = RunTask (iTask[deviceIndex]);
         if (err != Success) {
            printf("Bad Status from RunTask: %d\n", (int)err);
            return ARS_FAILURE;
         }
         *isrHandle = (CEI_UINT32)iTask[deviceIndex];
      }
   
      /* Set either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
       * Local Interrupt Input Enable, High to pass local bus interrupts
       */
      if (interruptBit != 0) {
         status = cei_read_32bit (deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
         intCtrlStatReg |= interruptBit;
         status = cei_write_32bit (deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
      }
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:     C E I _ T E R M I N A T E _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to 
 *              an ISR
 * 
 * PARAMETERS:  deviceIndex  -- (input)  device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *              ARS_FAILURE  -- Interrupt was not terminated successfully
 *
 * GLOBALS:     None
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function
 * 
 *===========================================================================*/
CEI_INT32 cei_terminate_interrupt
(
    CEI_INT32   deviceIndex, 
    CEI_UINT32  interruptBit,
    pCEI_UINT32 deviceBaseAddress
)
{
   IODevice cboard;
   char dbuf[40]; 
   Error err;
   CEI_UINT32 intCtrlStatReg;
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   CEI_INT32 status;
   
   /* Reset either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, Low to disable local bus interrupts
    */
   if (interruptBit != 0) {
      status = cei_read_32bit (deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg &= ~interruptBit;
      status = cei_write_32bit (deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }
   sprintf (dbuf, "CEIboard%d\n", deviceIndex);  
   err = RequestResource ((Object *)&cboard, dbuf, "!systempassword");       
   CommonCloseTask (iTask[deviceIndex]);
   hwInterruptDeviceAssignment = 99;
   if (err) ;
   return status;
}

