/*============================================================================*
 * FILE:                      C E I _ I N T _ P C I _ D R V . C
 *============================================================================*
 *
 * COPYRIGHT (C) 2006 - 2008 BY
 *          GE INTELLIGENT PLATFORMS, INC. SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:   GE INTELLIGENT PLATFORMS AVIONICS GENERIC Integrity PCI device driver
 *
 *             This module provides the interface between the GE Intelligent 
 *             Platforms Avionics APIs and the Integrity operating system.
 *             It performs the actual I/O and memory-mapping functions
 *             needed to support operation in an Integrity environment.
 *             Since this file is used by multiple products, use care when
 *             performing modifications.
 * 
 *
 * EXTERNAL ENTRY POINTS:
 * 
 *    CEIDevCreateIODevice
 *    CEIReadStatus
 *    CEIWriteStatus
 * 
 * EXTERNAL NON-USER ENTRY POINTS:
 * 
 *    cei_int_pci_drv_init    probes the PCI bus for supported GE IP Avionics
 *                            Avionics PCI devices, then intitializes
 *                            the device
 *    cei_int_pci_drv_match
 *    
 *    CEIDevHandler           the interrupt service routine for hardware
 *                            interrupts from a 1553 device. 
 *    CommercialDevHandler    the ISR for P-SER and CEI-830 hardware interrupts
 *  
 * INTERNAL ROUTINES: 
 * 
 *    CEIDeviceAllocate
 *
 *===========================================================================*
 *
 * Modification History:
 *
 *    Date     Modifications
 * ----------  ---------------------------------------------------------------
 * 11/23/2005  Added support for CEI-x30 devices.  Added support to read a 
 *             device's PCI config space registers.
 * 11/04/2008  BDW - Added CEI-430, RCEI-530, RAR-CPCI, R830RX, P-708 & P-SER 
 * 12/15/2008  BDW - Modified driver as follows:
 *                A) added a compiler directive NON_INTEL_WORD_ORDER around
 *                   Endian byte-swapping in CEIDevHandler
 *                B) corrected the type casting on destination for device PCI
 *                   configuration space read invocations in CEIReadStatus
 *                C) corrected the 1553 device recognition for h/w interrupt
 *                   initialization in cei_int_pci_drv_init().
 * 04/07/2009  BDW - Cleaned up the driver by...
 *                A) removed confusing overuse of the variable name "device"
 *                B) defined and used named GE IP Avionics PCI interface constants 
 *                C) finalized the P-SER ISR named PserDevHandler
 *                D) added lots of comment blocks
 * 07/30/2009  BDW - Renamed PserDevHandler to CommercialDevHandler to reflect
 *                   the common use amoung the GE IP Avionics Avionics commercial
 *                   bus interface boards: P-SER and CEI-830, so far.
 *
 *===========================================================================*/

#include <string.h>
#include "pci_config.h"
#include "bus.h"         /* for BusSpace API */
#include "interrupt.h"   /* for ChainedExceptions */


/*---------------------------------------------------------------------------*
 * Constants for the driver interface CEIReadStatus
 * A matching set of constants exists in the mem_*.c o/s support file
 *---------------------------------------------------------------------------*/
#define STATUS_DEVICE_ID         1
#define STATUS_HOST_INTERFACE    2
#define STATUS_INTERRUPT_DATA    3
#define STATUS_PCI_CONFIG_SPACE  4

#define MAX_BTA                  16
#define MAX_FIFO_LEN             32
#define INTERRUPT_QUEUE_SIZE     64
#define INTERRUPT_QUEUE_MASK INTERRUPT_QUEUE_SIZE - 1

/*---------------------------------------------------------------------------*
 * Constants for identifying supported GE IP Avionics Avionics devices
 *---------------------------------------------------------------------------*/
#define FUNCTION_ID_PMC_1553  0x155313c6
#define FUNCTION_ID_QPMC_1553 0x155413c6
#define FUNCTION_ID_QPC_1553  0x155513c6
#define FUNCTION_ID_QPCX_1553 0x155613c6
#define FUNCTION_ID_P_708     0x070813c6
#define FUNCTION_ID_CEI_430   0x043013c6
#define FUNCTION_ID_CEI_530   0x053013c6
#define FUNCTION_ID_RAR_CPCI  0x063013c6
#define FUNCTION_ID_CEI_830   0x083013c6
#define FUNCTION_ID_R830RX    0x083113c6
#define FUNCTION_ID_PMC_D1    0x100113c6
#define FUNCTION_ID_EPMC_1553 0x100313c6
#define FUNCTION_ID_P_SER     0x100413c6

#define DEVICE_ID_1553_MASK   0x1500
#define DEVICE_ID_PMC_1553    0x1553
#define DEVICE_ID_QPMC_1553   0x1554
#define DEVICE_ID_QPC_1553    0x1555
#define DEVICE_ID_QPCX_1553   0x1556
#define DEVICE_ID_P_708       0x0708
#define DEVICE_ID_CEI_430     0x0430
#define DEVICE_ID_CEI_530     0x0530
#define DEVICE_ID_RAR_CPCI    0x0630
#define DEVICE_ID_CEI_830     0x0830
#define DEVICE_ID_R830RX      0x0831
#define DEVICE_ID_PMC_D1      0x1001
#define DEVICE_ID_EPMC_1553   0x1003
#define DEVICE_ID_P_SER       0x1004

/*---------------------------------------------------------------------------*
 * Data Structures
 *---------------------------------------------------------------------------*/
typedef struct CEIPCIDeviceStruct {
   struct IODeviceVectorStruct    device;
   InterruptHandler               handler;
   BusSpace                       config_space;
   BusHandle                      config_handle;
   BusSpace                       memory_space;
   BusHandle                      memory_handle[6];
   char*                          name;
   UINT4                          address;
   UINT4                          mem_size[6];
   UINT4                          mem_region_base[6];
   UINT4                          mem_base;
   UINT4                          mem_sections;
   UINT2                          devid;
   UINT4                          host_interface;
   UINT1                          interrupt_queue[INTERRUPT_QUEUE_SIZE];
   unsigned char                  head;
   unsigned char                  tail;
   UINT2                          devnum;
   DynamicMemoryRegionReservation bar_reservation[6];
   char                           mem_res_name[6][40];
} CEIPCIDevice;

/*---------------------------------------------------------------------------*
 * exported function prototypes
 *---------------------------------------------------------------------------*/
static CEIPCIDevice * CEIDeviceAllocate(UINT2 pciDevId); 
static Error CEIDevCreateIODevice (IODeviceVector iodev);
static Error CEIReadStatus (IODeviceVector iodev, Value StatusNumber, void * Buffer, Address Length);
static Error CEIWriteStatus (IODeviceVector iodev, Value StatusNumber, void * Buffer, Address Length);

/*---------------------------------------------------------------------------*
 * modules used by the kernel 
 *---------------------------------------------------------------------------*/
static Error cei_int_pci_drv_match (PciData * data, PciFunction * function, Value * fitness);
static Error cei_int_pci_drv_init  (PciData * data, PciFunction * function);

PciDeviceDriver __ghsentry_pci_device_driver_cei_int_pci_drv = {
   "CEI Integrity PPC PCI Device Driver",  
   cei_int_pci_drv_match,  /* match */
   NULL,                   /* pre-init */
   NULL,                   /* requirements */
   cei_int_pci_drv_init,   /* init */
   NULL,                   /* device_name */
   NULL,                   /* device_bar */
};

/*---------------------------------------------------------------------------*
 * global variables
 *---------------------------------------------------------------------------*/
static char dname[MAX_BTA][80];
char devNameStr1[80];
char devNameStr2[80];
int debug = 0;
static int num_devices = 0;
static CEIPCIDevice deviceInfoArray[MAX_BTA];


/*===========================================================================*
 * ENTRY POINT:     C O M M E R C I A L   D E V   H A N D L E R
 *===========================================================================*
 *
 * FUNCTION:    This routine is the ISR for GE IP Avionics Avionics P-SER device.
 *
 *===========================================================================*/
static EVENT CommercialDevHandler (XFRAME *xframe, InterruptHandler *handler) 
{
   CEIPCIDevice * ceiPciDeviceStructPtr = (CEIPCIDevice *)handler->id;
   UINT4 temp;

   /* read the P-SER Control/Status Register, located in PCI BAR2 */
   temp = BusRead4B ( ceiPciDeviceStructPtr->memory_space,
                      ceiPciDeviceStructPtr->memory_handle[2], 
                      0 );

   /* if the P-SER H/W Interrupt is enabled and the active bit is set, clear the active
    * bit and notify the P-SER API Interrupt Process task an interrupt was encountered.
    */
   if ( temp & 0x0060 ) {
      BusWrite4B ( ceiPciDeviceStructPtr->memory_space,       /* clear the interrupt */
                   ceiPciDeviceStructPtr->memory_handle[2],
                   0,
                   temp);
      
      /* signal the API interrupt task there was an interrupt */
      ceiPciDeviceStructPtr->head++;
      ceiPciDeviceStructPtr->head &= MAX_FIFO_LEN-1;
      INTERRUPT_IODeviceNotifyTask(&ceiPciDeviceStructPtr->device);
      return EVENT_HANDLED;
   }

   return EVENT_UNKNOWN;
}


/*===========================================================================*
 * ENTRY POINT:            C E I  D E V  H A N D L E R
 *===========================================================================*
 *
 * FUNCTION:    This routine is the ISR for supported GE IP Avionics Avionics 
 *              MIL-STD-1553 devices.
 *
 *===========================================================================*/
static EVENT CEIDevHandler (XFRAME * xframe, InterruptHandler * handler) 
{
   CEIPCIDevice * ceiPciDeviceStrPtr = (CEIPCIDevice *)handler->id;
   UINT2 host_interface, temp;
   UINT4 nchan, handled=0;
   UINT1 indx;
   /* this is byte addressing */
   UINT4 chan_addr[] = {0x0,0x200000,0x400000,0x600000};
   UINT4 hw_reg_addr = 0x800;
   UINT4 clear_int_reg = 0x16;
   int bar;

   if (ceiPciDeviceStrPtr->mem_sections >= 2)
      bar = 2;
   else
      bar = 0;

   indx = ceiPciDeviceStrPtr->devnum;

   temp = BusRead2B(ceiPciDeviceStrPtr->memory_space,
                    ceiPciDeviceStrPtr->memory_handle[bar], 
                    0);
#ifdef NON_INTEL_WORD_ORDER
   host_interface = temp & 0xff00>>8 | temp & 0x00ff<<8;
#else
   host_interface = temp;
#endif          
   nchan = (host_interface & 0x7c0)>>6;
   if (nchan > 4)
      nchan=4;

   for (indx = 0;indx < nchan; indx++) {  /* need to check each channel on the board for interrupts pending */
      temp = BusRead2B(ceiPciDeviceStrPtr->memory_space,
                       ceiPciDeviceStrPtr->memory_handle[bar],
                       chan_addr[indx]+hw_reg_addr);     
      if ( temp & 0x200 ) {
         handled++;
         /* clear the interrupt */
         BusWrite2B(ceiPciDeviceStrPtr->memory_space,
                    ceiPciDeviceStrPtr->memory_handle[bar],
                    chan_addr[indx]+hw_reg_addr+clear_int_reg,
                    0);
         /* signal the user app that there was an interrupt */
         ceiPciDeviceStrPtr->interrupt_queue[ceiPciDeviceStrPtr->head] = indx;
         ceiPciDeviceStrPtr->head++;
         ceiPciDeviceStrPtr->head &= MAX_FIFO_LEN-1;
      }
   }

   if (handled) {
      INTERRUPT_IODeviceNotifyTask(&ceiPciDeviceStrPtr->device);
      handled = 0;
      return EVENT_HANDLED;
   }

   return EVENT_UNKNOWN;
}


/*===========================================================================*
 * ENTRY POINT:       C E I  D E V I C E  A L L O C A T E
 *===========================================================================*
 *
 * FUNCTION:    This routine provides tracks the device index used by the API 
 *              and assigns the appropriate name strings used by the kernel
 *              for identifying supported GE IP Avionics PCI devices.
 *
 *===========================================================================*/
static CEIPCIDevice * CEIDeviceAllocate(UINT2 pciDevId) 
{
   if (num_devices >= MAX_BTA)
      return NULL;

   if (debug)
      db_printf("\nCEIDEviceAllocate num_devices = %d\n", num_devices);

   sprintf (devNameStr1,"CEIboard%d (0x%x)", num_devices, pciDevId);
   strcpy  (dname[num_devices], devNameStr1);
   deviceInfoArray[num_devices].name = dname[num_devices];
   deviceInfoArray[num_devices].devnum = num_devices;

   if (debug)
      db_printf("%s\n",deviceInfoArray[num_devices].name);

   switch (pciDevId) {
      case DEVICE_ID_PMC_1553: /* QPMC-1553 */
         sprintf (devNameStr1, "QPMC1553_BASE%d", num_devices);
         strcpy  (devNameStr2, "");
         break;
      case DEVICE_ID_QPMC_1553: /* QPCI-1553 */
         sprintf (devNameStr1, "QPCI_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "QPCI1553_BASE%d", num_devices);
         break;
      case DEVICE_ID_QPC_1553: /* QCP-1553 */
         sprintf (devNameStr1, "QCP_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "QCP1553_BASE%d", num_devices);
         break;
      case DEVICE_ID_QPCX_1553: /* QPCX-1553 */
         sprintf (devNameStr1, "QPCX_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "QPCX1553_BASE%d", num_devices);
         break;     
      case DEVICE_ID_P_708:    /* P-708 */
         sprintf (devNameStr1, "P708_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "P708_BASE%d", num_devices);
         break;     
      case DEVICE_ID_CEI_430:  /* CEI-430 */
         sprintf (devNameStr1, "CEI430_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "CEI430_BASE%d", num_devices);
         break;
      case DEVICE_ID_CEI_530:  /* RCEI-530 */
         sprintf (devNameStr1, "CEI530_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "CEI530_BASE%d", num_devices);
         break;
      case DEVICE_ID_RAR_CPCI: /* RAR-CPCI */
         sprintf (devNameStr1, "RAR_CPCI_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "RAR_CPCI_BASE%d", num_devices);
         break;
      case DEVICE_ID_CEI_830:  /* CEI-830 */
         sprintf (devNameStr1, "CEI830_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "CEI830_BASE%d", num_devices);
         break;
      case DEVICE_ID_R830RX:   /* R830RX  */
         sprintf (devNameStr1, "R830RX_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "R830RX_BASE%d", num_devices);
         break;
      case DEVICE_ID_PMC_D1:   /* PMC-D-1 */
         sprintf (devNameStr1, "PMCD1_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "PMCD1_BASE%d", num_devices);
         break;
      case DEVICE_ID_EPMC_1553: /* EPMC-1553 */
         sprintf (devNameStr1, "EPMC1553_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "EPMC1553_BASE%d", num_devices);
         break;
      case DEVICE_ID_P_SER:    /* P-SER */
         sprintf (devNameStr1, "PSER_PLX_CONFIG_BASE%d", num_devices);
         sprintf (devNameStr2, "PSER_BASE%d", num_devices);
         break;     
      default:
         return NULL; /* maybe some other error? */
   };
   
   strcpy (deviceInfoArray[num_devices].mem_res_name[0], devNameStr1);
   strcpy (deviceInfoArray[num_devices].mem_res_name[2], devNameStr2);

   return &(deviceInfoArray[num_devices++]);
}


/*===========================================================================*
 * ENTRY POINT:       C E I _ I N T _ P C I _ D R V _ M A T C H
 *===========================================================================*
 *
 * FUNCTION:    This routine fulfills the positive identification of the
 *              supported GE IP Avionics Avionics devices.
 *
 *===========================================================================*/
static Error cei_int_pci_drv_match (PciData * data, PciFunction * function, Value * fitness) {
   /* Vendor and Device Id */
   switch (function->id) {
      case FUNCTION_ID_PMC_1553 :  /* PMC-1553  */
      case FUNCTION_ID_QPMC_1553:  /* QPMC-1553 */
      case FUNCTION_ID_QPC_1553 :  /* QPC-1553  */
      case FUNCTION_ID_QPCX_1553:  /* QPCX-1553 */
      case FUNCTION_ID_P_708    :  /* P-708     */
      case FUNCTION_ID_CEI_430  :  /* CEI-430   */
      case FUNCTION_ID_CEI_530  :  /* CEI-530   */
      case FUNCTION_ID_RAR_CPCI :  /* RAR-CPCI  */
      case FUNCTION_ID_CEI_830  :  /* CEI-830   */
      case FUNCTION_ID_R830RX   :  /* R830RX    */
      case FUNCTION_ID_PMC_D1   :  /* PMC-D-1   */
      case FUNCTION_ID_EPMC_1553:  /* EPMC-1553 */
      case FUNCTION_ID_P_SER    :  /* P-SER     */
         *fitness = 128;
         break;
      default:
         return Failure;
   };

   return Success;
}


/*===========================================================================*
 * ENTRY POINT:     C E I  D E V  C R E A T E  I O  D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    This routine initializes the specified device's IRQ pointers.
 *
 *===========================================================================*/
static Error CEIDevCreateIODevice(IODeviceVector iodev) {
   CEIPCIDevice * deviceInfoArrayPtr = (CEIPCIDevice *)iodev->DeviceSpecificInfo;
   deviceInfoArrayPtr->tail=0;
   deviceInfoArrayPtr->head=0;
   return Success;
}


/*===========================================================================*
 * ENTRY POINT:              C E I  R E A D  S T A T U S
 *===========================================================================*
 *
 * FUNCTION:    Returns device information requested by the calling application
 *
 * PARAMETERS:  PciData * data         -- (input) PCI header info on this device
 *              PciFunction * function -- (input) o/w function info on this device
 * 
 * RETURN VAL:  Error - Integrity-specific driver error codes
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine supplies detailed information regarding the 
 *              specified I/O device PCI Interface and local IRQ buffer tail ptr.
 *
 *===========================================================================*/
static Error CEIReadStatus (IODeviceVector iodev, Value StatusNumber, void * Buffer, Address Length) {
   CEIPCIDevice * deviceInfoArrayPtr = NULL;
   int bar = 0;
   int offset = 0;

   deviceInfoArrayPtr = (CEIPCIDevice *)iodev->DeviceSpecificInfo;

   switch (StatusNumber) {
      case STATUS_DEVICE_ID:  /* return the device ID */
         *(UINT2*)Buffer = deviceInfoArrayPtr->devid;
         break;

      case STATUS_HOST_INTERFACE:  /* returns the host interface register */
         if ((deviceInfoArrayPtr->devid == DEVICE_ID_PMC_D1) || (deviceInfoArrayPtr->devid == DEVICE_ID_EPMC_1553))
            offset = 0x200004;
         /* read from BAR 0 if only one BAR region, otherwise read BAR 2. */
         if (deviceInfoArrayPtr->mem_sections > 1)
            bar = 2;
         *(UINT2*)Buffer = BusRead4B(deviceInfoArrayPtr->memory_space, deviceInfoArrayPtr->memory_handle[bar], offset);
         break;

      case STATUS_INTERRUPT_DATA:
         if (deviceInfoArrayPtr->tail != deviceInfoArrayPtr->head) {
            *((UINT1*)Buffer) = deviceInfoArrayPtr->interrupt_queue[deviceInfoArrayPtr->tail];
            deviceInfoArrayPtr->tail++;
            deviceInfoArrayPtr->tail &= INTERRUPT_QUEUE_MASK;
         }
         break;

      case STATUS_PCI_CONFIG_SPACE:  /* returns the PCI config space register with the user supplied offset   */
         offset = (*((UINT4*)Buffer)*Length);    
         switch (Length) {
            case 1:
               *((UINT4*)Buffer) = BusRead1(deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, (Value)offset);
               break;
            case 2:
               *((UINT4*)Buffer) = BusRead2(deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, (Value)offset);
               break;
            case 4:       
               *((UINT4*)Buffer) = BusRead4(deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, (Value)offset);
               break;
            default:
               *((UINT4*)Buffer) = ~0;
         };
         break;

      default:
         db_printf("\nInvalid status number, %d\n", StatusNumber);
   };

   return Success;
}


/*===========================================================================*
 * ENTRY POINT:              C E I  W R I T E  S T A T U S
 *===========================================================================*
 *
 * FUNCTION:    Placeholder - Not currently implemented
 *
 *===========================================================================*/
static Error CEIWriteStatus (IODeviceVector iodev, Value StatusNumber, void * Buffer, Address Length) {
#if 0
   CEIPCIDevice * deviceInfoArrayPtr = NULL;

   deviceInfoArrayPtr = (CEIPCIDevice *)iodev->DeviceSpecificInfo;

   switch (StatusNumber) {
      default:
         db_printf("\nInvalid status number, %d\n", StatusNumber);
   };
#endif
   return Success;
}


/*===========================================================================*
 * ENTRY POINT:    C E I _ I N T _ P C I _ D R V _ I N I T
 *===========================================================================*
 *
 * FUNCTION:    Perform initialization on supported GE IP Avionics Avionics boards
 *
 * PARAMETERS:  PciData * data         -- (input) PCI header info on this device
 *              PciFunction * function -- (input) o/w function info on this device
 * 
 * RETURN VAL:  Error - Integrity-specific driver error codes
 *
 * GLOBALS:     deviceInfoArray (indirectly)
 *
 * DESCRIPTION: This routine is invoked by the kernel for each instance of a 
 *              detected PCI device.  For each supported Avionics PCI device,
 *              it registers the PCI Memory BARs and any required interrupt
 *              service routine with the kernel.
 *
 *===========================================================================*/
static Error cei_int_pci_drv_init (PciData * data, PciFunction * function) {
   UINT4 mask;
   UINT4 mem;
   UINT4 pciConfigCommand;
   int   barIndex;
   int   pciInterruptLine = 0;
   VECT  interrupt = 0;
   UINT2 pciDeviceId = 0;
   UINT4 bar_addr = 0;
   CEIPCIDevice * deviceInfoArrayPtr = NULL;

   if (debug)
      db_printf("\nCondor Engineering Integrity PCI Driver for PPC Processors\n");

   /* extract the PCI Device ID for this device */
   pciDeviceId = (UINT2)((function->id & 0xffff0000) >> 16);

   if (debug)
      db_printf("Condor Engineering device ID = %x vendor ID = %x \n", pciDeviceId, function->id & 0xffff);

   /* assign a device name and memory resource names to a valid GE IP Avionics Avionics device and its PCI memory BARs */
   if ((deviceInfoArrayPtr = CEIDeviceAllocate(pciDeviceId)) == NULL)
      return MaximumNumberOfResourcesExceeded;

   deviceInfoArrayPtr->devid         = pciDeviceId;
   deviceInfoArrayPtr->config_space  = data->config_space;
   deviceInfoArrayPtr->config_handle = function->handle;
   deviceInfoArrayPtr->memory_space  = data->memory_space;
   deviceInfoArrayPtr->head = 0;  /*initialize int FIFO head pointer */
   deviceInfoArrayPtr->tail = 0;  /*initialize int FIFO tail pointer */

   /* determine the PCI BAR configuration and map each implemented PCI memory region */
   for (barIndex = 0; barIndex < 6; barIndex++) {
      if (barIndex == 1) {
         continue;                       /* Skip BAR1 - I/O Region - mapping, it is not used */
      }
      bar_addr = (barIndex * 4) + 0x10;  /* Calculate the PCI Base Address Register's offset in PCI Config Space */

      /* Obtain this PCI Base Address Register content; if it is non-zero, this PCI memory region is implemented. */
      if ((mem = BusRead4B(deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, bar_addr)) != 0) {
         if (debug)
            db_printf("Found BAR%d, addr = %x\n", barIndex, mem);
   
         /* Using the old plug'n'play method, determine the PCI memory region size, limit it to a 4Kb minimum */
         BusWrite4B (deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, bar_addr, ~0);
         mask = BusRead4B (deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, bar_addr);
         deviceInfoArrayPtr->mem_size[barIndex] = ~mask+1;               /* Find the memory region size */
         if (deviceInfoArrayPtr->mem_size[barIndex] < 0x1000)
            deviceInfoArrayPtr->mem_size[barIndex] = 0x1000;             /* Limit BAR size smaller than 4Kb page */

         /* Restore the previously assigned physical BAR location to the respective register and log it */
         BusWrite4B (deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, bar_addr, mem);
         deviceInfoArrayPtr->mem_region_base[barIndex] = mem & ~0xf;     /* Strip non-address bits */
   
         if (debug)
            db_printf("Found BAR%d with size %x at phys addr %x\n",
                      barIndex, deviceInfoArrayPtr->mem_size[barIndex], deviceInfoArrayPtr->mem_region_base[barIndex]);
   
         /* Map each available and used BAR region  */
         if (BusMap (deviceInfoArrayPtr->memory_space, 
                     deviceInfoArrayPtr->mem_region_base[barIndex], 
                     deviceInfoArrayPtr->mem_size[barIndex], 
                     0, 
                     &deviceInfoArrayPtr->memory_handle[barIndex])) {
            db_printf ("Unable to create Bus Handle for BAR%d\n", barIndex);
            return Failure;
         }
   
         /* Record the information required to access these memory regions later, using memory_handle[BAR]
          * (virtual address) to reference the memory region's base address instead of mem_region_base[BAR]
          * (the physical address ?).
          */
         deviceInfoArrayPtr->bar_reservation[barIndex].First = (Address)deviceInfoArrayPtr->memory_handle[barIndex];
         deviceInfoArrayPtr->bar_reservation[barIndex].Last =  (Address)deviceInfoArrayPtr->memory_handle[barIndex] + 
                                                                        deviceInfoArrayPtr->mem_size[barIndex] - 1;
         deviceInfoArrayPtr->bar_reservation[barIndex].PhysicalSpaceNumber = 0;
         deviceInfoArrayPtr->bar_reservation[barIndex].RegistrationName = deviceInfoArrayPtr->mem_res_name[barIndex];
         
         /* Register each implemented memory region with the kernel */
         if (RegisterDynamicMemoryRegion(&deviceInfoArrayPtr->bar_reservation[barIndex]) != Success)
            db_printf("CEI PCI Driver: Failed to register region %d\n", barIndex);
   
         deviceInfoArrayPtr->mem_sections++;
      }
   }

   /* The devices supported by this driver have either 1 or 2 PCI memory regions, BAR0 and BAR2, respectively */
   switch (deviceInfoArrayPtr->mem_sections) {
      case 1:
         mem = 0;
         break;
      case 2:
      case 3:        
         mem = 2;
         break;
      default:
         db_printf("CEI PCI Driver Error: Invalid number of memory sections\n");
         return Failure;
   };

   deviceInfoArrayPtr->mem_base = deviceInfoArrayPtr->mem_region_base[mem];

   /* set the device PCI Command Word for Memory space access but no I/O space access */
   pciConfigCommand = BusRead4B (deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, 0x04);
   pciConfigCommand &= 0xFFFF; 
   pciConfigCommand |= 0x02;  /* 0x06 */
   BusWrite4B (deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, 0x04, pciConfigCommand);

   /* For All 1553 devices and the P-SER, setup the hardware interrupt */
   if ( ((deviceInfoArrayPtr->devid & DEVICE_ID_1553_MASK) == DEVICE_ID_1553_MASK) || 
         (deviceInfoArrayPtr->devid == DEVICE_ID_P_SER) || (deviceInfoArrayPtr->devid == DEVICE_ID_CEI_830) ) {
      deviceInfoArrayPtr->host_interface = BusRead4B (deviceInfoArrayPtr->memory_space, deviceInfoArrayPtr->memory_handle[mem], 0);

      if (debug)
         db_printf("\nHost_interface = %x\n", deviceInfoArrayPtr->host_interface);

      /* Read the Interrupt Line assignment from the PCI Config Space */
      pciInterruptLine = BusRead4B (deviceInfoArrayPtr->config_space, deviceInfoArrayPtr->config_handle, 0x3c) & 0xff;
      if (debug)
         db_printf("CEI PCI Driver IRQ Line = %d\n", pciInterruptLine);

      deviceInfoArrayPtr->head = 0; /*set the interrupt FIFO head pointer to 0; */
      
      /* Convert the Interrupt Line assignment to a Vector */
      if (BSP_PciIrqLineToVector (pciInterruptLine, &interrupt) != Success) {
         db_printf("\nCEI PCI Driver: Unable to map PCI interrupt line %d\n", pciInterruptLine);
         return Failure;
      }

      if (debug)
         db_printf("CEI PCI Driver: using interrupt %d\n", interrupt);

      /* Assign the ISR for the respective device type */
      if ((deviceInfoArrayPtr->devid == DEVICE_ID_P_SER) || (deviceInfoArrayPtr->devid == DEVICE_ID_CEI_830))
         deviceInfoArrayPtr->handler.handler = CommercialDevHandler;
      else
         deviceInfoArrayPtr->handler.handler = CEIDevHandler;

      deviceInfoArrayPtr->handler.vect = interrupt;
      deviceInfoArrayPtr->handler.id   = (Address)deviceInfoArrayPtr;

      /* finalize the ISR registration with the kernel */
      SetChainedExceptionHandler  (&deviceInfoArrayPtr->handler, deviceInfoArrayPtr->name);
      SetChainedExceptionPriority (&deviceInfoArrayPtr->handler, 1);
   }

   /* CORE-1553 devices, record the host interface value for ISR processing */
   if ((deviceInfoArrayPtr->devid == DEVICE_ID_PMC_D1) || (deviceInfoArrayPtr->devid == DEVICE_ID_EPMC_1553)) {
      deviceInfoArrayPtr->host_interface = BusRead2B(deviceInfoArrayPtr->memory_space, deviceInfoArrayPtr->memory_handle[mem], 0x200004);
      if (debug) {
         db_printf("Hardware ID = 0x%x\n", BusRead2B(deviceInfoArrayPtr->memory_space, deviceInfoArrayPtr->memory_handle[mem], 0x200000));
         db_printf("Hardware Version = 0x%x\n", BusRead2B(deviceInfoArrayPtr->memory_space, deviceInfoArrayPtr->memory_handle[mem], 0x200002));
         db_printf("Hardware Capablities = 0x%x\n", deviceInfoArrayPtr->host_interface);
         db_printf("Board Serial Number = 0x%x\n", BusRead2B(deviceInfoArrayPtr->memory_space, deviceInfoArrayPtr->memory_handle[mem], 0x20000A));
      }
   }

   deviceInfoArrayPtr->address = function->location;

   /* Register this new device with the system. */
   deviceInfoArrayPtr->device.Create                = CEIDevCreateIODevice;
   deviceInfoArrayPtr->device.ReadBlock             = NULL;
   deviceInfoArrayPtr->device.WriteBlock            = NULL;
   deviceInfoArrayPtr->device.ReadBuffers           = NULL;  /* CEIReadBuffers; */
   deviceInfoArrayPtr->device.WriteBuffers          = NULL;
   deviceInfoArrayPtr->device.ReadRegister          = NULL;
   deviceInfoArrayPtr->device.WriteRegister         = NULL;
   deviceInfoArrayPtr->device.ReadStatus            = CEIReadStatus;
   deviceInfoArrayPtr->device.WriteStatus           = CEIWriteStatus;
   deviceInfoArrayPtr->device.IOCoherentNotRequired = 1;
   deviceInfoArrayPtr->device.IOSynchronizationNotRequired = 0;
   deviceInfoArrayPtr->device.DeviceSpecificInfo    = (void *)deviceInfoArrayPtr;

   if (debug)
      db_printf("Registering -- %s\n", deviceInfoArrayPtr->name);

   return RegisterIODeviceVector(&deviceInfoArrayPtr->device, deviceInfoArrayPtr->name);
}
