/*============================================================================*
 * FILE:                      L O W L E V E L . H
 *============================================================================*
 *
 * COPYRIGHT (C) 1997 - 2009 BY
 *          GE INTELLIGENT PLATFORMS, INC. SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:    Header file for low-level device driver access.
 *
 *              This file defines the interface to the Integrity low-level  
 *              device interface routines.
 *
 *              Warning: this file is shared by many GE IP Avionics products, so 
 *              please exercise caution when making modifications.
 *
 *---------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *  Date        Revision
 * ----------   ---------------------------------------------------------------
 * 07/26/2005   Merged with CEI-x20 version.CEI-x20 v3.93.skb
 * 11/28/2005   Modified to remove CEI_LL specifics from lowlevel.h 
 * 02/26/2009   Moved Integrity-specific error codes from mem_integrity.c
 *===========================================================================*/

#ifndef _LOWLEVEL_H_
#define _LOWLEVEL_H_


/*--------------------------------------------------------------------------*
 * Constants common to all products supported by this low-level interface
 *--------------------------------------------------------------------------*/

#ifndef WIN32
   #ifndef HANDLE
      #define HANDLE int
   #endif
   #ifndef UINT
      #define UINT unsigned int
   #endif
   #ifndef DWORD
      #define DWORD unsigned long
   #endif
#endif


/*--------------------------------------------------------------------------*/
/* Error reporting constants                                                */
/*--------------------------------------------------------------------------*/

#define BTD_OK                0     /* success */
#define BTD_ERR_PARAM         1     /* invalid parameter */
#define BTD_ERR_BADADDR       4     /* invalid address */
#define BTD_ERR_BADADDRMAP    12    /* bad initial mapping of address */
#define BTD_NO_DRV_MOD        30    /* no Driver Module found */
#define BTD_IOCTL_SET_REG     32
#define BTD_IOCTL_REG_SIZE    33
#define BTD_IOCTL_GET_REG     34
#define BTD_BAD_SIZE          35
#define BTD_NO_HASH_ENTRY     38
#define BTD_WRONG_BOARD       39    /* Wrong board type for command          */
#define BTD_ERR_NOWINRT       50    /* WinRT driver not loaded/started */
#define BTD_ERR_BADREGISTER   51    /* WinRT parameters don't match registry */
#define BTD_ERR_BADOPEN       52    /* WinRT device open failed */
#define BTD_UNKNOWN_BUS       53    /* bus is not PCI, ISA or VME */
#define BTD_BAD_LL_VERSION    54    /* unsupported lowlevel driver installed */
#define BTD_BAD_INT_EVENT     55    /* unable to create interrupt event */
#define BTD_ISR_SETUP_ERROR   56    /* error setting up the ISR driver */
#define BTD_CREATE_ISR_THREAD 57    /* error creating the ISR thread */
#define BTD_NO_REGIONS_TO_MAP 58
#define BTD_BAD_DEVICE_NUM    59
#define BTD_DRV_READ_FAIL     60 
#define BTD_DRV_WRITE_FAIL    61
#define BTD_DRV_IOCTL_FAIL    62
#define BTD_SYSFS_FAIL        63        
#define BTD_SYSFS_ATTR_FAIL   64
#define BTD_TOO_MANY_RGNS     65    /* invalid hardware memory region count */
#define BTD_RESOURCE_ERR      70    /* Integrity Resource Error              */
#define BTD_READ_IODEV_ERR    71    /* Integrity IO Device Read Error        */
#define BTD_MEMREG_ERR        72    /* Integrity error getting memory region */
#define BTD_MEM_MAP_ERR       73    /* Integrity Memory Mapping error        */


/*--------------------------------------------------------------------------*/
/* Miscellaneous constants, types, and macros                               */
/*--------------------------------------------------------------------------*/

/* maximum number of memory regions supported by a device */
#define MAX_MEMORY   6     

/* maximum number of ports supported by a device */
#define MAX_PORTS    2

/* host bus types supported */
typedef enum _BUS_TYPE 
{
    BUS_INTERNAL,
    BUS_ISA,
    BUS_PCI,
    BUS_VME,
    BUS_PCMCIA,
    BUS_OTHER
} BUS_TYPE;           

/* maximum number of physical devices supported */
#define MAX_DEVICES   16
#define MAX_BOARDS MAX_DEVICES

/*===========================================================================*
 * ENTRY POINT: vbtGetPCIDeviceID
 *===========================================================================*
 *
 * FUNCTION:    Obtain the PCI device ID associated with the given board.
 *
 * RETURN VAL:  BTD_OK - operation completed successfully
 *              other  - refer to BTD_x error codes listed in lowlevel.h
 *
 * DESCRIPTION: Read the PCI device ID from the PCI configuration space
 *              associated with the given board index.
 *
 *===========================================================================*/
int vbtGetPCIDeviceID
(
   UINT board,                     /* (i) board index (0 - 9) */
   DWORD *deviceID                 /* (o) receives corresponding PCI device ID */
);


/*--------------------------------------------------------------------------*
 *
 * The following sections support the BusTools/1553-API
 *
 *--------------------------------------------------------------------------*/


#if defined(_LINUX_X86_) || defined(VXW_VME_PPC) || defined(VXW_Q104_X86) || defined(VXW_PCI_PPC) || defined(VXW_PCI_THALES) \
                         || defined(VXW_PCI_X86) || defined(LYNXOS) || defined (VXW_DY4_PCI_PPC) || defined(VXW_VME_X86)
#ifndef _UNIX_
#define _UNIX_
/*#define HANDLE int */
/*#define UINT unsigned */
#endif 
#endif

/* Define the max number of logical devices (cardnum's) we support */
#define MAX_CARDNUM   16            /* V4.06 */

/**********************************************************************
*  vbtHookInterrupt-invoked interrupt function return codes:
**********************************************************************/
#define BTD_NEXT_INTERRUPT    0
#define BTD_EXIT_INTERRUPT    1

/**********************************************************************
*  Lowlevel address mapping structure
**********************************************************************/
typedef struct _DEVMAP_T {
   int            busType;                    /* One of BUS_TYPE. */
   int            interruptNumber;            /* Interrupt number. */
   int            memSections;                /* Number of memory regions defined. */
   int            flagMapToHost[MAX_MEMORY];  /* Set to map region into host space. */
   char         * memHostBase[MAX_MEMORY];    /* Base address of region in host space. */
   unsigned       memStartPhy[MAX_MEMORY];    /* Physical base address of region. */
   unsigned       memLengthBytes[MAX_MEMORY]; /* Length of region in bytes. */
   int            portSections;               /* Number of I/O port regions. */
   unsigned       portStart[MAX_PORTS];       /* I/O Address of first byte. */
   unsigned       portLength[MAX_PORTS];      /* Number of bytes in region. */
   int            llDriverVersion;            /* Low Level driver version. */
   int            KernelDriverVersion;        /* Kernel driver version. */
   HANDLE         hKernelDriver;              /* Handle to the kernel driver. */
   unsigned int   VendorID;                   /* Vendor ID if PCI card */
   unsigned int   DeviceID;                   /* Device ID if PCI card */
   int            device;                     /* This the device */
   int            use_count;                  /* Number of user channels  */
} DEVMAP_T, *PDEVMAP_T;

typedef struct _DEVINFO_T {
   int            busType;                    /* One of BUS_TYPE. */
   int            nchan;                      /* number of channeld. */
   int            irig;                       /* IRIG option 0=no 1=yes. */
   int            mode;                       /* mode 0=single 1=multi. */
   int            memSections;
   unsigned int   VendorID;                   /* Vendor ID if PCI card */
   unsigned int   DeviceID;                   /* Device ID if PCI card */
} DEVINFO_T, *PDEVINFO_T;


#ifdef VXWORKS
static unsigned char irqvalue;
static unsigned int_count[4];
#endif

int  vbtGetInterrupt(UINT device);

void vbtFreeBoardAddresses (PDEVMAP_T pMap);      /* Pointer to region mapping structure */

int  vbtMap1553BoardAddresses (UINT      cardnum,
                               UINT      device, /* Device to open (0 - 9) */
                               PDEVMAP_T pMap);  /* Pointer to region mapping structure */

int  vbtMapBoardAddresses (UINT      device,     /* Device to open (0 - 9) */
                           PDEVMAP_T pMap);      /* Pointer to region mapping structure */

#if defined(__WIN32__)
int vbtHookInterrupt(UINT    device,
                     UINT    cardnum,         /* (i) BusTools card number */
                     void (_stdcall *fn)(int),/* (i) Address of user function to call */
                     int iPriority,           /* (i) Thread priority */
                     HANDLE  *phThread,       /* (o) Returned handle to thread */
                     HANDLE  *hIntEvent,      /* (o) Returned handle to interrupt event */
                     UINT  CurrentCardType,   /* (i) Board type for interrupt setup */
                     UINT  CurrentCardSlot,   /* (i) Carrier slot or board channel being setup */
                     UINT  CurrentCarrier);   /* (i) Type of IP carrier for interrupt setup */
#endif

void  vbtCloseInterrupt(UINT device, UINT cardnum);
#if defined(LYNXOS_VME_PPC)
int   vbtMapIoAddress(unsigned int cardnum, unsigned long* p_addr);  
int   vbtMapBoardAddress(unsigned int cardnum, unsigned long* p_addr, unsigned long* vme_addr);
#else
int   vbtMapIoAddress(unsigned base_address, char **addr);  /* for VME carriers under VxWorks  */
int   vbtMapBoardAddress(unsigned cardnum,          /* Card Number to open */
                         unsigned device,           /* device to open (0 - 9) */
                         char     **addr,           /* computed host address */
                         void     *MemorySize,      /* byte size of region mapped */
                         int      addressing_mode); /* for VME boars only. */
#endif


void vbtFreeBoardAddress(UINT cardnum);

#if !defined(VXW_VME_68K) || !defined(LYNXOS_VME_PPC)
int  vbtGetPCIConfigRegister(UINT    device,  /* WinRT device to open (0 - 9) */
                             UINT    offset,  /* Offset of PCI config register */
                             UINT    length,  /* Number of bytes to read (1,2,4) */
                             UINT   *value);  /* Returned value of the register */
#else
#define vbtGetPCIConfigRegister(a,b,c,d) 0
#endif

int vbtGetDevInfo(int device,  char* devInfo,  unsigned long* pData);

int vbtGetCEIDevInfo(UINT device,       /* WinRT device to open (0 - 15) */
                     DEVINFO_T *pInfo);   /* Pointer to device info structure */

#endif  /* #ifndef _LOWLEVEL_H_ */
