/*============================================================================*
 * FILE:                     M E M _ I N T E G R I T Y . C
 *============================================================================*
 *
 * COPYRIGHT (C) 2006 - 2009 BY 
 *          GE INTELLIGENT PLATFORMS, INC. SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:   GE IP Avionics Generic Integrity VME/PCIBus interface functions.
 *
 *             This module provides the interface between the GE Intelligent 
 *             Platforms Avionics APIs and the Integrity operating system.
 *             It performs the actual I/O and memory-mapping functions
 *             needed to support operation in an Integrity environment.
 *             Since this file is used by multiple products, use care when
 *             performing modifications.
 *
 * EXTERNAL ENTRY POINTS:
 *
 *    vbtMapBoardAddresses    Looks up the physical base address of the board
 *                            in the registery then creates a pointer which is
 *                            valid in the current 32-bit environment and may
 *                            be used to directly access the memory-mapped RAM
 *                            on the board.
 *
 *    vbtFreeBoardAddresses   Free any system resources which were obtained in
 *                            the mapping of the phyical board address.
 *
 *    vbtGetPCIConfigRegister Looks at the specified device's PCI config space
 *                            register. 
 *                            
 *    vbtGetDevInfo           Retrieves specific information concerning the 
 *                            device.
 *
 *    vbtMapBoardAddress      Looks up the physical base address of the board
 *                            in the registery then creates a pointer which is
 *                            valid in the current 32-bit environment and may
 *                            be used to directly access the memory-mapped RAM
 *                            on the board.
 *
 *    vbtFreeBoardAddress     Free any system resources which were obtained in
 *                            the mapping of the phyical board address.
 *
 * EXTERNAL NON-USER ENTRY POINTS:
 *
 * INTERNAL ROUTINES:
 *
 *===========================================================================*
 *
 * Modification History:
 *
 *    Date     Modifications
 * ----------  ---------------------------------------------------------------
 * 06/04/2001  Configure current API release
 * 11/25/2003  Add support for VME SUP and USR mapping
 * 11/15/2005  Add support for CEI-x30 API.  Added vbtGetPCIConfigRegister,
 *             vbtFreeBoardAddresses and vbtGetDevInfo. Added error checking for
 *             Integrity system calls.
 * 04/07/2009  Completed integration of CEI-830, P-SER, and P-708 devices.
 *
 *===========================================================================*/


/*---------------------------------------------------------------------------*
 *                    INCLUDES, DEFINES and GLOBALS
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <integrity.h>
#include <bsp.h>
#include <memory_region.h>
#include <util/error_string.h>
#include "lowlevel.h"


/* #define DEBUG_PRINT */



#ifdef INTEGRITY_VME_PPC
#include <vme/vme_api.h>
extern Error CreateWindow(Value num, Value am, Address vme_start, Address length, MemoryRegion* mr, Address* pci_start_p);
extern Error InitializeVMEBridge();
static int One_Time_Init=0;
static int vme_device_use[MAX_DEVICES];
static int vme_device_cnt=0;
MemoryRegion vme_mem[MAX_DEVICES][2];
#endif

/*---------------------------------------------------------------------------*
 * Constants for identifying supported GE IP Avionics Avionics devices
 *---------------------------------------------------------------------------*/
#define DEVICE_ID_PMC_1553    0x1553
#define DEVICE_ID_QPMC_1553   0x1554
#define DEVICE_ID_QPC_1553    0x1555
#define DEVICE_ID_QPCX_1553   0x1556
#define DEVICE_ID_P_708       0x0708
#define DEVICE_ID_CEI_430     0x0430
#define DEVICE_ID_CEI_530     0x0530
#define DEVICE_ID_RAR_CPCI    0x0630
#define DEVICE_ID_CEI_830     0x0830
#define DEVICE_ID_R830RX      0x0831
#define DEVICE_ID_PMC_D1      0x1001
#define DEVICE_ID_EPMC_1553   0x1003
#define DEVICE_ID_P_SER       0x1004

#define VENDOR_ID_GEF_AVIONICS  0x13C6

/*---------------------------------------------------------------------------*
 * Constants for the driver interface CEIReadStatus
 * A matching set of constants exists in the device driver file
 *---------------------------------------------------------------------------*/
#define STATUS_DEVICE_ID        1
#define STATUS_HOST_INTERFACE   2
#define STATUS_INTERRUPT_DATA   3
#define STATUS_PCI_CONFIG_SPACE 4

#ifdef INTEGRITY_PCI_PPC
extern MemoryRegionPool* __ghs_VirtualMemoryRegionPool;

/* holds the virtual memory addresses for each mapped BAR region */
MemoryRegion v_mem_regions[MAX_DEVICES][MAX_MEMORY];


/****************************************************************************
*
*  PROCEDURE - vbtMapBoardAddresses
*
*  FUNCTION
*     This routine looks up the physical base addresses of the board, then
*     creates pointers which are valid in the current 32-bit environment and
*     may be used to directly access the memory-mapped RAM on the board.  
*     Only memory regions specified by setting "memMapToHost" are mapped.  If
*     any such regions are mapped, call vbtFreeBoardAddress() to free the region
*     maps and close the handle.
*     If no regions were specified to be mapped, the handle is closed and it is
*     not necessary to call vbtFreeBoardAddress().
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
*     SYSTEM_ERROR    - any errors
*     
****************************************************************************/
int vbtMapBoardAddresses (UINT      device, /* Device to open (0 - 15)            */
                          PDEVMAP_T pMap)   /* Pointer to region mapping structur */
{
   int i;
   int status = BTD_OK;
   Error err;
   char dbuf[20] = "";
   char res_name[MAX_MEMORY][40];
   UINT2 boardType = 0;
   IODevice cboard;
   MemoryRegion p_mem_reg;
   unsigned long lpbase = 0;
   Value mem_size = 0;

   /* Only attempt to map a board that has not already been mapped */
   if (pMap->use_count == 0) {
      memset(res_name, 0, sizeof(res_name));
       
      /* obtain the resource handle from the kernel */
      sprintf (dbuf, "CEIboard%d\n", device);  
      err = RequestResource ((Object*)&cboard, dbuf, "!systempassword");

      if (err == Success) {
         /* read PCI configuration space word 1 - the PCI Device ID */
         err = ReadIODeviceStatus (cboard, 1, (void*)&boardType, sizeof(boardType));
         if (err != Success) {
#ifdef DEBUG_PRINT
            printf ("Fail:  vbtMapBoardAddresses. ReadIODeviceStatus - %s (%d)\n", ErrorString(err), err);
#endif
            status = BTD_READ_IODEV_ERR;
         }
      }
      else {
#ifdef DEBUG_PRINT
         printf ("Fail:  vbtMapBoardAddresses. RequestResource - %s (%d)\n", ErrorString(err), err);
#endif
         status = BTD_RESOURCE_ERR;
      }

      /* Reference the device memory resource by name based on the board type.
       * Note the name strings must match those used in the driver file CEI_INT_PCI_DRV.C
       */
      if (status == BTD_OK) {
         pMap->VendorID = VENDOR_ID_GEF_AVIONICS;
         pMap->memSections = 2;

         switch (boardType) {
            case DEVICE_ID_PMC_1553:
               pMap->memSections = 1;
               sprintf(res_name[0],"QPMC1553_BASE%d",device); 
               break;
            case DEVICE_ID_QPMC_1553:
               sprintf(res_name[0],"QPCI_PLX_CONFIG_BASE%d",device); 
               sprintf(res_name[1],"QPCI1553_BASE%d",device); 
               break;
            case DEVICE_ID_QPC_1553:
               sprintf(res_name[0],"QCP_PLX_CONFIG_BASE%d",device);
               sprintf(res_name[1],"QCP1553_BASE%d",device);
               break;
            case DEVICE_ID_QPCX_1553:
               sprintf(res_name[0],"QPCX_PLX_CONFIG_BASE%d",device);
               sprintf(res_name[1],"QPCX1553_BASE%d",device);
               break;
            case DEVICE_ID_CEI_830:
               sprintf(res_name[0],"CEI830_PLX_CONFIG_BASE%d",device);
               sprintf(res_name[1],"CEI830_BASE%d",device);
               break;
            case DEVICE_ID_R830RX:
               sprintf(res_name[0],"R830RX_PLX_CONFIG_BASE%d",device);
               sprintf(res_name[1],"R830RX_BASE%d",device);
               break;
            case DEVICE_ID_P_708:
               sprintf(res_name[0],"P708_PLX_CONFIG_BASE%d",device);
               sprintf(res_name[1],"P708_BASE%d",device);
               break;
            case DEVICE_ID_P_SER:
               sprintf(res_name[0],"PSER_PLX_CONFIG_BASE%d",device);
               sprintf(res_name[1],"PSER_BASE%d",device);
               break;
            default:
               status = BTD_WRONG_BOARD;
         } 
      }
   
      /* for any valid board, acquire the memory regions and map them in the virtual memory pool */
      for (i = 0; (i < pMap->memSections) && (status == BTD_OK); i++) {
         if (v_mem_regions[device][i] != 0) {
#ifdef DEBUG_PRINT
            printf("Fail:  memory region %d already mapped (vbtMapBoardAddresses)\n", i);
#endif
            continue;
         }

         err = RequestResource ((Object*)&p_mem_reg, res_name[i], "!systempassword");
         if (err == Success) {
#ifdef GHS_KERNEL
            {
               Address start;
               Address end;
               err = GetMemoryRegionAddresses (p_mem_reg, &start, &end);
               if (err == Success) {
                  lpbase = (unsigned long)start;
                  mem_size = (end - start) + 1;
               }
               else {
   #ifdef DEBUG_PRINT
                  printf("Fail:  vbtMapBoardAddresses. GetMemoryRegionAddresses - %s (%d)\n", ErrorString(err), err);        
   #endif
                  status = BTD_MEMREG_ERR;
               }
            }
#else
            err = MapMemory (__ghs_VirtualMemoryRegionPool, p_mem_reg, &v_mem_regions[device][i], &lpbase, &mem_size);
            if (err != Success) {
   #ifdef DEBUG_PRINT
               printf("Fail:  vbtMapBoardAddresses. MapMemory - %s (%d)\n", ErrorString(err), err);             
   #endif
               status = BTD_MEM_MAP_ERR;
            }
#endif    

            if (err == Success) {
               pMap->memHostBase[i] = (char*)lpbase;
               pMap->memLengthBytes[i] = (unsigned)mem_size;
               pMap->flagMapToHost[i] = 1;
            }
         }
         else {
#ifdef DEBUG_PRINT
            printf ("Fail:  vbtMapBoardAddresses. RequestResource - %s (%d)\n", ErrorString(err), err);
#endif
            status = BTD_RESOURCE_ERR;
         }
      }
   
      if (status == BTD_OK) {
         pMap->device   = device;
         pMap->DeviceID = boardType;
         pMap->busType  = BUS_PCI;
         pMap->use_count++;
      }
   }

   return status;
}

/****************************************************************
*
*  PROCEDURE - vbtFreeBoardAddresses
*
*  FUNCTION
*     This routine frees any system resources which were obtained
*     in the mapping of the phyical board address.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     Tothing
****************************************************************/
void vbtFreeBoardAddresses (PDEVMAP_T pMap)
{
   int i;
   Error err;

   if (pMap->use_count > 0) {
      pMap->use_count--;
   }
   
#ifndef GHS_KERNEL
   if (pMap->use_count == 0) {
      for (i = 0;i < MAX_MEMORY;i++) {
         if (pMap->flagMapToHost[i] == 1) {
            if ((err = UnmapMemory(__ghs_VirtualMemoryRegionPool, v_mem_regions[pMap->device][i])) == Success) {
               v_mem_regions[pMap->device][i] = 0;
            }
            else {
   #ifdef DEBUG_PRINT
               printf("Fail:  vbtFreeBoardAddresses. UnmapMemory - %s (%d)\n", ErrorString(err), err);
   #endif
            }
	      }
      }
   } 
   if (err) ;
#endif 
}

/****************************************************************
*
*  PROCEDURE - vbtGetPCIConfigRegister
*
*  FUNCTION
*     This routine takes the device number, an offset in PCI
*     configuration space, a byte length, and a pointer for the 
*     value of the specified PCI configuration space register.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK          - retrieved PCI config space register value
*     SYSTEM_ERROR    - any errors
****************************************************************/
int vbtGetPCIConfigRegister (UINT cardnum,
                             UINT offset,
                             UINT length,
                             UINT* value)
{
  Error err;
  IODevice ioDev;
  char dbuf[20]="";
  UINT4 val=offset;  
    
   sprintf(dbuf,"CEIboard%d\n",cardnum);  
   if ((err = RequestResource ((Object*)&ioDev,dbuf, "!systempassword")) != Success) {
#ifdef DEBUG_PRINT
      printf("Fail:  vbtGetPCIConfigRegister. RequestResource - %s (%d)\n", ErrorString(err), err);
#endif
      return BTD_RESOURCE_ERR;
   }

   if ((err = ReadIODeviceStatus (ioDev, STATUS_PCI_CONFIG_SPACE, &val, length)) != Success) {
#ifdef DEBUG_PRINT
      printf("Fail:  vbtGetPCIConfigRegister. ReadIODeviceStatus - %s (%d)\n", ErrorString(err), err);
#endif
      return BTD_READ_IODEV_ERR;
   }
   
   *value = (UINT)val;
  
   if (err) ;
   return BTD_OK;
}

/****************************************************************
*
*  PROCEDURE - vbtGetDevInfo
*
*  FUNCTION
*     This rountine retrieves specific information concerning the 
*     device.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     BTD_OK          - retrieved device information
*     SYSTEM_ERROR    - any errors
****************************************************************/
int vbtGetDevInfo (int device,
                   char * devInfo,
                   unsigned long * pData)
{
  Error err;
  IODevice ioDev;
  char dbuf[20]="";
  UINT2 val=0;
  int offset=0;
     
   if (strcmp(devInfo, "id") == 0)       /* the device id */
      offset = STATUS_DEVICE_ID;
   else if (strcmp(devInfo, "hir") == 0) /* host interface register */
      offset = STATUS_HOST_INTERFACE;
   else if (strcmp(devInfo, "iq") == 0)  /* interrupt queue */
      offset = STATUS_INTERRUPT_DATA;
   else
      return BTD_ERR_PARAM;    

   sprintf(dbuf,"CEIboard%d\n",device);  
   if ((err = RequestResource ((Object*)&ioDev, dbuf, "!systempassword")) != Success) {
#ifdef DEBUG_PRINT
      printf("Fail:  vbtGetDevInfo. RequestResource - %s (%d)\n", ErrorString(err), err);
#endif
      return BTD_RESOURCE_ERR;
   }
   if ((err = ReadIODeviceStatus (ioDev, offset, (void*)&val, sizeof(val))) != Success) {
#ifdef DEBUG_PRINT
      printf("Fail:  vbtGetDevInfo. ReadIODeviceStatus - %s (%d)\n", ErrorString(err), err);
#endif
      return BTD_READ_IODEV_ERR;
   }

   *pData = (unsigned long)val;
  
   if (err) ;
   return BTD_OK;
}
#endif /* INTEGRITY_PCI_PPC */



#ifdef INTEGRITY_VME_PPC

/****************************************************************
*
*  PROCEDURE - vbtMapBoardAddress(base_address,
*                                 map_length, 
*                                 char ** addr)
*
*  PARAMETERS
*     BT_U32BIT  base_address      - Card physical base address.
*     BT_U32BIT  map_length        - Size of region to map in bytes.
*     int        addressing_mode   - VME Addressing Mode.
*     char      ** addr            - Returned address of board.
*     void      *junk               - not used
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
****************************************************************/
int vbtMapBoardAddress(unsigned base_addr, unsigned map_length, char **addr, void *junk, int addressing_mode)
{
   Error err; 
   if (!One_Time_Init)
   {
      /* Can only be called once */
      err = InitializeVMEBridge();
      if (err !=Success)
         return 2222;
      
      One_Time_Init = 1;
   }

   if ( CreateWindow( 1, A32, base_addr, 0x800000, vme_mem[1], (Address *)addr[0]) != Success)
       return BTD_ERR_BADEXTMEM;
   else
       return 0;
}

/****************************************************************
*
*  PROCEDURE - vbtMapIoAddress(cardnum, base_address, char ** addr)
*
*  VxWorks Version - Flat VME addressing model
*
*  PARAMETERS
*     BT_U32BIT  base_address      - Card I/O base address.
*     char      ** addr            - Returned address of board.
*
*  RETURNS
*     BTD_OK          - Mapped address with no error
****************************************************************/
int vbtMapIoAddress(unsigned base_address, char **addr)
{
    int i;
    MemoryRegion mr;
      
    if ( CreateWindow( 0, A16, base_address,0x40, vme_mem[0], (Address *)addr[0]) == Success)
       return BTD_OK;
    else
       return BTD_ERR_BADADDR;
}

/****************************************************************
*
*  PROCEDURE - vbtFreeBoardAddress
*
*  FUNCTION
*     This routine frees any system resources which were obtained
*     in the mapping of the phyical board address.
*
*  PARAMETERS
*     See below.
*
*  RETURNS
*     Tothing
****************************************************************/
void vbtFreeBoardAddress(UINT cardnum) {}

#endif /*INTEGRITY_VME_PPC */
