/*===========================================================================*
 * FILE:                     C D E V _ V X W . C
 *===========================================================================*
 *
 * COPYRIGHT (C) 2004-2010 BY
 *          GE INTELLIGENT PLATFORMS, INC. SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:    Application Programmer's Interface for a CEI-x30 ARINC device.
 *              VxWorks 32-bit Operating System Specific Functionality
 *
 *    Internal CEI-x30 routines implemented herein:
 *
 *    timeGetTime                 - replacement for Windows O/S Get Timer call
 *    cei_acquire_device          - opens a session with the specified x30 device
 *    cei_acquire_device_identity - supplies the PCI configuration space ID values
 *    cei_acquire_thread_lock     - acquires sempahore/mutex lock
 *    cei_delay_milliseconds      - o/s specific delay utility in milliseconds
 *    cei_delay_seconds           - o/s specific delay utility in fractional seconds
 *    cei_get_host_system_time_x  - acquires host system time-of-year 64-bit uSec     
 *    cei_get_host_system_time    - acquires host system time-of-year 32-bit mSec     
 *    cei_get_irig_initial_time   - acquires host system time-of-year IRIG format
 *    cei_map_shared_memory       - maps memory region for sharing globals across processes
 *    cei_open_shared_lock        - creates a shared semaphore across processes
 *    cei_release_device          - closes a session with the specified x30 device
 *    cei_release_thread_lock     - releases sempahore/mutex lock     
 *    cei_read_32bit              - read a 32-bit value from an x30 device
 *    cei_write_32bit             - write a 32-bit value to an x30 device
 *    cei_init_interrupt          - configures an interrupt and ISR
 *    cei_terminate_interrupt     - terminate an interrupt
 *          
 *---------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ------------------------------------------------------------
 * 10/06/04  BDW  Created.
 * 04/06/06  BDW  Added support for the Motorola MCP750 cPCI controller BSP.
 *                Corrected a bug in the cei_delay_seconds routine for MVME2300.
 * 04/11/06  BDW  Removed utility routines cei_read_plx_32bit & cei_write_plx_32bit
 * 08/24/06  BDW  Incorporated CEI_TYPES.H data type definitions in 
 *                  support of 64-bit Linux operating systems. 
 *                Swapped the BAR reference order by the device structure 
 *                   members baseAddr and devicePtr
 * 02/27/07  BDW  Added cei_acquire_device_identity() to supply the PCI 
 *                   Manufacturer and Device IDs for x30 products 
 * 08/15/07  BDW  Added support for the CEI-530 and R830RX boards
 * 02/18/08  BDW  Removed the boardInitialized array, relocated as a structure
 *                   member in the data structure array "device"
 * 03/10/08  BDW  Removed the declaration of the following global arrays, 
 *                relocated to CDEV_GLB.H:
 *                   bypassWrapTest
 *                   irigDacThreshold
 *                   a717syncWords
 *                Removed declaration of basicConfiguration, it was unused.
 *                Added reference to device[] structure as a parameter to
 *                   cei_acquire_device and cei_release_device, so they no 
 *                   longer access the global data structure directly.
 * 04/15/08  BDW  Added support for initiation and termination of h/w interrupts
 *                via routines cei_init_interrupt and cei_terminate_interrupt.
 * 06/16/09  BDW  Added the function cei_open_preload_lock for multiprocess support.
 * 07/01/09  BDW  Added new function cei_delay_milliseconds (o/s thread suspend).
 * 10/28/10  BDW  Modified the parameter list for cei_init_interrupt to account for      
 *                   interrupt setup requirements in the Windows-only CEI_Install library
 *
 *===========================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <logLib.h>
#include <taskLib.h>
#include <intLib.h>
#include <sysLib.h>
#include <semLib.h>
#include <sigLib.h>
#include <vme.h>
#if defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include "target_defines.h"
#endif
#include "cdev_api.h"
#include "cdev_glb.h"

extern UINT VxW_GetCommercialInfo( UINT device, UINT *bar0addr, UINT *bar2addr, UINT *manufId, UINT *deviceId, UINT *irq );
extern void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32, pCEI_UINT32);

/* VxWorks Interrupt Support */
#define INT_NUM_IRQ0    0x20
#define INT_VEC_IRQ0		0x00	/* vector for IRQ0 */
#define IRQ_OFFSET 0                                         /* FFFFFFFFFFFFFFFFFFFFFFFFFFFIIIIIIXXX THIS */
#define MAX_MEMORY_REGIONS 6        /* max # of BAR regions on a device */

#ifdef NON_INTEL_WORD_ORDER
   #define CEI_HW_INTERRUPT_PENDING 0x20000000
#else
   #define CEI_HW_INTERRUPT_PENDING 0x00000010
#endif

#define MAX_DEV_OPEN_ERROR_STR_LEN 512

typedef struct device_attributes_type {
   CEI_CHAR  irq;                        /* Interrupt Request Level */
   CEI_INT32 deviceID;
   CEI_INT32 manufID;
} DEVICE_ATTRIBUTES_TYPE;

/*===========================================================================* 
 * VxWorks-based low-level interface data structure types
 *===========================================================================*/
/* VxWorks Interrupt Service Routine argument type (pointer to struct)       */
typedef struct VXW_INT_DATA {
   CEI_UINT32 baseAddress;                /* device BAR2 virtual address   */
   CEI_INT32  device;                     /* device index                  */
} VXW_INT_DEVICE_INFO_TYPE;

/* PCI Device information structure - all things PCI for VxWorks */
typedef struct pciregion_state_type 
{
   CEI_INT32  busNumber;                  /* PCI bus num of found device   */
   CEI_INT32  deviceNumber;               /* PCI dev num of found device   */
   CEI_INT32  functionNumber;             /* PCI func # of found device    */
   CEI_CHAR   irq;                        /* Device Interrupt Request Level*/
   CEI_UINT32 size[MAX_MEMORY_REGIONS];   /* size of each BAR for each dev */
   CEI_UINT32 membase[MAX_MEMORY_REGIONS];/* base addr of each BAR for each*/
   CEI_UINT32 initialStateMask;           /* mask parameter to vmStateSet  */
   CEI_UINT32 initialState;               /* state parameter to vmStateSet */
   CEI_INT16  deviceID;                   /* device ID for found device    */
   CEI_INT16  manufID;                    /* manuf ID for found device     */
} CEI_PCI_REGION_STATE;


/* Potentially common "non-irq assigned value" interrupt level assignments - to be defined by user for the respective BSP */
#if defined(INT_PROCESS_USE_INT_LEVEL)
   #define EXT_INTERRUPT_BASE	0x10
   #define PMC_INT_LVL1	( 0x09 + EXT_INTERRUPT_BASE )
   #define PMC_INT_LVL2	( 0x0a + EXT_INTERRUPT_BASE )
   #define PMC_INT_LVL3	( 0x0b + EXT_INTERRUPT_BASE )
   #define PMC_INT_LVL4	( 0x0c + EXT_INTERRUPT_BASE )

static CEI_INT32 pmcInterruptLevel[] = {
   PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1,
   PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1,
   PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1,
   PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1, PMC_INT_LVL1
};
#endif 

static DEVICE_ATTRIBUTES_TYPE deviceAttributes[MAX_BOARDS];

/* storage for Windows driver low-level board mapping error messages      */
static pCEI_CHAR deviceOpenErrorDesc = "Specified PCI Device Not Found";

/* Interrupt Service Routine structure - device index & BAR2 pointer */
/* static */ VXW_INT_DEVICE_INFO_TYPE isrDeviceInfo[MAX_BOARDS];

/* PCI Device information array */
CEI_PCI_REGION_STATE devicePciAttributes[MAX_BOARDS];


/* Interrupt Task Notification Semaphore */
static SEM_ID interruptSemaphoreAssignment[MAX_BOARDS] = { 
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL 
};

/* Global storage for the h/w interrupt(s) assigned to the device(s) */
static CEI_INT32 localIRQ[MAX_BOARDS] = {0,0,0,0,0,0,0,0};

/* Number of times the ISR has executed for each device - debug use mainly */
static CEI_INT32 interruptCounter[MAX_BOARDS] = {0,0,0,0,0,0,0,0};

/* Task ID for the interrupt processing task */
static CEI_INT32 interruptProcessTaskId[MAX_BOARDS] = {0,0,0,0,0,0,0,0};

#if defined (INT_PROCESS_USE_INT_LEVEL)
/* Used with the "interrupt connect" invocation, this var differs from the h/w int */
static CEI_INT32 interruptLevel[MAX_BOARDS] = {0,0,0,0,0,0,0,0};
#endif

/*===========================================================================*
 * INTERNAL ROUTINE:      L I T T L E _ E N D I A N _ 3 2
 *===========================================================================*
 *
 * FUNCTION:    Return the little-endian equivalent of the 4-byte integral
 *              argument (endian swap if host is big-endian).
 *
 * RETURN VAL:  The little-endian representation of the argument.
 *
 *===========================================================================*/
#define little_endian_32(longword) ( ((longword & 0x000000ff) << 24) | \
                                     ((longword & 0x0000ff00) << 8)  | \
                                     ((longword & 0xff0000)   >> 8)  | \
                                     ((longword & 0xff000000) >> 24) )

static SEM_ID ceiAccessMutex[MAX_BOARDS] = {         /* Device Access mutex  */
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL 
};


/*===========================================================================*
 * ENTRY POINT:            G E T   O P E N   E R R O R
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  openErrorString -- (output) reference to assign to error string
 * 
 * RETURN VAL:  ARS_NORMAL  -- Mutex was acquired successfully
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved
 *              when a device open error was encountered.
 *
 *===========================================================================*/
pCEI_CHAR cei_get_open_error (CEI_VOID)
{
  return deviceOpenErrorDesc;
}


/*========================================================================* 
 * Non-VxWorks Operating System specific functionality - here to resolve
 *========================================================================*/
CEI_INT32 cei_open_shared_lock (CEI_INT16 sharedLockId, pCEI_INT32 sharedLockState)
{
   return ARS_NORMAL;
}
CEI_INT32 cei_open_preload_lock (CEI_INT16 device)
{
   return ARS_NORMAL;
}

CEI_INT32 cei_map_shared_memory ( CEI_UCHAR    blockId,
                                  CEI_UINT32   blockSize,
                                  CEI_VOID  ** blockPtr,
                                  CEI_VOID  ** blockHandle,
                                  CEI_UINT32 * alreadyExists,
                                  CEI_UINT32 * isExclusive )
{
   return 1;
}

/*========================================================================* 
 * Operating System specific functionality
 *========================================================================*/
/*===========================================================================*
 * ENTRY POINT:            T I M E   G E T   T I M E
 *===========================================================================*
 *
 * FUNCTION:    Replicate the Windows O/S System Time Acquisition Method
 *
 * RETURN VAL:  32-bit 1 millisecond resolution system time
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the VxWorks clock_gettime invocation
 *              to obtain the current system timer value with a
 *              resolution of 1 millisecond, replicating the Windows O/S 
 *              system time acquisition method of the same name.
 *
 *===========================================================================*/
CEI_UINT32 timeGetTime()
{
   struct timespec ts;        /* Current time, use to get milliseconds */
   clock_gettime(CLOCK_REALTIME, &ts);
   return(CEI_UINT32)ts.tv_nsec/1000000.;
}

/*===========================================================================*
 * ENTRY POINT:        A C Q U I R E   T H R E A D   L O C K    
 *===========================================================================*
 *
 * FUNCTION:    Block on a semaphore previously reserved for the specific device 
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 7)
 * 
 * RETURN VAL:  ARS_NORMAL  -- semaphore was acquired successfully
 *              ARS_FAILURE -- Acquisition of the semaphore timed-out/failed
 *
 * GLOBALS:     ceiAccessMutex
 *
 * DESCRIPTION: This routine blocks execution, waiting for access to the 
 *              semaphore previously allocated for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_thread_lock (CEI_INT16 deviceIndex)
{
   CEI_INT32 status;
   if (ceiAccessMutex[deviceIndex] != NULL) {
      if (semTake(ceiAccessMutex[deviceIndex], WAIT_FOREVER) == OK)
         status = ARS_NORMAL;
      else
         status = ARS_FAILURE;
   }
   else
      status = ARS_FAILURE;
   
   return status;
}

/*===========================================================================*
 * ENTRY POINT:        R E L E A S E   T H R E A D   L O C K    
 *===========================================================================*
 *
 * FUNCTION:    Release a semaphore previously reserved/blocked by the 
 *              application/API for the specific device 
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 7)
 * 
 * RETURN VAL:  ARS_NORMAL  -- semaphore was acquired successfully
 *              ARS_FAILURE -- release of the semaphore timed-out/failed
 *
 * GLOBALS:     ceiAccessMutex
 *
 * DESCRIPTION: This routine releases the semaphore previously 
 *              reserved/blocked by the application/API.
 *
 *===========================================================================*/
CEI_INT32 cei_release_thread_lock (CEI_INT16 deviceIndex)
{
   CEI_INT32 status;
   if (ceiAccessMutex[deviceIndex] != NULL) {
      if (semGive(ceiAccessMutex[deviceIndex]) == OK)
         status = ARS_NORMAL;
      else
         status = ARS_FAILURE;
   }
   else
      status = ARS_FAILURE;
   
   return status;
}

/*===========================================================================*
 * ENTRY POINT:              D E L A Y   S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  numberOfSeconds -- (input) a floating point representation
 *                                         of the number (or fraction) of
 *                                         seconds to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the BSP-respective delay invocation
 *              as the basis for a delay defined in fractional seconds.
 *
 *===========================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds)
{
   CEI_UINT32 delay;

#if defined (DELAY_USE_SYS_US_DELAY)
   delay = (CEI_UINT32)((float)TIMER_CLOCK_TICKS_PER_SECOND * numberOfSeconds);
   sysUsDelay(delay);
#elif defined (DELAY_USE_SYS_MS_DELAY)
   delay = (CEI_UINT32)((float)TIMER_CLOCK_TICKS_PER_SECOND * numberOfSeconds);
   sysMsDelay(delay);
#else
   delay = (CEI_UINT32)(((float)sysClkRateGet() * numberOfSeconds) + 1);
   taskDelay((CEI_INT32)delay);
#endif

   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:              D E L A Y   M I L L I S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  numberOfMillieconds -- (input) the number of milliseconds
 *                                             to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the BSP-respective delay invocation
 *              as the basis to suspend thread execution for a specified
 *              number of milliseconds (resolution is platform dependent).
 *
 *===========================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds)
{
   CEI_UINT32 delay;
   CEI_UINT32 delayResolution;

#if defined (DELAY_USE_SYS_US_DELAY)
   delayResolution = (CEI_UINT32)(TIMER_CLOCK_TICKS_PER_SECOND/1000);
   if (delayResolution == 0)
      delayResolution = 1;
   delay = (CEI_UINT32)(delayResolution * numberOfMilliseconds);
   sysUsDelay(delay);
#elif defined (DELAY_USE_SYS_MS_DELAY)
   delayResolution = (CEI_UINT32)(TIMER_CLOCK_TICKS_PER_SECOND/1000);
   if (delayResolution == 0)
      delayResolution = 1;
   delay = (CEI_UINT32)(delayResolution * numberOfMilliseconds);
   sysMsDelay(delay);
#else
   delayResolution = (CEI_UINT32)(sysClkRateGet()/1000);
   if (delayResolution == 0)
      delayResolution = 1;
   delay = (CEI_UINT32)(delayResolution * numberOfMilliseconds);
   taskDelay((CEI_INT32)delay);
#endif

   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:       G E T   H O S T   S Y S T E M   T I M E
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host 32-bit 1ms resolution reference time value
 *              System Time in "seconds since midnight Jan 1 of this year"
 *
 * PARAMETERS:  hostTime -- (output) 32-bit 1 millisecond resolution time-tag
 *                                   seconds since midnight Jan 1 of this year
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current VxWorks System Time in
 *              "seconds since midnight Jan 1 of this year", then converts
 *              that value to a 32-bit 1 millisecond resolution value.
 *
 *===========================================================================*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime)
{
   time_t   now;             /* Current time, modify to get Time Reference Point. */
   struct   tm start;        /* Formatted "now" current time, modified for TRP.   */
   time_t   then;            /* Time Reference Point, Midnight or Jan 1.          */

   CEI_INT32 status;
   struct timespec ts;                     /* Current time, use to get milliseconds */
   
   CEI_UINT32 last_milli; /* This is the time we use to wait for time to change.    */
   TIME_TAG_TYPE diff;    /* This is the difference time we return in microseconds, */
                          /*  either time since Midnight or time since Jan 1.       */
   time(&now);                    /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now);      /* Local                                          */
   start.tm_sec  = 0;             /*   time                                         */
   start.tm_min  = 0;             /*     at                                         */
   start.tm_hour = 0;             /*       Jan 1                                    */
   start.tm_mday = 1;             /*         midnight (This is day 0)               */
   start.tm_mon  = 0;             /*           this                                 */
   start.tm_yday = 0;             /*             year                               */
   then = mktime(&start);         /* local Jan 1 to time in seconds since Jan 1     */
                                  /* We do IRIG time, which makes Jan 1 "day 0"     */

   status = clock_gettime(CLOCK_REALTIME, &ts);
   last_milli = (CEI_UINT32)ts.tv_nsec/1000000.;   /*  save millisecond */
   while (1) {                  /* Wait for milliseconds counter to change */
      clock_gettime(CLOCK_REALTIME, &ts);
      if ((CEI_UINT32)(ts.tv_nsec/1000000.) != last_milli)
         break;                 /* Break at exact time tick change */
   }
   diff = ts.tv_sec - then;     /* We only want the time since Midnight or Jan 1  */
   diff *= 1000000;             /* Number of Seconds converted to microseconds    */
   diff += (ts.tv_nsec/1000);   /* plus number of Milliseconds converted to       */
   *hostTime = (CEI_UINT32)diff;
   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:       G E T   H O S T   S Y S T E M   T I M E  X
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host 64-bit 1ms resolution reference time value
 *              System Time in "seconds since midnight Jan 1 of this year"
 *
 * PARAMETERS:  hostTime -- (output) 64-bit 1 millisecond resolution time-tag
 *                                   seconds since midnight Jan 1 of this year
 *                                   (value returned in microseconds)
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current VxWorks System Time in
 *              "seconds since midnight Jan 1 of this year", then converts
 *              that value to a 64-bit microseconds value (1ms resolution).
 *
 *===========================================================================*/
CEI_VOID cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime)
{
   CEI_UINT32 time;
   cei_get_host_system_time(&time);
   *hostTime = (TIME_TAG_TYPE)time;
}


/*===========================================================================*
 * ENTRY POINT:       G E T   I R I G   I N I T I A L   T I M E 
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host reference calendar time/date value, used
 *              for CEI-x30 IRIG timer initialization
 *
 * PARAMETERS:  time -- (output) 32-bit packed IRIG time value
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current Windows System Time in
 *              calendar time/date and formats it into an IRIG time value.
 *
 *===========================================================================*/
CEI_VOID cei_get_irig_initial_time (pCEI_UINT32 time)
{
   TIME_TAG_TYPE system_time;
   CEI_NATIVE_UINT seconds;
   CEI_NATIVE_UINT tens_of_seconds;
   CEI_NATIVE_UINT minutes;
   CEI_NATIVE_UINT tens_of_minutes;
   CEI_NATIVE_UINT hours;
   CEI_NATIVE_UINT tens_of_hours;
   CEI_NATIVE_UINT days;
   CEI_NATIVE_UINT tens_of_days;
   CEI_NATIVE_UINT hundreds_of_days;

   union
   {
      struct irigTimeFormat
           {
              CEI_NATIVE_UINT seconds          :4;
              CEI_NATIVE_UINT tens_of_seconds  :3;
              CEI_NATIVE_UINT minutes          :4;
              CEI_NATIVE_UINT tens_of_minutes  :3;
              CEI_NATIVE_UINT hours            :4;
              CEI_NATIVE_UINT tens_of_hours    :2;
              CEI_NATIVE_UINT days             :4;
              CEI_NATIVE_UINT tens_of_days     :4;
              CEI_NATIVE_UINT hundreds_of_days :2;
           } IRIG_TIME_FORMAT;
      CEI_UINT32 timeValue;
   } timeConvert;

   cei_get_host_system_time_x (&system_time);

   seconds = (CEI_UINT32)(system_time / (TIME_TAG_TYPE)1000000);
   if ( (system_time % (TIME_TAG_TYPE)1000000) > (TIME_TAG_TYPE)500000) {
      seconds++;
   }

   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;
   tens_of_seconds = seconds/10;
   seconds = seconds%10;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;
   tens_of_minutes = minutes/10;
   minutes = minutes%10;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;     
   tens_of_hours = hours/10;    
   hours = hours%10;    
   days++;                       /* account for no such thing as day 0 */
   hundreds_of_days = days/100;
   days = days%100;
   tens_of_days = days/10;
   days = days%10;

   timeConvert.IRIG_TIME_FORMAT.seconds          = seconds;
   timeConvert.IRIG_TIME_FORMAT.tens_of_seconds  = tens_of_seconds;
   timeConvert.IRIG_TIME_FORMAT.minutes          = minutes;
   timeConvert.IRIG_TIME_FORMAT.tens_of_minutes  = tens_of_minutes;
   timeConvert.IRIG_TIME_FORMAT.hours            = hours;
   timeConvert.IRIG_TIME_FORMAT.tens_of_hours    = tens_of_hours;
   timeConvert.IRIG_TIME_FORMAT.days             = days;
   timeConvert.IRIG_TIME_FORMAT.tens_of_days     = tens_of_days;
   timeConvert.IRIG_TIME_FORMAT.hundreds_of_days = hundreds_of_days;
   
   *time = timeConvert.timeValue;
}

#define PCI_MEMBASE_MASK        ~0xf     /* mask for memory base address */
#define PCI_BASE_ADDRESS_SPACE  0x01     /* 0 = memory, 1 = I/O */

/*===========================================================================*
 * ENTRY POINT:          A C Q U I R E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  deviceIndex   -- (input) device to reference (0 to 7)
 *              device        -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL    -- Session was opened successfully
 *              ARS_FAILURE   -- Driver Open Device failed
 *
 * GLOBALS:     deviceAttributes[]
 *              deviceMapped[]
 *              ceiAccessMutex[]
 *
 * DESCRIPTION: This routine searches for up to eight CEI-x30 devices, 
 *              mapping each CEI-x30 device found with the first execution 
 *              of this routine.  Subsequent executions will simply retrieve
 *              the mapped addresses for the specified device index, (with 
 *              the exception of some PPC SBCs such as Thales where the 
 *              "sysPciMemToLocalAdrs" translation will be applied to the 
 *              previously mapped addresses).
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_CHAR   device_name[40];
   CEI_INT32  status;
   UINT manufId;
   UINT deviceId;
   UINT irq;
   UINT bar2addr, localDeviceIndex, bar0addr;
   CEI_NATIVE_INT fd;

   localDeviceIndex = (CEI_UINT32)deviceIndex; 
   sprintf (device_name, "/gefesDev/%d", (CEI_NATIVE_INT)localDeviceIndex);
   fd = open (device_name, O_RDWR, 0);
   if (fd == -1) {
      status = ARS_FAILURE;
   }
   else { /* acquire the necessary device attributes for the API */
      status = (CEI_INT32)VxW_GetCommercialInfo ((UINT)deviceIndex, &bar0addr, &bar2addr, &manufId, &deviceId, &irq);

      deviceAttributes[deviceIndex].irq      = (CEI_INT32)irq;  
      deviceAttributes[deviceIndex].manufID  = (CEI_INT32)manufId;  
      deviceAttributes[deviceIndex].deviceID = (CEI_INT32)deviceId;  
      deviceAttributes[deviceIndex].irq   = (CEI_CHAR)irq;
      device->baseAddr  = (pCEI_INT32)bar0addr;
      device->devicePtr = (pCEI_INT32)bar2addr;
      devicePciAttributes[deviceIndex].manufID = (CEI_INT16)manufId;         
      devicePciAttributes[deviceIndex].deviceID = (CEI_INT16)deviceId;
      devicePciAttributes[deviceIndex].irq = (CEI_CHAR)irq;
      devicePciAttributes[deviceIndex].membase[0] = bar0addr;
      devicePciAttributes[deviceIndex].membase[2] = bar2addr;
      ceiAccessMutex[deviceIndex] = NULL;
      
      status = ARS_NORMAL;
   }
   close (fd);

   /* create thread protection on a first-in/first-out basis */
   ceiAccessMutex[deviceIndex] = semMCreate(SEM_Q_FIFO);

   return status;
}


/*===========================================================================*
 * ENTRY POINT:             R E L E A S E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 7)
 *              device      -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     ceiAccessMutex[8][3]
 *
 * DESCRIPTION: This routine releases a driver handle and unmaps the two PCI 
 *              memory regions previously mapped for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_release_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   /* release multi-thread protection */
   if (ceiAccessMutex[deviceIndex] != NULL) {
      semDelete( ceiAccessMutex[deviceIndex] );
   }
   deviceAttributes[deviceIndex].manufID = 0;
   deviceAttributes[deviceIndex].deviceID = 0;
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:   A C Q U I R E   D E V I C E   I D E N T I T Y
 *===========================================================================*
 *
 * FUNCTION:    Read the specified device PCI Configuration Space Identifiers
 *
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 7)
 *              vendorId    -- (output) ptr to storage of vendor PCI ID
 *              deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL  -- successful acquisition of the device ID's
 *              ARS_FAILURE -- low-level failed due to invalid device ID
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine acquires the specified device PCI Configuration  
 *              Space Vendor ID and Device ID fields via low-level function.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device_identity (
   CEI_INT32  deviceIndex,                /* device number to test (0-7)     */
   pCEI_INT32 vendorId,                   /* ptr to storage of vendor PCI ID */
   pCEI_INT32 deviceId                    /* ptr to storage of device PCI ID */
)
{
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   
   if (deviceAttributes[deviceIndex].manufID != 0) {
      *vendorId = deviceAttributes[deviceIndex].manufID;
      *deviceId = deviceAttributes[deviceIndex].deviceID;
   }
   else
      status = ARS_FAILURE;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:              R E A D   3 2   B I T
 *===========================================================================*
 *
 * FUNCTION:    Perform a 32-bit read from the specified device
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 7)
 *              address     -- (input) ptr to device address to read
 *              data        -- (input) ptr to location to store data
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit read access from one of the 
 *              device's two allocated PCI memory regions
 *
 *===========================================================================*/
CEI_INT32 cei_read_32bit (CEI_INT16 board, pCEI_UINT32 address, pCEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   CEI_UINT32 localData, modifiedData;
   localData = *address;
   modifiedData = little_endian_32(localData);
   *data = modifiedData;
#else
   *data = *address;
#endif          
   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:             W R I T E   3 2   B I T
 *===========================================================================*
 *
 * FUNCTION:    Perform a 32-bit write to the specified device
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 7)
 *              address     -- (input) ptr to device address to write
 *              data        -- (input) data to write
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit write access to one of the 
 *              device's two allocated PCI memory regions
 *
 *===========================================================================*/
CEI_INT32 cei_write_32bit (CEI_INT16 board, pCEI_UINT32 address, CEI_UINT32 data)
{
#ifdef NON_INTEL_WORD_ORDER
   *address = little_endian_32(data);
#else
   *address = data;
#endif          
   return ARS_NORMAL;
}


/*****************************************************************************
 *  L O C A L   I N T E R R U P T   P R O C E S S I N G   R O U T I N E S
 *****************************************************************************/

/*===========================================================================*
 * ENTRY POINT:   i n t e r r u p t S e r v i c e R o u t i n e
 *===========================================================================*
 *
 * DESCRIPTION: Invoked whenever an interrupt occurs, this routine clears 
 *              the hardware Interrupt Pending bit and invokes the ISR via 
 *              semaphore to launch "interruptProcessTask".
 *
 *===========================================================================*/
CEI_VOID interruptServiceRoutine (CEI_UINT32 isrDevInfo)
{
   VXW_INT_DEVICE_INFO_TYPE * deviceIntInfo;
   CEI_UINT32  csrRegisterValue;
   pCEI_UINT32 csrRegisterAddress; 
   CEI_INT32   board;

   deviceIntInfo = (VXW_INT_DEVICE_INFO_TYPE *)isrDevInfo;
   csrRegisterAddress = (pCEI_UINT32)deviceIntInfo->baseAddress;
   
   /* if this isn't our interrupt, ignore it */
   if (csrRegisterAddress != 0) {
      board = deviceIntInfo->device;   

      /* If set, clear the interrupt and invoke the ISR via semaphore usage */
      csrRegisterValue = *csrRegisterAddress;
   
      if (csrRegisterValue & CEI_HW_INTERRUPT_PENDING) {
         /* Clear the interrupt */        
         *csrRegisterAddress = csrRegisterValue;
   
         /* flag the interrupt process to invoke the ISR */   
         if (interruptSemaphoreAssignment[board] != NULL) {
            semGive (interruptSemaphoreAssignment[board]);
         }
      }
   }

   return;
}


/*===========================================================================*
 * ENTRY POINT:   i n t e r r u p t P r o c e s s T a s k
 *===========================================================================*
 *
 * DESCRIPTION: Spawned as a task from within "cei_interrupt_setup", this
 *              routine/task will block on the respective semaphore assigned 
 *              to this interrupt/device, then invoke the application/API
 *              assigned interrupt processing function when flagged to do so.
 *
 *===========================================================================*/
CEI_VOID interruptProcessTask ( CEI_INT32 deviceIndex,
                                CEI_INT32 b,
                                CEI_INT32 c,
                                CEI_INT32 d,
                                CEI_INT32 e,
                                CEI_INT32 f,
                                CEI_INT32 g,
                                CEI_INT32 h,
                                CEI_INT32 j,
                                CEI_INT32 k )
{
   while(1) {
      /* wait for an event to occur via "wait on semaphore" */ 
      semTake (interruptSemaphoreAssignment[deviceIndex], WAIT_FOREVER); 
      (x30InterruptFunctionPtr[deviceIndex])(deviceIndex, 0);
      interruptCounter[deviceIndex]++;
   }
}


CEI_VOID intStub (int deviceIndex, pCEI_UINT32 data)
{
   logMsg("No User Interrupt function Attached for deviceIndex %d\n",deviceIndex,0,0,0,0,0);
}


/*===========================================================================*
 * ENTRY POINT:     C E I _ I N T E R R U P T _ S E T U P
 *===========================================================================*
 *
 * FUNCTION:    Setup for interrupt processing.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 7)
 * 
 * RETURN VAL:  ARS_NORMAL   -- successful interrupt setup
 *              ARS_INVBOARD -- invalid device index
 *              ARS_FAILURE  -- VxWorks interrupt connect failure
 *                              VxWorks semaphore create failure
 *                              VxWorks task create failure
 *
 * GLOBALS:     
 *              interruptCounter
 *              interruptLevel
 *              interruptProcessTaskId
 *              interruptSemaphoreAssignment
 *              isrDeviceInfo
 *              localIRQ
 *              devicePciAttributes
 *              pmcInterruptLevel
 *
 * DESCRIPTION: Setup the hardware interrupt service routine thread - task
 *              activated via semaphore.
 *
 *===========================================================================*/
CEI_INT32 cei_interrupt_setup (CEI_UINT32 deviceIndex)
{
   CEI_INT32 status = ARS_NORMAL;                  /* Function return status */
   CEI_INT32 irqValue;
   CEI_CHAR  intProcessName[16];

#if defined (INT_PROCESS_USE_INT_LEVEL)
   VOIDFUNCPTR * vector;
#endif
   if (deviceIndex > MAX_BOARDS)
      status = ARS_INVBOARD;
   else {
      interruptCounter[deviceIndex] = 0;
   
      /* create the semaphore used by the interrupt processing task */ 
      interruptSemaphoreAssignment[deviceIndex] = semBCreate (SEM_Q_FIFO, SEM_EMPTY);
      if (interruptSemaphoreAssignment[deviceIndex] == NULL)
         status = ARS_FAILURE;
      else {
         sprintf (intProcessName,"t_cei_isr_%d",(int)deviceIndex);
         /* with the semaphore created, now create the interrupt processing task */      
         interruptProcessTaskId[deviceIndex] = taskSpawn ( intProcessName,
                                                           16 + deviceIndex,
                                                           0, 
                                                           20000,
                                                           (FUNCPTR)interruptProcessTask, 
                                                           deviceIndex, 
                                                           0,0,0,0,0,0,0,0,0 );
      
         if (interruptProcessTaskId[deviceIndex] == ERROR)
            status = ARS_FAILURE;
      }
   }
   
   if (status == ARS_NORMAL) {
      irqValue = (CEI_INT32)devicePciAttributes[deviceIndex].irq;
      localIRQ[deviceIndex] = irqValue;
   
      isrDeviceInfo[deviceIndex].device = deviceIndex;
      isrDeviceInfo[deviceIndex].baseAddress = devicePciAttributes[deviceIndex].membase[2];

      /******************************* NOTE ********************************
       * Check the BSP reference manual for the correct interrupt value.
       * some boards use interrupt level or offsets to the level while 
       * others use the assigned IRQ or offsets to the IRQ value.
       * The interrupt level for the respective PMC slot should be defined
       * for the BSP in the pmcInterruptLevel[] array at the top of this file.
       *                                                                   
       * Make sure you are using the correct interrupt level otehrwise the 
       * interrupt will not get processed correctly or the system will hang.
       **********************************************************************/
#if defined (INT_PROCESS_USE_INT_LEVEL)

   #if defined (INT_PROCESS_INT_LEVEL_IS_IRQ_VALUE)
      interruptLevel[deviceIndex] = irqValue + IRQ_OFFSET;
   #else
      interruptLevel[deviceIndex] = pmcInterruptLevel[deviceIndex];
   #endif 
      
      status = intDisable (interruptLevel[deviceIndex]);

      if (status == ERROR)
         status = ARS_FAILURE;
      else {
         vector = INUM_TO_IVEC(INT_VEC_IRQ0 + interruptLevel[deviceIndex]);

         if ( intConnect (INUM_TO_IVEC(INT_VEC_IRQ0 + interruptLevel[deviceIndex]),
                          (VOIDFUNCPTR)interruptServiceRoutine,
                          (unsigned int)&isrDeviceInfo[deviceIndex]) )
            status = ARS_FAILURE;
         else {
            status = intEnable (interruptLevel[deviceIndex]);
            if (status == ERROR)
               status = ARS_FAILURE;
         }
      }

#elif defined (INT_PROCESS_USE_IRQ_VALUE)
      
      status = intDisable (irqValue);

      if (status == ERROR)
         status = ARS_FAILURE;
      else {
         if ( pciIntConnect (INUM_TO_IVEC((unsigned int)irqValue), 
                             (VOIDFUNCPTR)interruptServiceRoutine, 
                             (unsigned int)&isrDeviceInfo[deviceIndex]) == ERROR )
            status = ARS_FAILURE;
         else {
            status = intEnable(irqValue);
            if (status == ERROR)
               status = ARS_FAILURE;
         }
      }

#elif defined (VXW_PCI_X86)   
        
      if ( sysIntDisablePIC (irqValue) == ERROR )
         status = ARS_FAILURE;
      else {
   #if defined (USE_INT_NUM_GET)
         if ( pciIntConnect (INUM_TO_IVEC(INT_NUM_GET(irqValue)),
                             (VOIDFUNCPTR)interruptServiceRoutine,
                             (unsigned int)&isrDeviceInfo[deviceIndex]) == ERROR )
   #else
         if ( pciIntConnect (INUM_TO_IVEC(irqValue + INT_NUM_IRQ0 + IRQ_OFFSET),
                             (VOIDFUNCPTR)interruptServiceRoutine,
                             (unsigned int)&isrDeviceInfo[deviceIndex]) == ERROR )
   #endif
            status = ARS_FAILURE;
         else {
            if ( sysIntEnablePIC(irqValue) == ERROR )
               status = ARS_FAILURE;
         }
      }
#else
      #error Error defining board type.   
#endif

   }

   /* if an error occurred, remove anything created */
   if (status == ARS_FAILURE) {
      semDelete(interruptSemaphoreAssignment[deviceIndex]);
      interruptSemaphoreAssignment[deviceIndex] = 0;
      taskDelete (interruptProcessTaskId[deviceIndex]);
      interruptProcessTaskId[deviceIndex] = 0;
   }
   else
      status = ARS_NORMAL;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:         C E I _ I N I T _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Configures an interrupt and associates an ISR with it
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 15)
 *              pCEI_VOID   isr               -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 *              CEI_INT32   memRegion         -- (input)  not used
 *              CEI_UINT32  disOffset         -- (input)  not used
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 isrEnabled        -- (output) thread handle if successfully defined
 * 
 * RETURN VAL:  ARS_NORMAL        -- Interrupt was initialized successfully
 *              ARS_FAILURE       -- Interrupt was not initialized successfully
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device.
 *
 *===========================================================================*/
CEI_INT32 cei_init_interrupt (
   CEI_INT32   deviceIndex,
   pCEI_VOID   isr, 
   CEI_INT32   priority,
   pCEI_UINT32 deviceBaseAddress,
   CEI_INT32   memRegion,
   CEI_UINT32  disOffset,
   CEI_UINT32  interruptBit,
   pCEI_UINT32 isrHandle
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the low-level global virtual base address array     */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;

   /* setup the interrupt task and semaphore, then connect to the h/w int */
   status = cei_interrupt_setup ((CEI_UINT32)deviceIndex);

   /* Set either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, High to pass local bus interrupts
    */
   if ((status == ARS_NORMAL) && (interruptBit != 0)) {
      status = cei_read_32bit (deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg |= interruptBit;
      status = cei_write_32bit (deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }
   *isrHandle = 1;


   return status;
}


/*===========================================================================*
 * ENTRY POINT:     C E I _ T E R M I N A T E _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to 
 *              an ISR
 * 
 * PARAMETERS:  deviceIndex  -- (input)  device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *              ARS_FAILURE  -- Interrupt was not terminated successfully
 *
 * GLOBALS:     None
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function
 * 
 *===========================================================================*/
CEI_INT32 cei_terminate_interrupt
(
    CEI_INT32   deviceIndex, 
    CEI_UINT32  interruptBit,
    pCEI_UINT32 deviceBaseAddress
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   
   /* Reset either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, Low to disable local bus interrupts
    */
   if (interruptBit != 0) {
      status = cei_read_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg &= ~interruptBit;
      status = cei_write_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }
   /* next, delete the interrupt processing task */  
   if (interruptProcessTaskId[deviceIndex] != 0) {
      taskDelete (interruptProcessTaskId[deviceIndex]);
	}

   /* next, delete the interrupt "informer" semaphore */
   if (interruptSemaphoreAssignment[deviceIndex] != 0) {
      semDelete(interruptSemaphoreAssignment[deviceIndex]);
	}
   
   interruptProcessTaskId[deviceIndex] = 0;
   interruptSemaphoreAssignment[deviceIndex] = 0;
   
   /* finally, disconnect the h/w interrupt */

#if defined (VXW_PCI_X86)
   if ( sysIntDisablePIC (localIRQ[deviceIndex]) == ERROR )
      status = ARS_FAILURE;
#elif defined (INT_PROCESS_USE_IRQ_VALUE)
   status = intDisable (localIRQ[deviceIndex]);
   if (status == ERROR)
      status = ARS_FAILURE;
#elif defined (INT_PROCESS_USE_INT_LEVEL)

   #if defined (INT_TERM_INCL_CTL_DISCONNECT)
   intCtrlDisconnect(INUM_TO_IVEC(INT_VEC_IRQ0 + interruptLevel[deviceIndex]));
   #endif
   status = intDisable (interruptLevel[deviceIndex]);
   if (status == ERROR)
      status = ARS_FAILURE;
#endif
   
   return status;
}
