/*===========================================================================*
 * FILE:                     C D E V _ W I N . C
 *===========================================================================*
 *
 * COPYRIGHT (C) 2004-2011 BY
 *          GE INTELLIGENT PLATFORMS, INC. SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:    Low-level Application Programmer's Interface for a CEI-x30
 *              ARINC device operating with all Windows Operating Systems
 *
 *    Internal CEI-x30 routines implemented herein:
 *
 *    cei_acquire_device          - opens a session with the specified x30 device
 *    cei_acquire_device_identity - supplies the PCI configuration space ID values
 *    cei_acquire_thread_lock     - acquires sempahore/mutex lock
 *    cei_delay_milliseconds      - o/s specific delay utility in milliseconds
 *    cei_delay_seconds           - o/s specific delay utility in fractional seconds
 *    cei_get_host_system_time_x  - acquires host system time-of-year 64-bit uSec     
 *    cei_get_host_system_time    - acquires host system time-of-year 32-bit mSec     
 *    cei_get_irig_initial_time   - acquires host system time-of-year IRIG format
 *    cei_map_shared_memory       - maps memory region for sharing globals across processes
 *    cei_open_shared_lock        - creates a shared semaphore across processes
 *    cei_release_device          - closes a session with the specified x30 device
 *    cei_release_thread_lock     - releases sempahore/mutex lock     
 *    cei_read_32bit              - read a 32-bit value from an x30 device
 *    cei_write_32bit             - write a 32-bit value to an x30 device
 *    cei_init_interrupt          - configures an interrupt and ISR
 *    cei_terminate_interrupt     - terminate an interrupt
 *          
 *    internal routines previously implemented in CEI_LL library:      
 *
 *    WaitForInterrupts           - isr thread awoken by device driver generated event
 *    ceill_initInterrupt         - associates the isr with the above thread, sets up interrupt for CEI_Install
 *    ceill_terminateInterrupt    - disconnects the isr from the driver
 *
 *---------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ------------------------------------------------------------
 * 11/06/03  BDW  Created.
 * 04/11/06  BDW  Removed utility routines cei_read_plx_32bit & cei_write_plx_32bit
 * 08/24/06  BDW  Incorporated CEI_TYPES.H data type definitions in 
 *                   support of 64-bit Linux operating systems. 
 *                Swapped the BAR reference order by the device structure 
 *                   members baseAddr and devicePtr
 * 02/22/07  BDW  Added cei_acquire_device_identity() to supply the PCI 
 *                   Manufacturer and Device IDs for x30 products 
 * 10/29/07  BDW  Added routine for open error handling, cei_get_open_error
 * 02/18/08  BDW  Removed the boardInitialized array, relocated as a structure
 *                   member in the data structure array "deviceGlobal"
 *                Added new routines for multi-process support: 
 *                   cei_open_shared_lock and cei_map_shared_memory
 * 03/10/08  BDW  Removed the declaration of the following global arrays, 
 *                relocated to CDEV_GLB.H:
 *                   bypassWrapTest
 *                   irigDacThreshold
 *                   a717syncWords
 *                Removed declaration of basicConfiguration, it was unused.
 *                Added reference to device[] structure as a parameter to
 *                   cei_acquire_device & cei_release_device so they no longer access 
 *                   the global data structure directly.
 * 04/15/08  BDW  Added support for initiation and termination of h/w interrupts
 *                via routines cei_init_interrupt and cei_terminate_interrupt.
 * 06/16/09  BDW  Added the function cei_open_preload_lock for multiprocess support.
 * 07/01/09  BDW  Added new function cei_delay_milliseconds (o/s thread suspend).
 * 10/28/10  BDW  Drastic alterations to switch from CEI_LL library to CEI_Install
 *                   for all driver interface operations.
 *                Modified the parameter list for cei_init_interrupt to account for
 *                   interrupt setup requirements in the Windows-only CEI_Install library
 * 01/05/11  BDW  Corrected successful interrupt initialization return status 
 *                   to force the value ARS_NORMAL from ceill_initInterrupt
 *
 *===========================================================================*/
/* include installation library dynamic link support */
#define CEI_INSTALL_DYNAMIC_LINK_SUPPORT
#include "cei_install.h"
#include "cdev_glb.h"
#include <mmsystem.h>
#include <sys/types.h>
#include <sys/timeb.h>

/*========================================================================* 
 * Local data and defines
 *========================================================================*/
#define MAX_DEV_OPEN_ERROR_STR_LEN 512

static CEI_UINT64 virtualBaseAddress[MAX_BOARDS][3];
static CEI_DEV_HANDLE ceiAccessLock[MAX_BOARDS] = {   /* Device Access mutex  */
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
/* storage for Windows driver low-level board mapping error messages      */
static CEI_INSTALL_ERR deviceOpenErrorDesc = "No error description available";

/* global: dev_handle                         */
/* usage:  Stores handle to each open device. */
CEI_DEV_HANDLE dev_handle[MAX_BOARDS];

/* global: cei_dll                                             */
/* usage:  Pointer to CEI-Install library dynamic link module. */
CEI_INSTALL_MODULE *cei_dll = NULL;

/* Structure of information passed to the hardware interrupt thread */
typedef struct _THREAD_DATA_STRUCT {
   CEI_DEV_HANDLE handle;    /* Handle to current device.                      */
   CEI_DEV_HANDLE hIntThrd;  /* Handle to the interrupt thread                 */
   pCEI_CHAR HWaddress;      /* Save the device BAR0 base address.             */
   pCEI_CHAR configAddr;     /* Save the device BAR2 address.                  */
   CEI_INT32 fExitThread;    /* Set TRUE to cause the threads to exit.         */
   CEI_INT32 DeviceIntOpen;  /* WinRT device Interrupt open count.             */
   CEI_INT32 board;          /* Board number being processed.                  */
   void (_stdcall * interrupt_function)(CEI_INT32 deviceIndex); /* Delayed ISR */
   CEI_UINT32 vendorID;      /* Vendor ID if PCI card                          */
   CEI_UINT32 deviceID;      /* Device ID if PCI card                          */
} THREAD_DATA_STRUCT, *PTHREAD_DATA_STRUCT;

THREAD_DATA_STRUCT pData[MAX_BOARDS];


/*========================================================================* 
 * Windows Operating System specific functionality
 *========================================================================*/

/*===========================================================================*
 * INTERNAL ROUTINE:               e r r o r _ b o x 
 *===========================================================================*
 *
 * FUNCTION:    Display error description.
 *
 * RETURN VAL:  none
 *
 * DESCRIPTION: Used to display a low-level error message to the user when 
 *              propagation of the error message to the caller is impractical.
 *              The message text is displayed using the MessageBox routine.
 *
 *===========================================================================*/
CEI_VOID error_box(const CEI_CHAR *format_str,...)
{
    /* receives message string */
    CEI_CHAR text[1024];

    /* store error description */
    va_list argp;
    va_start(argp,format_str);
    _vsnprintf(text,sizeof(text) - 1,format_str,argp);
    text[sizeof(text)-1] = 0;

    /* display a message box  */
    MessageBox(NULL,text,"CEI-x30-SW Error",MB_SYSTEMMODAL|MB_OK|MB_ICONERROR);
}


/*===========================================================================*
 *  PROCEDURE - WaitForInterrupts
 *===========================================================================*
 *
 *  FUNCTION
 *     This thread waits on an interrupt event from WinRT, calling the 
 *     API-supplied interrupt service routine when an interrupt occurs.
 *
 *  PARAMETERS
 *     Param - THREAD_DATA_STRUCT struct pointer.
 *
 *   Notes:  To kill this thread, signal the hCloseEvent event.
 *
 *  RETURNS
 *     Thread/Interrupt Blocking fail status.
 *===========================================================================*/
static DWORD WINAPI WaitForInterrupts (LPVOID Param) 
{
   PTHREAD_DATA_STRUCT pData;
   CEI_INT32           board;
   CEI_UINT64          wait_on_regs = 0; /* bitfield specifying which int status register(s) to wait on */
   CEI_UINT32          wait_result;      /* receives wait operation result */
   CEI_UINT64          int_on_regs;      /* receives bitfield indicating which int status register(s) generated the int */
   CEI_INSTALL_ERR     cei_err;         /* receives error description upon failure */

   pData = (PTHREAD_DATA_STRUCT)Param;  /* Convert the input parameter. */
   board = pData->board;

   /* only one int status register in this example, wait on it */
   wait_on_regs = 1;

   /* handle interrupts until the process calls LL_UnhookInterrupt 22 */
   while (TRUE) {
      /* wait for a pending interrupt */
      if (cei_dll->LL_WaitForInterrupt (pData->handle, wait_on_regs, &wait_result, &int_on_regs, NULL, cei_err) == FALSE)
         return ARS_INT_ISR;

      /* bail out if the process called LL_UnhookInterrupt */
      if (wait_result == CEI_WIR_INT_DISABLED)
         break;
      else if (wait_result != CEI_WIR_INT_OCCURRED)
         return ARS_INT_ISR;

      /* an interrupt has occurred, handle it  */
      pData->interrupt_function(board);
   }
   
   /* signal the main thread this thread is done */
   ExitThread(0);
   return 0;
}


/*===========================================================================*
 * INTERNAL ROUTINE: c e i l l _ i n i t I n t e r r u p t
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to associate an ISR with an interrupt
 * 
 * PARAMETERS:  CEI_INT32 deviceId     -- Device Number to map
 *              CEI_INT32 _stdcall isr -- Address of interrupt function to be called
 *              CEI_INT32 priority     -- Interrupt Service Routine thread priority
 *              CEI_INT32 memRegion    -- PCI BAR region/index for interrupt clear bit (0 - BAR0 or 1 - BAR2)
 *              CEI_UINT32 disOffset   -- PCI Interrupt Disable register (PLX reg offset, 0 for no PLX interface)
 *              pCEI_UINT32 isrHandle  -- Thread successfully defined state
 * 
 * RETURN VAL:  CEI_INT32 STATUS -        ARS_DRIVERFAIL
 *                                        ARS_NO_INT_SUPPORT
 *                                        ARS_NO_INT_ENABLED
 *                                        ARS_INT_ISR
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: Connects the user-specified interrupt function to the  
 *              interrupt assigned to the respective device.
 * 
 *===========================================================================*/
CEI_INT32 ceill_initInterrupt 
(
   CEI_INT32   deviceId,
   CEI_INT32   (_stdcall *isr)(CEI_INT32), 
   CEI_INT32   priority,
   CEI_INT32   memRegion,
   CEI_UINT32  disOffset,
   pCEI_UINT32 isrHandle
)
{
   CEI_INT32 status = ARS_NORMAL;
   DWORD     ThreadID;                  /* Unique id of the interrupt thread */
   HANDLE    hThread;                   /* Handle to the interrupt thread    */

   CEI_UINT32 ints_supported; /* receives hardware interrupt support status    */
   CEI_INT_CONTROL_V1 my_ic;  /* initialize to configure interrupt management  */
   CEI_UINT32 preexisted;     /* receives true if another process is already using the devices interrupt */
   CEI_INSTALL_ERR cei_err;   /* receives error description upon failure */
   CEI_DEV_HANDLE devHandle = dev_handle[deviceId];

   /* query hardware interrupt support capability */
   if (cei_dll->LL_GetCapability(devHandle, CEI_CAP_HW_INTS, &ints_supported, cei_err) == FALSE) {
      error_box("Failed to acquire capabilities for Device '%ld' (details: %s).", deviceId, cei_err);
      status = ARS_DRIVERFAIL;
   }
   
   /* bail out if hardware interrupts are not supported */
   else if (!ints_supported) {
      error_box("Hardware interrupts are not supported for Device '%ld'.", deviceId);
      status = ARS_NO_INT_SUPPORT;
   }
   else {
      /* initialize the interrupt control structure */
      if (cei_dll->LL_InitInterruptControlStruct(devHandle, CEI_INT_CONTROL_VER_V1, sizeof(my_ic), &my_ic, cei_err) == FALSE) {
         error_box("Failed to initialize an interrupt for Device '%ld' (details: %s).", deviceId, cei_err);
         status = ARS_NO_INT_ENABLED;
      }
   }

   if (status == ARS_NORMAL) {
      /* set main interrupt control parameters */
      my_ic.chk_cap_reg = FALSE; /* no capability register checking           */
      my_ic.add_chk_cnt = 0;     /* no special ISA processing                 */
      my_ic.num_status_reg = 1;  /* device uses a single int status register  */
      if (disOffset) {
         my_ic.num_disable_reg = 1; /* device uses a single int disable register */
      }
      else {
         my_ic.num_disable_reg = 0; /* device has no int disable register */
      }
   
      /* All devices having h/w interrupt support via CEI_LL have a single 4-byte global enable */
      /* register at byte offset 0 in the memory region specified in the priority parameter. */
      /* Bit 4 is non-zero upon interrupt, no chk_reg_opt option flags needed */
      my_ic.status_reg[0].chk_reg.rgn_usr_addr = virtualBaseAddress[deviceId][memRegion];
      my_ic.status_reg[0].chk_reg.offset_bytes = 0;
      my_ic.status_reg[0].chk_reg.len_bytes = 4;
      my_ic.status_reg[0].chk_reg_opt = CEI_XRO_NONE;
      my_ic.status_reg[0].chk_reg_mask = 0x00000010;
   
      /* upon interrupt, bitwise AND the same register with 0xFFFFFFFF */
      my_ic.status_reg[0].clr_reg_opt = CEI_CRO_USE_CHK_REG | CEI_CRO_AND;
      my_ic.status_reg[0].clr_reg_mask = 0xFFFFFFFF;
   
      /* the device in this example has a single 4-byte interrupt disable  */
      /* register at byte offset 0x4C in the first memory region, bitwise AND */
      /* the register with 0xFFFFFFFF to disable device interrupts  */
      if (disOffset) {
         my_ic.disable_reg[0].dis_reg.rgn_usr_addr = virtualBaseAddress[deviceId][0];
         my_ic.disable_reg[0].dis_reg.offset_bytes = disOffset;
         my_ic.disable_reg[0].dis_reg.len_bytes = 4;
         my_ic.disable_reg[0].dis_reg_opt = CEI_DRO_AND;
         my_ic.disable_reg[0].dis_reg_mask = 0xFFFFFFFF;
      }
      /* install the interrupt handler */
      if (cei_dll->LL_HookInterrupt(devHandle, &my_ic, &preexisted, cei_err) == FALSE) {
         error_box("Failed to install the interrupt handler for Device '%ld' (details: %s).", deviceId, cei_err);
         status = ARS_INT_ISR;
      }
      else {
         pData[deviceId].board = deviceId;
         pData[deviceId].handle = devHandle;
         pData[deviceId].interrupt_function = isr;
      
         /* Create the interrupt thread suspended so we can set its priority */
         hThread = CreateThread ( NULL,
                                  8192, 
                                  (void *)WaitForInterrupts, 
                                  (void *)&pData[deviceId],
                                  CREATE_SUSPENDED, 
                                  &ThreadID );
      
         /* Make sure the thread was created properly */
         if ( hThread != NULL ) {
            pData[deviceId].hIntThrd = hThread;
         
            /* Set the thread priority to user-specified value */
            status = SetThreadPriority(hThread, priority);
      
            /* Everything's set, let it run.  It will execute "startup". */
            status = ResumeThread(hThread);
            status = ARS_NORMAL;
            *isrHandle = TRUE;
         }
         else {
            error_box("Failed to create the interrupt handler thread for Device '%ld'.", deviceId);
            status = ARS_INT_ISR;
            *isrHandle = FALSE;
         }
      }
   }
   return status;
}


/*===========================================================================*
 * INTERNAL ROUTINE: c e i l l _ t e r m i n a t e I n t e r r u p t
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to 
 *              an ISR
 * 
 * PARAMETERS:  CEI_INT32 hwDevId        -- Device Number to map
 * 
 * RETURN VAL:  fixed LL_OK 
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: terminates the specified device interrupt function
 * 
 *===========================================================================*/
CEI_INT32 ceill_terminateInterrupt (CEI_DEV_HANDLE devHandle)
{
   CEI_UINT32 persists;       /* receives true if another process is still using the devices interrupt */
   CEI_INSTALL_ERR cei_err;   /* receives error description upon failure */
   CEI_INT32 status;

   /* indicate that this process is done monitoring the device interrupt - */
   /* any pending LL_WaitForInterrupt calls associated with threads */
   /* belonging to this process return with the result parameter set to */
   /* CEI_WIR_INT_DISABLED - if this is the last process to unhook, device */
   /* interrupts are disabled and the interrupt handler is uninstalled */
   if (cei_dll->LL_UnhookInterrupt(devHandle, &persists, cei_err) == FALSE) {
      status = ARS_INT_ISR;
   }
   else {
      status = ARS_NORMAL;
   }
   return status;   
}


/*===========================================================================*
 * ENTRY POINT:            G E T   O P E N   E R R O R
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  None
 * 
 * RETURN VAL:  Reference to assign to error string
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved
 *              when a device open error was encountered.
 *
 *===========================================================================*/
pCEI_CHAR cei_get_open_error (CEI_VOID)
{
  return deviceOpenErrorDesc;
}


/*===========================================================================*
 * ENTRY POINT:             O P E N   S H A R E D   L O C K
 *===========================================================================*
 *
 * FUNCTION:    Opens a shared (global) mutex 
 *
 * PARAMETERS:  sharedLockId -- (input) index of the specific semaphore to open
 *              sharedLockState -- (input) location to store semaphore handle
 *
 * RETURN VAL:  ARS_NORMAL  -- semaphore was opened successfully
 *              ARS_INVARG  -- semaphore id too large
 *              ARS_BOARD_MUTEX -- failed to create/open the semaphore
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Opens the specified shared (global) semaphore for use in
 *              multiple process applications
 *
 *===========================================================================*/
CEI_INT32 cei_open_shared_lock (CEI_INT16 sharedLockId, pCEI_INT32 sharedLockState)
{
   CHAR mutexName[32];             /* unique mutex name                   */
   CEI_UINT32 preexisted_ignore;   /* whether the lock already existed    */
   CEI_INSTALL_ERR cei_err;        /* receives installation library error */
   CEI_INT32 status = ARS_NORMAL;  /* function return status              */
   
   /* build unique mutex name */
   if (cei_dll != NULL) {
      sprintf (mutexName, "ceix30_shared_lock_%d", sharedLockId);
   
      cei_dll->OpenGlobalNamedObject (CEI_GSO_MUTEX, FALSE, mutexName, &preexisted_ignore, NULL, (HANDLE*)&ceiAccessLock[sharedLockId], cei_err);
      if (ceiAccessLock[sharedLockId] == NULL) {
         status = ARS_BOARD_MUTEX;
      }
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:   O P E N   P R E L O A D   L O C K
 *===========================================================================*
 *
 * FUNCTION:    Requests the initiation of a shared (global) semaphore 
 *              for multiprocess application usage
 *
 * PARAMETERS:  device -- (input) index of the device to create lock based on
 *
 * RETURN VAL:  ARS_NORMAL  -- semaphore was opened successfully
 *              ARS_FAILURE -- failed to create/open the semaphore
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Create the specified shared (global) semaphore for use in
 *              multiple process applications
 *
 *===========================================================================*/
CEI_INT32 cei_open_preload_lock (CEI_INT16 device)
{
   CEI_INT32 status;
   CEI_INT32 temp;
   CEI_INSTALL_ERR cei_err;        /* receives installation library error     */

      /* dynamically link to the installation library and open the device */
   cei_dll = CEI_DLL_GetDynamicLink (0, cei_err);
   if (cei_dll == NULL) {
      status = ARS_WINRTFAIL;
   }
   else {
      status = cei_open_shared_lock(device, &temp);
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:        A C Q U I R E   T H R E A D   L O C K    
 *===========================================================================*
 *
 * FUNCTION:    Block on a Mutex previously reserved for the specific device 
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL      -- Mutex was acquired successfully
 *              ARS_BOARD_MUTEX -- Acquisition of the Board Lock timed-out/failed
 *
 * GLOBALS:     ceiAccessLock
 *
 * DESCRIPTION: This routine blocks execution, waiting for access to the 
 *              Windows Mutex previously allocated for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_thread_lock (CEI_INT16 deviceIndex)
{
   DWORD status;
   status = WaitForSingleObject(ceiAccessLock[deviceIndex], INFINITE);
   if (status != WAIT_FAILED)
      return ARS_NORMAL;
   else
      return ARS_BOARD_MUTEX;
}


/*===========================================================================*
 * ENTRY POINT:        R E L E A S E   T H R E A D   L O C K    
 *===========================================================================*
 *
 * FUNCTION:    Release a Mutex previously reserved/blocked by the 
 *              application/API for the specific device 
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL      -- Mutex was acquired successfully
 *              ARS_BOARD_MUTEX -- Release of the Board Lock timed-out/failed
 *
 * GLOBALS:     ceiAccessLock
 *
 * DESCRIPTION: This routine releases the Windows Mutex previously 
 *              reserved/blocked by the application/API.
 *
 *===========================================================================*/
CEI_INT32 cei_release_thread_lock (CEI_INT16 deviceIndex)
{
   if (ReleaseMutex(ceiAccessLock[deviceIndex]))
      return ARS_NORMAL;
   else
      return ARS_BOARD_MUTEX;
}


/*===========================================================================*
 * ENTRY POINT:              D E L A Y   M I L L I S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  numberOfMillieconds -- (input) the number of milliseconds
 *                                             to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the BSP-respective delay invocation
 *              as the basis to suspend thread execution for a specified
 *              number of milliseconds (resolution is platform dependent).
 *
 *===========================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds)
{
   Sleep(numberOfMilliseconds);
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:              D E L A Y   S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  numberOfSeconds -- (input) a floating point representation
 *                                         of the number (or fraction) of
 *                                         seconds to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the Windows timeGetTime invocation
 *              as the basis for a delay-polling process having a typical
 *              resolution of 1 millisecond.
 *
 *===========================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds)
{
   CEI_UINT32 start, now, delay;

   delay = (CEI_UINT32)((float)TIMER_CLOCK_TICKS_PER_SECOND * numberOfSeconds);
   start = timeGetTime();
   now = start;
   while ((now - start) < delay) {
      now = timeGetTime();
   }
   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:       G E T   H O S T   S Y S T E M   T I M E  X
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host 64-bit 1us resolution reference time value
 *              System Time in "seconds since midnight Jan 1 of this year"
 *
 * PARAMETERS:  hostTime -- (output) 64-bit 20 nanosecond resolution time-tag
 *                                   seconds since midnight Jan 1 of this year
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current Windows System Time in
 *              "seconds since midnight Jan 1 of this year", then converts
 *              that value to a 64-bit 1 microsecond resolution value.
 *
 *===========================================================================*/
CEI_VOID cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime)
{
   time_t now;           /* Current time, modify to get Time Reference Point. */
   struct   tm start;    /* Formatted "now" current time, modified for TRP.   */
   time_t   then;        /* Time Reference Point, Midnight or Jan 1.          */
   struct   timeb ts;    /* Current time, use to get milliseconds.            */
   CEI_UINT32 last_milli;/* This is time we use to wait for time to change.   */
   TIME_TAG_TYPE diff;   /* This is difference time we return in microseconds */

   time(&now);               /* Time in seconds since midnight Jan 1,1970 GMT */
   start = *localtime(&now); /* Local                                         */
   start.tm_sec  = 0;        /*   time                                        */
   start.tm_min  = 0;        /*     at                                        */
   start.tm_hour = 0;        /*       Jan 1                                   */
   start.tm_mday = 1;        /*         midnight (This is day 0)              */
   start.tm_mon  = 0;        /*           this                                */
   start.tm_yday = 0;        /*             year                              */
   then = mktime(&start);    /* local Jan 1 to time in seconds since Jan 1    */
                             /* We do IRIG time, which makes Jan 1 "day 0"    */ 

   ftime(&ts);                             /* Current time                    */
   last_milli = ts.millitm;                /*  save milliseconds              */
   while (TRUE) {                          /* Wait for milliseconds to change */
      ftime(&ts);                          /* Current time                    */
      if (ts.millitm != last_milli)
         break;                            /* Break at exact time tick change */
   }
   diff = ts.time-then;       /* We only want the time since Midnight on Jan 1*/
   diff *= 1000000;           /* Number of Seconds converted to microseconds  */
   diff += ts.millitm * 1000; /* plus number of Milliseconds converted to     */
   *hostTime = diff;          /* microseconds.  Passed back to the caller.    */
}

/*===========================================================================*
 * ENTRY POINT:       G E T   H O S T   S Y S T E M   T I M E 
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host 32-bit 1ms reference time value, supplied
 *              as the Windows System Time
 *
 * PARAMETERS:  hostTime -- (output) 32-bit 1 millisecond resolution time
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current Windows System Time in
 *              milliseconds
 *
 *===========================================================================*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime)
{
   *hostTime = timeGetTime();
   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:       G E T   I R I G   I N I T I A L   T I M E 
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host reference calendar time/date value, used
 *              for CEI-x30 IRIG timer initialization
 *
 * PARAMETERS:  time -- (output) 32-bit packed IRIG time value
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current Windows System Time in
 *              calendar time/date and formats it into an IRIG time value.
 *
 *===========================================================================*/
CEI_VOID cei_get_irig_initial_time (pCEI_UINT32 time)
{
   TIME_TAG_TYPE system_time;
   CEI_NATIVE_UINT seconds;
   CEI_NATIVE_UINT tens_of_seconds;
   CEI_NATIVE_UINT minutes;
   CEI_NATIVE_UINT tens_of_minutes;
   CEI_NATIVE_UINT hours;
   CEI_NATIVE_UINT tens_of_hours;
   CEI_NATIVE_UINT days;
   CEI_NATIVE_UINT tens_of_days;
   CEI_NATIVE_UINT hundreds_of_days;

   union
   {
      struct irigTimeFormat
           {
              CEI_NATIVE_UINT seconds          :4;
              CEI_NATIVE_UINT tens_of_seconds  :3;
              CEI_NATIVE_UINT minutes          :4;
              CEI_NATIVE_UINT tens_of_minutes  :3;
              CEI_NATIVE_UINT hours            :4;
              CEI_NATIVE_UINT tens_of_hours    :2;
              CEI_NATIVE_UINT days             :4;
              CEI_NATIVE_UINT tens_of_days     :4;
              CEI_NATIVE_UINT hundreds_of_days :2;
           } IRIG_TIME_FORMAT;
      CEI_UINT32 timeValue;
   } timeConvert;

   cei_get_host_system_time_x (&system_time);

   seconds = (CEI_UINT32)(system_time / (TIME_TAG_TYPE)1000000);
   if ( (system_time % (TIME_TAG_TYPE)1000000) > (TIME_TAG_TYPE)500000) {
      seconds++;
   }

   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;
   tens_of_seconds = seconds/10;
   seconds = seconds%10;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;
   tens_of_minutes = minutes/10;
   minutes = minutes%10;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;     
   tens_of_hours = hours/10;    
   hours = hours%10;    
   days++;                       /* account for no such thing as day 0 */
   hundreds_of_days = days/100;
   days = days%100;
   tens_of_days = days/10;
   days = days%10;

   timeConvert.IRIG_TIME_FORMAT.seconds          = seconds;
   timeConvert.IRIG_TIME_FORMAT.tens_of_seconds  = tens_of_seconds;
   timeConvert.IRIG_TIME_FORMAT.minutes          = minutes;
   timeConvert.IRIG_TIME_FORMAT.tens_of_minutes  = tens_of_minutes;
   timeConvert.IRIG_TIME_FORMAT.hours            = hours;
   timeConvert.IRIG_TIME_FORMAT.tens_of_hours    = tens_of_hours;
   timeConvert.IRIG_TIME_FORMAT.days             = days;
   timeConvert.IRIG_TIME_FORMAT.tens_of_days     = tens_of_days;
   timeConvert.IRIG_TIME_FORMAT.hundreds_of_days = hundreds_of_days;
   
   *time = timeConvert.timeValue;
}

/*========================================================================* 
 * Device Access Routines for Windows Operating Systems 
 *========================================================================*/

/*===========================================================================*
 * ENTRY POINT:   A C Q U I R E   D E V I C E   I D E N T I T Y
 *===========================================================================*
 *
 * FUNCTION:    Read the specified device PCI Configuration Space Identifiers
 *
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 15)
 *              vendorId    -- (output) ptr to storage of vendor PCI ID
 *              deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL  -- successful acquisition of the device ID's
 *              ARS_FAILURE -- low-level failed due to invalid device ID
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine acquires the specified device PCI Configuration  
 *              Space Vendor ID and Device ID fields via low-level function.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device_identity (
   CEI_INT32  deviceIndex,                /* device number to test (0-15)    */
   pCEI_INT32 vendorId,                   /* ptr to storage of vendor PCI ID */
   pCEI_INT32 deviceId                    /* ptr to storage of device PCI ID */
)
{
   CEI_INT32 status;
   CEI_UINT32 idValue;
   CEI_INSTALL_ERR err;

   if (cei_dll != NULL) {
      if ( (status = cei_dll->LL_ReadConfigSpace(dev_handle[deviceIndex], 0, 2, &idValue, err)) == FALSE) {
         status = ARS_INVBOARD;
      }
      else {
         *vendorId = (CEI_INT32)idValue;
         status = cei_dll->LL_ReadConfigSpace(dev_handle[deviceIndex], 2, 2, &idValue, err);
         *deviceId = (CEI_INT32)idValue;
         status = ARS_NORMAL;
      }
   }
   else {
      status = ARS_FAILURE;
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:          A C Q U I R E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  deviceIndex    -- (input)  device to reference (0 to 15)
 *              device         -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL     -- Session was opened successfully
 *              ARS_WINRTFAIL  -- Driver Open Device failed
 *
 * GLOBALS:     virtualBaseAddress[16][3]
 *              device[16]
 *
 * DESCRIPTION: This routine aquires a device handle and maps the two
 *              PCI memory regions for the specified hardware.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INT32 status = ARS_NORMAL;
   CEI_INT32 sharedLockState;
   TIMECAPS tc;
   CEI_DEV_HANDLE new_dev_handle;  /* receives new device handle */
   CEI_DEVICE_MEM dev_mem;         /* receives device memory region info */
   CEI_INSTALL_ERR cei_err_ignore; /* ignored installation library error     */
   CEI_UINT32 preexisted;          /* receives device mapping preexistence */
   CEI_UINT32 board = (CEI_UINT32)deviceIndex;

   /* dynamically link to the installation library and open the device */
   cei_dll = CEI_DLL_GetDynamicLink (0, deviceOpenErrorDesc);
   if (cei_dll == NULL) {
      status = ARS_WINRTFAIL;
   }
   else {
      /* attempt to open the specified device */
      if (cei_dll->LL_OpenDeviceByLogicalId (board, &new_dev_handle, deviceOpenErrorDesc) == FALSE) {
         status = ARS_WINRTFAIL;
      }
   
      /* attempt to map device memory */
      else if (cei_dll->LL_MapMemoryRegions (new_dev_handle, CEI_DEVICE_MEM_VER, 0, &dev_mem, &preexisted, deviceOpenErrorDesc) == FALSE) {
         /* close the device and bail out */
         cei_dll->LL_CloseDevice (new_dev_handle, cei_err_ignore);
         status = ARS_WINRTFAIL;
      }
   
      /* store host addresses of all memory regions */
      if (status == ARS_NORMAL) {
         virtualBaseAddress[board][0] = (CEI_UINT64)dev_mem.rgn_usr_addr[0];
         virtualBaseAddress[board][1] = (CEI_UINT64)dev_mem.rgn_usr_addr[1];
         device->baseAddr = (pCEI_UINT32)dev_mem.rgn_usr_addr[0];
         device->devicePtr = (pCEI_UINT32)dev_mem.rgn_usr_addr[1];
         /* store the device handle for later */
         dev_handle[board] = new_dev_handle;
      
         /* if not already defined, request multi-thread/process protection */
         if (ceiAccessLock[board] == NULL) {
            status = cei_open_shared_lock (deviceIndex, &sharedLockState);
         }
   
         /* Attempt to set the Windows timer resolution to 1 millisecond */
         if (timeGetDevCaps(&tc, sizeof(TIMECAPS)) == TIMERR_NOERROR) {
            timeBeginPeriod(min(max(tc.wPeriodMin, 1), tc.wPeriodMax));
         }
      }
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:             R E L E A S E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 15)
 *              device      -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL     -- Session was opened successfully
 *              ARS_WINRTFAIL  -- Driver Close Device failed
 *
 * GLOBALS:     virtualBaseAddress[16][3]
 *
 * DESCRIPTION: This routine releases a driver handle and unmaps the two PCI 
 *              memory regions previously mapped for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_release_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INSTALL_ERR cei_err;        /* receives installation library error     */
   CEI_UINT32 persists_ignore;     /* receives device mapping persistence */
   CEI_UINT32 board = (CEI_UINT32)deviceIndex;
   
   if ((dev_handle[board] != NULL) && (cei_dll != NULL)) {
      /* unmap device memory */
      if (cei_dll->LL_UnmapMemoryRegions (dev_handle[board], &persists_ignore, cei_err) == FALSE)
         error_box("Failed to unmap device memory on board '%ld' (details: %s).",board,cei_err);
   
      /* close the device and flag the handle as closed */
      if (cei_dll->LL_CloseDevice(dev_handle[board],cei_err) == FALSE)
         error_box("Failed to close board '%ld' (details: %s).",board,cei_err);
   
      /* indicate no regions are still mapped and flag the handle as closed */
      dev_handle[board] = INVALID_HANDLE_VALUE;
      
      /* release multithread mutex protection */
      cei_dll->CloseGlobalNamedObject(ceiAccessLock[board], &persists_ignore, cei_err); 
   
      /* relinquish the resources */
      ceiAccessLock[board] = NULL;
      return ARS_NORMAL;
   }
   else
      return ARS_WINRTFAIL;
}


/*===========================================================================*
 * ENTRY POINT: cei_map_shared_memory
 *===========================================================================*
 *
 * FUNCTION:    Map the desired shared global memory block into the calling
 *              process's address space, allocating the block if it doesn't
 *              already exist.  
 *
 * PARAMETERS:  CEI_UCHAR  blockId        -- (input)  unique global block id (0-15)
 *              CEI_UINT32 blockSize      -- (input)  memory block size (in bytes)
 *              CEI_VOID   **blockPtr     -- (output) receives pointer to memory block
 *              CEI_VOID   **blockHandle  -- (output) receives platform-specific handle 
 *                                                    to the shared memory block
 *              CEI_UINT32 *alreadyExists -- (output) receives true if the block already
 *                                                    exists, false otherwise
 *              CEI_UINT32 *isExclusive   -- (output) receives true if no other processes
 *                                                    currently have the block mapped, 
 *                                                    false otherwise
 *
 * RETURN VAL:  ARS_NORMAL          - operation completed successfully
 *              ARS_ERR_SH_MEM_MAP  - failed to map shared memory block
 *              ARS_ERR_SH_MEM_OBJ  - failed to allocate shared memory block
 *              ARS_INVARG          - board index is invalid
 *
 *
 * DESCRIPTION: If a shared global memory block associated with the given 
 *              application-specific id has already been allocated on the 
 *              target system, map the memory block into the calling 
 *              process's address space and set 'already_exists' to true.  
 *              Otherwise, allocate and map the specified global memory block
 *              and set 'already_exists' to false.
 *
 *              If the memory block is being allocated for the first time, 
 *              zero out the entire block.
 *
 *===========================================================================*/
CEI_INT32 cei_map_shared_memory ( CEI_UCHAR    blockId,
                                  CEI_UINT32   blockSize,
                                  CEI_VOID  ** blockPtr,
                                  CEI_VOID  ** blockHandle,
                                  CEI_UINT32 * alreadyExists,
                                  CEI_UINT32 * isExclusive )
{
   /* build unique block name */
   CHAR blockName[32];
   
   /* make sure the block id is valid */
   if (blockId >= MAX_BOARDS)
      return ARS_INVARG;

   if (cei_dll != NULL) {
      sprintf(blockName, "ceix30_shmem_block_%d", blockId);
   
      /* initialize the block pointer and handle */
      *blockPtr = NULL;
      *blockHandle = NULL;
   
      /* create a file mapping object of the desired size */
      if (cei_dll->OpenGlobalNamedObject (CEI_GSO_SHMEM, (CEI_UINT64)blockSize, blockName, alreadyExists, blockPtr, blockHandle, deviceOpenErrorDesc) == FALSE)
         return ARS_ERR_SH_MEM_OBJ;
   
      /* no other procs have block mapped if it didn't already exist (under windows) */
      *isExclusive = (*alreadyExists == FALSE);
   
      /* obtain pointer to the shared memory block */
      if (*blockPtr == NULL)
         return ARS_ERR_SH_MEM_MAP;
   
      /* operation completed successfully */
      return ARS_NORMAL;
   }
   else {
      return ARS_FAILURE;
   }
}


/*========================================================================* 
 * Device Access Routines for Windows Operating Systems 
 * using WinRT Device Driver via the CEI Low Level Library.
 *========================================================================*/

/*===========================================================================*
 * ENTRY POINT:              R E A D   3 2   B I T
 *===========================================================================*
 *
 * FUNCTION:    Perform a 32-bit read from the specified device
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15)
 *              address     -- (input) ptr to device address to read
 *              data        -- (input) ptr to location to store data
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit read access from one of the 
 *              device's two allocated PCI memory regions
 *
 *===========================================================================*/
CEI_INT32 cei_read_32bit (CEI_INT16 deviceIndex, pCEI_UINT32 address, pCEI_UINT32 data)
{
   *data = *address;
   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:             W R I T E   3 2   B I T
 *===========================================================================*
 *
 * FUNCTION:    Perform a 32-bit write to the specified device
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 15)
 *              address     -- (input) ptr to device address to write
 *              data        -- (input) data to write
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine performs a 32-bit write access to one of the 
 *              device's two allocated PCI memory regions
 *
 *===========================================================================*/
CEI_INT32 cei_write_32bit (CEI_INT16 deviceIndex, pCEI_UINT32 address, CEI_UINT32 data)
{
   *address = data;
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:         C E I _ I N I T _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Configures an interrupt and associates an ISR with it
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 15)
 *              CEI_INT32   _stdcall isr      -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 *              CEI_INT32   memRegion         -- (input)  PCI BAR region/index for interrupt clear bit (0 - BAR0 or 1 - BAR2)
 *              CEI_UINT32  disOffset         -- (input)  PCI Interrupt Disable register (PLX reg offset, 0 for no PLX interface)
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 isrEnabled        -- (output) thread handle if successfully defined
 * 
 * RETURN VAL:  ARS_NORMAL        -- Interrupt was initialized successfully
 *              ARS_DRIVERFAIL    -- Device driver session never opened
 *              ARS_INT_ISR       -- Interrupt was not initialized successfully
 *                                -- Driver interrupt event creation failure
 *                                -- Windows thread create failure
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device.
 *
 *===========================================================================*/
CEI_INT32 cei_init_interrupt (
   CEI_INT32   deviceIndex,
   CEI_INT32   (_stdcall *isr)(CEI_INT32), 
   CEI_INT32   priority,
   pCEI_UINT32 deviceBaseAddress,
   CEI_INT32   memRegion,
   CEI_UINT32  disOffset,
   CEI_UINT32  interruptBit,
   pCEI_UINT32 isrEnabled
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the low-level global virtual base address array     */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   
   if (interruptBit != 0) {
      status = cei_read_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg &= ~interruptBit;
      status = cei_write_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }

   /* initialize the interrupt */
   status = ceill_initInterrupt (deviceIndex, isr, priority, memRegion, disOffset, isrEnabled);   
   
   if (status == ARS_NORMAL) {
      /* Set either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
       * Local Interrupt Input Enable, High to pass local bus interrupts
       */
      if (interruptBit != 0) {
         status = cei_read_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
         intCtrlStatReg |= interruptBit;
         status = cei_write_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
      }
   }
   return status;
}

/*===========================================================================*
 * ENTRY POINT:     C E I _ T E R M I N A T E _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to 
 *              an ISR
 * 
 * PARAMETERS:  deviceIndex  -- (input)  device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *              ARS_INT_ISR  -- Interrupt was not terminated successfully
 *
 * GLOBALS:     None
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function
 * 
 *===========================================================================*/
CEI_INT32 cei_terminate_interrupt
(
    CEI_INT32   deviceIndex, 
    CEI_UINT32  interruptBit,
    pCEI_UINT32 deviceBaseAddress
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status = ARS_NORMAL;         /* local function return status    */
   /* local reference to the supplied device BAR0 virtual base address       */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
   
   /* Reset either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, Low to disable local bus interrupts
    */
   if (interruptBit != 0) {
      status = cei_read_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg &= ~interruptBit;
      status = cei_write_32bit ((CEI_INT16)deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }

   if (status == ARS_NORMAL) {
      status = ceill_terminateInterrupt(dev_handle[deviceIndex]);
   }

   return status;
}



