/*===========================================================================*
 * FILE:                     C D E V _ L N X . C
 *===========================================================================*
 *
 * COPYRIGHT (C) 2005-2011 BY
 *          GE FANUC INTELLIGENT PLATFORMS, INC. SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE FANUC 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:    Application Programmer's Interface for a CEI-x30 ARINC device.
 *              Linux 2.4.x & 2.6.x Operating System Specific Functionality
 *
 *    Internal CEI-x30 routines implemented herein:
 *
 *    cei_acquire_device          - opens a session with the specified x30 device
 *    cei_acquire_device_identity - supplies the PCI configuration space ID values
 *    cei_acquire_thread_lock     - acquires sempahore/mutex lock
 *    cei_delay_milliseconds      - o/s specific delay utility in milliseconds
 *    cei_delay_seconds           - o/s specific delay utility in fractional seconds
 *    cei_get_host_system_time_x  - acquires host system time-of-year 64-bit uSec     
 *    cei_get_host_system_time    - acquires host system time-of-year 32-bit mSec     
 *    cei_get_irig_initial_time   - acquires host system time-of-year IRIG format
 *    cei_map_shared_memory       - maps memory region for sharing globals across processes
 *    cei_open_shared_lock        - creates a shared semaphore across processes
 *    cei_release_device          - closes a session with the specified x30 device
 *    cei_release_thread_lock     - releases sempahore/mutex lock     
 *    cei_read_32bit              - read a 32-bit value from an x30 device
 *    cei_write_32bit             - write a 32-bit value to an x30 device
 *    cei_init_interrupt          - configures an interrupt and ISR
 *    cei_terminate_interrupt     - terminate an interrupt
 *          
 *---------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ------------------------------------------------------------
 * 11/28/05  BCH  Created.
 * 04/11/06  BDW  Removed utility routines cei_read_plx_32bit & cei_write_plx_32bit
 * 08/24/06  BDW  Incorporated CEI_TYPES.H data type definitions in 
 *                  support of 64-bit Linux operating systems. 
 *                Swapped the BAR reference order by the device structure 
 *                   members baseAddr and devicePtr
 * 02/27/07  BDW  Added cei_acquire_device_identity() to supply the PCI 
 *                   Manufacturer and Device IDs for x30 products 
 * 07/10/07  BCH  Added support for CEI-530
 * 01/11/08  BDW  Added multi-process global semophore protection support, see
 *                the use of the compiler directive MULTI_PROCESS_SUPPORT below
 * 01/18/08  BCH  Added support for RAR-CPCI, RAR-EC, and R830RX
 * 02/18/08  BDW  Removed the boardInitialized array, relocated as a structure
 *                   member in the data structure array "deviceGlobal"
 *                Added new routines for multi-process support: 
 *                   cei_open_shared_lock and cei_map_shared_memory
 * 03/10/08  BDW  Removed the declaration of the following global arrays, 
 *                relocated to CDEV_GLB.H:
 *                   bypassWrapTest
 *                   irigDacThreshold
 *                   a717syncWords
 *                Removed declaration of basicConfiguration, it was unused.
 *                Added reference to device[] structure as a parameter to
 *                   cei_acquire_device & cei_release_device so they no longer access 
 *                   the global data structure directly.
 * 04/15/08  BDW  Added support for initiation and termination of h/w interrupts
 *                via routines cei_init_interrupt and cei_terminate_interrupt.
 * 06/16/09  BDW  Added the function cei_open_preload_lock for multiprocess support.
 * 07/01/09  BDW  Added new function cei_delay_milliseconds (o/s thread suspend).
 * 10/28/10  BDW  Modified the parameter list for cei_init_interrupt to account for      
 *                   interrupt setup requirements in the Windows-only CEI_Install library
 * 09/28/11  BDW  Added support for CEI-430A in cei_acquire_device.               
 *
 *===========================================================================*/
#define IN_LNX_SOURCE 1

#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h> 
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/syscall.h>
#include "cdev_glb.h"
#include "lowlevel.h"
#include "../../../Lowlevel/ll_posix.h"

#if defined (_LINUX_PPC_)
   #define TARGET_BIG_ENDIAN
#endif

/* Maximum number of global semaphores to create for use in multiprocess applications */
#define MAX_CFG_LOCKS 32

/* define: MAX_SHMEM_RGNS                                          */
/* usage:  The maximum number of supported shared memory regions.  */
#define MAX_SHMEM_RGNS  16

/* define: CEIX30_IPC_KEY_PATH                                            */
/* usage:  This path is used to generate unique IPC keys under Linux in   */
/*         conjunction with the ftok() routine.                           */
#define CEIX30_IPC_KEY_PATH    "/tmp/ceix30_ipc_keygen"

#define MAX_DEV_OPEN_ERROR_STR_LEN 512

/* Structure for storing application-assigned BAR virtual base addresses */
typedef struct RESOURCE_ALLOC {
   CEI_UINT32 bar0BaseAddress;
   CEI_UINT32 bar2BaseAddress;
} RESOURCE_ALLOC_TYPE;

extern void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32);

/*========================================================================* 
 * Linux-based globals (Board-based items only supporting 8 devices)
 *========================================================================*/
static DEVMAP_T devObj[MAX_BOARDS];
static CEI_UINT32 embeddedResourcesAssigned[MAX_BOARDS] = 
   { FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE };
static RESOURCE_ALLOC_TYPE assignedResources[MAX_BOARDS];

/*========================================================================* 
 * Operating System specific functionality
 *========================================================================*/
/*===========================================================================*
 * INTERNAL ROUTINE:      L I T T L E _ E N D I A N _ 3 2
 *===========================================================================*
 *
 * FUNCTION:    Return the little-endian equivalent of the 4-byte integral
 *              argument (endian swap if host is big-endian).
 *
 * RETURN VAL:  The little-endian representation of the argument.
 *
 *===========================================================================*/
#define little_endian_32(longword) ( ((longword & 0x000000ff) << 24) | \
                                     ((longword & 0x0000ff00) << 8)  | \
                                     ((longword & 0xff0000)   >> 8)  | \
                                     ((longword & 0xff000000) >> 24) )

/* external references for low-level interrupt processing */
CEI_INT vbtWaitForInterrupt(CEI_UINT device);
CEI_INT vbtInterruptMode(CEI_UINT device, CEI_INT mode);

/* interrupt processing local variables */
static CEI_THREAD int_thread;
static CEI_INT int_thread_running = 0;
static CEI_VOID (*pIsr)(int val) = NULL;

/* Device Access mutex  */
#ifdef MULTI_PROCESS_SUPPORT
CEI_INT32 ceiAccessLock[MAX_BOARDS] = 
{ 
   (CEI_INT32)NULL, 
   (CEI_INT32)NULL, 
   (CEI_INT32)NULL, 
   (CEI_INT32)NULL, 
   (CEI_INT32)NULL, 
   (CEI_INT32)NULL, 
   (CEI_INT32)NULL, 
   (CEI_INT32)NULL
};
#else
pthread_mutex_t ceiAccessLock[MAX_BOARDS];
pthread_mutexattr_t m_attr;
#endif

/* storage for Windows driver low-level board mapping error messages      */
static pCEI_CHAR deviceOpenErrorDesc = "Specified PCI Device Not Found";


/*===========================================================================*
 * INTERNAL ROUTINE: get_ipc_key
 *===========================================================================*
 *
 * FUNCTION:    Obtain unique IPC key.
 *
 * RETURN VAL:  unique IPC key, or -1 upon failure
 *
 * DESCRIPTION: Generate a unique system-global IPC key using the given 
 *              unique application-specific IPC identifier.
 *
 *===========================================================================*/
key_t get_ipc_key(CEI_UCHAR ipc_id)
{
   /* make sure the file used to generate unique IPC keys exists */
   int key_file = open(CEIX30_IPC_KEY_PATH, O_RDWR | O_CREAT | O_TRUNC, 0666);
   if (key_file == -1)
      return -1;

   /* make the key file global readable/writable and close the file */
   fchmod(key_file,0666);
   close(key_file);

   /* obtain system-global IPC key */
   return ftok(CEIX30_IPC_KEY_PATH,ipc_id);
}

/*===========================================================================*
 * ENTRY POINT:   O P E N   S H A R E D   L O C K
 *===========================================================================*
 *
 * FUNCTION:    Opens a shared (global) semaphore 
 *
 * PARAMETERS:  sharedLockId -- (input) index of the specific semaphore to open
 *              sharedLockState -- (input) location to store semaphore handle
 *
 * RETURN VAL:  ARS_NORMAL  -- semaphore was opened successfully
 *              ARS_INVARG  -- semaphore id too large
 *              ARS_FAILURE -- failed to create/open the semaphore
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Opens the specified shared (global) semaphore for use in
 *              multiple process applications
 *
 *===========================================================================*/
CEI_INT32 cei_open_shared_lock (CEI_INT16 sharedLockId, pCEI_INT32 sharedLockState)
{
   /* make sure the lock id is valid */
   if (sharedLockId >= MAX_BOARDS)
      return ARS_INVARG;

   /* obtain unique IPC key (shared config locks use unique ids 16-31) */
   key_t lock_key = get_ipc_key(MAX_SHMEM_RGNS + sharedLockId);
   if (lock_key == -1)
      return ARS_FAILURE;

   /* attempt to open the shared semaphore (fail if it doesn't exist) */
   *sharedLockState = semget(lock_key, 1, 0);
   if (*sharedLockState == -1)
   {
      /* bail out for all failures except 'semaphore doesn't exist' */
      if (errno != ENOENT)
         return ARS_FAILURE;

      /* attempt to create the shared semaphore (fail if it exists) */
      *sharedLockState = semget(lock_key, 1, 0666 | IPC_CREAT | IPC_EXCL);
      if (*sharedLockState != -1)
      {
         /* semaphore created successfully - set the semaphore value */
         if (semctl(*sharedLockState, 0, SETVAL, (int)1) == -1)
            return ARS_FAILURE;
      }
      else
      {
         /* attempt open again (another thread may have created between semget calls) */
         *sharedLockState = semget(lock_key,1,0);
         if (*sharedLockState == -1)
            return ARS_FAILURE;
      }
   }
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:   O P E N   P R E L O A D   L O C K
 *===========================================================================*
 *
 * FUNCTION:    Requests the initiation of a shared (global) semaphore 
 *              for multiprocess application usage
 *
 * PARAMETERS:  device -- (input) index of the device to create lock based on
 *
 * RETURN VAL:  ARS_NORMAL  -- semaphore was opened successfully
 *              ARS_FAILURE -- failed to create/open the semaphore
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: Create the specified shared (global) semaphore for use in
 *              multiple process applications
 *
 *===========================================================================*/
CEI_INT32 cei_open_preload_lock (CEI_INT16 device)
{
   CEI_INT32 status;
   status = cei_open_shared_lock(device, (pCEI_INT32)&ceiAccessLock[device]);
   return status;
}


/*===========================================================================*
 * ENTRY POINT:            G E T   O P E N   E R R O R
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the error description for a device open error
 *
 * PARAMETERS:  None
 * 
 * RETURN VAL:  Reference to assign to error string
 *
 * GLOBALS:     deviceOpenErrorDesc
 *
 * DESCRIPTION: This routine retrieves the error description previously saved
 *              when a device open error was encountered.
 *
 *===========================================================================*/
pCEI_CHAR cei_get_open_error (CEI_VOID)
{
  return deviceOpenErrorDesc;
}

/*===========================================================================*
 * ENTRY POINT:            T I M E   G E T   T I M E
 *===========================================================================*
 *
 * FUNCTION:    Replicate the Windows O/S System Time Acquisition Method
 *
 * RETURN VAL:  32-bit 1 millisecond resolution system time
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the Linux clock invocation
 *              to obtain the current system timer value with a
 *              resolution of 1 millisecond, replicating the Windows O/S 
 *              system time acquisition method of the same name.
 *
 *===========================================================================*/
CEI_UINT32 timeGetTime()
{
   return(CEI_UINT32)clock();
}

/*===========================================================================*
 * ENTRY POINT:        A C Q U I R E   T H R E A D   L O C K    
 *===========================================================================*
 *
 * FUNCTION:    Block on a mutex or semaphore previously reserved for the 
 *              specific device.  A mutex is used for single process multi-
 *              threading, a global semaphore used for multi-process.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 7)
 * 
 * RETURN VAL:  ARS_NORMAL  -- semaphore was acquired successfully
 *              ARS_FAILURE -- Acquisition of the semaphore timed-out/failed
 *
 * GLOBALS:     ceiAccessLock
 *
 * DESCRIPTION: This routine blocks execution, waiting for access to the 
 *              Mutex previously allocated for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_thread_lock (CEI_INT16 deviceIndex)
{
#ifdef MULTI_PROCESS_SUPPORT
   CEI_INT32 status;
   static struct sembuf sem_lock = { 0, -1, SEM_UNDO };
   
   status = semop(ceiAccessLock[deviceIndex], &sem_lock, 1);
   if (status != -1)
      status = ARS_NORMAL;
   
   return status;
#else
   pthread_mutex_lock(&ceiAccessLock[deviceIndex]);
   return ARS_NORMAL;
#endif
}

/*===========================================================================*
 * ENTRY POINT:        R E L E A S E   T H R E A D   L O C K    
 *===========================================================================*
 *
 * FUNCTION:    Release a mutex or semaphore previously reserved/blocked by  
 *              the application/API for the specific device.  A mutex is used
 *              for single process multi-threading, a global semaphore used 
 *              for multi-process.
 *
 * PARAMETERS:  deviceIndex -- (input) device to reference (0 to 7)
 * 
 * RETURN VAL:  ARS_NORMAL  -- semaphore was acquired successfully
 *              ARS_FAILURE -- release of the semaphore timed-out/failed
 *
 * GLOBALS:     ceiAccessLock
 *
 * DESCRIPTION: This routine releases the Mutex previously 
 *              reserved/blocked by the application/API.
 *
 *===========================================================================*/
CEI_INT32 cei_release_thread_lock (CEI_INT16 deviceIndex)
{
#ifdef MULTI_PROCESS_SUPPORT
   CEI_INT32 status;
   static struct sembuf sem_unlock = { 0, 1, SEM_UNDO };

   status = semop(ceiAccessLock[deviceIndex], &sem_unlock, 1);
   if (status != -1)
      status = ARS_NORMAL;

   return status;
#else
   pthread_mutex_unlock(&ceiAccessLock[deviceIndex]);
   return ARS_NORMAL;
#endif
}

/*===========================================================================*
 * ENTRY POINT:              D E L A Y   M I L L I S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified number of milliseconds
 *
 * PARAMETERS:  numberOfMillieconds -- (input) the number of milliseconds
 *                                             to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the BSP-respective delay invocation
 *              as the basis to suspend thread execution for a specified
 *              number of milliseconds (resolution is platform dependent).
 *
 *===========================================================================*/
CEI_INT32 cei_delay_milliseconds (CEI_UINT32 numberOfMilliseconds)
{
   usleep(numberOfMilliseconds * 1000);
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:              D E L A Y   S E C O N D S
 *===========================================================================*
 *
 * FUNCTION:    Delay for a specified (fractional) number of seconds
 *
 * PARAMETERS:  numberOfSeconds -- (input) a floating point representation
 *                                         of the number (or fraction) of
 *                                         seconds to delay execution
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine utilizes the target-respective delay invocation
 *              as the basis for a delay defined in fractional seconds.
 *
 *===========================================================================*/
CEI_INT32 cei_delay_seconds (CEI_FLOAT numberOfSeconds)
{
   CEI_UINT32 start, now, delay;

   if((delay = (CEI_UINT32)(CLOCKS_PER_SEC * numberOfSeconds)) < 1)
     delay = 1;

   start = clock ();
   now = start;
   while ((now - start) < delay) {
      now = clock ();
   }

   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:       G E T   H O S T   S Y S T E M   T I M E  X
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host 64-bit 1ms resolution reference time value
 *              System Time in "seconds since midnight Jan 1 of this year"
 *
 * PARAMETERS:  hostTime -- (output) 64-bit 1 millisecond resolution time-tag
 *                                   seconds since midnight Jan 1 of this year
 *                                   (value returned in microseconds)
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current VxWorks System Time in
 *              "seconds since midnight Jan 1 of this year", then converts
 *              that value to a 64-bit microseconds value (1ms resolution).
 *
 *===========================================================================*/
CEI_VOID cei_get_host_system_time_x (TIME_TAG_TYPE * hostTime)
{
   time_t now;               /* Current time, modify to get Time Reference Point. */
   struct   tm start;        /* Formatted "now" current time, modified for TRP.   */
   time_t   then;            /* Time Reference Point, Midnight or Jan 1.          */
   struct   timeb ts;        /* Current time, use to get milliseconds.            */
   CEI_UINT32 last_milli;    /* This is the time we use to wait for time to change.    */
   TIME_TAG_TYPE diff;       /* This is the difference time we return in microseconds, */

   time(&now);               /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now); /* Local                                          */
   start.tm_sec  = 0;        /*   time                                         */
   start.tm_min  = 0;        /*     at                                         */
   start.tm_hour = 0;        /*       Jan 1                                    */
   start.tm_mday = 1;        /*         midnight (This is day 0)               */
   start.tm_mon  = 0;        /*           this                                 */
   start.tm_yday = 0;        /*             year                               */
   then = mktime(&start);    /* local Jan 1 to time in seconds since Jan 1     */
                             /* We do IRIG time, which makes Jan 1 "day 0"     */ 

   ftime(&ts);                              /* Current time                    */
   last_milli = ts.millitm;                 /*  save milliseconds              */
   while (1) {                              /* Wait for milliseconds to change */
      ftime(&ts);                           /* Current time                    */
      if (ts.millitm != last_milli)
         break;                             /* Break at exact time tick change */
   }
   diff = ts.time-then;       /* We only want the time since Midnight on Jan 1  */
   diff *= 1000000;           /* Number of Seconds converted to microseconds    */
   diff += ts.millitm * 1000; /* plus number of Milliseconds converted to       */
   *hostTime = diff;          /* microseconds.  Passed back to the caller.      */
}

/*===========================================================================*
 * ENTRY POINT:       G E T   H O S T   S Y S T E M   T I M E
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host 32-bit 1ms resolution reference time value
 *              System Time in "seconds since midnight Jan 1 of this year"
 *
 * PARAMETERS:  hostTime -- (output) 32-bit 1 millisecond resolution time-tag
 *                                   seconds since midnight Jan 1 of this year
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current System Time in
 *              "seconds since midnight Jan 1 of this year", then converts
 *              that value to a 32-bit 1 millisecond resolution value.
 *
 *===========================================================================*/
CEI_INT32 cei_get_host_system_time (pCEI_UINT32 hostTime)
{
   *hostTime = timeGetTime();
   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT:       G E T   I R I G   I N I T I A L   T I M E 
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the host reference calendar time/date value, used
 *              for CEI-x30 IRIG timer initialization
 *
 * PARAMETERS:  time -- (output) 32-bit packed IRIG time value
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine retrieves the current Windows System Time in
 *              calendar time/date and formats it into an IRIG time value.
 *
 *===========================================================================*/
CEI_VOID cei_get_irig_initial_time (pCEI_UINT32 time)
{
   TIME_TAG_TYPE system_time;
   CEI_NATIVE_UINT seconds;
   CEI_NATIVE_UINT tens_of_seconds;
   CEI_NATIVE_UINT minutes;
   CEI_NATIVE_UINT tens_of_minutes;
   CEI_NATIVE_UINT hours;
   CEI_NATIVE_UINT tens_of_hours;
   CEI_NATIVE_UINT days;
   CEI_NATIVE_UINT tens_of_days;
   CEI_NATIVE_UINT hundreds_of_days;

   union
   {
      struct irigTimeFormat
           {
              CEI_NATIVE_UINT seconds          :4;
              CEI_NATIVE_UINT tens_of_seconds  :3;
              CEI_NATIVE_UINT minutes          :4;
              CEI_NATIVE_UINT tens_of_minutes  :3;
              CEI_NATIVE_UINT hours            :4;
              CEI_NATIVE_UINT tens_of_hours    :2;
              CEI_NATIVE_UINT days             :4;
              CEI_NATIVE_UINT tens_of_days     :4;
              CEI_NATIVE_UINT hundreds_of_days :2;
           } IRIG_TIME_FORMAT;
      CEI_UINT32 timeValue;
   } timeConvert;

   cei_get_host_system_time_x (&system_time);

   seconds = (CEI_UINT32)(system_time / (TIME_TAG_TYPE)1000000);
   if ( (system_time % (TIME_TAG_TYPE)1000000) > (TIME_TAG_TYPE)500000) {
      seconds++;
   }

   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;
   tens_of_seconds = seconds/10;
   seconds = seconds%10;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;
   tens_of_minutes = minutes/10;
   minutes = minutes%10;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;     
   tens_of_hours = hours/10;    
   hours = hours%10;    
   days++;                       /* account for no such thing as day 0 */
   hundreds_of_days = days/100;
   days = days%100;
   tens_of_days = days/10;
   days = days%10;

   timeConvert.IRIG_TIME_FORMAT.seconds          = seconds;
   timeConvert.IRIG_TIME_FORMAT.tens_of_seconds  = tens_of_seconds;
   timeConvert.IRIG_TIME_FORMAT.minutes          = minutes;
   timeConvert.IRIG_TIME_FORMAT.tens_of_minutes  = tens_of_minutes;
   timeConvert.IRIG_TIME_FORMAT.hours            = hours;
   timeConvert.IRIG_TIME_FORMAT.tens_of_hours    = tens_of_hours;
   timeConvert.IRIG_TIME_FORMAT.days             = days;
   timeConvert.IRIG_TIME_FORMAT.tens_of_days     = tens_of_days;
   timeConvert.IRIG_TIME_FORMAT.hundreds_of_days = hundreds_of_days;
   
   *time = timeConvert.timeValue;
}

/*========================================================================* 
 * Device Access Routines for Linux Operating Systems 
 *========================================================================*/

/*===========================================================================*
 * ENTRY POINT:   A C Q U I R E   D E V I C E   I D E N T I T Y
 *===========================================================================*
 *
 * FUNCTION:    Read the specified device PCI Configuration Space Identifiers
 *
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 7)
 *              vendorId    -- (output) ptr to storage of vendor PCI ID
 *              deviceId    -- (output) ptr to storage of device PCI ID
 * 
 * RETURN VAL:  ARS_NORMAL  -- successful acquisition of the device ID's
 *              ARS_FAILURE -- low-level failed due to invalid device ID
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine acquires the specified device PCI Configuration  
 *              Space Vendor ID and Device ID fields via low-level function.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device_identity (
   CEI_INT32  deviceIndex,                /* device number to test (0-7)     */
   pCEI_INT32 vendorId,                   /* ptr to storage of vendor PCI ID */
   pCEI_INT32 deviceId                    /* ptr to storage of device PCI ID */
)
{
   CEI_INT32 status;                      /* local function return status    */
   CEI_INT16 pciDeviceId;

   if (embeddedResourcesAssigned[deviceIndex]) {
      *vendorId = 0x000013C6;
      *deviceId = 0x00000530; 
   }
   else {
      if((status = vbtGetPCIConfigRegister (deviceIndex, 0, 2, &pciDeviceId)) == 0) {
        *vendorId = pciDeviceId;
        if((status = vbtGetPCIConfigRegister (deviceIndex, 1, 2, &pciDeviceId)) == 0) {
          *deviceId = pciDeviceId; 
	  status = ARS_NORMAL;
	}
      }
   }

   return status;
}


/*===========================================================================*
 * ENTRY POINT:          A C Q U I R E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Open a session with a device
 *              
 * PARAMETERS:  deviceIndex   -- (input) device to reference (0 to 7)
 *              device        -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  ARS_NORMAL    -- Session was opened successfully
 *              ARS_FAILURE   -- Driver Open Device failed
 *
 * GLOBALS:     pci_state[8]
 *              deviceMapped[8][3]
 *              ceiAccessLock[8][3]
 *
 * DESCRIPTION: This routine searches for up to eight CEI-x30 devices, 
 *              mapping each CEI-x30 device found with the first execution 
 *              of this routine.  Subsequent executions will simply retrieve
 *              the mapped addresses for the specified device index.
 *
 *===========================================================================*/
CEI_INT32 cei_acquire_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   CEI_INT32 status;
   CEI_INT16 pciDeviceId;

   if (embeddedResourcesAssigned[deviceIndex]) {
      device->baseAddr  = (pCEI_UINT32)assignedResources[deviceIndex].bar0BaseAddress;
      device->devicePtr = (pCEI_UINT32)assignedResources[deviceIndex].bar2BaseAddress;
      status = ARS_NORMAL;
   }
   else {
      memset (&devObj[deviceIndex], 0, sizeof(DEVMAP_T));
   
      devObj[deviceIndex].flagMapToHost[0] = TRUE;
      devObj[deviceIndex].flagMapToHost[1] = TRUE;
   
      status = vbtMapBoardAddresses (deviceIndex, &devObj[deviceIndex]);
   
      if (status == BTD_OK) {
         status = vbtGetPCIConfigRegister (deviceIndex, 1, 2, &pciDeviceId);
         if (status == BTD_OK) {
            if ( (pciDeviceId == PCI_DEVICE_ID_830  ) || 
                 (pciDeviceId == PCI_DEVICE_ID_830RX) ||
                 (pciDeviceId == PCI_DEVICE_ID_630  ) || 
                 (pciDeviceId == PCI_DEVICE_ID_530  ) ||
                 (pciDeviceId == PCI_DEVICE_ID_430  ) || 
                 (pciDeviceId == PCI_DEVICE_ID_430A ) || 
                 (pciDeviceId == PCI_DEVICE_ID_A30  ) ||
                 (pciDeviceId == PCI_DEVICE_ID_RAREC) ||
                 (pciDeviceId == PCI_DEVICE_ID_RAR_PCIE) ) {
               device->baseAddr  = (pCEI_UINT32)devObj[deviceIndex].memHostBase[0];
               device->devicePtr = (pCEI_UINT32)devObj[deviceIndex].memHostBase[1];
               status = ARS_NORMAL;
            }
            else {
               status = ARS_WINRTFAIL;
            }
         }
         else {
            status = ARS_WINRTFAIL;
         }
      }
      else {
         status = ARS_WINRTFAIL;
      }
   }

   /* create thread/process protection on a first-in/first-out basis */
   if (status == ARS_NORMAL) {
#ifdef MULTI_PROCESS_SUPPORT
      if (ceiAccessLock[deviceIndex] == NULL) {
         status = cei_open_shared_lock(deviceIndex, &ceiAccessLock[deviceIndex]);
      }
#else
      pthread_mutex_init(&ceiAccessLock[deviceIndex], &m_attr);
#endif
   }

   return status;
}


/*===========================================================================*
 * ENTRY POINT:             R E L E A S E   D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Close a session with a device
 *              
 * PARAMETERS:  deviceIndex -- (input)  device to reference (0 to 7)
 *              device      -- (output) reference to ARINC_LOCAL_TYPE struct
 * 
 * RETURN VAL:  fixed - ARS_NORMAL
 *
 * GLOBALS:     ceiAccessLock[8][3]
 *
 * DESCRIPTION: This routine releases a driver handle and unmaps the two PCI 
 *              memory regions previously mapped for the specific device.
 *
 *===========================================================================*/
CEI_INT32 cei_release_device (CEI_INT16 deviceIndex, ARINC_LOCAL_TYPE * device)
{
   if (embeddedResourcesAssigned[deviceIndex]) {
      embeddedResourcesAssigned[deviceIndex] = FALSE;
   }
   else {
      vbtFreeBoardAddresses(&devObj[deviceIndex]);
   }
#ifndef MULTI_PROCESS_SUPPORT
   /* release multi-thread protection */
   pthread_mutex_destroy(&ceiAccessLock[deviceIndex]);
#endif

   return ARS_NORMAL;
}

/*===========================================================================*
 * ENTRY POINT: cei_map_shared_memory
 *===========================================================================*
 *
 * FUNCTION:    Map the desired shared global memory block into the calling
 *              process's address space, allocating the block if it doesn't
 *              already exist.  
 *
 * PARAMETERS:  CEI_UCHAR  blockId        -- (input)  unique global block id (0-15)
 *              CEI_UINT32 blockSize      -- (input)  memory block size (in bytes)
 *              CEI_VOID   **blockPtr     -- (output) receives pointer to memory block
 *              CEI_VOID   **blockHandle  -- (output) receives platform-specific handle 
 *                                                    to the shared memory block
 *              CEI_UINT32 *alreadyExists -- (output) receives true if the block already
 *                                                    exists, false otherwise
 *              CEI_UINT32 *isExclusive   -- (output) receives true if no other processes
 *                                                    currently have the block mapped, 
 *                                                    false otherwise
 *
 * RETURN VAL:  ARS_NORMAL          - operation completed successfully
 *              ARS_ERR_SH_MEM_MAP  - failed to map shared memory block
 *              ARS_ERR_SH_MEM_OBJ  - failed to allocate shared memory block
 *              ARS_INVARG          - board index is invalid
 *
 *
 * DESCRIPTION: If a shared global memory block associated with the given 
 *              application-specific id has already been allocated on the 
 *              target system, map the memory block into the calling 
 *              process's address space and set 'alreadyExists' to true.  
 *              Otherwise, allocate and map the specified global memory block
 *              and set 'alreadyExists' to false.
 *
 *              If the memory block is being allocated for the first time, 
 *              zero out the entire block.
 *
 *===========================================================================*/
CEI_INT32 cei_map_shared_memory ( CEI_UCHAR    blockId,
                                  CEI_UINT32   blockSize,
                                  CEI_VOID  ** blockPtr,
                                  CEI_VOID  ** blockHandle,
                                  CEI_UINT32 * alreadyExists,
                                  CEI_UINT32 * isExclusive )
{
   /* receives block key */
   key_t blockKey;

   /* receives block attributes */
   struct shmid_ds blockInfo;

   /* initialize the block pointer and handle */
   *blockPtr = NULL;
   *blockHandle = NULL;

   /* make sure the block id is valid */
   if (blockId >= MAX_SHMEM_RGNS)
      return ARS_INVARG;

   /* obtain unique IPC key (shared mem blocks use unique ids 0-15) */
   blockKey = get_ipc_key(blockId);
   if (blockKey == -1)
      return ARS_FAILURE;

   /* assume (for now) that the block already exists */
   *alreadyExists = 1;

   /* attempt to open the shared block (fail if it doesn't exist) */
   *blockHandle = (CEI_VOID *)shmget(blockKey, blockSize, 0);

   if (*blockHandle == (CEI_VOID *)-1) {
      /* bail out for all failures except 'block doesn't exist' */
      if (errno != ENOENT)
         return ARS_ERR_SH_MEM_OBJ;

      /* attempt to create the block (fail if it exists) */
      *blockHandle = (CEI_VOID *)shmget(blockKey, blockSize, 0666 | IPC_CREAT | IPC_EXCL);
      if (*blockHandle != (CEI_VOID *)-1) {
         /* this process created the block */
         *alreadyExists = 0;
      }
      else {
         /* attempt open again (another thread may have created between shmget calls) */
         *blockHandle = (CEI_VOID *)shmget(blockKey, blockSize, 0);
         if (*blockHandle == (CEI_VOID *)-1)
            return ARS_ERR_SH_MEM_MAP;
      }
   }

   /* attach to the shared block */
   *blockPtr = shmat((int)*blockHandle, (void *)0, 0);
   if (*blockPtr == (CEI_VOID *)-1)
      return ARS_ERR_SH_MEM_MAP;

   /* initialize the block attributes structure */
   memset(&blockInfo, 0, sizeof(blockInfo));

   /* obtain the attached process count */
   if (shmctl((int)*blockHandle, IPC_STAT, &blockInfo) == -1)
      return ARS_ERR_SH_MEM_MAP;

   /* make sure the attached process count looks valid */
   if (blockInfo.shm_nattch == 0)
      return ARS_ERR_SH_MEM_MAP;

   /* determine if no other processes are attached to the block */
   *isExclusive = (blockInfo.shm_nattch == 1);

   /* operation completed successfully */
   return ARS_NORMAL;
}


/*------------------------------------------------------------------------* 
 * Perform a 32-bit read access
 *------------------------------------------------------------------------*/
CEI_INT32 cei_read_32bit (CEI_INT16 board, pCEI_UINT32 address, pCEI_UINT32 data)
{
#ifdef TARGET_BIG_ENDIAN
   CEI_UINT32 localData, modifiedData;
   localData = *address;
   modifiedData = little_endian_32(localData);
   *data = modifiedData;
#else
   *data = *address;
#endif          
   return ARS_NORMAL;
}

/*------------------------------------------------------------------------* 
 * Perform a 32-bit write access
 *------------------------------------------------------------------------*/
CEI_INT32 cei_write_32bit (CEI_INT16 board, pCEI_UINT32 address, CEI_UINT32 data)
{
#ifdef TARGET_BIG_ENDIAN
   *address = little_endian_32(data);
#else
   *address = data;
#endif          
   return ARS_NORMAL;
}


/*===========================================================================*
 * ENTRY POINT:         C E I _ I N I T _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    This routine invokes the API ISR routine.
 *              
 * PARAMETERS:  val -- (input) pointer to the device handle structure
 * 
 * RETURN VAL:  None
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine executes as a thread created by cei_init_interrupt.
 *              It invokes the user-specified interrupt function when an interrupt
 *              is processed by the device driver and this thread is awakened.
 *
 *===========================================================================*/
pCEI_VOID cei_lnx_int_function (pCEI_VOID val) 
{
   CEI_UINT dev = *((CEI_UINT *)val);

   while (int_thread_running == 1) {
      if(vbtWaitForInterrupt(dev) < 0)
         break;
      if(int_thread_running == 1)
         (*pIsr)(dev);      
   }

   return 0;
}


/*===========================================================================*
 * ENTRY POINT:         C E I _ I N I T _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Configures an interrupt and associates an ISR with it
 *              
 * PARAMETERS:  CEI_INT32   deviceIndex       -- (input)  device to reference (0 to 15)
 *              pCEI_VOID   isr               -- (input)  ptr to the interrupt service routine
 *              CEI_INT32   priority          -- (input)  desired priority of the ISR thread
 *              pCEI_UINT32 deviceBaseAddress -- (input)  virtual address of BAR2 region
 *              CEI_INT32   memRegion         -- (input)  not used
 *              CEI_UINT32  disOffset         -- (input)  not used
 *              CEI_UINT32  interruptBit      -- (input)  Local Interrupt Input Enable bit for PLX interface control register
 *              pCEI_UINT32 isrEnabled        -- (output) thread handle if successfully defined
 * 
 * RETURN VAL:  ARS_NORMAL        -- Interrupt was initialized successfully
 *              ARS_FAILURE       -- Interrupt was not initialized successfully
 *
 * GLOBALS:     None
 *
 * DESCRIPTION: This routine connects the user-specified interrupt function   
 *              to the interrupt assigned to the respective device.
 *
 *===========================================================================*/
CEI_INT32 cei_init_interrupt (
   CEI_INT32   deviceIndex,
   pCEI_VOID   isr, 
   CEI_INT32   priority,
   pCEI_UINT32 deviceBaseAddress,
   CEI_INT32   memRegion,
   CEI_UINT32  disOffset,
   CEI_UINT32  interruptBit,
   pCEI_UINT32 isrHandle
)
{
   CEI_UINT32 intCtrlStatReg;             /* PLX Int Control/Status Reg data */
   CEI_INT32 status;                      /* local function return status    */
   /* local reference to the low-level global virtual base address array     */
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;

   if (x30InterruptFunctionPtr[deviceIndex] != NULL) { 
      pIsr = x30InterruptFunctionPtr[deviceIndex];
      int_thread_running = 1;
      status = CEI_THREAD_CREATE(&int_thread, priority, cei_lnx_int_function, &deviceIndex);
      if (status != 0) {
         printf("Error: cei_init_interrupt/CEI_THREAD_CREATE - %d\n", status);
         status = ARS_FAILURE;
      }
      else {
         /* wait for thread create to complete */
         usleep(10);

         status = vbtInterruptMode(deviceIndex, 1);
         if (status != 0) {
           printf("Error: cei_init_interrupt/vbtInterruptMode - %d\n", status);
           status = ARS_FAILURE;
         }
       
         /* Set either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
          * Local Interrupt Input Enable, High to pass local bus interrupts
          */
         else if (interruptBit != 0) {
            status = cei_read_32bit (deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
            intCtrlStatReg |= interruptBit;
            status = cei_write_32bit (deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
         }
         else {
            status = ARS_NORMAL;
         }
      }
   }
   
   *isrHandle = 1;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:     C E I _ T E R M I N A T E _ I N T E R R U P T
 *===========================================================================*
 *
 * FUNCTION:    Provides the method to terminate an interrupt association to 
 *              an ISR
 * 
 * PARAMETERS:  deviceIndex  -- (input)  device to reference (0 to 15)
 * 
 * RETURN VAL:  ARS_NORMAL   -- Interrupt was terminated successfully
 *              ARS_FAILURE  -- Interrupt was not terminated successfully
 *
 * GLOBALS:     None
 * 
 * DESCRIPTION: This routine terminates the specified device interrupt function
 * 
 *===========================================================================*/
CEI_INT32 cei_terminate_interrupt
(
    CEI_INT32   deviceIndex, 
    CEI_UINT32  interruptBit,
    pCEI_UINT32 deviceBaseAddress
)
{
   CEI_INT32 status;
   CEI_UINT32 intCtrlStatReg;
   int_thread_running = 0;
   pCEI_UINT32 virtualBaseAddressPtr = deviceBaseAddress;
 
   /* Reset either the PLX PCI9056 or PCI9030 Interrupt Control/Status Register
    * Local Interrupt Input Enable, Low to disable local bus interrupts
    */
   if (interruptBit != 0) {
      status = cei_read_32bit (deviceIndex, virtualBaseAddressPtr, &intCtrlStatReg);
      intCtrlStatReg &= ~(interruptBit);
      status = cei_write_32bit (deviceIndex, virtualBaseAddressPtr, intCtrlStatReg);
   }
   status = vbtInterruptMode ((CEI_UINT)deviceIndex, 0);
   if (status != 0) {
      printf("Error: cei_terminate_interrupt/vbtInterruptMode - %d\n", status);
   }
 
   status = CEI_THREAD_DESTROY (&int_thread);
   if (status != 0) {
      printf("Error: cei_terminate_interrupt/CEI_THREAD_DESTROY - %d\n", status);
      status = ARS_FAILURE;
   }
   else
      status = ARS_NORMAL;
 
   return status;
}


/*===========================================================================*
 * ENTRY POINT:     C E I _ A S S I G N _ D E V I C E _ R E S O U R C E S
 *===========================================================================*
 *
 * FUNCTION:    This routine provides a method to supply the BAR0 and BAR2
 *              virtual address pointers to the base address assigned
 *              to the device acquired from a custom embedded host device driver.
 *
 * PARAMETERS:  CEI_INT16  board    -- (input) board to reference (0 to 15)
 *              CEI_UINT32 bar0Addr -- (input) virtual address for BAR0
 *              CEI_UINT32 bar2Addr -- (input) virtual address for BAR2
 *
 * RETURN VAL:  ARS_INVBOARD  board parm invalid
 * 
 * GLOBALS:     device[board].devicePtr
 * 
 * DESCRIPTION: This routine provides a method for any embedded platform
 *              application to bypass the standard device driver/host interface
 *              to acquire the device memory resources and supply the device
 *              BAR0 and BAR2 virtual base addresses to the API.
 *
 *===========================================================================*/
CEI_VOID cei_assign_device_resources (CEI_INT16 deviceIndex, CEI_UINT32 bar0Addr, CEI_UINT32 bar2Addr)
{
   embeddedResourcesAssigned[deviceIndex] = TRUE;
   assignedResources[deviceIndex].bar0BaseAddress = bar0Addr;
   assignedResources[deviceIndex].bar2BaseAddress = bar2Addr;
}

