/*============================================================================*
 * FILE:                      U C E I P C I _ 1 5 5 3. H
 *============================================================================*
 *
 * COPYRIGHT (C) 2005 - 2011 BY
 *          GE INTELLIGENT PLATFORMS, INC., GOLETA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 * FUNCTION:   uceipci driver subroutines for MIL-STD-1553 devices
 *
 *             Supplements the universal PCI driver for GE Intelligent
 *             Platforms avionics MIL-STD-1553 devices.
 *
 *
 * ENTRY POINTS:
 *    initialize_1553       called by uceipci_probe 
 *
 *    uninitialize_1553     called by uceipci_remove
 * 
 *    intrpt_cntrl_1553     called by intrpt_cntrl
 *
 *    uceipci_1553_isr      interrupt service routine for UCA boards
 *
 *===========================================================================*/

/* $Revision:  1.13 Release $
   Date        Revision
  ----------   ---------------------------------------------------------------
  03/30/2005   initial. bch
  05/25/2006   replaced kill_proc with send_sig_info in uceipci_1553_isr. bch
  09/13/2006   added support for AMC-1553. bch
  02/19/2007   added support for EPMC-1553 and QPCX-1553. modified to support
                kernel changes in 2.6.19. bch
  05/30/2007   added intrpt_cntrl_1553. modified to support multiple processes. bch
  08/20/2007   modified to support kernel changes in 2.6.22. bch
  12/11/2007   added R15-EC. added AMC-1553 to intrpt_cntrl_1553. bch
  11/18/2008   added support for the R15-AMC. modified uceipci_1553_isr to
                support kernel changes in 2.6.26. bch
  02/27/2009   modified initialize_1553 by removing the probing and mapping of 
                BAR regions. modified uninitialize_1553 by removing the
                unmapping of BAR regions. added hardware interrupt status to 
                intrpt_cntrl_1553. modified debug output in uceipci_1553_isr. bch
  07/21/2009   added support for the RXMC-1553. bch	
  10/20/2009   added support for the RPCIe-1553. added attribute bus_type for sysfs.
                modified uceipci_1553_isr to support kernel changes in 2.6.31 - bch
  04/13/2011   added support for the R15-LPCIE and RXMC2-1553. bch
  08/12/2011   modified Endian check. bch
  10/14/2011   modified uceipci_1553_isr for get_pid_task and added lock/unlock. bch  
*/


#include <linux/sched.h>
#ifndef NO_HW_INTERRUPTS
 #include <linux/interrupt.h>
 // PCI, cPCI, PMC
 #define HWR_OFFSET         0x00400000
 #define HWR_CHAN_OFFSET    0x01000000
 // Q104-P, QPCI, QPMC, QcP 
 #define Q_HWR_OFFSET       0x00000800
 #define Q_HWR_CHAN_OFFSET  0x00200000
 #define GLOBAL_REG_OFFSET  0x00200000
 #define CHAN_REG_OFFSET    0x00020000
 #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19))
  static irqreturn_t uceipci_1553_isr(int irq, void* data);
 #else
  static irqreturn_t uceipci_1553_isr(int irq, void* data, struct pt_regs* regs);
 #endif
#endif  
 
#ifndef NO_SYSFS_SUPPORT  
 // setup the attributes for the sysfs
 bd_info_attr(mode, "%d\n")
 bd_info_attr(ch, "%d\n")
 bd_info_attr(irig, "%d\n")
 bd_info_attr(bus_type, "%d\n")
 static struct attribute* bt1553_attrs[] = {
	&dev_attr_boardname.attr,
	&dev_attr_mode.attr,
	&dev_attr_ch.attr,
	&dev_attr_irig.attr,
	&dev_attr_bus_type.attr,
	&dev_attr_id.attr,
	&dev_attr_board_type.attr,
	&dev_attr_status.attr,
	NULL,
 };
 static struct attribute_group cei_1553_attr_group = {
	.name = "board",
	.attrs = bt1553_attrs,
 };
#endif


static int initialize_1553(struct pci_dev* pdev) {
  int i, status=0;
  DEV_DATA* pData;
  unsigned short valW1=0; 
  unsigned short valW2=0; 
  unsigned short valW3=0;

  if((pData = (DEV_DATA*) pci_get_drvdata(pdev)) == NULL) {
    printk(KERN_ERR "uceipci:  failed to get device data (PCI: bus %d, dev %d).\n", pdev->bus->number,(pdev->devfn)>>3); 
    return -EFAULT;
  }
  
  switch(pdev->device) {
  case 0x1008:
  case 0x1553:
  case 0x1554:
  case 0x1555:
  case 0x1556:
  case 0x1557:
  case 0x1558:
  case 0x1559:	
  case 0x155A:  
  case 0x155B:
  case 0x155C:
    // read the CSC from the board's host interface register
    valW1 = readw((unsigned short*)pData->pci_bar_laddr_membase);
    if(uceipci_debug >= 1)
      printk(KERN_INFO "uceipci:  HIR - 0x%x\n", valW1);
    break;
  case 0x1003:
    // configure PLX9056 Local Configuration Register with default values
    writel(0x100C767E, ((u32*)pData->pci_bar_conf_membase) + 27); // CNTRL register
    writel(0x42030343, ((u32*)pData->pci_bar_conf_membase) + 6);  // LBRD0 register
    // read Hardware Capabilities register
    valW1 = readw((u16*)(pData->pci_bar_laddr_membase + 0x200004));
    if(uceipci_debug >= 1)
      printk(KERN_INFO "uceipci:  HCAP - 0x%x\n", valW1);
    // read Hardware ID register
    valW2 = readw((u16*)(pData->pci_bar_laddr_membase + 0x200000));
    if(uceipci_debug >= 1)
      printk(KERN_INFO "uceipci:  HWID - 0x%x\n", valW2);
  };

  // read/set board configuration
  switch(pdev->device) {
  case 0x1553:
    if(pci_resource_len(pdev, 0) >= 0x2000000) {  // board is a PCI-1553, cPCI-1553, or PMC-1553 (BAR0-32MB)
      valW1 = valW1 >> 8;
      pData->board_type = (valW1 & 0xC0);
      if(valW1 | 0x3)  // CH 1 and CH 2 multiple mode
        pData->mode = 1;
      pData->ch++;
      if(valW1 & 0x10)  // CH 1 present
        pData->ch++;	
      if(valW1 & 0x20)  // CH 2 present 
        pData->ch++;	
      pData->id = 0x40;
      pData->irig = 0;
      pData->bus_type = 2;
      pData->status = 0;
      strcpy(pData->boardname, "PCI-1553, cPCI-1553 or PMC-1553");
     #ifndef NO_HW_INTERRUPTS
      for(i=0;i<pData->ch;i++)
        pData->hwr_chan_addr[i] = (pData->pci_bar_laddr_membase + HWR_OFFSET + (i*HWR_CHAN_OFFSET));
     #endif
    }
    else if(pci_resource_len(pdev, 0) < 0x2000000) {
      pData->board_type = ((valW1 & 0x1F) >> 1);
      switch(pData->board_type) {
      case 1:  // QPMC-1553 (BAR0-8MB)
        pData->id = 0x110;
        strcpy(pData->boardname, "QPMC-1553");
        break;
      case 5:  // Q104-1553P (BAR0-8MB)
        pData->id = 0x180;
        strcpy(pData->boardname, "Q104-1553P");
        break;
      default:
        strcpy(pData->boardname, "unsupported");
        printk(KERN_ERR "uceipci:  invalid board type - %d.\n", pData->board_type);
        break;
      };
      pData->ch = (valW1 & 0x3E0) >> 6;	
      pData->irig = (valW1 & 0x1000) >> 12;
      pData->mode = (valW1 & 0x800) >> 11;	
      pData->bus_type = 2;
      pData->status = 0;
     #ifndef NO_HW_INTERRUPTS  
      for(i=0;i<pData->ch;i++)  
        pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
     #endif
    }
    break;
  case 0x1554:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    switch(pData->board_type) {
    case 3:  // QPCI-1553 (BAR0-12b, BAR2-8MB)
      pData->id = 0x160;
      strcpy(pData->boardname, "QPCI-1553");
      break;
    default:
      strcpy(pData->boardname, "unsupported");
      printk(KERN_ERR "uceipci:  invalid board type - %d.\n", pData->board_type);
      break;
    };
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x1555:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    switch(pData->board_type) {
    case 7:  // QCP-1553 (BAR0-512b, BAR1-256b, BAR2-8MB)	      
      pData->id = 0x210;
      strcpy(pData->boardname, "QCP-1553");
      break;
    default:
      strcpy(pData->boardname, "unsupported");
      printk(KERN_ERR "uceipci:  invalid board type - %d.\n", pData->board_type);
      break;
    };
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x1008:  // AMC-1553 (BAR2-8MB)
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x110;
    strcpy(pData->boardname, "AMC-1553");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;	
    pData->bus_type = 2;
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x1003:
    strcpy(pData->boardname, "EPMC-1553");
    pData->board_type = pdev->device;
    pData->id = pdev->device;
    pData->ch = 0;
    for(i=0;i<8;i++)
      if(valW1 & (0x1<<i)) pData->ch++;  // if((valW1 >> i) & 0x1) pData->ch++;
    pData->irig = ((valW1 >> 9) & 0x1);
    pData->mode = ((valW2 >> 12) & 0x1);
    pData->bus_type = 2;
    pData->status = 0;
    break;
  case 0x1556:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x220;
    strcpy(pData->boardname, "QPCX-1553");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;	
    pData->bus_type = 2;
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x1557:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x230;
    strcpy(pData->boardname, "R15-EC");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x1558:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x240;
    strcpy(pData->boardname, "R15-AMC");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x1559:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x260;
    strcpy(pData->boardname, "RXMC-1553");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x155A:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x250;
    strcpy(pData->boardname, "RPCIe-1553");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x155B:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x320;
    strcpy(pData->boardname, "R15-LPCIe");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  case 0x155C:
    pData->board_type = ((valW1 & 0x1F) >> 1);
    pData->id = 0x300;
    strcpy(pData->boardname, "RXMC2-1553");
    pData->ch = (valW1 & 0x3E0) >> 6;	
    pData->irig = (valW1 & 0x1000) >> 12;
    pData->mode = (valW1 & 0x800) >> 11;
    pData->bus_type = 2;	
    pData->status = 0;
   #ifndef NO_HW_INTERRUPTS  
    for(i=0;i<pData->ch;i++)  
      pData->hwr_chan_addr[i] = pData->pci_bar_laddr_membase + Q_HWR_OFFSET + (i*Q_HWR_CHAN_OFFSET);
   #endif
    break;
  default:
    strcpy(pData->boardname, "unsupported");
    printk(KERN_ERR "uceipci:  invalid PCI device ID - 0x%x.\n", pdev->device);
    pci_set_drvdata(pdev, pData);
    return -EFAULT;
  };

  if(pdev->device != 0x1003) {
    if(uceipci_debug >= 1) {
      // display the ACR from the board's host interface register
      if((pData->id == 0x110) || (pData->id == 0x220))
        printk(KERN_INFO "uceipci:  ACR - 0x%X\n", readw(((unsigned short*)pData->pci_bar_laddr_membase)+1) & 0xF);
      printk(KERN_INFO "uceipci:  WCS_ver: 0x%X, LPU_ver: 0x%X, LPU_build: 0x%X\n", readw(((unsigned short*)pData->pci_bar_laddr_membase)+0x18) & 0xFFFF, readw(((unsigned short*)pData->pci_bar_laddr_membase)+0x9) & 0xFFFF, readw(((unsigned short*)pData->pci_bar_laddr_membase)+0x1A) & 0xFFFF);
    }
    // determine Endian
    if(uceipci_debug >= 2) {
      valW1 = readw((unsigned short*)(pData->pci_bar_laddr_membase + 8));
      writew(0x2A15, (unsigned short*)(pData->pci_bar_laddr_membase + 8));
      valW2 = readw((unsigned short*)(pData->pci_bar_laddr_membase + 8));
      valW3 = *((unsigned short*)(pData->pci_bar_laddr_membase + 8));
      writew(valW1, (unsigned short*)(pData->pci_bar_laddr_membase + 8));
      if(valW2 == valW3) 
        printk("uceipci:  Endian - Little\n");
      else if(valW2 == (((valW3&0xFF)<<8)|(valW3>>8)))
        printk("uceipci:  Endian - Big\n");
      else
        printk("uceipci:  Endian - unknown\n");
    }
  }

  pci_set_drvdata(pdev, pData);

  // default, enable wait queue for all boards
  pData->status |= STATUS_HWINT_WQ;

 #ifndef NO_SYSFS_SUPPORT  
  if((status=sysfs_create_group(&pdev->dev.kobj, &cei_1553_attr_group)) != 0)
    printk(KERN_ERR "uceipci(%d):  failed to create sysfs group, status %d\n", pData->minor, status);
 #endif

  return status;
}


static int uninitialize_1553(struct pci_dev* pdev) {
 #ifndef NO_SYSFS_SUPPORT  
  sysfs_remove_group(&pdev->dev.kobj, &cei_1553_attr_group);
 #endif

  return 0;
}

#ifndef NO_HW_INTERRUPTS
static int intrpt_cntrl_1553(DEV_DATA* pData, int mode) {
  int i, status=0;
  unsigned short wVal=0;
  int device = ucei_pdev[pData->minor]->device;

  switch(device) {
  case 0x1008:
  case 0x1553:
  case 0x1554:
  case 0x1555:
  case 0x1556:
  case 0x1557:
  case 0x1558:
  case 0x1559:
  case 0x155A:
  case 0x155B:
  case 0x155C:
    // control each available channel for hardware interrupts
    for(i=0;i<pData->ch;i++) {
      wVal = readw((unsigned short*)pData->hwr_chan_addr[i]);
      writew(0, (unsigned short*)(pData->hwr_chan_addr[i] + 0x16));  // clear register
      if(mode == 0) // disable
        writew(wVal & ~0x4000, (unsigned short*)(pData->hwr_chan_addr[i]));  // disable
      else if(mode == 1)  // enable
        writew(wVal | 0x4000, (unsigned short*)(pData->hwr_chan_addr[i]));
    }
    break;
  case 0x1003:
    // disable hardware interrupts on all channels (Interrupt Mask Register)
    writeb(0, (unsigned short*)(pData->pci_bar_laddr_membase + GLOBAL_REG_OFFSET + 0x14));
    // clear hardware interrupts on all channels
    for(i=0;i<pData->ch;i++) 
      writeb(0, (unsigned short*)(pData->pci_bar_laddr_membase + (CHAN_REG_OFFSET * i) + 0x4));
    break;
  default:
    return -EINVAL;
  };

  if(mode == 0) {
    free_irq(pData->irq, (void*)pData);  // returns when all executing interrupts are done
    pData->status &= ~STATUS_INT_ENABLE;
    if(uceipci_debug >= 2)
      printk(KERN_INFO "uceipci(%d):  unregistered interrupt handler IRQ (%d), dev_id (0x%lx).\n",pData->minor, pData->irq, (unsigned long) pData);
   #ifdef HW_INTERRUPTS_WAITQUEUE
    if(pData->status & STATUS_HWINT_WQ) {
      for(i=0;i<pData->ch;i++) {
        pData->cur_intrpt[i]++; 
        wake_up_interruptible(&(pData->isr_wait_q[i]));  // wake the user process held in uceipci_read
      }
    }
   #endif
   #ifdef HW_INTERRUPTS_SIGNAL
    if(pData->status & STATUS_HWINT_SIGNAL) {
      pData->status &= ~STATUS_HWINT_SIGNAL;
     #ifdef HW_INTERRUPTS_WAITQUEUE
      pData->status |= STATUS_HWINT_WQ;  // re-enable wait queue
     #endif
    }
   #endif
  }
  else if(mode == 1) {
   #ifdef HW_INTERRUPTS_WAITQUEUE
    if(pData->status & STATUS_HWINT_WQ) {
      for(i=0;i<pData->ch;i++)
        pData->last_intrpt[i] = pData->cur_intrpt[i] = 0;
    }
   #endif
   #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22))
    if((status = request_irq(pData->irq, uceipci_1553_isr, IRQF_DISABLED|IRQF_SHARED, DRIVER_NAME, (void*)pData)) != 0) {
   #else
    if((status = request_irq(pData->irq, uceipci_1553_isr, SA_INTERRUPT|SA_SHIRQ, DRIVER_NAME, (void*)pData)) != 0) {
   #endif
      printk(KERN_ERR "uceipci(%d):  unable to register interrupt handler IRQ (%d), status (%d).\n",pData->minor, pData->irq, status);
      return -1;
    }
    else {
      if(uceipci_debug >= 2)
        printk(KERN_INFO "uceipci(%d):  registered interrupt handler IRQ (%d), dev_id (0x%lx).\n",pData->minor, pData->irq, (unsigned long) pData);
    }
    pData->status |= STATUS_INT_ENABLE;
  }

  return 0;
}
#endif

#ifndef NO_HW_INTERRUPTS 
// Handle the hardware interrupts generated by the board
// Notes:
//  1. when an interrupt is detected on any channel, all IDs attached to the
//      hardware interrupt will receive a signal. So the user application will
//      need to determine which channel to check for
//  2. if using kernel 2.6.21 or later with CONFIG_DEBUG_SHIRQ defined, and
//      IRQF_SHARED flag is set, then the ISR will be call (no PCI assertion)
//      when requesting and freeing the interrupt line. 
#if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19))
static irqreturn_t uceipci_1553_isr(int irq, void* data) {
#else
static irqreturn_t uceipci_1553_isr(int irq, void* data, struct pt_regs* regs) {
#endif   
   int i, int_count=0;
   unsigned short cntrl_reg=0;
   DEV_DATA* pData=(DEV_DATA*)data;
  #ifdef HW_INTERRUPTS_SIGNAL
   int indx_mask, indx_id=0;
   ID_DATA* pId;
   struct siginfo siginfo;
   struct task_struct* ts;
  #endif

   if(pData == NULL) {
     if(uceipci_debug >= 2)
       printk(KERN_ERR "uceipci:  uceipci_1553_isr - no data pointer, IRQ(%d)\n", irq);
     return IRQ_NONE;
   }

   if(uceipci_debug >= 2)
     printk(KERN_INFO "uceipci(%d): uceipci_1553_isr\n", pData->minor);

   // return if hardware interrupts are not enabled
   if(!(pData->status & STATUS_INT_ENABLE)) {
    #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)) && defined (CONFIG_DEBUG_SHIRQ)
     if(uceipci_debug >= 2)
       printk(KERN_INFO "uceipci(%d):  ISR called before interrupts enabled - IRQ (%d), could be the kernel check for a valid ISR\n", pData->minor, irq);
     return IRQ_HANDLED;
    #endif
     if(uceipci_debug >= 2)
       printk(KERN_ERR "uceipci(%d):  received an interrupt before interrupts enabled - IRQ (%d), possible shared IRQ\n", pData->minor, irq);
     return IRQ_NONE;
   }
   
   if(pData->irq != irq) {
     if(uceipci_debug >= 2)
       printk(KERN_ERR "uceipci(%d):  IRQ mismatch (%d, %d)\n", pData->minor, pData->irq, irq);
     return IRQ_NONE;
   }

  #ifdef HW_INTERRUPTS_SIGNAL
   siginfo.si_code = SI_QUEUE;
   siginfo.si_errno = 0;
   siginfo.si_signo = pData->signal;  // set by the user
  #endif

   switch(pData->id) {
//     case 0x1001:
   case 0x1003:  // EPMC-1553
     // read the Interrupt Status Register (global register) instead of each channel register
     cntrl_reg = readw((unsigned short*)(pData->pci_bar_laddr_membase + GLOBAL_REG_OFFSET + 0x12));
     for(i=0;i<pData->ch;i++) {
       // check for interrupt per channel
       if((cntrl_reg >> i) & 0x1) {
         // clear the interrupt in the channel register
         writew(0, (unsigned short*)(pData->pci_bar_laddr_membase + (CHAN_REG_OFFSET * i) + 0x4));
         if(uceipci_debug >= 3)
           printk(KERN_INFO "uceipci(%d):  valid interrupt - IRQ (%d), chan (%d), cntrl_reg (0x%x).\n", pData->minor, pData->irq, i, cntrl_reg);
         int_count++;
        #ifdef HW_INTERRUPTS_SIGNAL
         // check if signals have been enabled
	 if((pData->status & STATUS_HWINT_SIGNAL) == 0)
           continue;
         indx_mask = 1;
         indx_id = 0;
         // checks for IDs marked to receive signal for hardware interrupts
         while(indx_mask <= pData->ids_mask_intrpt) {
           pId = &(pData->ids[indx_id]);
           if(pId->sigval == -1)  // send the devnum (device and channel)
             siginfo.si_int = (pData->minor*8) + i;
           else  // user provided
             siginfo.si_int = pId->sigval;
           // check if the ID is marked to receive signals and has a valid PID
           if((indx_mask & pData->ids_mask_intrpt) && (pId->sigpid > 1)) {
            #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39))
	     ts = get_pid_task(find_vpid(pId->sigpid), PIDTYPE_PID);
            #else
	     rcu_read_lock();
            #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
	     ts = pid_task(find_vpid(pId->sigpid), PIDTYPE_PID);
            #else
             ts = find_task_by_pid(pId->sigpid); 
            #endif
	     rcu_read_unlock();
            #endif
             if(!ts)
               continue; // -ESRCH;
             send_sig_info(pData->signal, &siginfo, ts);
             if(uceipci_debug >= 2)
               printk(KERN_INFO "uceipci(%d):  signal(%d) to PID(%d) with val(%d)\n", pData->minor, siginfo.si_signo, pId->sigpid, siginfo.si_int);
           }
           indx_mask = (1<<++indx_id);
         };
        #endif
       }
     }
     break;
   default:  // UCA boards
     for(i=0;i<pData->ch;i++) {
       // read the 1553_control_register on channel
       cntrl_reg = readw((unsigned short*)pData->hwr_chan_addr[i]);
       // check for interrupt
       if(cntrl_reg & 0x200) {
         // write a 0 to the write_interrupt_bit
         writew(0, (unsigned short*)(pData->hwr_chan_addr[i] + 0x16));
         if(uceipci_debug >= 3)
           printk(KERN_INFO "uceipci(%d):  valid interrupt - IRQ (%d), chan (%d), cntrl_reg (0x%x).\n", pData->minor, pData->irq, i, cntrl_reg);
         int_count++;
        #ifdef HW_INTERRUPTS_WAITQUEUE
         if(pData->status & STATUS_HWINT_WQ) {
           pData->cur_intrpt[i]++;
           wake_up_interruptible(&(pData->isr_wait_q[i]));  // wake any process/thread waiting to be notified that a hardware interrupt occured
           if(uceipci_debug >= 2)
             printk(KERN_INFO "uceipci(%d):  wake wait queue (%d) - count (%d)\n", pData->minor, i, pData->cur_intrpt[i]);
	   continue;
	 }
        #endif
        #ifdef HW_INTERRUPTS_SIGNAL
         if((pData->status & STATUS_HWINT_SIGNAL) == 0)
           continue;
         indx_mask = 1;
         indx_id = 0;
         // checks for IDs marked to receive signal for hardware interrupts
         while(indx_mask <= pData->ids_mask_intrpt) {
           pId = &(pData->ids[indx_id]);
           if(pId->sigval == -1)
             siginfo.si_int = pData->minor;  // send the device number
           else
             siginfo.si_int = pId->sigval;  // user provided
           // check if the ID is marked to receive signals and has a valid PID
           if((indx_mask & pData->ids_mask_intrpt) && (pId->sigpid > 1)) {
            #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39))
	     ts = get_pid_task(find_vpid(pId->sigpid), PIDTYPE_PID);
            #else
	     rcu_read_lock();
            #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
	     ts = pid_task(find_vpid(pId->sigpid), PIDTYPE_PID);
            #else
             ts = find_task_by_pid(pId->sigpid); 
            #endif
	     rcu_read_unlock();
            #endif
             if(!ts)
               continue; // -ESRCH;
             send_sig_info(pData->signal, &siginfo, ts);
             if(uceipci_debug >= 2)
               printk(KERN_INFO "uceipci(%d):  signal(%d) to PID(%d) with val(%d)\n", pData->minor, siginfo.si_signo, pId->sigpid, siginfo.si_int);
           }
           indx_mask = (1<<++indx_id);
         };
        #endif
       }
     }
   };
 
   if(int_count == 0) {
     if(uceipci_debug >= 3) {
      #if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)) && defined (CONFIG_DEBUG_SHIRQ)
       printk(KERN_ERR "uceipci(%d):  spurious interrupt - IRQ (%d,%d), could be the kernel check for a valid ISR\n",pData->minor,irq,pData->irq);
      #else
       printk(KERN_ERR "uceipci(%d):  spurious interrupt - IRQ (%d,%d)\n",pData->minor,irq,pData->irq);
      #endif
     }
     return IRQ_NONE;
   }

   return IRQ_HANDLED;
}
#endif
