/*===========================================================================*
 * FILE:                     C D E V _ A P I . C                    
 *===========================================================================*
 *
 * COPYRIGHT (C) 2002-2011 BY
 *          GE INTELLIGENT PLATFORMS, INC., SANTA BARBARA, CALIFORNIA
 *          ALL RIGHTS RESERVED.
 *
 *          THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND
 *          COPIED ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH
 *          THE INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY
 *          OTHER COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE
 *          AVAILABLE TO ANY OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF THE
 *          SOFTWARE IS HEREBY TRANSFERRED.
 *
 *          THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT
 *          NOTICE AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY GE 
 *          INTELLIGENT PLATFORMS.
 *
 *===========================================================================*
 *
 * FUNCTION:    Application Programmer's Interface for a CEI-x30 ARINC device.
 *
 * DESCRIPTION: These routines are provided to demonstrate the basic
 *              programming primitives for th CEI-x30 API using the C
 *              language.  Each routine utilizes a particular function; 
 *              however, a user application program may need to combine
 *              many of these functions in different ways.  The user may wish 
 *              to combine the function of several of these routines into
 *              larger routines or macros for greater efficiency.
 *
 *              All routines utilize a board number argument, based on the 
 *              Device ID/Board Number assigned during the installation of 
 *              API and device driver.
 *
 * EXTERNALLY DEFINED ROUTINE ENTRY POINTS:         
 *
 *    CEI-x30 ARINC API routines:
 *
 *    ar_board_test               - init-time internal functionality test
 *    ar_bypass_wrap_test         - controls execution of init wrap test
 *    ar_close                    - relinquishes WinRT device resources
 *    ar_define_msg               - defines a message for scheduled transmission
 *    ar_define_msg_block         - defines a block of messages for scheduled transmission
 *    ar_enh_label_filter         - enable/disable enhanced label filtering on the device
 *    ar_execute_bit              - performs BIT functionality testing of the device
 *    ar_get_base_addr            - returns the virtual address of the device
 *    ar_getblock                 - read a block of messages from a receiver FIFO
 *    ar_getblock_t               - read a block of messages from a receiver FIFO
 *    ar_getfilter                - retrieve the device filter table contents
 *    ar_get_boardname            - returns a device description
 *    ar_get_boardnameLV          - returns the target board name string
 *    ar_get_boardtype            - returns the target board type
 *    ar_get_config               - returns 429 config register bit states
 *    ar_get_device_config        - returns ARINC 429 & I/O chan attributes  
 *    ar_get_573_config           - returns 573 config register bit states
 *    ar_get_data                 - retrieves data and time-tag from a receive FIFO
 *    ar_get_data_xt              - retrieves data and time-tag from a receive FIFO
 *    ar_get_error                - returns error string for a status code
 *    ar_get_429_message          - retrieves an ARINC 429 message
 *    ar_get_573_frame            - retrieves an ARINC 573 data frame
 *    ar_get_label_filter         - returns label filter state on all rx chans
 *    ar_get_latest               - retrieves latest data from the snapshot buffer
 *    ar_get_latest_t             - retrieves latest data and time-tag from the snapshot buffer
 *    ar_get_snap_data            - retrieves latest data from the snapshot buffer
 *    ar_get_status               - returns FIFO Data Avail status for all rx
 *    ar_get_storage_mode         - returns current receiver storage mode
 *    ar_get_time                 - returns the current 64-bit device timer value
 *    ar_get_timercntl            - returns a 32-bit rep. of current device timer value
 *    ar_get_transmitter_mode     - returns current 573/CSDB transmitter mode
 *    ar_getnext                  - wait for next data word from a receiver
 *    ar_getnextt                 - wait for next data, return with 32-bit timetag
 *    ar_getnext_xt               - wait for next data, return with 64-bit timetag
 *    ar_getword                  - read a single message data from a receiver FIFO
 *    ar_getwordt                 - read data and 32-bit timetag from a receiver FIFO
 *    ar_getword_xt               - read data and 64-bit timetag from a receiver FIFO
 *    ar_go                       - enable data processing all channels on the board 
 *    ar_hw_interrupt_buffer_read - reads the API h/w interrupt buffer
 *    ar_initialize_api           - initializes the device API, downloads firmware
 *    ar_initialize_device        - initializes the device
 *    ar_interrupt_queue_read     - reads the device interrupt queue
 *    ar_label_filter             - defines data filtering via ARINC label
 *    ar_loadlsv                  - initializes the API and device
 *    ar_modify_msg               - modifies a message value (scheduled transmission)
 *    ar_modify_msg_block         - modifies a block of message values (scheduled transmission)
 *    ar_num_rchans               - returns the no. of recv channels
 *    ar_num_xchans               - returns the no. of xmit channels
 *    ar_put_573_frame            - transmit an ARINC 573 frame
 *    ar_put_429_message          - transmit an ARINC 429 message
 *    ar_putblock                 - write a block of messages to a transmitter FIFO
 *    ar_putfilter                - define the device filter table content
 *    ar_putword                  - send an ARINC word on a channel   
 *    ar_query_device             - queries the PCI IDs and returns the board type
 *    ar_reset                    - reset the device to an initial state
 *    ar_set_config               - configure using x20-generic 429 def's
 *    ar_set_device_config        - set ARINC 429 and I/O chan attributes  
 *    ar_set_573_config           - set ARINC 573-specific chan attributes  
 *    ar_set_isr_function         - reassigns the default ISR to host app
 *    ar_set_multithread_protect  - turn on specific hardware access protection
 *    ar_set_raw_mode             - enables/disables parity for a channel
 *    ar_set_storage_mode         - select buffered/merged receive modes
 *    ar_set_time                 - sets the current 64-bit device timer value
 *    ar_set_timerrate            - sets an x20-compatible device timer resolution
 *    ar_set_transmitter_mode     - set ARINC 573/CSDB encoding (BPRZ vs. HBP)
 *    ar_sleep                    - suspends call thread by milliseconds
 *    ar_stop                     - disable the device via the global-enable
 *    ar_version                  - get the API version string.
 *    ar_wait                     - delay a specified number of seconds
 *
 *---------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date    By   Comments
 * -------- ----- ------------------------------------------------------------
 * 01/03/04  BDW  Taken from CEI-715, modified to add support for the CEI-x30 
 *                enhanced features and hardware interface
 * 10/06/04  BDW  Added compiler directives to allow 32-bit integer processing
 *                for VxWorks BSPs and LabVIEW R/T not supporting 64-bit ints
 * 11/09/04  BDW  Modified memory_test(), performing an optional full 
 *                SRAM memory test and unused SRAM memory test
 * 12/14/04  BDW  Added the routine ar_getblock_t() to receive blocks of
 *                ARINC 429 data with 64-bit time-tag support
 * 12/23/04  BDW  Added the routine ar_putblock_multi_chan() for transmission
 *                of multiple one-shot messages over multiple channels
 * 01/13/05  BDW  Modified ar_get_status() to support status reporting for
 *                up to 16 CEI-830 individual receiver FIFO's.
 * 02/17/05  BDW  Modified logic in a429_utl_store_message_scheduler_entry() to disable
 *                the message first, write all member values & write rate last
 * 07/26/05  BDW  Removed SDI pre-filtering logic attempt to replicate x20 SDI
 *                   filtering, relying only on x30 Enhanced Label Filtering
 *                Corrected a bug in ar_define_msg_block
 *                Added "Read FIFO Count" to ar_get_device_config & _573_config
 *                Removed the XC2V1000 FPGA load capabilities
 * 09/14/05  BDW  Added support for Linux via inclusion of CDEV_LNX.H
 *                Included relocating some compiler directive defines to 
 *                CDEV_API.H
 * 10/21/05  BDW  Added support for the CEI-430 firmware load:        
 *                 - renamed programX30device() to program_xilinx_9056()
 *                 - added program_xilinx_9030 to support 430 FPGA loading
 *                 - modified ar_initialize_api() to test PLX Local Config
 *                   Register to detect target board as either 830 or 430
 *                 - added differential "discrete" input channel type logging
 *                   to a429_utl_determine_channel_support()
 * 11/22/05  BDW  Added more support for the CEI-430:
 *                - Added support for the Differential In/Out Channel Types -
 *                  ena/disable and state set/get to ar_get/set_device_config()
 *                - Added support for the Differential In/Out Channel Types 
 *                - Incorp initialize_enh_channels() into ar_initialize_device()
 *                - Added CEI430 board type support to ar_get_boardtype()
 * 11/28/05  BDW  Added support for the Linux and Integrity operating systems
 * 12/06/05  BDW  Added channel count report options for AR_Get_Device_Config
 *                   ARU_CHAN_COUNT_429        448
 *                   ARU_CHAN_COUNT_573        449
 *                   ARU_CHAN_COUNT_DISC       450
 *                   ARU_CHAN_COUNT_DIFF       451
 * 08/25/06  BDW  Modified ar_getblock & ar_getblock_t to only access the 
 *                   board for time-tags when timetag parameter is not null.
 *                Fixed CEI-430/CEI-830 config check bug in ar_initialize_api.
 *                Added ARU_TX_FIFO_COUNT option validation to the routine
 *                   validate_channel_request().
 *                Added the compiler directive TARGET_EMBEDDED to remove
 *                   execution of parameter checking for an embedded target
 *                Added the routine ar_modify_msg_block() to support the
 *                   modification of message values on a previously defined 
 *                   block of scheduled messages.
 *                Incorporated CEI_TYPES.H data type definitions in support of
 *                   64-bit Linux operating systems - this modified every
 *                   data declaration and prototype in the API.
 *                Updated the CEI-430 firmware version to 2.0
 *                Updated the CEI-830 firmware version to 4.3
 *                Added support for the AMC-A30
 *                   Added the CEI-A30 firmware declaration
 *                   Added the CEI-A30 download routine program_xilinx_9056_32bit
 *                   Added AMC-A30 board type support to ar_get_boardtype()
 *                Swapped the BAR reference order by the device structure 
 *                   members baseAddr and devicePtr
 * 02/22/07  BDW  Added invocation of cei_acquire_device_identity() to positively ID
 *                   x30 products based on PCI Manufacturer and Device IDs 
 * 05/11/07  BDW  Added support for the RAR-EC, RAR-CPCI, CEI-530, and R830RX
 *                   Modified the parameter lists for the f/w load routines 
 *                   so the f/w load arrays are passed to the resp. routine:
 *                     - program_xilinx_9030
 *                     - program_xilinx_9056
 *                     - programXilinx905632Bit
 *                   Added declarations for RAR-EC, CEI-530 & R830RX f/w loads
 *                   Added processing for RAR-EC, CEI-530 & R830RX in 
 *                      ar_initialize_api
 *                Added processing in the f/w load routines to handle compressed
 *                   f/w array definitions; old uncompressed loading requires
 *                   definition of the compiler directive UNCOMPRESSED_FW.
 * 07/25/07  BDW  Support for R830RX ARU_IRIG_OUTPUT_ENABLE option added
 *                   to ar_set_device_config/ar_get_device_config
 * 10/15/07  BDW  Added the RAR-EC firmware download routine and invocation of
 *                   program_xilinx_9030_32bit, from ar_initialize_api
 * 12/07/07  BDW  Added the RAR-CPCI firmware download case to ar_initialize_api
 *                Moved the firmware array declarations and header include
 *                   statements to a new header file cdev_fw.h.
 * 02/18/08  BDW  Moved the data structure array channelIdent[] structure 
 *                   members to the ARINC_DEVICE_TYPE struture array device[]
 * 02/19/08  BDW  Added RAR-EC board type reporting to ar_query_device and
 *                   ar_get_boardtype
 * 02/25/08  BDW  Added the option ARU_CONC_MULTIPROC to AR_ON/AR_OFF as a
 *                valid value for the ar_set_multithread_protect routine state 
 *                parameter when concurrent mulitple processes accessing the 
 *                card are desired.
 * 03/10/08  BDW  Moved the source for the following routines from CDEVTIME.C
 *                and LVRTTIME.C and removed both files from the distribution:
 *                   acquire_formatted_time_tag  (API local)
 *                   is_irig_calibrated          (API local)
 *                   ar_convert_time_to_string
 *                   ar_get_time
 *                   ar_get_data
 *                   ar_set_time
 *                Moved all API local data arrays from CDEV_GLB.H to this file.
 * 03/19/08  BDW  Modified the prototype for ar_get_boardnameLV & implemented it.
 *                Added the internal board name string array, boardNames[].
 * 03/26/08  BDW  Modified the FPGA programming routines to default to the
 *                uncompressed firmware build.  The compiliation must now define 
 *                the directive COMPRESSED_FW to use compressed firmware download 
 *                procedure.
 * 06/02/08  BDW  Modified the multi-process counter logic in ar_loadslv().
 * 11/07/08  BDW  Removed multi-process semaphore acquisition in ar_loadslv().
 *                Replaced the VXWORKS compilation directive with one for
 *                each VxWorks processor type - VXW_PCI_X86 & VXW_PCI_PPC;
 *                with other target specific directives now defined in the 
 *                header file target_defines.h instead of via build settings.
 * 01/16/09  BDW  Corrected an invalid access of deviceGlobal[board] when
 *                ar_close is invoked prior to ever invoking ar_loadslv.
 * 02/02/09  BDW  Modified ar_error.h array references.
 * 04/15/09  BDW  Incorporated support for the CEI-x30 Enhanced Firmware Design.
 *                - Modified all methods to access Channel Registers and Buffers,
 *                  Message Schedule Table, Label Filter Table, and Snapshot Buffer.
 *                - Added Interrupt Queue access routines and generic interrupt
 *                  handler (with provisions for designating a user-defined handler):
 *                      cei_utl_interrupt_queue_read
 *                      cei_utl_interrupt_handler
 *                      ar_hw_interrupt_buffer_read
 *                      ar_interrupt_queue_read
 *                      ar_set_isr_function
 *                - Added the routine ar_get_latest_t to provide a method to 
 *                  retrieve the label-only snapshot mode message & time-stamp
 *                - Implemented the functional support within the following API 
 *                  routines for backward compatibility to the CEI-x20 product 
 *                  line/API:
 *                     ar_get_boardname     ar_init_dual_port
 *                     ar_get_label_filter  ar_init_slave
 *                     ar_get_timercnt      ar_int_control
 *                     ar_sleep
 *                - Added support for the ARS_BAD_MESSAGE return status from 
 *                  the following API routines, indicating a "bad msg length" 
 *                  ARINC 429 message was encountered on the respective receiver:
 *                     ar_getword       ar_getwordt       ar_getword_xt
 *                     ar_getnext       ar_getnextt       ar_getnext_xt
 *                     ar_getblock      ar_getblock_t     ar_get_429_message
 *                     ar_get_data      ar_get_data_xt
 * 06/16/09  BDW  Replaced call to cei_open_shared_lock in ar_set_preload_config
 *                   with the new function called cei_open_preload_lock.
 *                Removed support for compressed firmware download.
 * 06/23/09  BDW  Added a test for the Enhanced CEI-x30 Interface to 
 *                ar_get_device_config and the invocation to update any 
 *                standard device to the enhanced i/f in ar_set_device_config.
 *                Added the utility routines to support the EEPROM update.
 *                Renamed all local subroutines to have an "a429_utl_" prefix.
 * 09/08/09  BDW  Fixed merged mode rx channel reporting issue in ar_getblock_t().
 * 09/30/09  BDW  Added support for ar_set_base_addresses routine for the host
 *                to assign device BAR0 and BAR2 resources for embedded linux.
 * 05/19/10  BDW  Added support for the RAR-PCIE to the following items:
 *                   ar_initialize_api (bypass f/w download, copy BAR0 ptr to BAR2 ptr)
 *                   ar_query_device (identify RAR-PCIE board type)
 *                   boardNames array adds next board type name RAR-PCIE
 * 09/16/10  BDW  Renamed interruptFunctionPtr to x30InterruptFunctionPtr
 *                to eliminate duplicate name with the PIO product line API
 * 11/01/10  BDW  Modified the return type for ar_get_base_addr to be a pointer
 * 01/05/11  BDW  Corrected slew rate assignment for custom baud rate selections
 *                   in ar_set_device_config.
 * 02/01/11  BDW  Modified the operation of the routines ar_modify_msg_block and 
 *                a429_utl_store_message_scheduler_entry to allow a message 
 *                scheduler table entry's transmit count to be modified if
 *                it had been created with or previously modified to have a
 *                limited transmit count and the current transmit count is zero.
 * 02/15/11  BDW  Corrected the CEI-x20 rate and start offset "programmable timer"
 *                   resolution conversion in a429_utl_store_message_scheduler_entry.
 *                Corrected the snapshot buffer table content returned from
 *                   ar_get_latest when the entire table is requested
 *                   using the label parameter ARU_ALL_LABELS option.
 *                Corrected the function of ar_init_slave so it only
 *                   resets/flushes the transmit and receive buffers and
 *                   does not initialize the snapshot buffer, label filter
 *                   table, or message scheduler table (more x20-like).
 * 04/05/11  BDW  Replaced LABVIEW_RT directive use for PLX accesses
 *                   for LabVIEW Real-Time o/s with the directive NIVISA_PXI.
 *                Added support for creating a 1 second host time format,
 *                   AR_TIMETAG_HOST_SEC_32BIT in ar_convert_time_to_string
 *                Corrected an issue in ar_getblock_t, where in merged mode
 *                   the actual receive channel isn't returned unless time-tags
 *                   are returned (time-tag array pointers are not NULL)
 * 04/29/11  BDW  Replaced use of cei_read/write_32bit function invocations
 *                   in the utl_plx_eeprom_* routines with the respective
 *                   READ/WRITE_32_BIT macros to support PROM updates under
 *                   LabVIEW Real-Time.
 * 08/15/11  BDW  Introduced support for CEI-x20 discrete I/O set/get options
 *                   ARU_DISCRETE_VALUES, ARU_DISCRETE_INPUTS and
 *                   ARU_DISCRETE_OUTPUTS in ar_set_config and ar_get_config.
 * 09/16/11  BDW  Added CEI-430A support with additional boardNames element
 *                   and modifications to the following functions:
 *                   a429_utl_enh & a429_utl_enh_check (formality, not necessary)
 *                   ar_initialize_api to support identification and f/w download
 *                   ar_query_device to report the device type
 *
 *===========================================================================*/ 

/*---------------------------------------------------------------------------*
 *                      GENERAL INCLUDE STATEMEMTS
 *---------------------------------------------------------------------------*/

#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   #include <unistd.h>
   #include <fcntl.h>
   #include <string.h>
   #include <pthread.h>
   #include <errno.h>
   #include <sys/stat.h> 
   #include <sys/shm.h>
   #include <sys/sem.h>
   #include <sys/syscall.h>
   #ifndef max
      #define max(a,b) (a>b?a:b)
   #endif
   #define ARU_OS_THREAD_PRIORITY 2
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
	#include "target_defines.h"
   #define ARU_OS_THREAD_PRIORITY 0
#elif defined (NIVISA_PXI)
   #ifdef LABVIEW_RT
      #include <ansi_c.h> 
   #endif
   #include <visa.h>  
   #define ARU_OS_THREAD_PRIORITY 0
#elif defined (INTEGRITY_PCI_PPC)
   #include <string.h>
   #define max(a,b) (a>b?a:b)
   #define ARU_OS_THREAD_PRIORITY 0
#else
   #define ARU_OS_THREAD_PRIORITY THREAD_PRIORITY_TIME_CRITICAL 
#endif

/* include local header files, CDEV_API.H included in CDEV_GLB.H */
#include "cdev_glb.h"
#include "Ar_error.h"
#include "cdev_fw.h"

/*---------------------------------------------------------------------------*
 * Interrupt Service Routine prototype differs for different platforms
 *---------------------------------------------------------------------------*/
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   /* All Linux distributions */
   void cei_utl_interrupt_handler ( CEI_INT32 deviceIndex );
   void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32) = {
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler
   };
#elif defined (VXW_PCI_X86) || defined (VXW_PCI_PPC) || defined (INTEGRITY_PCI_PPC)
   /* VxWorks and Integrity */
   void cei_utl_interrupt_handler (CEI_INT32 deviceIndex, pCEI_UINT32 data);

   void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32, pCEI_UINT32) = {
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler, 
      cei_utl_interrupt_handler, cei_utl_interrupt_handler
   };
#else
   /* All Windows o/s */
   void (*x30InterruptFunctionPtr[MAX_BOARDS])(CEI_INT32) = {
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
      NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
   };
#endif


/*********************************************************************************
 *********************************************************************************
 * O P E R A T I N G   S Y S T E M   C O M P I L A T I O N   D I R E C T I V E S *
 *********************************************************************************
 *********************************************************************************/

/*===========================================================================* 
 *  For LabVIEW Real-Time with full VISA support, writes to BAR0 require a 
 *  different argument to the viOut32 invocation, so we use a different write 
 *  access routine, see cdev_lrt.h for details; select via compiler directive.
 *  For all other operating systems BAR0 is accessed like BAR2.
 *===========================================================================*/
#if defined (NIVISA_PXI) /* && !defined (_DEBUG) */
   #define WRITE_32_BIT cei_write_plx_32bit
   #define READ_32_BIT  cei_read_plx_32bit
#else
   #define WRITE_32_BIT cei_write_32bit
   #define READ_32_BIT  cei_read_32bit
#endif
#ifndef LABVIEW_RT
   static pCEI_CHAR thresholdText[2] = {"lower", "upper"};
#endif
/*===========================================================================*/


/*===========================================================================* 
 *  Compiler Directives select o/s-specific header files to include 
 *===========================================================================*/
#if defined DEBUG_PROGRAMMING
   #include "irigdecl.h"
#elif defined NO_INT64
   #include "math.h"
#endif
/*===========================================================================*/

/* Macros for general board index and board initialized verification required for most API routines */
#define VALID_BOARD_AND_INIT (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) && (deviceGlobal[board]->boardInitialized)
#define VALID_BOARD_AND_INIT_II (board >= MAX_BOARDS) || (deviceGlobal[board] == NULL) || !(deviceGlobal[board]->boardInitialized)

/*****************************************************************************
 *****************************************************************************
 *                       A P I   P R I V A T E   D A T A                     *
 *****************************************************************************
 *****************************************************************************/

/* API local PCI interrupt entry storage & controls, maintaining a copy of the h/w  
 * interrupt info, used by the default API-supplied ISR cei_utl_interrupt_handler.
 */
CEI_INT_QUEUE_TYPE interruptQueue[MAX_BOARDS];
CEI_UINT32 lastReadIntQueueLocation[MAX_BOARDS];

/* Application x30 h/w interrupt queue last read location, used in ar_interrupt_queue_read
 * by the host when polling the interrupt queue with PCI interrupts disabled 
 */
CEI_UINT32 lastApplicationReadIntQueueLocation[MAX_BOARDS];

/* pointer to process-local board parameters */
ARINC_LOCAL_TYPE device[MAX_BOARDS];

/* shared board parameters for single-process only implementation */
static ARINC_GLOBAL_TYPE deviceSingleProcessShared[MAX_BOARDS] = {0,};

/* pointer to global shared board parameters */
static volatile ARINC_GLOBAL_TYPE * deviceGlobal[MAX_BOARDS];

/* miscellaneous global data and arrays */
static CEI_INT32 baseConfig = 0;

/* bypass wrap test by default, altered via ar_bypass_wrap_test() invocation */
static CEI_INT32 bypassWrapTest[MAX_BOARDS];

/* programmable IRIG DAC threshold voltage, initial value set here */
static CEI_INT32 irigDacThreshold[MAX_BOARDS];

/* local storage of ARINC 717 sync words for app retrieval */
static CEI_UINT32 a717syncWords[MAX_BOARDS][NUM_SUBFRAME_SYNC_WORDS];

/* used in ar_get_573_config() to translate config register bit field values to parm type */
static CEI_INT16 a717busSpeedOptions[A717_NUM_SUPPORTED_BUS_SPEEDS] = {
   ARU_573_RATE_SIZE_384_32,    
   ARU_573_RATE_SIZE_768_64,    
   ARU_573_RATE_SIZE_1536_128,  
   ARU_573_RATE_SIZE_3072_256,  
   ARU_573_RATE_SIZE_6144_512,  
   ARU_573_RATE_SIZE_12288_1024,
   ARU_573_RATE_SIZE_24576_2048,
   ARU_573_RATE_SIZE_49152_4096 };

/* used in ar_get_573_config() to translate config register bit field values to parm type */
static CEI_INT16 a717frameSizeOptions[A717_NUM_SUPPORTED_BUS_SPEEDS] = {
   ARU_573_SUB_FRAME_32,    
   ARU_573_SUB_FRAME_64,    
   ARU_573_SUB_FRAME_128,  
   ARU_573_SUB_FRAME_256,  
   ARU_573_SUB_FRAME_512,  
   ARU_573_SUB_FRAME_1024,
   ARU_573_SUB_FRAME_2048,
   ARU_573_SUB_FRAME_4096 };

/* parameters used in the SRAM Memory Test */
static CEI_UINT32 patternTest[AR_NUM_PATTERNS] =   {       1, 0xFFFFFFFE, 0xAAAAAAAA, 0x55555555,      0,         0};
static CEI_UINT32 patternInc[AR_NUM_PATTERNS] =    {       0,          1,          0,          0,      1,         0};
static CEI_INT32 patternProcess[AR_NUM_PATTERNS] = {AR_SHIFT,   AR_SHIFT,  AR_STATIC,  AR_STATIC, AR_ADD, AR_STATIC};
static CEI_UINT32 unusedSram[AR_NUM_UNUSED_SRAM_LOCATIONS] = {0x28000, 0x29555, 0x2A666, 0x2B999, 0x2CAAA, 0x2DFFF};
   
/* Board names used by ar_get_boardname */
static pCEI_CHAR boardNames[CEIDEV_LAST_CEIX30_CONFIG - CEIDEV_FIRST_CEIX30_CONFIG + 1] = {  
   "CEI-830", "CEI-430", "AMC-A30", "CEI-530", "R830RX", "RAR-CPCI", "RAR-EC", "RAR-PCIE", "CEI-430A"
};

static CEI_INT32 a717frameSizeRxDelayDuration[A717_NUM_SUPPORTED_BUS_SPEEDS] = { 32,   16,   8,    4,    2,    1,    1,    1};
static CEI_INT32 a717frameSizeRxTimeOutCount[A717_NUM_SUPPORTED_BUS_SPEEDS]  = { 128, 256, 512, 1024, 2048, 4096, 4096, 4096};

/*--------------------------------------------------------------------------
  Even/odd parity conversion table
  --------------------------------------------------------------------------*/
#define ODD     0x80             /* indicates odd parity                 */
#define EVN     0x00             /* indicates even parity                */
                                 /* indicates parity for byte vals 0-255 */
static CEI_UCHAR byte_parity[] = {
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN, EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD,
   EVN, ODD, ODD, EVN, ODD, EVN, EVN, ODD, ODD, EVN, EVN, ODD, EVN, ODD, ODD, EVN
};


/*****************************************************************************
 *****************************************************************************
 *                  A P I   P R I V A T E   F U N C T I O N S                *
 *****************************************************************************
 *****************************************************************************/

/*===========================================================================*
 * ENTRY POINT:     C E I _ I N T E R R U P T _ Q U E U E _ R E A D
 *===========================================================================*
 *
 * FUNCTION:    Reads the hardware interrupt queue filled buffer
 *
 * PARAMETERS:  deviceIndex   -- (input)  device number to reference (0-15)
 *              numberOfWords -- (in/output) number of words to read & no. read
 *              data          -- (output) the queue entry value(s) read
 *
 * RETURN VAL:  ARS_GOTDATA    -- Data was successfully retrieved
 *              ARS_NODATA     -- No data was retrieved
 *
 * GLOBALS:     device[].devicePtr
 *              lastReadIntQueueLocation[]
 *
 * DESCRIPTION: This routine provides read access to the interrupt queue 
 *              filled by processing h/w interrupts using the default API ISR.
 *              For each invocation, the specified number of queue entries
 *              will be read from the buffer region starting at the location
 *              last referenced by the API and ending at the location
 *              referenced by the interrupt queue pointer.
 *
 *===========================================================================*/
CEI_INT32 cei_utl_interrupt_queue_read (
   CEI_INT32   deviceIndex,                  /* device number (0-15)         */
   pCEI_UINT32 numberOfWords,                /* number of words to read      */
   pCEI_UINT32 data                          /* data read from the buffer    */
)
{
   CEI_UINT32  bufferHead = 0;               /* temp register data parm      */
   CEI_UINT32  bufferTail = 0;               /* temp API tail ptr tracker    */
   CEI_UINT32  unreadBufferSize = 0;         /* unused space in interrupt q  */
   CEI_UINT32  wordsRead = 0;                /* words to write index         */
   CEI_INT32   rxStatus = ARS_NODATA;        /* API routine status           */
   CEI_INT32   status = ARS_NORMAL;          /* local API routine status     */
   pCEI_UINT32 appDataPtr;                   /* ptr to the app data array    */
   pCEI_UINT32 bufferDataPtr;                /* general purpose pointer var  */

   /* Acquire the current interrupt queue head pointer */
   status = cei_read_32bit((CEI_INT16)deviceIndex, &device[deviceIndex].devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], &bufferHead);
   bufferTail = lastReadIntQueueLocation[deviceIndex];

   /* Determine how many unread queue entries are available in the h/w interrupt queue */
   if (status == ARS_NORMAL) {
      if (bufferTail != bufferHead) {
         if (bufferTail < bufferHead)
            unreadBufferSize = bufferHead - bufferTail;
         else
            unreadBufferSize = (CEIDEV_INT_QUEUE_SIZE - bufferTail) + bufferHead + 1;

         if (unreadBufferSize > 0)
            rxStatus = ARS_GOTDATA;
      }
   }

   if (rxStatus == ARS_GOTDATA) {
      appDataPtr = data;
      /* Read the lesser of the number of entries requested or the number of
       * unread entries available in the queue, starting from the tail ptr */
      bufferDataPtr = (pCEI_UINT32)(device[deviceIndex].devicePtr + CEIDEV_INT_QUEUE_START + bufferTail);

      for (wordsRead = 0; (wordsRead < *numberOfWords) && (wordsRead < unreadBufferSize); wordsRead++) {
         status = cei_read_32bit ((CEI_INT16)deviceIndex, bufferDataPtr, appDataPtr++);
         /* If we reach the end of the h/w interrupt queue, wrap around to the start */
         if (bufferTail == CEIDEV_INT_QUEUE_SIZE) {
            bufferTail = 0;
            bufferDataPtr = (pCEI_UINT32)(device[deviceIndex].devicePtr + CEIDEV_INT_QUEUE_START);
         }
         else {
            bufferTail++;
            bufferDataPtr++;
         }
      }

      /* Update the API-tracked interrupt queue tail ptr to upon completion */
      lastReadIntQueueLocation[deviceIndex] = bufferTail;
   }

   *numberOfWords = wordsRead;

   if (status != ARS_NORMAL)
      rxStatus = status;

   return rxStatus;
}


/*===========================================================================*
 * ENTRY POINT:   C E I _ U T L _ I N T E R R U P T _ H A N D L E R
 *===========================================================================*
 *
 * FUNCTION:    The default hardware interrupt delayed processing routine
 *
 * PARAMETERS:  deviceIndex   -- (input)  device number to reference (0-15)
 *
 * RETURN VAL:  None
 *
 * GLOBALS:     interruptQueue[] - all members
 *
 * DESCRIPTION: This routine provides the process by which the interrupt
 *              queue is handled when hardware interrupts are enabled.
 *              When invoked, this routine will flush the interrupt queue
 *              and store the entries in a local interrupt storage buffer.
 *              The local buffer is treated as a circular buffer, and
 *              requires constant flushing by the application in order to
 *              eliminate any potential loss of interrupt information.
 *
 *              Under Windows, all processing described above may be
 *              superceeded by a user-defined function, defined as follows:
 *
 *                   void _stdcall function ( CEI_INT32 argument )
 *
 *              The user-defined function can be specified via invocation of
 *              the routine ar_set_isr_function().  Once this routine is
 *              invoked, the assigned invocation will remain as the invoked
 *              routine for the duration the API DLL remains in system RAM.
 *
 *===========================================================================*/
#if defined (_WIN32)
   void _stdcall cei_utl_interrupt_handler ( CEI_INT32 deviceIndex )
#elif defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   void cei_utl_interrupt_handler ( CEI_INT32 deviceIndex )
#else
   void cei_utl_interrupt_handler ( CEI_INT32 deviceIndex, pCEI_UINT32 data )
#endif
{
   CEI_UINT32 numIntQueueEntriesToRead;         /* the amt of space in the   */
   CEI_UINT32 numIntQueueEntriesToFill;         /* buffer available to fill  */
   CEI_UINT32 tempTailPtr;                      /* API int buffer tail ptr   */
   CEI_UINT32 tempHeadPtr;                      /* API int buffer head ptr   */
   CEI_INT32  status = ARS_NORMAL;              /* local API routine status  */


#if defined (_WIN32)
   /* for Windows, a user-defined ISR is invoked here - for VxWorks the user
    * replaces this function "cei_utl_interrupt_handler" with their own
    */
   if (x30InterruptFunctionPtr[deviceIndex] != NULL)
      x30InterruptFunctionPtr[deviceIndex] ( deviceIndex );
   else {
#endif
   /* get the local interrupt buffer head/tail pointers, determine how
    * much room is available in this buffer for storage of h/w interrupt
    * queue entries that have yet to be read.
    */
   tempTailPtr = interruptQueue[deviceIndex].intQueueTailPtr;
   tempHeadPtr = interruptQueue[deviceIndex].intQueueHeadPtr;

   /* If there is room in the buffer, request to read as much as possible */
   if (tempTailPtr == tempHeadPtr) {
      numIntQueueEntriesToRead = CEIDEV_INT_QUEUE_SIZE - 1;
   }
   else if (tempTailPtr > (tempHeadPtr+1) ) {
      numIntQueueEntriesToRead = tempTailPtr - tempHeadPtr - 1;
   }
   else if (tempTailPtr < tempHeadPtr) {
      numIntQueueEntriesToRead = CEIDEV_INT_QUEUE_SIZE - 1 - tempHeadPtr + tempTailPtr;
   }
   else {
      numIntQueueEntriesToRead = 0;
   }
   if (numIntQueueEntriesToRead) {
      /* If the capacity of the local buffer may require a buffer wrap when updating from the h/w int q, handle a fill */
      if ((tempHeadPtr + numIntQueueEntriesToRead) > CEIDEV_INT_QUEUE_SIZE - 1) {
         numIntQueueEntriesToFill = CEIDEV_INT_QUEUE_SIZE - tempHeadPtr;
         status = cei_utl_interrupt_queue_read ( deviceIndex,
                                                 &numIntQueueEntriesToFill,
                                                 &interruptQueue[deviceIndex].intQueueEntries[tempHeadPtr] );
         tempHeadPtr += numIntQueueEntriesToFill;
         if (tempHeadPtr >= CEIDEV_INT_QUEUE_SIZE) {
            numIntQueueEntriesToFill = numIntQueueEntriesToRead - numIntQueueEntriesToFill;
            status = cei_utl_interrupt_queue_read ( deviceIndex,
                                                    &numIntQueueEntriesToFill,
                                                    &interruptQueue[deviceIndex].intQueueEntries[0] );
            tempHeadPtr = numIntQueueEntriesToFill;
         }
      }
      else {
         /* The capacity of the local buffer won't require a buffer wrap to fill */
         status = cei_utl_interrupt_queue_read ( deviceIndex,
                                                 &numIntQueueEntriesToRead,
                                                 &interruptQueue[deviceIndex].intQueueEntries[tempHeadPtr] );
   
         tempHeadPtr += numIntQueueEntriesToRead;
      }
   }

   /* Set the local buffer head pointer based on how many entries were
    * actually read from the hardware interrupt queue
    */
   interruptQueue[deviceIndex].intQueueHeadPtr = tempHeadPtr;
#if defined (_WIN32)
   }
#endif
#ifdef INTEGRITY_PCI_PPC
   if (status) ;
#endif
   return;
}

/*===========================================================================* 
 * Compiler Directive selection for available integer support 32 vs 64 bit
 * on the following API Local Routines:
 *                                        a429_utl_acquire_formatted_time_tag
 *                                        a429_utl_is_irig_calibrated
 *
 * Typically only used for older embedded kernels such as VxWorks 5.5 and
 * Pharlap DOS that do not support 64-bit integer data types.
 *===========================================================================*/

#if defined (NO_INT64)

/* First is the 32-bit only integer support for limited operating systems    */

/*=====================================================================================*
 * ENTRY POINT:  A 4 2 9 _ U T L _ A C Q U I R E _ F O R M A T T E D _ T I M E _ T A G
 *=====================================================================================*
 *
 * FUNCTION:    Format the active receive buffer time-tag in the specified
 *              format, return the 64-bit time-tag in two 32-bit variables
 *
 * PARAMETERS:  CEI_INT16 board         -- (input) Board number (0 to 15)  
 *
 *              pCEI_UINT32 chanMemPtr      -- (input) Pointer to channel memory area
 *
 *              CEI_INT32 timeSource    -- (input) Source (timer or receiver FIFO)
 *                                           AR_TIMER_REGISTER_SET
 *                                           AR_RECEIVER_FIFO_TIMETAG
 *                                           AR_BYPASS_TIMER_READ
 *                                           
 *              CEI_INT32 timeTagFormat -- (input) Selects the format of the
 *                                           time-tag value returned:
 *                                           AR_TIMETAG_EXT_IRIG_64BIT
 *                                           AR_TIMETAG_EXT_IRIG_REF_64BIT
 *                                           AR_TIMETAG_INT_USEC_64BIT
 *                                           AR_TIMETAG_INT_MSEC_32BIT *
 *                                           AR_LOG_EXT_IRIG_SAMPLE
 *                                           AR_TIMETAG_REF_IRIG_64BIT
 *                                           
 *              CEI_INT32 mergedMode    -- (input) buffering mode (merged/dedicated)
 *                                           For 32-bit merged buffer usage, *
 *                                           the channel number is inserted in
 *                                           the upper 5 bits of the time-tag
 *                                              
 *              pAR_TIMETAG_TYPE timeTagRef -- (output) 64-bit time-tag structure
 *                                              
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid timeSource or timeTagFormat
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     
 *              device[].devicePtr[]
 *              boardInitialized[]
 * 
 * DESCRIPTION: This routine assumes the desired receive channel has been
 *              selected by the calling routine.  It reads the 64-bit
 *              microsecond resolution time-tag from the receive buffer,
 *              then reformats it based on either the requested resolution
 *              and size or the actively selected resolution/size.
 * 
 *===========================================================================*/
CEI_INT32 a429_utl_acquire_formatted_time_tag
(
   CEI_INT16 board,              /* (input)  Board number (0 to 15)          */
   pCEI_UINT32 chanMemPtr,       /* (input)  Pointer to channel memory area  */
   CEI_INT32 timeSource,         /* (input)  Source (timer or receiver FIFO) */
   CEI_INT32 timeTagFormat,      /* (input)  Format of the time-tag returned */
   CEI_INT32 mergedMode,         /* (input)  Buffering mode merged/dedicated */
   pAR_TIMETAG_TYPE timeTagRef   /* (output) 64-bit time-tag structure       */
)
{
   CEI_UINT32 tempLowTimeTag;           /* 32-bit LSLW for 64-bit time-tag   */
   CEI_UINT32 tempHighTimeTag;          /* 32-bit MSLW for 64-bit time-tag   */
   CEI_UINT32 appProgTimeTag;           /* Application specified time-tag res*/
   CEI_UINT32 remainderRoundUp = 0;     /* round-up increment for low res.   */
   CEI_INT32 localChannel;              /* merged/individual receive channel */
   CEI_INT32 localTimeTagFormat;        /* local version of time tag format  */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   TIME_TAG_TYPE timeTag64LowValue;     /* temp 64-bit time-tag placeholder  */
   TIME_TAG_TYPE timeTag64HighValue;    /* temp 64-bit time-tag placeholder  */
   pCEI_UINT32 boardPointer;            /* local pointer to chan memory area */
   CEI_INT32 numberOfDays;              /* temp IRIG time conversion vars    */
   CEI_INT32 numberOfMinutes;           /* temp IRIG time conversion vars    */
   CEI_INT32 numberOfSeconds;           /* temp IRIG time conversion vars    */
   double floatingTimeValue;            /* temp IRIG time conversion vars    */
   double boardReferenceTimeTag;        /* temp IRIG time conversion vars    */
   double irigOffsetTime;               /* temp IRIG time conversion vars    */
   double upperTime;                    /* temp IRIG time conversion vars    */
   double lowerTime;                    /* temp IRIG time conversion vars    */
   double tempTime;
   CEI_INT32 status;                    /* operation status value            */

   /* Process the source of this pass' time-tag, either the receive buffer or the timer */
   switch (timeSource) {
      case AR_RECEIVER_FIFO_TIMETAG:
         /* retrieve the merged channel identifier when in merged receive mode */
         if (mergedMode != CEIDEV_MAX_CHANNELS) {
            localChannel = mergedMode;
            localChannel = localChannel << MERGED_CHANNEL_SHIFT;
         }
         
         /* The low order 32-bits of time tag for the ARINC message is in the Receive 
          * Buffer FIFO Long Word 0 location; high order 32-bits in FIFO Long Word 1. 
          * Location of Buffer FIFO Long Word 0 supplied by the calling routine. 
          */
         boardPointer = chanMemPtr;
         accessStatus = cei_read_32bit (board, boardPointer++, &tempLowTimeTag);
         if (accessStatus == ARS_NORMAL)
            accessStatus = cei_read_32bit (board, boardPointer, &tempHighTimeTag);
         break;

      case AR_TIMER_REGISTER_SET:
         /* if multi-thread protection requested, semaphore/mutex-lock device access */
         if (device[board].multiThreadProtect) {
            status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
            if (status != ARS_NORMAL)
               return (status);
         }
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_LOW], &tempLowTimeTag);
         if (accessStatus == ARS_NORMAL)
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_HIGH], &tempHighTimeTag);

         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect)
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         break;

      default:
         accessStatus = ARS_INVARG;
      case AR_BYPASS_TIMER_READ: ;
   }
   
   if (accessStatus == ARS_NORMAL) {
      timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
      timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
      boardReferenceTimeTag = (double)timeTag64LowValue + ((double)timeTag64HighValue * (double)0x10000000 * (double)0x10);

      if (timeTagFormat == AR_USE_PROGRAMMED_TIMETAG_MODE)
         localTimeTagFormat = deviceGlobal[board]->deviceTimeTagReference;
      else
         localTimeTagFormat = timeTagFormat;

      switch (localTimeTagFormat) {
         case AR_TIMETAG_INT_USEC_64BIT:           /* referenceTimeTag contains the upper 32 bits of the time-tag */
            timeTagRef->referenceTimeTag = timeTag64HighValue;
            break;

         case AR_TIMETAG_EXT_IRIG_REF_64BIT:
            /* read the current IRIG time and the device time at which this IRIG time was received */
            /* if multi-thread protection requested, semaphore/mutex-lock device access */
            if (device[board].multiThreadProtect) {
               status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
               if (status != ARS_NORMAL)
                  return (status);
            }
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);

            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
            if (accessStatus == ARS_NORMAL)
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);

            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect)
               accessStatus = cei_release_thread_lock(LOCK_IDENT);

            timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
            timeTagRef->referenceTimeTag = timeTag64HighValue;
            break;

         case AR_TIMETAG_EXT_IRIG_64BIT:
         case AR_LOG_EXT_IRIG_SAMPLE:
            /* if multi-thread protection requested, semaphore/mutex-lock device access */
            if (device[board].multiThreadProtect) {
               status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
               if (status != ARS_NORMAL)
                  return (status);
            }
            /* read the current IRIG time and the device time at which this IRIG time was received */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);

            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
            if (accessStatus == ARS_NORMAL)
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);

            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect)
               accessStatus = cei_release_thread_lock(LOCK_IDENT);

            /* convert the IRIG time to "number of seconds since 12:00AM of Jan 01 */
            numberOfDays     = ((appProgTimeTag >> IRIG_SHIFT_HUNDREDS_OF_DAYS) & IRIG_MASK_HUNDREDS_OF_DAYS) * 100;
            numberOfDays    += ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_DAYS    ) & IRIG_MASK_TENS_OF_DAYS)     * 10;
            numberOfDays    +=  (appProgTimeTag >> IRIG_SHIFT_ONES_OF_DAYS    ) & IRIG_MASK_ONES_OF_DAYS;
            numberOfMinutes  = ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_HOURS   ) & IRIG_MASK_TENS_OF_HOURS)    * 600;
            numberOfMinutes += ((appProgTimeTag >> IRIG_SHIFT_ONES_OF_HOURS   ) & IRIG_MASK_ONES_OF_HOURS)    * 60;
            numberOfMinutes += ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_MINUTES ) & IRIG_MASK_TENS_OF_MINUTES)  * 10;
            numberOfMinutes +=  (appProgTimeTag >> IRIG_SHIFT_ONES_OF_MINUTES ) & IRIG_MASK_ONES_OF_MINUTES;
            numberOfSeconds  = ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_SECONDS ) & IRIG_MASK_TENS_OF_SECONDS)  * 10;
            numberOfSeconds += appProgTimeTag & IRIG_MASK_ONES_OF_SECONDS;
            
            numberOfSeconds += numberOfMinutes * 60;
            floatingTimeValue = (double)numberOfSeconds;
            numberOfDays = max(numberOfDays-1,0);
            floatingTimeValue += (double)numberOfDays * (double)86400;
            deviceGlobal[board]->currentIrigTimeUsec = floatingTimeValue;

            timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
            timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
            deviceGlobal[board]->irigReferenceTimeTag = (double)timeTag64LowValue + 
                                                        ((double)timeTag64HighValue * (double)0x10000000 * (double)0x10);
            
            /* if this invocation was only to read the IRIG time sample, exit here */
            if (localTimeTagFormat == AR_LOG_EXT_IRIG_SAMPLE)
               break;
            
         case AR_TIMETAG_REF_IRIG_64BIT:
            /* Offset the IRIG source time-tag by the difference of the time-tag reference 
             * and the time at which the IRIG update was received */
            irigOffsetTime = boardReferenceTimeTag - deviceGlobal[board]->irigReferenceTimeTag;

            /* now calculate respective IRIG time based on the IRIG reference offset */
            floatingTimeValue = (double)(1000000) * deviceGlobal[board]->currentIrigTimeUsec;
            floatingTimeValue += irigOffsetTime;
            upperTime = floatingTimeValue / ((double)0x10000000 * (double)0x10);
            timeTagRef->referenceTimeTag = (TIME_TAG_TYPE)upperTime;
            upperTime = (double)timeTagRef->referenceTimeTag;
            lowerTime = floatingTimeValue - (upperTime * ((double)0x10000000 * (double)0x10));
            timeTagRef->referenceTimeTag = (TIME_TAG_TYPE)upperTime;
            timeTag64LowValue = (TIME_TAG_TYPE)(lowerTime);
            timeTag64LowValue += deviceGlobal[board]->userProgrammableIrigOffset + AR_IRIG_TIMETAG_BIAS;

            if ( (timeTagRef->referenceTimeTag > (TIME_TAG_TYPE)(7342)) && (timeTag64LowValue > (TIME_TAG_TYPE)(2350112768)) ) {
               timeTagRef->referenceTimeTag = (TIME_TAG_TYPE)(0);
               timeTag64LowValue -= (TIME_TAG_TYPE)(2350112768);
            }
            break;

         case AR_TIMETAG_INT_MSEC_32BIT:
            floatingTimeValue = (double)timeTag64LowValue + ((double)timeTag64HighValue * (double)0x10000000 * (double)0x10);
            tempTime = (floatingTimeValue/(double)1000) * (double)1000;
            if ( (floatingTimeValue - tempTime) > (double)499)
               remainderRoundUp = 1;

            floatingTimeValue /= (double)1000;
            timeTag64LowValue = (TIME_TAG_TYPE)floatingTimeValue;
            timeTag64LowValue += (TIME_TAG_TYPE)remainderRoundUp;

            /*  If Merged Receive Mode is active, replace the upper 3 bits of 
             *  the time-tag with the channel number on which the data was received. */
            if (mergedMode == AR_ON) {
               timeTag64LowValue &= MERGED_CHANNEL_MASK;
               timeTag64LowValue |= (TIME_TAG_TYPE)localChannel;
            }
      }
      
      /* combine the low and high time-tag values into a single 64-bit time-tag */
      timeTagRef->timeTag = timeTag64LowValue;
      timeTagRef->timeTagFormat = localTimeTagFormat;
   }
   return accessStatus;
}


/*===========================================================================*
 * ENTRY POINT:    A 4 2 9 _ U T L _ I S _ I R I G _ C A L I B R A T E D
 *===========================================================================*
 *
 * FUNCTION:    Determine if the given board is synched to an IRIG signal.
 *
 * DESCRIPTION: Check if the IRIG base time has been updated in the last one
 *              second.  If it has, return; otherwise, wait a second  
 *              for the reference IRIG time-tag to update.
 *
 * RETURN VAL:  TRUE or ARS_NORMAL indicates valid IRIG signal reception
 *              ARS_INVHARVAL or other non-zero value indicates bad API status
 *              FALSE or 0 indicates failure to detect valid IRIG input
 *
 *===========================================================================*/
CEI_INT32 a429_utl_is_irig_calibrated (CEI_INT16 board)
{
   AR_TIMETAG_TYPE updateStart;         /* initial IRIG update timetag       */
   AR_TIMETAG_TYPE updateCur;           /* current IRIG update timetag       */
   AR_TIMETAG_TYPE hostCur;             /* current host timer value          */
   CEI_INT32 delay_count;               /* tracks delay iterations           */
   CEI_INT32 status = ARS_NORMAL;       /* operation status value            */
   CEI_INT32 calibrationStatus = FALSE; /* return IRIG valid signal indicat. */
   TIME_TAG_TYPE referenceTimeDifference;      /* temp delta time variable   */
   TIME_TAG_TYPE boardTimeDifference;          /* temp delta time variable   */

   /* make sure IRIG is available on this board */
   if (deviceGlobal[board]->irigInstalled) {
      
      /* read the time of the last IRIG update */
      if (status == ARS_NORMAL)
         status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &updateStart);
      if (status == ARS_NORMAL)
         status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &hostCur);
   
      /* not calibrated if the last IRIG update time is more than 1 second ago */
      if (status == ARS_NORMAL) {
         if (updateStart.timeTag == 0)
            status = FALSE;
         else {
            /* when 64-bit integer math isn't available, use the lower 32-bits to determine IRIG validity */
            if (hostCur.timeTag > updateStart.timeTag)
               boardTimeDifference = hostCur.timeTag - updateStart.timeTag;
            else if (hostCur.timeTag == updateStart.timeTag)
               boardTimeDifference = 0;
            else   /* in this case, the lower word of the timer must have rolled over since the IRIG sample was received */
               boardTimeDifference = (CEI_UINT32)(0xFFFFFFFF - updateStart.timeTag) + hostCur.timeTag - 1;
            
            if (boardTimeDifference > 1000900)
               status = FALSE;
         }
      }
   
      /* wait for one second to obtain a valid IRIG update */
      for (delay_count = 200; (delay_count > 0) && (status == ARS_NORMAL); delay_count--)
      {
         /* read the time of the last IRIG update */
         status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &updateCur);
   
         /* we are calibrated if there is one second between updates (+/- 1ms) */
         if (status == ARS_NORMAL) {
            if (updateCur.timeTag > updateStart.timeTag)
               referenceTimeDifference = updateCur.timeTag - updateStart.timeTag;
            else if (updateCur.timeTag == updateStart.timeTag)
               referenceTimeDifference = 0;
            else
               referenceTimeDifference = (CEI_UINT32)(0xFFFFFFFF - updateStart.timeTag) + updateCur.timeTag - 1;
            
            if ( (referenceTimeDifference < 1001000) && (referenceTimeDifference > 999000) ) {
               calibrationStatus = TRUE;
               break;
            }
         
            /* check if we timed out waiting for the board time to increment */
            cei_delay_milliseconds(5);
         }
      }
   }
   else
      status = ARS_INVHARVAL;

   if ((!calibrationStatus) && (status != ARS_NORMAL))
      calibrationStatus = status;
   
   return calibrationStatus;
}

#else
/* Next is the 64-bit integer support for most operating systems             */

/*=====================================================================================*
 * ENTRY POINT:  A 4 2 9 _ U T L _ A C Q U I R E _ F O R M A T T E D _ T I M E _ T A G
 *=====================================================================================*
 *
 * FUNCTION:    Read and potentially convert the active receive buffer 
 *              or IRIG time-stamp to the specified format
 *
 * PARAMETERS:  CEI_INT16 board             -- (input) Board number (0 to 15)  
 *
 *              pCEI_UINT32 chanMemPtr      -- (input) Pointer to channel memory area
 *
 *              CEI_INT32 timeSource        -- (input) Source (timer or receiver FIFO)
 *                                              AR_TIMER_REGISTER_SET
 *                                              AR_RECEIVER_FIFO_TIMETAG
 *                                              AR_BYPASS_TIMER_READ
 *                                              
 *              CEI_INT32 timeTagFormat     -- (input) Selects the format of the
 *                                              time-tag value returned:
 *                                              AR_TIMETAG_EXT_IRIG_64BIT
 *                                              AR_TIMETAG_INT_USEC_64BIT
 *                                              AR_TIMETAG_INT_20USEC_32BIT *
 *                                              AR_TIMETAG_INT_MSEC_32BIT   *
 *                                              AR_TIMER_X20_COMPAT_32BIT   *
 *                                              AR_USE_PROGRAMMED_TIMETAG_MODE
 *                                              AR_LOG_EXT_IRIG_SAMPLE
 *                                              AR_TIMETAG_REF_IRIG_64BIT
 *                                              
 *              CEI_INT32 mergedMode        -- (input) buffering mode (merged/dedicated)
 *                                              When mergedMode < CEIDEV_MAX_CHANNELS,
 *                                              merged mode is active and then if
 *                                              a 32-bit time-stamp is requested, 
 *                                              the channel number is inserted in
 *                                              the upper 5 bits of the time-stamp (legacy)
 *                                              
 *              pAR_TIMETAG_TYPE timeTagRef -- (output) 64-bit time-tag structure
 *                                              
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid timeSource or timeTagFormat
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine assumes the desired receive channel has been
 *              selected by the calling routine.  It reads the 64-bit
 *              microsecond resolution time-tag from the receive buffer,
 *              then reformats it based on either the requested resolution
 *              and size or the actively selected resolution/size.
 * 
 *===========================================================================*/
CEI_INT32 a429_utl_acquire_formatted_time_tag
(
   CEI_INT16 board,              /* (input)  Board number (0 to 15)          */
   pCEI_UINT32 chanMemPtr,       /* (input)  Pointer to channel memory area  */
   CEI_INT32 timeSource,         /* (input)  Source (timer or receiver FIFO) */
   CEI_INT32 timeTagFormat,      /* (input)  Format of the time-tag returned */
   CEI_INT32 mergedMode,         /* (input)  Buffering mode merged/dedicated */
   pAR_TIMETAG_TYPE timeTagRef   /* (output) 64-bit time-tag structure       */
)
{
   CEI_UINT32 tempLowTimeTag = 0;       /* 32-bit LSLW for 64-bit time-tag   */
   CEI_UINT32 tempHighTimeTag = 0;      /* 32-bit MSLW for 64-bit time-tag   */
   CEI_UINT32 appProgTimeTag = 0;       /* Application specified time-tag res*/
   CEI_UINT32 remainderRoundUp = 0;     /* round-up increment for low res.   */
   CEI_UINT32 localChannel = 0;         /* merged/individual receive channel */
   CEI_INT32 localTimeTagFormat;        /* local version of time tag format  */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   TIME_TAG_TYPE timeTag64LowValue;     /* temp 64-bit time-tag placeholder  */
   TIME_TAG_TYPE timeTag64HighValue;    /* temp 64-bit time-tag placeholder  */
   TIME_TAG_TYPE irigOffsetTime;        /* temp 64-bit time-tag placeholder  */
   CEI_INT32 numberOfDays;              /* temp IRIG time conversion vars    */
   CEI_INT32 numberOfMinutes;           /* temp IRIG time conversion vars    */
   CEI_INT32 numberOfSeconds;           /* temp IRIG time conversion vars    */
   time_t   now;                        /* Current time                      */
   struct   tm start;                   /* Formatted "now" current time      */
   CEI_INT32 year;                      /* Current year (since 1900)         */
   TIME_TAG_TYPE maxMicroSecondsPerYear;/* Max number of usec in this year   */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   CEI_INT32 status;                    /* operation status value            */

   /* Process the source of this pass' time-tag, either the receive buffer or the timer */
   switch (timeSource) {
      case AR_RECEIVER_FIFO_TIMETAG:
         /* retrieve the merged channel identifier when in merged receive mode */
         if (mergedMode != CEIDEV_MAX_CHANNELS) {
            localChannel = mergedMode;
            localChannel = localChannel << MERGED_CHANNEL_SHIFT;
         }

         /* The low order 32-bits of time tag for the ARINC message is in the Receive 
          * Buffer FIFO Long Word 0 location; high order 32-bits in FIFO Long Word 1.
          * Location of Buffer FIFO Long Word 0 supplied by the calling routine. 
          */
         boardPointer = chanMemPtr;
         accessStatus = cei_read_32bit (board, boardPointer++, &tempLowTimeTag);
         if (accessStatus == ARS_NORMAL)
            accessStatus = cei_read_32bit (board, boardPointer, &tempHighTimeTag);
         break;

      case AR_TIMER_REGISTER_SET:
         /* if multi-thread protection requested, semaphore/mutex-lock device access */
         if (device[board].multiThreadProtect) {
            status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
            if (status != ARS_NORMAL)
               return (status);
         }

         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_LOW], &tempLowTimeTag);
         if (accessStatus == ARS_NORMAL)
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_HIGH], &tempHighTimeTag);

         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect)
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         break;

      default:
         accessStatus = ARS_INVARG;
      case AR_BYPASS_TIMER_READ: ;
   }
   
   if (accessStatus == ARS_NORMAL) {
      timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
      timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
      timeTagRef->referenceTimeTag = timeTag64LowValue + (timeTag64HighValue << 32);
   
      if (timeTagFormat == AR_USE_PROGRAMMED_TIMETAG_MODE)
         localTimeTagFormat = deviceGlobal[board]->deviceTimeTagReference;
      else
         localTimeTagFormat = timeTagFormat;
      
      switch (localTimeTagFormat) {
         case AR_TIMETAG_INT_USEC_64BIT:           /* time-tag already in usec 64-bit format   */
            timeTag64LowValue += (timeTag64HighValue << 32);
            break;

         case AR_TIMETAG_EXT_IRIG_64BIT:
         case AR_LOG_EXT_IRIG_SAMPLE:
            /* if multi-thread protection requested, semaphore/mutex-lock device access */
            if (device[board].multiThreadProtect) {
               status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
               if (status != ARS_NORMAL)
                  return (status);
            }
            /* read the current IRIG time and the device time at which this IRIG time was received */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);

            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
            if (accessStatus == ARS_NORMAL)
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);

            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect)
               accessStatus = cei_release_thread_lock(LOCK_IDENT);

            /* convert the IRIG time to "number of seconds since 12:00AM of Jan 01 */
            numberOfDays     = ((appProgTimeTag >> IRIG_SHIFT_HUNDREDS_OF_DAYS) & IRIG_MASK_HUNDREDS_OF_DAYS) * 100;
            numberOfDays    += ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_DAYS    ) & IRIG_MASK_TENS_OF_DAYS)     * 10;
            numberOfDays    +=  (appProgTimeTag >> IRIG_SHIFT_ONES_OF_DAYS    ) & IRIG_MASK_ONES_OF_DAYS;
            numberOfMinutes  = ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_HOURS   ) & IRIG_MASK_TENS_OF_HOURS)    * 600;
            numberOfMinutes += ((appProgTimeTag >> IRIG_SHIFT_ONES_OF_HOURS   ) & IRIG_MASK_ONES_OF_HOURS)    * 60;
            numberOfMinutes += ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_MINUTES ) & IRIG_MASK_TENS_OF_MINUTES)  * 10;
            numberOfMinutes +=  (appProgTimeTag >> IRIG_SHIFT_ONES_OF_MINUTES ) & IRIG_MASK_ONES_OF_MINUTES;
            numberOfSeconds  = ((appProgTimeTag >> IRIG_SHIFT_TENS_OF_SECONDS ) & IRIG_MASK_TENS_OF_SECONDS)  * 10;
            numberOfSeconds += appProgTimeTag & IRIG_MASK_ONES_OF_SECONDS;
            
            numberOfSeconds += numberOfMinutes * 60;
            numberOfSeconds += max(numberOfDays-1,0) * AR_SECONDS_PER_DAY; /* 0 days is invalid IRIG */
            deviceGlobal[board]->currentIrigTimeUsec = numberOfSeconds;

            timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag;
            timeTag64HighValue = (TIME_TAG_TYPE)tempHighTimeTag;
            deviceGlobal[board]->irigReferenceTimeTag = timeTag64LowValue + (timeTag64HighValue << 32);
            
            /* if this invocation was only to read the IRIG time sample, exit here */
            if (localTimeTagFormat == AR_LOG_EXT_IRIG_SAMPLE)
               break;
            
         case AR_TIMETAG_REF_IRIG_64BIT:
            /* Offset the IRIG source time-tag by the difference of the time-tag 
             * reference and the time at which the IRIG update was received.
             *
             * Once we start applying offsets, the actual IRIG conversion could
             * cause the actual resulting time-of-year to exceed the number of
             * microseconds available in a year; therefore, we must force a
             * calendar rollover after applying the offsets.
             */
            irigOffsetTime = timeTagRef->referenceTimeTag - deviceGlobal[board]->irigReferenceTimeTag;

            /* now calculate respective IRIG time based on the IRIG reference offset */
            timeTag64LowValue = deviceGlobal[board]->currentIrigTimeUsec * AR_MICROSECONDS_PER_SECOND;
            timeTag64LowValue += irigOffsetTime;
            timeTag64LowValue += deviceGlobal[board]->userProgrammableIrigOffset + AR_IRIG_TIMETAG_BIAS;

            /* Acquire the current year */
            time(&now);               /* Time in seconds since 12AM Jan 1, 1970 GMT */
            start = *localtime(&now); /* Local time in sec, min, hr, day, mo, yr */
            year = (CEI_INT32)start.tm_year + AR_LOCAL_TIME_YEAR_BASIS;
         
            /* Determine if this is a leap year for max # of usecs allowed this year */
            if ( ((year % AR_LEAP_YEAR_DIVISOR) == 0) && (year != AR_LEAP_YEAR_EXCEPTION) )
               maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_LEAP_YEAR;
            else
               maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_NON_LEAP_YEAR;
            
            /* handle an IRIG rollover once the offsets are applied */
            if (timeTag64LowValue > maxMicroSecondsPerYear) {
               timeTag64LowValue -= maxMicroSecondsPerYear;
            }
            /*
            timeTagRef->referenceTimeTag = deviceGlobal[board]->irigReferenceTimeTag; */
            break;

         case AR_TIMETAG_EXT_IRIG_REF_64BIT:
            /* if multi-thread protection requested, semaphore/mutex-lock device access */
            if (device[board].multiThreadProtect) {
               status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
               if (status != ARS_NORMAL)
                  return (status);
            }
            /* read the current IRIG time and the device time at which this IRIG time was received */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);

            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
            if (accessStatus == ARS_NORMAL)
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);

            timeTag64LowValue = (TIME_TAG_TYPE)tempLowTimeTag + ((TIME_TAG_TYPE)tempHighTimeTag << 32);

            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect)
               accessStatus = cei_release_thread_lock(LOCK_IDENT);

            break;

         case AR_TIMETAG_EXT_IRIG_RAW:
            /* if multi-thread protection requested, semaphore/mutex-lock device access */
            if (device[board].multiThreadProtect) {
               status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
               if (status != ARS_NORMAL)
                  return (status);
            }
            /* read the current IRIG time and the device time at which this IRIG time was received */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME], &appProgTimeTag);

            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_LOW], &tempLowTimeTag);
            if (accessStatus == ARS_NORMAL)
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_TIME_TAG_HIGH], &tempHighTimeTag);

            timeTag64LowValue = (TIME_TAG_TYPE)appProgTimeTag;
            timeTagRef->referenceTimeTag = (TIME_TAG_TYPE)tempLowTimeTag + ((TIME_TAG_TYPE)tempHighTimeTag << 32);

            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect)
               accessStatus = cei_release_thread_lock(LOCK_IDENT);

            break;

         case AR_TIMETAG_INT_20USEC_32BIT:
            timeTag64LowValue = timeTag64LowValue + (timeTag64HighValue << 32);
            if (tempLowTimeTag%20 > 9)
               remainderRoundUp = 1;

            timeTag64LowValue /= (TIME_TAG_TYPE)20;
            timeTag64LowValue += (TIME_TAG_TYPE)remainderRoundUp;

            if (mergedMode != CEIDEV_MAX_CHANNELS) {
               timeTag64LowValue &= MERGED_CHANNEL_MASK;
               timeTag64LowValue |= (TIME_TAG_TYPE)localChannel;
            }
            break;

         case AR_TIMETAG_INT_MSEC_32BIT:
            timeTag64LowValue = timeTag64LowValue + (timeTag64HighValue << 32);
            if (tempLowTimeTag%1000 > 499)
               remainderRoundUp = 1;

            timeTag64LowValue /= AR_MICROSEC_PER_MILLISEC;
            timeTag64LowValue += (TIME_TAG_TYPE)remainderRoundUp;

            /*  If Merged Receive Mode is active, replace the upper 3 bits of 
             *  the time-tag with the channel number on which the data was received. */
            if (mergedMode != CEIDEV_MAX_CHANNELS) {
               timeTag64LowValue &= MERGED_CHANNEL_MASK;
               timeTag64LowValue |= (TIME_TAG_TYPE)localChannel;
            }
            break;

         case AR_TIMER_X20_COMPAT_32BIT:
            timeTag64LowValue = timeTag64LowValue + (timeTag64HighValue << 32);
            if (timeTag64LowValue < AR_X20_MAX_TIMETAG_VALUE) {
               timeTag64LowValue /= (TIME_TAG_TYPE)(deviceGlobal[board]->deviceX20CompatibleResolution/AR_X20_SCALE_TIMETAG_VALUE);
            }
            
            /*  If Merged Receive Mode is active, replace the upper 3 bits of 
             *  the time-tag with the channel number on which the data was received. */
            if (mergedMode != CEIDEV_MAX_CHANNELS) {
               timeTag64LowValue &= MERGED_CHANNEL_MASK;
               timeTag64LowValue |= (TIME_TAG_TYPE)localChannel;
            }
            break;
      }
      
      /* combine the low and high time-tag values into a single 64-bit time-tag */
      timeTagRef->timeTag = timeTag64LowValue;
      timeTagRef->timeTagFormat = (TIME_TAG_TYPE)localTimeTagFormat;
   }
   return accessStatus;
}


/*===========================================================================*
 * ENTRY POINT:     A 4 2 9 _ U T L _ I S _ I R I G _ C A L I B R A T E D
 *===========================================================================*
 *
 * FUNCTION:    Determine if the given board is synched to an IRIG signal.
 *
 * DESCRIPTION: Check if the IRIG base time has been updated in the last one
 *              second.  If it has, return; otherwise, wait a second  
 *              for the reference IRIG time-tag to update.
 *
 * PARAMETERS:  CEI_INT16 board  -- (input)  Board number (0 to 15)  
 * 
 * RETURN VAL:  TRUE or ARS_NORMAL indicates valid IRIG signal reception
 *              ARS_INVHARVAL or other non-zero value indicates bad API status
 *              FALSE or 0 indicates failure to detect valid IRIG input
 *
 *===========================================================================*/
CEI_INT32 a429_utl_is_irig_calibrated (CEI_INT16 board)
{
   AR_TIMETAG_TYPE updateStart;         /* initial IRIG update timetag       */
   AR_TIMETAG_TYPE updateCur;           /* current IRIG update timetag       */
   TIME_TAG_TYPE timeDiff;              /* delta time between samples        */
   CEI_INT32 delay_count;               /* tracks delay iterations           */
   CEI_INT32 status = ARS_NORMAL;       /* operation status value            */
   CEI_INT32 calibrationStatus = FALSE; /* return IRIG valid signal indicat. */

   /* make sure IRIG is available on this board */
   if (deviceGlobal[board]->irigInstalled) {
      
      /* read the time of the last IRIG update */
      status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &updateStart);
   
      /* not calibrated if the last IRIG update time is more than 1 second ago */
      if (status == ARS_NORMAL) {
         if (updateStart.timeTag == 0) {
            status = FALSE;
         }
         else {
            if (updateStart.referenceTimeTag >= updateStart.timeTag)
               timeDiff = updateStart.referenceTimeTag - updateStart.timeTag;
            else
               timeDiff = (AR_MAX_TIMETAG_VALUE - updateStart.timeTag) + updateStart.referenceTimeTag;
            
            if (timeDiff > 1000900) {
               status = FALSE;
            }
         }
      }
   
      /* wait for one second to obtain a valid IRIG update */
      for (delay_count = 200; (delay_count > 0) && (status == ARS_NORMAL); delay_count--)
      {
         /* read the time of the last IRIG update */
         status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &updateCur);
   
         /* we are calibrated if there is one second between updates (+/- 1ms) */
         if (status == ARS_NORMAL) {
            if (updateCur.timeTag >= updateStart.timeTag)
               timeDiff = updateCur.timeTag - updateStart.timeTag;
            else
               timeDiff = (AR_MAX_TIMETAG_VALUE - updateStart.timeTag) + updateCur.timeTag;

            if ( (timeDiff < 1001000) && (timeDiff >  997000) ) {
               calibrationStatus = TRUE;
               break;
            }
         
            /* check if we timed out waiting for the board time to increment */
            if (updateCur.referenceTimeTag >= updateStart.referenceTimeTag)
               timeDiff = updateCur.referenceTimeTag - updateStart.referenceTimeTag;
            else
               timeDiff = (AR_MAX_TIMETAG_VALUE - updateStart.referenceTimeTag) + updateCur.referenceTimeTag;
               
            if (timeDiff > 1000900) {
               status = FALSE;
            }
            else
               cei_delay_milliseconds(5);
         }
      }
   }
   else
      status = ARS_INVHARVAL;

   if ((!calibrationStatus) && (status != ARS_NORMAL))
      calibrationStatus = status;
   
   return calibrationStatus;
}

#endif

/*===========================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ A D J U S T _ I R I G _ T H R E S H O L D
 *===========================================================================*
 *
 * FUNCTION:    Performs a thorough adjustment of the IRIG DAC for an external 
 *              input IRIG signal, using valid IRIG-B 32-bit time reception for 
 *              verification of signal presence.
 *
 * DESCRIPTION: This routine sets the DAC to Vmin + (0.825 * (Vmax-Vmin)).  
 *              Vmin is the lower peak value and Vmax is the upper peak value.
 *
 * PARAMETERS:  CEI_INT16 board  -- (input)  Board number (0 to 15)  
 * 
 * RETURN VAL:  status   --  Any value from 0 to x7FFFFFFF is the DAC value
 *                           ARS_FAILURE (-1) Any issue encountered with
 *                                            incoming IRIG signal level
 *
 *===========================================================================*/
CEI_INT32 a429_utl_adjust_irig_threshold (CEI_INT16 board, CEI_INT32 displayDacSettings)
{
   CEI_UINT32 dacMaxVoltage = 0xff;     /* temp max DAC value for thold test */
   CEI_UINT32 dacValue;                 /* dynamic DAC value during test     */
   CEI_UINT32 initialDacValue = 0x80;   /* initial DAC value                 */
   CEI_UINT32 initialDacAdjustment = 0x10;  /* initial DAC increment         */
   CEI_UINT32 minimumVoltageValue;      /* lower bounds of final DAC value   */
   CEI_UINT32 maximumVoltageValue;      /* upper bounds of final DAC value   */
   CEI_UINT32 irigRegValue;             /* temp contents of IRIG Receiver    */
   CEI_UINT32 dacBoudaryCushionValue;   /* minor adj to final DAC value      */
   CEI_INT32 signalFound = FALSE;       /* flag ind IRIG signal edge detected*/
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of board read/write access */
   CEI_INT32 status = ARS_NORMAL;       /* status of threshold detection     */
   CEI_UINT32 prevIrigRegValue = 0;     /* IRIG Time sample value            */
   pCEI_UINT32 dacControlRegPtr = &device[board].devicePtr[CEIDEV_DAC_CONTROL];
   pCEI_UINT32 irigReceiverRegPtr = &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME];
   
   dacValue = CEIDEV_IRIG_DAC_OFFSET;

#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("Initial DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
#endif
   /*=======================================================================
    * FIRST FIND THE INCOMING SIGNAL
    * Attempt to initially detect the incoming IRIG signal between the 
    * mid-upper peak by setting the DAC to 128 (0.0 Volts) and increasing
    * the threshold until the signal is found.
    *=======================================================================*/
   accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
   cei_delay_milliseconds(5);             /* Wait 5 Msec for signal    */
   
   /* Read IRIG value, clearing "edge received" bit, then  */
   /* Wait 30 Msec for an incoming signal edge transition. */
   accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
   cei_delay_milliseconds(30);
   
   /* Read IRIG value again, check the "edge received" bit to be set */
   accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
   
   cei_delay_milliseconds(2000);
   if (irigRegValue != prevIrigRegValue) {
      prevIrigRegValue = irigRegValue;
      signalFound = TRUE;
   }
   else {
      /* Signal wasn't immediately detected, raise the threshold until it's found */
      for (dacValue = 0x80; (dacValue < 0x100) && (!signalFound); dacValue += 0x10) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
         
#ifndef LABVIEW_RT
         if (displayDacSettings)
            printf ("\rFinding the signal above the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         cei_delay_milliseconds(5);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
         cei_delay_milliseconds(30);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);

         cei_delay_milliseconds(2000);
         if (irigRegValue != prevIrigRegValue) {
            prevIrigRegValue = irigRegValue;
            dacMaxVoltage = 0xffL;
            initialDacValue = dacValue;
            initialDacAdjustment = 0x20L;
            signalFound = TRUE;
         }
      }

#ifndef LABVIEW_RT
      if (displayDacSettings)
         printf ("\n");
#endif
      /* Signal wasn't detected by raising the threshold? Lower it now until it's found */
      for (dacValue = 0x80; (dacValue > 0x0) && (!signalFound); dacValue -= 0x10) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);

#ifndef LABVIEW_RT
         if (displayDacSettings)
            printf ("\rFinding the signal below the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         cei_delay_milliseconds(5);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
         cei_delay_milliseconds(30);
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
         
         cei_delay_milliseconds(2000);
         if (irigRegValue != prevIrigRegValue) {
            prevIrigRegValue = irigRegValue;
            dacMaxVoltage = 128L;
            initialDacAdjustment = 0x20L;
            initialDacValue = dacValue;
            signalFound = TRUE;
         }
      }

      if (!signalFound) {
#ifndef LABVIEW_RT
         if (displayDacSettings)
            printf ("\nNever found a signal -- final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
#endif
         return ARS_FAILURE;
      }
   }

   /*=======================================================================
    * SECOND FIND THE LOW THRESHOLD AT WHICH SIGNAL BECOMES PRESENT
    * Starting with the theshold DAC at level the signal was detected above,
    * decrement the DAC voltage until the signal is no longer present;
    * then perform a fine adjustment until DAC setting is just above
    * the lower threshold where the signal becomes present.
    *=======================================================================*/
   dacValue = initialDacValue;
   status = a429_utl_find_dac_threshold (board, 
                              TRUE, 
                              displayDacSettings, 
                              AR_LOWER, 
                              dacMaxVoltage, 
                              initialDacAdjustment, 
                              &dacValue, 
                              &dacBoudaryCushionValue); 
   if (status == ARS_FAILURE)
      return status;

   minimumVoltageValue = dacValue + dacBoudaryCushionValue;
   
#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("\nLower transition final value = %3d\n", (CEI_NATIVE_INT)minimumVoltageValue);
#endif
   /*=======================================================================
    * THIRD FIND THE AR_UPPER THRESHOLD AT WHICH SIGNAL BECOMES LOST
    * Starting with the theshold DAC at level the signal was detected first,
    * increment the DAC voltage until the signal is no longer present;
    * then perform a fine adjustment until DAC setting is just below
    * the upper threshold where the signal is lost.
    *=======================================================================*/
   dacValue = initialDacValue + initialDacAdjustment;
   status = a429_utl_find_dac_threshold (board, 
                              TRUE, 
                              displayDacSettings, 
                              AR_UPPER, 
                              minimumVoltageValue, 
                              initialDacAdjustment, 
                              &dacValue, 
                              &dacBoudaryCushionValue); 
   if (status == ARS_FAILURE)
      return status;

   maximumVoltageValue = dacValue - dacBoudaryCushionValue;
   
#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("\nUpper transition final value = %3d\n", (CEI_NATIVE_INT)maximumVoltageValue);
#endif
   
   /* validate the span between the upper and lower thresholds */
   if ((maximumVoltageValue < minimumVoltageValue) || ((maximumVoltageValue-minimumVoltageValue) < MIN_DAC_SPAN)) {
#ifndef LABVIEW_RT
      if (displayDacSettings)
         printf ("Error, Signal Detection Range is invalid, Vmax = %d  Vmin = %d\n", 
                 (CEI_NATIVE_INT)maximumVoltageValue, (CEI_NATIVE_INT)minimumVoltageValue);
#endif
      accessStatus = cei_write_32bit (board, dacControlRegPtr, 0x80L);
      return ARS_FAILURE;
   }

   dacValue = minimumVoltageValue + (CEI_UINT32)((825*(maximumVoltageValue-minimumVoltageValue))/1000);
   accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
   
#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("Final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
#endif

   if (dacValue <= CEIDEV_IRIG_DAC_OFFSET) {
      irigDacThreshold[board] = 0;
      dacValue = 0;
   }
   else {
      irigDacThreshold[board] = (CEI_INT32)dacValue;
      dacValue = (CEI_UINT32)(CEIDEV_IRIG_DAC_MV_CONVERSION * (float)(dacValue - CEIDEV_IRIG_DAC_OFFSET));
   }
   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
   return (CEI_INT32)(dacValue);
}

/*===========================================================================*
 * ENTRY POINT:             A 4 2 9 _ U T L _ E N H
 *===========================================================================*
 *
 * FUNCTION:    Open a session to the board and update the EEPROM for 
 *              CEI-x30 Enhanced firmware operation
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *
 *              CEI_INT32 action -- (input) operation to perform:
 *                                          PLX_ENHANCED_CHECK - check to 
 *                                             see if BAR2 is 512Kb
 *                                          PLX_ENHANCED_UPDATE - check
 *                                             to see if BAR2 is 512Kb and
 *                                             reprogram it if it is only 4Kb
 *                                          PLX_ORIGINAL_RESTORE - if BAR2
 *                                             is not 4kb, this option will
 *                                             result in reprogramming BAR2
 *                                             to the original x30 4Kb size
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL     Success updating or already done
 *                              ARS_INVBOARD   Invalid board parameter
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_FAILURE    Invalid device (not x30)
 *                              ARS_MEMWRERR   PROM update failed
 * 
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: This routine is a special utility routine to upgrade a
 *              CEI-x30 board for use with the enhanced firmware and API.
 *              It modifies the PLX PROM BAR2 size from 4Kb to 512Kb.
 *
 *===========================================================================*/
CEI_INT16 a429_utl_enh (CEI_INT16 board, CEI_INT32 action)
{
   CEI_INT32   status = ARS_NORMAL;      /* function return status           */
   CEI_INT32   tempStatus = ARS_NORMAL;  /* function return status           */
   CEI_INT32   plxType = 0;              /* type of PLX PCI intfc 9056/9030  */
   CEI_INT32   bar2SizeWord1 = 0;        /* PCI Cfg offset for PCI BAR2 Size */
   CEI_INT32   bar2SizeWord2 = 0;        /* PCI Cfg offset for PCI BAR2 Size */
   CEI_INT32   vendorIdent;              /* PCI Vendor ID for opened device  */
   CEI_INT32   deviceIdent;              /* PCI Device ID for opened device  */
   CEI_UINT32  promContent[PLX_EEPROM_SIZE]; /* contents of PLX EEPROM       */
   pCEI_UINT32 pciConfigMemPtr = 0;          /* BAR1 offset for PLX EEPROM       */

   /* open a session with the specified device */
   status = cei_acquire_device (board, &device[board]);
   if (status == ARS_NORMAL) {
      /* force exclusive access to the board for the duration of this routine */
#ifndef LABVIEW_RT
      status = cei_acquire_thread_lock(LOCK_IDENT);
#endif      
      if (status == ARS_NORMAL) {
         /* assure this is a CEI-x30 board before continuing */
         status = cei_acquire_device_identity ((CEI_INT32)board, &vendorIdent, &deviceIdent);
         if ( (status != ARS_FAILURE) && (vendorIdent == PCI_X30_MANUF_ID) ) {
            switch (deviceIdent) {
               case PCI_DEVICE_ID_530:
               case PCI_DEVICE_ID_630:
               case PCI_DEVICE_ID_830:
               case PCI_DEVICE_ID_830RX:
               case PCI_DEVICE_ID_A30:
                  plxType       = PLX_9056;
                  bar2SizeWord1 = PLX_9056_BAR2_SIZE_WORD_1;
                  bar2SizeWord2 = PLX_9056_BAR2_SIZE_WORD_2;
                  pciConfigMemPtr = device[board].baseAddr + PLX_CONFIG_MEM_CONTROL_REG;
                  break;
               case PCI_DEVICE_ID_430:
               case PCI_DEVICE_ID_430A:
               case PCI_DEVICE_ID_RAREC:
                  plxType       = PLX_9030;
                  bar2SizeWord1 = PLX_9030_BAR2_SIZE_WORD_1;
                  bar2SizeWord2 = PLX_9030_BAR2_SIZE_WORD_2;
                  pciConfigMemPtr = device[board].baseAddr + PLX9030_PCI_TARGET_RESPONSE;
                  break;
               case PCI_DEVICE_ID_RAR_PCIE:
                  status = ARS_INVBOARD;
                  break;
               default:
                  status = ARS_FAILURE;
            }

            if (status == ARS_NORMAL) {
               /* read the current content of the PROM */
               status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
               /* if the PROM is currently programmed for the old CEI-x30 BAR2 size, update it to 512Kb */
               if ((promContent[bar2SizeWord1] != PLX_BAR2_ENHANCED_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ENHANCED_SIZE_2)) {
                  if (action == PLX_ENHANCED_CHECK) {
                     status = ARS_HW_CONSISTENCY;
                  }
                  else if (action == PLX_ENHANCED_UPDATE) {
                     status = a429_utl_plx_eeprom_write_enable(board, plxType, pciConfigMemPtr);
                     ar_sleep(52);
                     status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord1, PLX_BAR2_ENHANCED_SIZE_1);
                     ar_sleep(52);
                     status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord2, PLX_BAR2_ENHANCED_SIZE_2);
                     ar_sleep(52);
                     status = a429_utl_plx_eeprom_write_disable(board, plxType, pciConfigMemPtr);
                     ar_sleep(52);  
      
                     /* read/verify the updated content of the PROM */
                     status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
                     if ((promContent[bar2SizeWord1] != PLX_BAR2_ENHANCED_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ENHANCED_SIZE_2)) {
                        status = ARS_MEMWRERR;
                     }
                     else {
                        status = ARS_NORMAL;
                     }
                  }
                  else {
                     status = ARS_INVARG;
                  }
               }
               else if (action == PLX_ORIGINAL_RESTORE) {
                  status = a429_utl_plx_eeprom_write_enable(board, plxType, pciConfigMemPtr);
                  ar_sleep(52);
                  status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord1, PLX_BAR2_ORIGINAL_SIZE_1);
                  ar_sleep(52);
                  status = a429_utl_plx_eeprom_write(board, plxType, pciConfigMemPtr, bar2SizeWord2, PLX_BAR2_ORIGINAL_SIZE_2);
                  ar_sleep(52);
                  status = a429_utl_plx_eeprom_write_disable(board, plxType, pciConfigMemPtr);
                  ar_sleep(52);  
   
                  /* read/verify the updated content of the PROM */
                  status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
                  if ((promContent[bar2SizeWord1] != PLX_BAR2_ORIGINAL_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ORIGINAL_SIZE_2)) {
                     status = ARS_MEMWRERR;
                  }
                  else {
                     status = ARS_NORMAL;
                  }
               }
               else
                  status = ARS_NORMAL;
            }
         }
         else
            status = ARS_FAILURE;

#ifndef LABVIEW_RT
         tempStatus = cei_release_thread_lock(LOCK_IDENT);
#endif      
      }
      tempStatus = cei_release_device(board, &device[board]);
   }
   #ifdef INTEGRITY_PCI_PPC
      if (tempStatus) ;
   #endif
   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:             A 4 2 9 _ U T L _ E N H _ C H E C K
 *===========================================================================*
 *
 * FUNCTION:    Check the PLX EEPROM for CEI-x30 Enhanced firmware support
 *
 * PARAMETERS:  CEI_INT16 board        -- (input) board to reference (0 to 15)
 *              CEI_INT32  deviceIdent -- (input) PCI Device ID for device
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL     Success updating or already done
 *                              ARS_INVBOARD   Invalid board parameter
 *                              ARS_FAILURE    Invalid device (not x30)
 * 
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: This routine is a special utility routine to check
 *              CEI-x30 board support for the enhanced firmware and API.
 *
 *===========================================================================*/
CEI_INT16 a429_utl_enh_check (CEI_INT16 board, CEI_INT32 deviceIdent)
{
   CEI_INT32   status = ARS_NORMAL;      /* function return status           */
   CEI_INT32   plxType = 0;              /* type of PLX PCI intfc 9056/9030  */
   CEI_INT32   bar2SizeWord1 = 0;        /* PCI Cfg offset for PCI BAR2 Size */
   CEI_INT32   bar2SizeWord2 = 0;        /* PCI Cfg offset for PCI BAR2 Size */
   CEI_UINT32  promContent[PLX_EEPROM_SIZE]; /* contents of PLX EEPROM       */
   pCEI_UINT32 pciConfigMemPtr = 0;          /* BAR1 offset for PLX EEPROM   */

   switch (deviceIdent) {
      case PCI_DEVICE_ID_530:
      case PCI_DEVICE_ID_630:
      case PCI_DEVICE_ID_830:
      case PCI_DEVICE_ID_830RX:
      case PCI_DEVICE_ID_A30:
         plxType       = PLX_9056;
         bar2SizeWord1 = PLX_9056_BAR2_SIZE_WORD_1;
         bar2SizeWord2 = PLX_9056_BAR2_SIZE_WORD_2;
         pciConfigMemPtr = device[board].baseAddr + PLX_CONFIG_MEM_CONTROL_REG;
         break;
      case PCI_DEVICE_ID_430:
      case PCI_DEVICE_ID_430A:
      case PCI_DEVICE_ID_RAREC:
         plxType       = PLX_9030;
         bar2SizeWord1 = PLX_9030_BAR2_SIZE_WORD_1;
         bar2SizeWord2 = PLX_9030_BAR2_SIZE_WORD_2;
         pciConfigMemPtr = device[board].baseAddr + PLX9030_PCI_TARGET_RESPONSE;
         break;
      case PCI_DEVICE_ID_RAR_PCIE:
         status = ARS_INVBOARD;
         break;
      default:
         status = ARS_FAILURE;
   }

   if (status == ARS_NORMAL) {
      /* read the current content of the PROM */
      status = a429_utl_plx_eeprom_read (board, plxType, pciConfigMemPtr, promContent);
      /* if the PROM is currently programmed for the old CEI-x30 BAR2 size, update it to 512Kb */
      if ((promContent[bar2SizeWord1] != PLX_BAR2_ENHANCED_SIZE_1) || (promContent[bar2SizeWord2] != PLX_BAR2_ENHANCED_SIZE_2)) {
         status = ARS_HW_CONSISTENCY;
      }
   }
   
   
   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:   A 4 2 9 _ U T L _ F I N D _ D A C _ T H R E S H O L D
 *===========================================================================*
 *
 * FUNCTION:    Given a threshold type (lower or upper) and the bounds of the
 *              threshold search, find the DAC setting at which the IRIG 
 *              signal is lost, then return a DAC value just under that value.
 *
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL       threshold was detected
 *                           ARS_FAILURE (-1) Any issue encountered with
 *                                            incoming IRIG signal level
 *
 *===========================================================================*/
CEI_INT32 a429_utl_find_dac_threshold (
   CEI_INT16 board,                         /* the board id                  */
   CEI_INT32 useActualTimeSample,           /* IRIG edge det. vs time cmpare */
   CEI_INT32 displayDacSettings,            /* conditional flag printf invoc */
   CEI_INT32 thresholdType,                 /* find AR_LOWER or AR_UPPER threshold */ 
   CEI_UINT32 voltageLimit,                 /* outer bounds for the search   */
   CEI_UINT32 initialDacAdjustment,         /* initial DAC increment         */
   pCEI_UINT32 finalDacValue,               /* dynamic DAC value during test */
   pCEI_UINT32 dacBoudaryCushionValue       /* minor adj to final DAC value  */
)
{
   CEI_UINT32 adjustmentValue;          /* dynamic increment to DAC voltage  */
   CEI_UINT32 dacValue;                 /* dynamic DAC value during test     */
   CEI_UINT32 irigRegValue;             /* temp contents of IRIG Receiver    */
   CEI_INT32 sampleReceived = TRUE;     /* flag indicating IRIG signal recvd */
   CEI_INT32 adjustmentDirection = -1;  /* direction to apply increment      */
   CEI_INT32 signalFound = FALSE;       /* flag ind IRIG signal edge detected*/
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of board read/write access */
   CEI_UINT32 prevIrigRegValue = 0;     /* IRIG Time sample value            */
   pCEI_UINT32 irigReceiverRegPtr;      /* pointers to registers on the board*/
   pCEI_UINT32 dacControlRegPtr = &device[board].devicePtr[CEIDEV_DAC_CONTROL];
   
   if (useActualTimeSample)
      irigReceiverRegPtr = &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME];
   else
      irigReceiverRegPtr = &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE];
   
   signalFound = FALSE;
   adjustmentValue = initialDacAdjustment;
   dacValue = *finalDacValue;
   while (accessStatus == ARS_NORMAL) {
      if (useActualTimeSample)
         accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &prevIrigRegValue);
                        
      accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
      
#ifndef LABVIEW_RT
      if (displayDacSettings)
         printf ("\rFinding the %s transition -- DAC Setting = %3d", 
                 thresholdText[thresholdType], (CEI_NATIVE_INT)dacValue);
#endif
      /* Read IRIG value, clearing "edge received" bit, then  */
      /* Wait 30 Msec for an incoming signal edge transition. */
      if (useActualTimeSample)
         cei_delay_milliseconds(2000);
      else
         cei_delay_milliseconds(5);

      accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
      cei_delay_milliseconds(30);
      accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
      
      /* if the cal register is set assure it not just some noise */
      if (useActualTimeSample)
         sampleReceived = (irigRegValue == prevIrigRegValue) ? FALSE : TRUE;
      else
         sampleReceived = irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT;

      if (sampleReceived) {
         prevIrigRegValue = irigRegValue;

         /* if the cal register is set make sure its not just some noise */
         if (thresholdType == AR_LOWER)
            dacValue -= adjustmentValue;
         else
            dacValue += adjustmentValue;

         if (adjustmentValue <= 4) {
            adjustmentDirection = -1;        /* start search in a neutral direction */
            adjustmentValue = 1L;            /* with a fine granularity adjustment  */

#ifndef LABVIEW_RT
            if (displayDacSettings)
               printf ("\n");
#endif
            do {
               accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);

               if (useActualTimeSample)
                  cei_delay_milliseconds(2000);
                                                
#ifndef LABVIEW_RT
               if (displayDacSettings)
                  printf("\rFine tuning %s transition -- DAC Setting = %3d", 
                         thresholdText[thresholdType], (CEI_NATIVE_INT)dacValue);
#endif
               /* Read IRIG value, clearing "edge received" bit, then  */
               /* Wait 30 Msec for an incoming signal edge transition. */
               accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
               cei_delay_milliseconds(30);
               accessStatus = cei_read_32bit (board, irigReceiverRegPtr, &irigRegValue);
               
               if (useActualTimeSample)
                  sampleReceived = (irigRegValue == prevIrigRegValue) ? FALSE : TRUE;
               else
                  sampleReceived = irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT;
         
               if (sampleReceived) {
                  prevIrigRegValue = irigRegValue;
                  if (adjustmentDirection == 1) {
                     signalFound = TRUE;
                     *dacBoudaryCushionValue = 1;
                  } 
                  else {
                     adjustmentDirection = 0; /*false */
                     if (thresholdType == AR_LOWER)
                        dacValue -= adjustmentValue;
                     else
                        dacValue += adjustmentValue;
                  }
               } 
               else {
                  if (adjustmentDirection == 0) {
                     signalFound = TRUE;
                     *dacBoudaryCushionValue = 2;
                  } 
                  else {
                     adjustmentDirection = 1; /*true */
                     if (thresholdType == AR_LOWER)
                        dacValue += adjustmentValue;
                     else
                        dacValue -= adjustmentValue;
                  }
               }
            } while (signalFound == FALSE);
            
            break;
         }
      }
      else {
         if ( ((thresholdType == AR_LOWER) && (dacValue >= voltageLimit)) ||
              ((thresholdType == AR_UPPER) && (dacValue <= voltageLimit)) ) {
#ifndef LABVIEW_RT
            if (displayDacSettings)
               printf("\nCould not find a %s transition -- DAC Setting = %3d\n", 
                      thresholdText[thresholdType], (CEI_NATIVE_INT)dacValue);
#endif
            return ARS_FAILURE;
         }

         if (thresholdType == AR_LOWER)
            dacValue += adjustmentValue;
         else
            dacValue -= adjustmentValue;
         
      }
      if (thresholdType == AR_LOWER)
         adjustmentValue /= 2;
      else
         adjustmentValue -= 2;

      if (adjustmentValue < 4)
         adjustmentValue = 4;
   }
   *finalDacValue = dacValue;
   return ARS_NORMAL;
}


/*=================================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ D E T E R M I N E _ C H A N N E L _ S U P P O R T
 *=================================================================================*
 *
 * FUNCTION:    Determine the channel type for each channel available on the
 *              specified device
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 * 
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL    Normal succesful completion.
 *                              ARS_INVBOARD  board parameter value invalid or
 *                                            board not initialized or CSDB
 *                                            not supported on this device
 * 
 * GLOBALS:     device[]
 *
 * DESCRIPTION: This function determines the number of channels installed
 *              on the board for each channel/protocol type detected.
 *
 *===========================================================================*/
CEI_VOID  a429_utl_determine_channel_support (CEI_INT16 board)
{
   CEI_INT32   accessStatus = ARS_NORMAL; /* status of device r/w access     */
   CEI_INT32   channelIndex;              /* local channel index             */
   CEI_UINT32  channelType;               /* channel type for each channel   */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */

   deviceGlobal[board]->alternateJconfiguration = 0;
   
   /* initialize the channel counters to their "uninitialized" state */
   for (channelIndex = 0; channelIndex < NUMBER_OF_CHANNEL_TYPES; channelIndex++) {
      deviceGlobal[board]->numberOfChannels[channelIndex] = 0;
      deviceGlobal[board]->firstChannel[channelIndex] = CEIDEV_MAX_CHANNELS;  
      deviceGlobal[board]->lastChannel[channelIndex] = CEIDEV_MAX_CHANNELS;
   }

   deviceGlobal[board]->numTx = 0;
   deviceGlobal[board]->numRx = 0;

   /* identify each receive channel and setup API local channel validation information */
   boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START - CEIDEV_CHANNEL_REGION_SIZE);
   for (channelIndex = 0; channelIndex < CEIDEV_MAX_CHANNELS; channelIndex++) {
      boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
      
      /* retrieve the channel type (and protocol) for this channel index */
      accessStatus = cei_read_32bit (board, boardPointer, &channelType);
      channelType = (channelType & CEIDEV_CSR_CHAN_TYPE_MASK) >> CEIDEV_CSR_CHAN_TYPE_SHIFT;

      /* identify the receive channels and tally the protocol channel count */
      if (channelType != CEIDEV_CH_TYPE_NONE) {

         switch (channelType) {

            /* identify the serial data channels and tally total device receive channel count */
            /* then fall through to next case to tally individual protocol channel counts     */
            case CEIDEV_CH_TYPE_ARINC573_RX:
               deviceGlobal[board]->alternateJconfiguration = CEIDEV_ALT_CONFIGURATION_717;

            case CEIDEV_CH_TYPE_ARINC429_RX:
            case CEIDEV_CH_TYPE_ARINC561_RX:
            case CEIDEV_CH_TYPE_SERIAL_RX:
               deviceGlobal[board]->numRx++;

            /* identify state-type channels and tally only individual discrete/digital channel count */
            case CEIDEV_CH_TYPE_DISC_IN:
            case CEIDEV_CH_TYPE_DIGL_IN:
            case CEIDEV_CH_TYPE_DIFF_IN:
               deviceGlobal[board]->numberOfChannels[channelType]++; 
               if (deviceGlobal[board]->firstChannel[channelType] == CEIDEV_MAX_CHANNELS) {
                  deviceGlobal[board]->firstChannel[channelType] = channelIndex;  
                  deviceGlobal[board]->lastChannel[channelType] = channelIndex;  
               }
               else
                  deviceGlobal[board]->lastChannel[channelType]++;  
               break;
   
            case CEIDEV_CH_TYPE_MERGED_RX:
               deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] = channelIndex;
               break;

            /* identify the serial data channels and tally total device transmit channel count */
            /* then fall through to next case to tally individual protocol channel counts      */
            case CEIDEV_CH_TYPE_ARINC573_TX:
               deviceGlobal[board]->alternateJconfiguration = CEIDEV_ALT_CONFIGURATION_717;

            case CEIDEV_CH_TYPE_ARINC429_TX:
            case CEIDEV_CH_TYPE_ARINC561_TX:
            case CEIDEV_CH_TYPE_SERIAL_TX:
               deviceGlobal[board]->numTx++;
               
            /* identify state-type channels and tally only individual discrete/digital channel count */
            case CEIDEV_CH_TYPE_DISC_OUT:
            case CEIDEV_CH_TYPE_DIGL_OUT:
            case CEIDEV_CH_TYPE_DIFF_OUT:
               deviceGlobal[board]->numberOfChannels[channelType]++; 
               if (deviceGlobal[board]->firstChannel[channelType] == CEIDEV_MAX_CHANNELS) {
                  deviceGlobal[board]->firstChannel[channelType] = channelIndex;  
                  deviceGlobal[board]->lastChannel[channelType] = channelIndex;  
               }
               else
                  deviceGlobal[board]->lastChannel[channelType]++;  
               break;
      
            default:
               ;
         }
      }
   }
   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
}


/*===========================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ I N I T I A L I Z E _ G L O B A L _ D A T A
 *===========================================================================*
 *
 * FUNCTION:    Function that initializes most global data
 * 
 * PARAMETERS:  CEI_INT16 board       -- (input) board to reference (0 to 15)
 * 
 * RETURN VAL:  none
 *
 * GLOBALS:     deviceGlobal[board]->firstChannel
 *
 * DESCRIPTION: This routine is used to initialize global data structures
 *              and arrays shared between the API and the operating system
 *              specifically-sized arrays based on MAX_BOARDS.
 *
 *===========================================================================*/
CEI_VOID  a429_utl_initialize_global_data 
(
   CEI_INT16 board                      /* board to reference (0 to 15)      */
)
{
   CEI_INT32 index;                                     /* general use index */

   interruptQueue[board].intQueueTailPtr = 0;
   interruptQueue[board].intQueueHeadPtr = 0;
   bypassWrapTest[board] = TRUE;
   irigDacThreshold[board] = CEIDEV_IRIG_DAC_INIT_THRESHOLD;
   for (index = 0; index < NUM_SUBFRAME_SYNC_WORDS; index++) {
      a717syncWords[board][index] = 0;
   }
}


/*===========================================================================*
 * ENTRY POINT:      A 4 2 9 _ U T L _ M E M O R Y _ T E S T
 *===========================================================================*
 *
 * FUNCTION:    Perform a memory test of the CEI-x30 SRAM device
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 * 
 *              CEI_INT32 testType -- (input) size of memory test to execute
 *                                      FULL_SRAM   0 test all SRAM
 *                                      UNUSED_SRAM 1 test only unused SRAM (Tx Chan 16-31)
 *                                      SELECT_SRAM 2 test a portion of SRAM
 * 
 *              CEI_INT32 testOption -(input) when the testType is SELECT_SRAM, 
 *                                    the test option parameter specifies which 
 *                                    block of SRAM will be verified. Block select
 *                                    is based on a length of 512 locations 
 *                                    starting at SRAM address = 512*testOption
 * 
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL     Normal succesful completion.
 *                              ARS_INVARG     Invalid memory test request
 *                              ARS_MEMWRERR   Memory test failure
 * GLOBALS:     none
 *
 * DESCRIPTION: This function performs SRAM memory tests, either on a small
 *              unused portion (active processing test) a selected portion of, 
 *              or on the entire SRAM (destructive test that should only be 
 *              executed when the board is disabled, duration is six seconds
 *              to test all 2Mb of SRAM).
 *
 *===========================================================================*/
CEI_INT16 a429_utl_memory_test (CEI_INT16 board, CEI_INT32 testType, CEI_INT32 testOption)
{
   CEI_INT16 status = ARS_NORMAL;       /* status for the memory test        */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_UINT32 writeValue;               /* the data pattern written to memory*/
   CEI_UINT32 readValue;                /* the data pattern read from memory */
   CEI_UINT32 memoryIndex = 0;          /* local index into memory block     */
   CEI_UINT32 minMemoryIndex = 0;       /* starting location of memory test  */
   CEI_UINT32 maxMemoryAddress = 0;     /* ending location of memory test    */
   CEI_UINT32 testAddress;              /* temp address to write/read        */
   CEI_INT32 patternIndex;              /* index into data patterns (6 total)*/
   
   switch (testType) {
      /*  During active CEI-x30 processing, only test six locations in unused Tx Channel 16-31
       *  FIFO Buffer SRAM, verifying specific Address Bit access with varying data patterns.
       */
      case UNUSED_SRAM:
         minMemoryIndex = 0;
         maxMemoryAddress = AR_NUM_UNUSED_SRAM_LOCATIONS;
         break;

      case FULL_SRAM:
         status = ar_stop (board);
         minMemoryIndex = 0;
         maxMemoryAddress = CEIDEV_SRAM_MAX;
         break;

      case SELECT_SRAM:
         if ( (testOption >= AR_BIT_SELECT_SRAM_MIN) && (testOption <= AR_BIT_SELECT_SRAM_MAX) ) {
            testOption -= AR_BIT_SELECT_SRAM_MIN;
            minMemoryIndex = (CEI_UINT32)testOption * SELECT_SRAM_TEST_BLOCK_SIZE;
            maxMemoryAddress = minMemoryIndex + SELECT_SRAM_TEST_BLOCK_SIZE;
            status = ar_stop (board);
         }
         else
            status = ARS_INVARG;
         break;

      default:
         status = ARS_INVARG;
   }

   /* Loop through all of the various patterns to test SRAM */
   for (patternIndex = 0; (patternIndex < AR_NUM_PATTERNS) && (status == ARS_NORMAL); patternIndex++) {
      writeValue = patternTest[patternIndex];

      /* Write the pattern out to SRAM in CEI_INT32-word accesses */
      for (memoryIndex = minMemoryIndex; (memoryIndex < maxMemoryAddress) && (accessStatus == ARS_NORMAL); memoryIndex++) {

         /* manipulate the pattern as requested */
         if (patternProcess[patternIndex] == AR_SHIFT) {
            if ( ((memoryIndex+1) % 32) == 0 ) {
               writeValue = patternTest[patternIndex];
            }
            else {
               writeValue = writeValue << 1;
               writeValue += patternInc[patternIndex];
            }
         }
         else if (patternProcess[patternIndex] == AR_ADD) 
            writeValue += patternInc[patternIndex];

         testAddress = (testType == UNUSED_SRAM) ? unusedSram[memoryIndex] : memoryIndex;
      
         /* load the SRAM CEI_INT32-word address index */
         /* write the data pattern to this location */
         accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], testAddress);
         accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], writeValue);
      }
      
      /* Read and verify the written pattern */
      writeValue = patternTest[patternIndex];
      for (memoryIndex = minMemoryIndex; (memoryIndex < maxMemoryAddress) && (accessStatus == ARS_NORMAL); memoryIndex++) {

         /* manipulate the pattern as requested */
         if (patternProcess[patternIndex] == AR_SHIFT) {
            if ( ((memoryIndex+1) % 32) == 0 ) {
               writeValue = patternTest[patternIndex];
            }
            else {
               writeValue = writeValue << 1;
               writeValue += patternInc[patternIndex];
            }
         }
         else if (patternProcess[patternIndex] == AR_ADD) 
            writeValue += patternInc[patternIndex];

         testAddress = (testType == UNUSED_SRAM) ? unusedSram[memoryIndex] : memoryIndex;
      
         /* load the SRAM CEI_INT32-word address index */
         /* read and verify the contents of this location */
         accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], testAddress);
#if defined (PPC_SYNC)
         IO_SYNC;
#endif
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], &readValue);
         
         if (readValue != writeValue) {
            status = ARS_MEMWRERR;
            break;
         }
      }
   }

   if ( (status == ARS_NORMAL) && (accessStatus != ARS_NORMAL) )
      status = (CEI_INT16)accessStatus;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:        A 4 2 9 _ U T L _ W R A P _ T E S T
 *===========================================================================*
 *
 * FUNCTION:    Perform an internal wrap test using the specified number of 
 *              ARINC 429 messages on each transmit/receive channel pairs
 *              installed on the device.
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 * 
 *              CEI_INT32 wrap     -- (input) type of wrap to enable: 
 *                                      AR_WRAP_ON  for internal wrap 
 *                                      AR_WRAP_OFF for external wrap
 * 
 *              CEI_INT32 msgCount -- (input) number of messages to transmit 
 *                                      limited to AR_WRAP_TEST_MSG_COUNT
 * 
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL          Normal succesful completion.
 *                           ARS_INVHARVAL       Item was invalid or channel
 *                                               not supported on this device
 *                           ARS_INVBOARD        Board parameter value invalid or
 *                                               board has not been initialized
 *                           ARS_BOARD_MUTEX     Access to the Board Lock timed-out/failed
 *                           ARS_WRAP_DROP_FAIL  Wrap Data Missing
 *                           ARS_WRAP_DATA_FAIL  Wrap Data Pattern Mismatch
 *                           ARS_WRAP_FLUSH_FAIL Unknown External data received
 *                           ARS_XMITOVRFLO      Transmit buffer overrun has occurred
 * GLOBALS:     none
 *
 * DESCRIPTION: This function performs an internal wrap test based on a  
 *              specified number of ARINC 429 messages on each transmit/
 *              receive channel pair installed on the device.  Configuration
 *              parameters are set for no parity and 100Kbps.
 *
 *===========================================================================*/
CEI_INT16 a429_utl_wrap_test (CEI_INT16 board, CEI_INT32 wrap, CEI_INT32 msgCount)
{
   CEI_INT16 numTxChannels;                   /* number of transmitters/rcvrs*/
   CEI_INT16 channelIndex;                    /* index of transmitters/rcvrs */
   CEI_INT16 status = ARS_NORMAL;             /* function return status      */
   CEI_UINT32 currentTxConfig[MAX_CHANNEL];   /* saved Chan Config on entry  */
   CEI_UINT32 currentRxConfig[MAX_CHANNEL];   /* saved Chan Config on entry  */
   CEI_INT32 bitFieldValue;                   /* temp Chan Config variable   */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_UINT16 testIndex = 0;                  /* message counter/pattern gen */
   CEI_UINT32 deviceChannel;                  /* general channel index       */
   CEI_UINT32 channelStatus;            /* Rx channel FIFO data avail status */
   CEI_UINT32 globalEnableRegData;      /* contents of the Global Enable reg */
   ARINC testOutput;                          /* Output test data            */
   ARINC testInput[AR_WRAP_TEST_MSG_COUNT+2]; /* Received data               */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer;        /* local pointer to channel memory area */

   /* limit the wrap test message count */
   if (msgCount > AR_WRAP_TEST_MSG_COUNT) {
      msgCount = AR_WRAP_TEST_MSG_COUNT;
   }

   numTxChannels = (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]; 

   /* if multi-thread protection requested, acquire semaphore/mutex-locking */
   if (device[board].multiThreadProtect) {
      accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
   }
   
   /* save the existing channel global enable register state, disable data processing temporarily */
   accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
   accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegData & CEIDEV_GLOBAL_ENABLE_MASK);

   /* save the existing channel configuration for all transmit and receive channels */
   boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE));
   for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX], channelIndex = 0; 
        deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel++, channelIndex++) {
      accessStatus = cei_read_32bit (board, boardPointer, &currentTxConfig[channelIndex]);
      boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
   }
   boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE));
   for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX], channelIndex = 0; 
        deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel++, channelIndex++) {
      accessStatus = cei_read_32bit (board, boardPointer, &currentRxConfig[channelIndex]);
      boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
   }
   
   /* Configure each transmit/receive channel pair for 100Kbps with parity disabled 
    * Additionally configure the wrap, for internal wrap disable external transmission.
    */
   boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE));
   bitFieldValue = CEIDEV_FAST_SLEW_RATE + CEIDEV_CHANNEL_ENABLE + (CEIDEV_BAUD_RATE_100K << CEIDEV_BAUD_RATE_SHIFT);
   if (wrap == AR_WRAP_ON)
      bitFieldValue += CEIDEV_TRANSMIT_DISABLE;
   for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel++) {
      accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
      boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
   }
   boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE));
   bitFieldValue = CEIDEV_FAST_SLEW_RATE + CEIDEV_CHANNEL_ENABLE + (CEIDEV_BAUD_RATE_100K << CEIDEV_BAUD_RATE_SHIFT);
   if (wrap == AR_WRAP_ON)
      bitFieldValue += CEIDEV_INTERNAL_WRAP_ENABLE;
   for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel++) {
      accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
      boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
   }
      
   /* enable ARINC data processing */
   accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegData | CEIDEV_GLOBAL_ENABLE_ON);
   cei_delay_milliseconds(1);

   for (channelIndex = 0; (channelIndex < numTxChannels) && (status == ARS_NORMAL); channelIndex++) {
      /* check for an actively transmitting connection - bail if so */
      status = ARS_NORMAL;
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + 
                                   CEIDEV_CHANNEL_STATUS + (channelIndex * CEIDEV_CHANNEL_REGION_SIZE));
      for (testIndex = 0; (testIndex < 10) && (status == ARS_NORMAL); testIndex++) {
         accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
         if ( (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL) ) {
            status = ARS_WRAP_FLUSH_FAIL;
            break;
         }
      }
   
      /*---------------------------------------------------------------*
       *  Test ARINC 429 tranmsit & receive
       *---------------------------------------------------------------*/
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
         ((channelIndex + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]) * CEIDEV_CHANNEL_REGION_SIZE));
      bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
      for (testIndex = 0; (testIndex < msgCount) && (status == ARS_NORMAL); testIndex++) { 
         testOutput = (CEI_UINT32)(testIndex + ((testIndex+1) << 16) + 1);
         testInput[0] = 0;
         /* check the transmit channel status, if it indicates "NOT FULL"
          * then stuff the data into the queue; otherwise, it's an overflow.
          */
         accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
         if (accessStatus == ARS_NORMAL) {
            if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
               status = ARS_XMITOVRFLO;
            }
            else {
               /* stuff the data into the transmit FIFO */
               accessStatus = cei_write_32bit (board, bufferPointer, testOutput);
               status = ARS_NORMAL;
            }
         }
      }
      /* delay for the initial transmission to begin then more for each message transmitted */
      if (status == ARS_NORMAL) {
         cei_delay_milliseconds(5 + msgCount);
      }
      /* attempt to receive the specified burst count */
      if (status == ARS_NORMAL) {
         status = ARS_NORMAL;
         testIndex = 0;
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      (channelIndex * CEIDEV_CHANNEL_REGION_SIZE));
         bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
         while ( (status == ARS_NORMAL) && (testIndex < msgCount+1) ) {
            accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
            if ( (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL) ) {
               accessStatus = cei_read_32bit (board, bufferPointer, &testInput[testIndex]);
               testIndex++;
            }
            else {
               status = ARS_NODATA;
            }
         }
         
         /* check to see we received the correct number of messages */
         if (testIndex < msgCount)
            status = ARS_WRAP_DROP_FAIL;
         else if (testIndex > msgCount)
            status = ARS_WRAP_FLUSH_FAIL;
         else
            status = ARS_NORMAL;
      }

      /* if all messages were received, verify the data content for all received messages */
      if (status == ARS_NORMAL) {
         for ( testIndex = 0; (testIndex < msgCount) && (status == ARS_NORMAL); testIndex++ ) { 
            if (testInput[testIndex] != (CEI_UINT32)(testIndex + ((testIndex+1) << 16) + 1)) {
               status = ARS_WRAP_DATA_FAIL;
            }
         } 
      }
   }

   /* restore the previous global enable register setting */
   accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegData);

   /* restore the previous channel configuration for all transmit and receive channels */
   boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE));
   for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX], channelIndex = 0; 
        deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
        deviceChannel++, channelIndex++) {
      accessStatus = cei_write_32bit (board, boardPointer, currentTxConfig[channelIndex]);
      boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
   }
   boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE));
   for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX], channelIndex = 0; 
        deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
        deviceChannel++, channelIndex++) {
      accessStatus = cei_write_32bit (board, boardPointer, currentRxConfig[channelIndex]);
      boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
   }
   /* if multi-thread protection requested, free semaphore/mutex-locking */
   if (device[board].multiThreadProtect) {
      accessStatus = cei_release_thread_lock(LOCK_IDENT);
   }

   return status;
}

/*===========================================================================*
 * ENTRY POINT:      A 4 2 9 _ U T L _ P L X _ E E P R O M _ R E A D
 *===========================================================================*
 *
 * FUNCTION:    Read and return the contents of the PLX EEPROM
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr -- (input) ptr to the BAR0 memory region
 * 

 *              pCEI_UINT32 promValues -- (output) ptr to array to store values
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function reads the PLX PROM and returns the values.
 *
 *===========================================================================*/
CEI_INT32 a429_utl_plx_eeprom_read 
(
   CEI_INT16   board, 
   CEI_INT32   pciInterfaceType, 
   pCEI_UINT32 pciConfigPtr, 
   pCEI_UINT32 promValues
)
{
   CEI_UINT32 copy;
   CEI_UINT32 readValue;
   CEI_UINT32 x;
   CEI_UINT32 promData;
   CEI_UINT32 addr;
   pCEI_UINT32 promValuesPtr = promValues;
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */

   promData = 0;
   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056) 
      copy &= ~EEDS;

   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Set the EEDI bit high (Start bit) */
   copy |=  EEDI;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (5) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (7) Clear EEDI */
   copy &= ~EEDI;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (8) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (9) Toggle EESK high, then low, 8 times (serial EEPROM address 0) */
   for (x = 0; x < 8; x++) {
      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }

   /* (10) Set the EEDO Input Enable to 1 to float the EEDO input for reading */
   copy |= (pciInterfaceType == PLX_9056) ? EEDS : EEDO;

   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (11) Toggle EESK high, then low, 16 times, (12) reading in a bit after each clock */
   for (addr = 0; addr < PLX_EEPROM_SIZE; addr++) {
      promData = 0;
      for (x = 0; x <= 15; x++) {
         copy |=  EESK;
         status = WRITE_32_BIT (board, pciConfigPtr, copy);
         copy &= ~EESK;
         status = WRITE_32_BIT (board, pciConfigPtr, copy);

         status = READ_32_BIT (board, pciConfigPtr, &readValue);
         if (readValue & EEDO)
            promData |= 0x0001;
         else
            promData &= 0xFFFE;

         if (x!=15)
            promData = promData << 1;
         else {
            *promValuesPtr = promData;
            promValuesPtr++;
         }
      }
   }

   /* (13) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (14) Toggle EESK high, then low.  The read is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   return status;
}


/*=============================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E _ E N A B L E
 *=============================================================================*
 *
 * FUNCTION:    Enable write access to the PLX PROM
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr -- (input) ptr to the BAR0 memory region
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function enables write access to the PLX PROM.
 *
 *===========================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write_enable 
(
   CEI_INT16 board, 
   CEI_INT32 pciInterfaceType, 
   pCEI_UINT32 pciConfigPtr
)
{
   CEI_UINT32 copy;
   CEI_INT32  x;
   CEI_INT32  y[11] = {1,0,0,1,1,0,0,0,0,0,0};
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */

   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056) 
      copy &= ~EEDS;

   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Clock in the 11 bits (Start bit, cmd bits, etc.) */

   for (x = 0; x <= 10; x++) {
      if (y[x] == 0) 
         copy &= ~EEDI;
      else
         copy |=  EEDI;

      status = WRITE_32_BIT (board, pciConfigPtr, copy);

      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }


   /* (5) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low.  The command is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   return status;
}


/*===============================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E _ D I S A B L E
 *===============================================================================*
 *
 * FUNCTION:    Disable write access to the PLX PROM
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr -- (input) ptr to the BAR0 memory region
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function disables write access to the PLX PROM.
 *
 *===============================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write_disable 
(
   CEI_INT16 board, 
   CEI_INT32 pciInterfaceType, 
   pCEI_UINT32 pciConfigPtr
)
{
   CEI_UINT32 copy;
   CEI_INT32  x;
   CEI_INT32  y[11] = {1,0,0,0,0,0,0,0,0,0,0};
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */

   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056) 
      copy &= ~EEDS;

   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Clock in the 11 bits (Start bit, cmd bits, etc.) */

   for (x = 0; x <= 10; x++) {
      if (y[x] == 0) 
         copy &= ~EEDI;
      else
         copy |=  EEDI;
      
      status = WRITE_32_BIT (board, pciConfigPtr, copy);

      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }


   /* (5) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low.  The command is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   return status;
}

/*=============================================================================*
 * ENTRY POINT:      A 4 2 9 _ U T L _ P L X _ E E P R O M _ W R I T E
 *=============================================================================*
 *
 * FUNCTION:    Write a value to the specified PLX PROM location
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 * 
 *              CEI_INT32 pciInterfaceType -- (input) PLX Part (9030 or 9056)
 * 
 *              pCEI_UINT32 pciConfigPtr -- (input) ptr to the BAR0 memory region
 * 
 *              CEI_UINT32 address -- (input) offset into the PLX PROM to write
 * 
 *              CEI_UINT32 data    -- (input) value to write to the PLX PROM
 * 
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL     Normal succesful completion.
 *                           other values   Platform specific device access failures
 * 
 * GLOBALS:     none
 *
 * DESCRIPTION: This function writes a value to the specified PLX PROM location.
 *
 *===========================================================================*/
CEI_INT32 a429_utl_plx_eeprom_write 
(
   CEI_INT16 board, 
   CEI_INT32 pciInterfaceType, 
   pCEI_UINT32 pciConfigPtr, 
   CEI_UINT32 address, 
   CEI_UINT32 data
)
{
   CEI_UINT32 copy;
   CEI_INT32  x;
   CEI_INT32  y[27] = {1, 0,1, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0};
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */

   if (address & 0x0080)  y[3]  = 1;
   if (address & 0x0040)  y[4]  = 1;
   if (address & 0x0020)  y[5]  = 1;
   if (address & 0x0010)  y[6]  = 1;
   if (address & 0x0008)  y[7]  = 1;
   if (address & 0x0004)  y[8]  = 1;
   if (address & 0x0002)  y[9]  = 1;
   if (address & 0x0001)  y[10] = 1;
   if (data    & 0x8000)  y[11] = 1;
   if (data    & 0x4000)  y[12] = 1;
   if (data    & 0x2000)  y[13] = 1;
   if (data    & 0x1000)  y[14] = 1;
   if (data    & 0x0800)  y[15] = 1;
   if (data    & 0x0400)  y[16] = 1;
   if (data    & 0x0200)  y[17] = 1;
   if (data    & 0x0100)  y[18] = 1;
   if (data    & 0x0080)  y[19] = 1;
   if (data    & 0x0040)  y[20] = 1;
   if (data    & 0x0020)  y[21] = 1;
   if (data    & 0x0010)  y[22] = 1;
   if (data    & 0x0008)  y[23] = 1;
   if (data    & 0x0004)  y[24] = 1;
   if (data    & 0x0002)  y[25] = 1;
   if (data    & 0x0001)  y[26] = 1;

   status = READ_32_BIT (board, pciConfigPtr, &copy);

   /* (1) Clear the Input Enable Bits */
   copy &= ~EECS;
   copy &= ~EESK;
   copy &= ~EEDI;
   if (pciInterfaceType == PLX_9056) 
      copy &= ~EEDS;
      
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (2) Set EECS high */
   copy |=  EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (3) Toggle EESK high, then low */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (4) Clock in the 27 bits (Start bit, cmd bits, addr, data) */

   for (x = 0; x <= 26; x++) {
      if (y[x] == 0) 
         copy &= ~EEDI;
      else
         copy |=  EEDI;

      status = WRITE_32_BIT (board, pciConfigPtr, copy);

      copy |=  EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
      copy &= ~EESK;
      status = WRITE_32_BIT (board, pciConfigPtr, copy);
   }


   /* (5) Clear the EECS bit */
   copy &= ~EECS;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);

   /* (6) Toggle EESK high, then low.  The command is complete */
   copy |=  EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   copy &= ~EESK;
   status = WRITE_32_BIT (board, pciConfigPtr, copy);
   ar_sleep (15);  /* Write time */
   return status;
}


/*===========================================================================* 
 * ENTRY POINT:  A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 3 0
 *===========================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9030
 *              PCI interface and Xilinx FPGA, supporting 8-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===========================================================================*/
CEI_INT32 a429_utl_program_xilinx_9030
(
   CEI_INT16   board,            /* (input)  board to reference (0 to 15)    */
   CEI_INT32   deviceIdent,      /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,      /* (input)  size of FPGA program load       */
   pCEI_UCHAR  fwloadPtr         /* (input)  pointer to the FPGA program     */
)
{
#if defined(FW_FILE_LOAD)
   CEI_UINT32 ucProgramIndex;           /* index into the fpga load array    */
   CEI_UCHAR  fwDataCount;              /* the count for the next sequence   */
#endif   
   CEI_UINT32 tempData;                 /* temp variable for read/write      */
   CEI_UINT32 programIndex;             /* temporary FPGA program load index */
   CEI_UCHAR  programData;              /* temporary FPGA program data       */
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */
   pCEI_UINT32 pciConfigPtr;            /* dev PCI config virt base address  */
   pCEI_UINT32 localMemPtr;             /* dev local memory virt base address*/
   
#ifdef FW_FILE_LOAD
   unsigned char data, temp;
   int state = 0;
   FILE *bit_file_ptr;                  /* input file pointer */
   CEI_UINT32 fdata[] = {0xFF,0xFF,0xFF,0xFF,0x55,0x99};
#endif

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;

   /**************************************************************
    * Initialize PLX PCI9030 Local Configuration Register Set
    **************************************************************/
   /* Perform a software reset w/delay */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_SET_SW_RESET); 
   cei_delay_milliseconds(20);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_DEFAULT); 
   
   /* Delay for 100 milliseconds to allow all configuration memory to clear */
   cei_delay_milliseconds(100);

   status = cei_read_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], &tempData);
   
   /* set the local addr space 0 bus region descriptor */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], PLX9030_LOC_ADDR_SPACE_0_DESC); 

   /* Setup the General Purpose I/O Control register to program the Xilinx  */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   cei_delay_milliseconds(10);
   
#ifndef FW_FILE_LOAD
   /* download the FPGA program file */
   for (programIndex=0; programIndex < programSize; programIndex++) {
      programData = *fwloadPtr++;
      status = cei_write_32bit (board, &localMemPtr[0], (CEI_UINT32)programData);
   }
#else
   if ((bit_file_ptr = fopen("cei430.bit", "rb")) != NULL) {
      status = cei_write_32bit (board, &localMemPtr[0], fdata[0]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[1]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[2]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[3]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[4]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[5]);
      while (fread(&data, 1, 1, bit_file_ptr) == 1) {
         switch (state) {
            case 0: if      (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 1: if      (data==0xFF) state=2;
                    else                 state=0;
                    break;
            case 2: if      (data==0xFF) state=3;
                    else                 state=0;
                    break;
            case 3: if      (data==0xFF) state=4;
                    else                 state=0;
                    break;
            case 4: if      (data==0xAA) state=5;
                    else if (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 5: if      (data==0x99) state=6;
                    else if (data==0xFF) state=1;
                    else                 state=0;
                    break;
            case 6: /* we need to reverse bit ordering for parallel loading */
                    temp = 0;
                    if (data & 0x80)
                       temp |= 0x01;
                    if (data & 0x40)
                       temp |= 0x02;
                    if (data & 0x20)
                       temp |= 0x04;
                    if (data & 0x10)
                       temp |= 0x08;
                    if (data & 0x08)
                       temp |= 0x10;
                    if (data & 0x04)
                       temp |= 0x20;
                    if (data & 0x02)
                       temp |= 0x40;
                    if (data & 0x01)
                       temp |= 0x80;
 
                    status = cei_write_32bit (board, &localMemPtr[0], temp);
                    break;
         }
      }
      
      fclose(bit_file_ptr);
   }
#endif
   
   /* flush the PLX write fifo */
   for (programIndex=0; programIndex < 3; programIndex++) {  
      status = cei_write_32bit (board, &localMemPtr[0], (CEI_UINT32)0);
   }

   cei_delay_milliseconds(3);
   status = cei_write_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], tempData);
   
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);

   /* Delay for 10 milliseconds to allow the FPGA to complete load/execution */
   cei_delay_milliseconds(10);
   status = cei_read_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], &tempData);
   if ((tempData & PLX9030_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      return ARS_BADLOAD;
   }
   
   /* read the FPGA program load checksum, write it back for load confirmation */
   status = cei_read_32bit (board, &pciConfigPtr[PLX9030_CHIP_SEL_0_BASE_ADD], &tempData);
#if defined DEBUG_PROGRAMMING
   #include "dbg_30defs.h"
#else
   deviceIdent = 0;
#endif
   tempData &= PLX9030_CONFIG_MEM_CSM_MASK;
   status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], tempData);
   cei_delay_milliseconds(3);
   #ifdef INTEGRITY_PCI_PPC
      if (deviceIdent) ;
   #endif
   return status;
}


/*===================================================================================* 
 * ENTRY POINT: A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 3 0 _ 3 2 b i t
 *===================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9030
 *              PCI interface and Xilinx FPGA, supporting 32-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9030_32bit
(
   CEI_INT16    board,           /* (input)  board to reference (0 to 15)    */
   CEI_INT32    deviceIdent,     /* (input)  PCI Configuration Device ID     */
   CEI_UINT32   programSize,     /* (input)  size of FPGA program load       */
   pCEI_UINT32  fwloadPtr        /* (input)  pointer to the FPGA program     */
)
{
#if defined(FW_FILE_LOAD)
   CEI_UINT32 fwDataWord;               /* the first data word in sequence   */
   CEI_UINT32 fwDataCount;              /* the count for the next sequence   */
   CEI_UINT32 ucProgramIndex;           /* index into the fpga load array    */
   CEI_NATIVE_INT lineCtr = 5;
#endif   
   CEI_UINT32 tempData;                 /* temp variable for read/write      */
   CEI_UINT32 programIndex;             /* temporary FPGA program load index */
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */
   pCEI_UINT32 pciConfigPtr;            /* dev PCI config virt base address  */
   pCEI_UINT32 localMemPtr;             /* dev local memory virt base address*/
   
#ifdef FW_FILE_LOAD
   unsigned char data, byte3, byte2, byte1, byte0;
   int state = 0;
   FILE *bit_file_ptr;                  /* input file pointer */
   CEI_UINT32 fdata[] = {0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xAA995566};
#endif

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;

   /**************************************************************
    * Initialize PLX PCI9030 Local Configuration Register Set
    **************************************************************/
   /* Perform a software reset w/delay */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_SET_SW_RESET); 
   cei_delay_milliseconds(20);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_PCI_TARGET_RESPONSE], PLX9030_CTRL_REG_DEFAULT); 
   
   /* Delay for 100 milliseconds to allow all configuration memory to clear */
   cei_delay_milliseconds(100);

   status = cei_read_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], &tempData);
   
   /* set the local addr space 0 bus region descriptor */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], PLX9030_LOC_ADDR_SPACE_0_DESC); 

   /* Setup the General Purpose I/O Control register to program the Xilinx  */
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   cei_delay_milliseconds(10);
   
#ifndef FW_FILE_LOAD
   /* load the firmware by writing to any location in the local bus memory region */
   for (programIndex=0; programIndex < programSize; programIndex++) {
      status = cei_write_32bit (board, &localMemPtr[0], *fwloadPtr++);
   }
#else
   if ((bit_file_ptr = fopen("fpga_ec.bit", "rb")) == NULL) {
      /* load the firmware by writing to any location in the local bus memory region */
      for (programIndex=0; programIndex < programSize; programIndex+=2) {
         fwDataCount = *fwloadPtr++;
         fwDataWord = *fwloadPtr++;
         lineCtr++;
         if (fwDataWord == 0) {
            for (ucProgramIndex=0; ucProgramIndex < fwDataCount; ucProgramIndex++) {
               status = cei_write_32bit (board, &localMemPtr[0], 0);
            }
         }
         else {
            status = cei_write_32bit (board, &localMemPtr[0], fwDataWord);
            if (--fwDataCount > 0) {
               for (ucProgramIndex=0; ucProgramIndex < fwDataCount; ucProgramIndex++) {
                  status = cei_write_32bit (board, &localMemPtr[0], *fwloadPtr++);
                  programIndex++;
               }
            }
         }
      }
   }
   else {
      status = cei_write_32bit (board, &localMemPtr[0], fdata[0]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[1]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[2]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[3]);
      status = cei_write_32bit (board, &localMemPtr[0], fdata[4]);
      while (fread(&data, 1, 1, bit_file_ptr) == 1) {
         switch (state) {
             case 0: if      (data==0xFF) state=1;
                     else                 state=0;
                     break;
             case 1: if      (data==0xFF) state=2;
                     else                 state=0;
                     break;
             case 2: if      (data==0xFF) state=3;
                     else                 state=0;
                     break;
             case 3: if      (data==0xFF) state=4;
                     else                 state=0;
                     break;
             case 4: if      (data==0xAA) state=5;
                     else if (data==0xFF) state=1;
                     else                 state=0;
                     break;
             case 5: if      (data==0x99) state=8;
                     else if (data==0xFF) state=1;
                     else                 state=0;
                     break;
             case 6: state = 7;
                     byte3 = data;
                     break;
             case 7: state = 8;
                     byte2 = data;
                     break;
             case 8: state = 9;
                     byte1 = data;
                     break;
             case 9: state = 6;
                     byte0 = data;
                     fwDataWord = (byte3 << 24) + (byte2 << 16) + (byte1 << 8) + byte0;
      
                     status = cei_write_32bit (board, &localMemPtr[0], fwDataWord);
                     byte3 = 0;
                     byte2 = 0;
                     byte1 = 0;
                     byte0 = 0;
                     break;
         }
      }
      
      fclose(bit_file_ptr);
   }
#endif
   
   /* flush the PLX write fifo */
   for (programIndex=0; programIndex < 3; programIndex++) {  
      status = cei_write_32bit (board, &localMemPtr[0], (CEI_UINT32)0);
   }

   cei_delay_milliseconds(3);
   status = cei_write_32bit  (board, &pciConfigPtr[PLX9030_LOC_AD_SP_BUS0_DESC], tempData);
   
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_3);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_2);
   cei_delay_milliseconds(10);
   status = cei_write_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], PLX9030_GPIO_CTRL_REG_1);

   /* Delay for 10 milliseconds to allow the FPGA to complete load/execution */
   cei_delay_milliseconds(10);
   status = cei_read_32bit (board, &pciConfigPtr[PLX9030_GEN_PURP_IO_CTRL], &tempData);
   if ((tempData & PLX9030_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      return ARS_BADLOAD;
   }
   
   /* read the FPGA program load checksum, write it back for load confirmation */
   status = cei_read_32bit (board, &pciConfigPtr[PLX9030_CHIP_SEL_0_BASE_ADD], &tempData);
#if defined DEBUG_PROGRAMMING
   #include "dbg_30defs.h"
#else
   deviceIdent = 0;
#endif
   tempData &= PLX9030_CONFIG_MEM_CSM_MASK;
   status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], tempData);
   cei_delay_milliseconds(3);
   status = cei_read_32bit (board, &pciConfigPtr[PLX9030_CHIP_SEL_1_BASE_ADD], &tempData);
   deviceGlobal[board]->serialNumber = (tempData >> 4) & 0x00FFFFFF;
   #ifdef INTEGRITY_PCI_PPC
      if (deviceIdent) ;
   #endif
   return status;
}


/*===========================================================================* 
 * ENTRY POINT: A 4 2 9 _ U T L _ P R O G R A M    X I L I N X   9 0 5 6
 *===========================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9056 
 *              PCI interface and Xilinx FPGA, supporting 8-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *===========================================================================*/
CEI_INT32 a429_utl_program_xilinx_9056 
(
   CEI_INT16   board,            /* (input)  device ID                       */
   CEI_INT32   deviceIdent,      /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,      /* (input)  size of FPGA program load       */
   pCEI_UCHAR  fwloadPtr         /* (input)  pointer to the FPGA program     */
)
{
   CEI_UINT32 configuration;            /* device channel configuration      */
   CEI_UINT32 programCsum;              /* device programming checksum       */
   CEI_UINT32 programIndex;             /* temporary FPGA program load index */
   CEI_UCHAR  programData;              /* temporary FPGA program data       */
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */
   pCEI_UINT32 pciConfigPtr;            /* dev PCI config virt base address  */
   pCEI_UINT32 localMemPtr;             /* dev local memory virt base address*/

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;
   
   /* Assign the startup value for the PLX CONTROL register */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   
   /* Set the Software Reset b30 and delay for at least 1ms */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_SW_RESET);
   cei_delay_milliseconds(3);

   /* Clear Software Reset b30
    * Delay for 100 milliseconds to allow all configuration memory to clear */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   cei_delay_milliseconds(100);

   /* Restore the local configuration registers by reloading from the serial 
    * EEPROM (cleared by the software reset) */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_EE_RELOAD);
   cei_delay_milliseconds(10);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   cei_delay_milliseconds(10);

   /* Assign the local address space 0 bus region descriptor */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_SEL);

   /* set USERO pin low */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_USER0_LOW);

   /* download the FPGA program file */
   for (programIndex=0; programIndex < programSize; programIndex++) {
      programData = *fwloadPtr++;
      status = cei_write_32bit (board, &localMemPtr[5], (CEI_UINT32)programData);
   }
   
   /* flush the PLX write fifo */
   for (programIndex=0; programIndex < 3; programIndex++) {
      status = cei_write_32bit (board, &localMemPtr[5], (CEI_UINT32)0);
   }

   cei_delay_milliseconds(3);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   cei_delay_milliseconds(3);
   
   /* check if DONE signal is high */
   status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], &configuration);
   if ((configuration & PLX9056_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      return ARS_BADLOAD;
   }
   
   /* Restore the Bus 0 descriptor register to it's normal operational value */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_NORM);
   
   /* write/validate the checksum to local memory */
   status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CSUM_READ], &programCsum);
#if defined DEBUG_PROGRAMMING
   #include "dbg_defs.h"
#else
   programCsum &= PLX9056_CONFIG_MEM_CSUM_MASK;
   deviceIdent = 0;
#endif
   status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], programCsum);
   cei_delay_milliseconds(3);
   #ifdef INTEGRITY_PCI_PPC
      if (deviceIdent) ;
   #endif
   return status;
}


/*==================================================================================* 
 * ENTRY POINT: A 4 2 9 _ U T L _ P R O G R A M _ X I L I N X _ 9 0 5 6 _ 3 2 B I T
 *==================================================================================*
 *
 * FUNCTION:    Download the FPGA program to an x30 device having a PLX9056 
 *              PCI interface and Xilinx FPGA, supporting 32-bit programming.
 *
 * RETURN VAL:  CEI_INT32  -- ARS_NORMAL      Success
 *                            ARS_BADLOAD     FPGA load process failed
 *
 *==================================================================================*/
CEI_INT32 a429_utl_program_xilinx_9056_32bit
(
   CEI_INT16   board,            /* (input)  device ID                       */
   CEI_INT32   deviceIdent,      /* (input)  PCI Configuration Device ID     */
   CEI_UINT32  programSize,      /* (input)  size of FPGA program load       */
   pCEI_UINT32 fwloadPtr         /* (input)  pointer to the FPGA program     */
)
{
   CEI_UINT32 configuration;            /* device channel configuration      */
   CEI_UINT32 programCsum;              /* device programming checksum       */
   CEI_UINT32 programIndex;             /* temporary FPGA program load index */
   pCEI_UINT32 pciConfigPtr;            /* dev PCI config virt base address  */
   pCEI_UINT32 localMemPtr;             /* dev local memory virt base address*/
   CEI_INT32 status = ARS_NORMAL;       /* function return status            */

   pciConfigPtr = device[board].baseAddr;
   localMemPtr  = device[board].devicePtr;
   
   /* Assign the startup value for the PLX CONTROL register */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   
   /* Set the Software Reset b30 and delay for at least 1ms */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_SW_RESET);
   cei_delay_milliseconds(3);

   /* Clear Software Reset b30
    * Delay for 100 milliseconds to allow all configuration memory to clear */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   cei_delay_milliseconds(100);

   /* Restore the local configuration registers by reloading from the serial 
    * EEPROM (cleared by the software reset) */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_EE_RELOAD);
   cei_delay_milliseconds(10);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   cei_delay_milliseconds(10);

   /* Assign the local address space 0 bus region descriptor */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_SEL);

   /* set USERO pin low */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_SET_USER0_LOW);
   
   /* load the firmware by writing to any location in the local bus memory region */
   for (programIndex=0; programIndex < programSize; programIndex++) {
      status = cei_write_32bit (board, &localMemPtr[5], *fwloadPtr++);
   }

   cei_delay_milliseconds(3);
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], PLX9056_CTRL_REG_DEFAULT);
   cei_delay_milliseconds(3);

   /* check if DONE signal is high */
   status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CONTROL_REG], &configuration);
   if ((configuration & PLX9056_FPGA_LOAD_COMPLETE_MASK) == 0) { 
      return ARS_BADLOAD;
   }
   
   /* Restore the Bus 0 descriptor register to it's normal operational value */
   status = WRITE_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_DESCRIPTOR_REG], PLX9056_MEM_DESC_REG_BUS0_NORM);
   
   /* write/validate the checksum to local memory */
   status = READ_32_BIT (board, &pciConfigPtr[PLX_CONFIG_MEM_CSUM_READ], &programCsum);
   
#if defined DEBUG_PROGRAMMING
   #include "dbg_32defs.h"
#else
   deviceIdent = 0;
#endif
   status = cei_write_32bit (board, &localMemPtr[CEIDEV_CONFIG_MEM_CSUM], programCsum);
   cei_delay_milliseconds(3);
   #ifdef INTEGRITY_PCI_PPC
      if (deviceIdent) ;
   #endif
   return status;
}


/*=====================================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ Q U I C K _ A D J U S T _ I R I G _ T H R E S H O L D
 *=====================================================================================*
 *
 * FUNCTION:    Performs a quick adjustment of the IRIG DAC for an external 
 *              input IRIG signal, using signal edge detection for verification
 *              of signal presence.
 *
 * DESCRIPTION: This routine sets the DAC to Vmin + (0.825 * (Vmax-Vmin)).  
 *              Vmin is the lower peak value and Vmax is the upper peak value.
 *
 * RETURN VAL:  CEI_INT32    -- Any value from 0 to x7FFFFFFF is the DAC value
 *                              ARS_FAILURE (-1) Any issue encountered with
 *                                               incoming IRIG signal level
 *
 *=====================================================================================*/
CEI_INT32 a429_utl_quick_adjust_irig_threshold (CEI_INT16 board, CEI_INT32 displayDacSettings)
{
   CEI_UINT32 dacMaxVoltage = 0xff;     /* temp max DAC value for thold test */
   CEI_UINT32 dacValue;                 /* dynamic DAC value during test     */
   CEI_UINT32 initialDacValue = 0x80;   /* initial DAC value                 */
   CEI_UINT32 initialDacAdjustment = 0x10;  /* initial DAC increment         */
   CEI_UINT32 minimumVoltageValue;      /* lower bounds of final DAC value   */
   CEI_UINT32 maximumVoltageValue;      /* upper bounds of final DAC value   */
   CEI_UINT32 irigRegValue;             /* temp contents of IRIG Receiver    */
   CEI_UINT32 dacBoudaryCushionValue;   /* minor adj to final DAC value      */
   CEI_INT32 signalFound = FALSE;       /* flag indicating IRIG Recvd time   */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of board read/write access */
   CEI_INT32 status = ARS_NORMAL;       /* status of threshold detection     */
   pCEI_UINT32 dacControlRegPtr = &device[board].devicePtr[CEIDEV_DAC_CONTROL];
   pCEI_UINT32 globalEnableRegPtr = &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE];
    
   dacValue = 0x80L;
#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("Initial DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
#endif
   
   /*=======================================================================
    * FIRST FIND THE INCOMING SIGNAL
    * Attempt to initially detect the incoming IRIG signal between the 
    * mid-upper peak by setting the DAC to 128 (0.0 Volts) and increasing
    * the threshold until the signal is found.
    *=======================================================================*/
   accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
   /* Wait 5 Msec for signal    */
   cei_delay_milliseconds(5);
   
   /* Read IRIG value, clearing "edge received" bit, then  */
   /* Wait 30 Msec for an incoming signal edge transition. */
   accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
   cei_delay_milliseconds(30);
   
   /* Read IRIG value again, check the "edge received" bit to be set */
   accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
   
   if (irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT) {
      signalFound = TRUE;
   }
   else {
      /* Signal wasn't immediately detected, raise the threshold until it's found */
      for (dacValue = 0x80; (dacValue < 0x100) && (!signalFound); dacValue += 0x10) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
         
#ifndef LABVIEW_RT
         if (displayDacSettings)
            printf ("\rFinding the signal above the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         cei_delay_milliseconds(5);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
         cei_delay_milliseconds(30);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);

         if (irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT) {
            dacMaxVoltage = 0xffL;
            initialDacValue = dacValue;
            initialDacAdjustment = 0x20L;
            signalFound = TRUE;
         }
      }

#ifndef LABVIEW_RT
      if (displayDacSettings)
         printf ("\n");
#endif
      /* Signal wasn't detected by raising the threshold? Lower it now until it's found */
      for (dacValue = 0x80; (dacValue > 0x0) && (!signalFound); dacValue -= 0x10) {
         accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);

#ifndef LABVIEW_RT
         if (displayDacSettings)
            printf ("\rFinding the signal below the mid point -- DAC Setting = %3d", (CEI_NATIVE_INT)dacValue);
#endif
         /* Read IRIG value, clearing "edge received" bit, then  */
         /* Wait 30 Msec for an incoming signal edge transition. */
         cei_delay_milliseconds(5);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
         cei_delay_milliseconds(30);
         accessStatus = cei_read_32bit (board, globalEnableRegPtr, &irigRegValue);
         
         if (irigRegValue & CEIDEV_IRIG_EDGE_DETECT_BIT) {
            dacMaxVoltage = 128L;
            initialDacAdjustment = 0x20L;
            initialDacValue = dacValue;
            signalFound = TRUE;
         }
      }

      if (!signalFound) {
#ifndef LABVIEW_RT
         if (displayDacSettings)
            printf ("\nNever found a signal -- final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
#endif
         return ARS_FAILURE;
      }
   }

   /*=======================================================================
    * SECOND FIND THE LOW THRESHOLD AT WHICH SIGNAL BECOMES PRESENT
    * Starting with the theshold DAC at level the signal was detected above,
    * decrement the DAC voltage until the signal is no longer present;
    * then perform a fine adjustment until DAC setting is just above
    * the lower threshold where the signal becomes present.
    *=======================================================================*/
   dacValue = initialDacValue;
   status = a429_utl_find_dac_threshold (board, 
                              FALSE, 
                              displayDacSettings, 
                              AR_LOWER, 
                              dacMaxVoltage, 
                              initialDacAdjustment, 
                              &dacValue, 
                              &dacBoudaryCushionValue); 
   if (status == ARS_FAILURE)
      return status;
   
   minimumVoltageValue = dacValue + dacBoudaryCushionValue;
   
#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("\nLower transition final value = %3d\n", (CEI_NATIVE_INT)minimumVoltageValue);
#endif
   /*=======================================================================
    * THIRD FIND THE AR_UPPER THRESHOLD AT WHICH SIGNAL BECOMES LOST
    * Starting with the theshold DAC at level the signal was detected first,
    * increment the DAC voltage until the signal is no longer present;
    * then perform a fine adjustment until DAC setting is just below
    * the upper threshold where the signal is lost.
    *=======================================================================*/
   dacValue = initialDacValue;
   status = a429_utl_find_dac_threshold (board, 
                              FALSE, 
                              displayDacSettings, 
                              AR_UPPER, 
                              minimumVoltageValue, 
                              initialDacAdjustment, 
                              &dacValue, 
                              &dacBoudaryCushionValue); 
   if (status == ARS_FAILURE)
      return status;
   
   maximumVoltageValue = dacValue - dacBoudaryCushionValue;
   
#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("\nUpper transition final value = %3d\n", (CEI_NATIVE_INT)maximumVoltageValue);
#endif

   /* validate the span between the upper and lower thresholds */
   if ((maximumVoltageValue < minimumVoltageValue) || ((maximumVoltageValue-minimumVoltageValue) < MIN_DAC_SPAN)) {
#ifndef LABVIEW_RT
      if (displayDacSettings)
         printf ("Error, Signal Detection Range is invalid, Vmax = %d  Vmin = %d\n", 
                 (CEI_NATIVE_INT)maximumVoltageValue, (CEI_NATIVE_INT)minimumVoltageValue);
#endif
      accessStatus = cei_write_32bit (board, dacControlRegPtr, 0x80L);
      return ARS_FAILURE;
   }

   dacValue = minimumVoltageValue + (CEI_UINT32)((825*(maximumVoltageValue-minimumVoltageValue))/1000);
   accessStatus = cei_write_32bit (board, dacControlRegPtr, dacValue);
   
   irigDacThreshold[board] = (CEI_INT32)dacValue;

#ifndef LABVIEW_RT
   if (displayDacSettings)
      printf ("Final DAC Setting = %3d\n", (CEI_NATIVE_INT)dacValue);
#endif
   if (dacValue <= CEIDEV_IRIG_DAC_OFFSET) {
      irigDacThreshold[board] = 0;
      dacValue = 0;
   }
   else
      dacValue = (CEI_UINT32)(CEIDEV_IRIG_DAC_MV_CONVERSION * (float)(dacValue - CEIDEV_IRIG_DAC_OFFSET));

   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
   return (CEI_INT32)(dacValue);
}


/*==========================================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ S T O R E _ M E S S A G E _ S C H E D U L E R _ E N T R Y
 *==========================================================================================*
 *
 * FUNCTION:    Define or modify a single message scheduler table entry
 *
 * PARAMETERS:  CEI_INT32 modifyEntryFlag -- (input) indicates whether to 
 *                                              create a new scheduled message 
 *                                              or modify an existing message
 *                                              FALSE - new : TRUE - modify
 *              CEI_INT32 numberOfEntries -- (input) the number of entries to
 *                                              define or modify, used as a
 *                                              reference for the number of 
 *                                              structure elements defined
 *                                              via the messageEntry pointer
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE
 *                        messageEntry    -- (input) structure of message 
 *                                              content defined as follows:
 *                                              struct {
 *                                                 CEI_UINT32 messageIndex;                  
 *                                                 CEI_UINT32 board;                         
 *                                                 CEI_UINT32 channel;                       
 *                                                 CEI_UINT32 rate;                          
 *                                                 CEI_UINT32 start;                         
 *                                                 CEI_UINT32 txCount;               
 *                                                 CEI_UINT32 data; } 
 *
 * RETURN VAL:  CEI_INT16  -- ARS_NORMAL      Success
 *                            ARS_FAILURE     message scheduler table full
 *
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: This routine defines the supplied message(s) in the device
 *              message scheduler table.
 * 
 *==========================================================================================*/
CEI_INT32 a429_utl_store_message_scheduler_entry 
(
   CEI_INT32 modifyEntryFlag,                /* (input) modify existing flag */
   CEI_INT32 numberOfEntries,                /* (input) number of entries    */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry /* (input) ptr to struct of msgs*/
)
{
   CEI_INT16 board;                     /* temp board index                  */
   CEI_INT32 msgTableCounter;           /* counter for multiple msg entry/mod*/
   CEI_UINT32 msgIndex;                 /* temp index for table search       */
   CEI_UINT32 msgRate;                  /* temp table entry rate setting     */
   CEI_UINT32 localStart;               /* temp table entry start setting    */
   CEI_UINT32 localRate;                /* temp table entry rate setting     */
   CEI_UINT32 appProgTimerResolution;   /* Application specified time-tag res*/
   CEI_UINT32 currentRateValue;         /* modify entry current rate value   */
   CEI_UINT32 currentTxCountValue;      /* modify entry current tx cnt value */
   CEI_UINT32 currentChannelValue;      /* modify entry current channel value*/
   CEI_INT32 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_INT32 disableMessage = FALSE;    /* flag indicatn need to disable msg */
   pCEI_UINT32 msgEntryBasePtr;         /* local pointer to Msg Sched table  */
   pCEI_UINT32 boardPointer;            /* local pointer to Msg Sched table  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntryPtr = messageEntry;
   
   for (msgTableCounter = 0; msgTableCounter < numberOfEntries; msgTableCounter++) {
      board = (CEI_INT16)messageEntryPtr->board;
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_MSG_SCHED_TABLE_START);
   
      /* if this invocation is to modify an existing entry, bypass search for first available */
      if (modifyEntryFlag) {
         msgIndex = messageEntryPtr->messageIndex;
         boardPointer += msgIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE;

         /* check for a change to this entry's rate value */
         accessStatus = cei_read_32bit (board, boardPointer, &currentRateValue);

         /* for non-zero rates, check for CEI-x20 timer compatibility */
         localRate = messageEntryPtr->rate;
         if (localRate && (deviceGlobal[board]->deviceTimeTagReference == AR_TIMER_X20_COMPAT_32BIT)) {
            appProgTimerResolution = deviceGlobal[board]->deviceX20CompatibleResolution;
            if (appProgTimerResolution >= AR_X20_QTR_MICROSEC_PER_MSEC)
               localRate *= appProgTimerResolution/AR_X20_QTR_MICROSEC_PER_MSEC;
         }
         /* when modifying an existing entry, disable the message to alter any
          * attribute (rate or channel) other than just the data value.
          */
         if (localRate != currentRateValue)
            disableMessage = TRUE;
         else {
            /* check for a change to this entry's channel value */
            boardPointer++;
            accessStatus = cei_read_32bit (board, boardPointer, &currentChannelValue);
            if (messageEntryPtr->channel != currentChannelValue)
               disableMessage = TRUE;
         }
      }
      else {
         /* Search for the first unused message scheduler table entry (rate = 0) */
         for (msgIndex = 0; 
              (msgIndex < CEIDEV_CHAN_MSG_SCH_ENTRY_MAX) && (accessStatus == ARS_NORMAL); 
              msgIndex++ ) {
            
            accessStatus = cei_read_32bit (board, boardPointer, &msgRate);
            
            /* unused table entries are indicated via rate value of zero */
            if ((accessStatus == ARS_NORMAL) && (msgRate == 0))
               break;

            boardPointer += CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE;
         }
      }
      
      if ((accessStatus == ARS_NORMAL) && (msgIndex < CEIDEV_CHAN_MSG_SCH_ENTRY_MAX)) {
         msgEntryBasePtr = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_MSG_SCHED_TABLE_START +
                                         (msgIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE));
         boardPointer = msgEntryBasePtr;
         if (disableMessage) {
            /* first disable the message by zeroing the entry rate */
            accessStatus = cei_write_32bit (board, boardPointer, (CEI_UINT32)0);
         }
         
         /* define the entry channel */
         boardPointer++;
         accessStatus = cei_write_32bit (board, boardPointer, messageEntryPtr->channel);
      
         /* only allow modification of the transmit count if this message was created
          * with a limited transmit count and not in an active f/w countdown condition */
         /* define the entry transmit count */
         boardPointer++;
         accessStatus = cei_read_32bit (board, boardPointer, &currentTxCountValue);
         if (!modifyEntryFlag || (modifyEntryFlag && (currentTxCountValue == 0)) ) {
            accessStatus = cei_write_32bit (board, boardPointer, messageEntryPtr->txCount);
         }
         
         /* only allow assignment of the start offset at message creation */
         if (!modifyEntryFlag) {
            /* define the entry start offset */
            localStart = messageEntryPtr->start;
            
            /* check for CEI-x20 timer compatibility */
            if (deviceGlobal[board]->deviceTimeTagReference == AR_TIMER_X20_COMPAT_32BIT) {
               appProgTimerResolution = deviceGlobal[board]->deviceX20CompatibleResolution;
               if (appProgTimerResolution < AR_X20_QTR_MICROSEC_PER_MSEC)
                  localStart = 0;
               else {
                  appProgTimerResolution /= AR_X20_QTR_MICROSEC_PER_MSEC;
                  localStart *= appProgTimerResolution;
               }
            }
            
            boardPointer++;
            accessStatus = cei_write_32bit (board, boardPointer, localStart);
         }

         /* define the entry data */
         boardPointer = msgEntryBasePtr + CEIDEV_MSG_SCHED_TBL_DATA_OFFSET;
         accessStatus = cei_write_32bit (board, boardPointer, messageEntryPtr->data);
            
         /* after all other entries are written, now define the entry rate */
         localRate = messageEntryPtr->rate;
            
         /* check for CEI-x20 timer compatibility */
         if (deviceGlobal[board]->deviceTimeTagReference == AR_TIMER_X20_COMPAT_32BIT) {
            appProgTimerResolution = deviceGlobal[board]->deviceX20CompatibleResolution;
            if (appProgTimerResolution >= AR_X20_QTR_MICROSEC_PER_MSEC) {
               appProgTimerResolution /= AR_X20_QTR_MICROSEC_PER_MSEC;
               localRate *= appProgTimerResolution;
            }
         }
         
         accessStatus = cei_write_32bit (board, msgEntryBasePtr, localRate);
         
         /* on message creation, assign the message index to the respective structure member */
         if ((!modifyEntryFlag) && (accessStatus == ARS_NORMAL))
            messageEntryPtr->messageIndex = msgIndex;
      }
      else
         status = ARS_FAILURE;
      
      if ((accessStatus != ARS_NORMAL) || (status != ARS_NORMAL))
         break;
      else
         messageEntryPtr++;
   }
   
   if (accessStatus != ARS_NORMAL)
      status = accessStatus;
   
   return status;
}


/*================================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _ V A L I D A T E _ C H A N N E L _ R E Q U E S T 
 *================================================================================*
 *
 * FUNCTION:    Validate the channel / item combination specified by the
 *              host application.
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 *              CEI_UINT32 channel -- (input) h/w channel select value
 *              CEI_INT16 item     -- (input) config register attribute to check
 * 
 * RETURN VAL:  CEI_INT16     -- ARS_NORMAL     Normal succesful completion.
 *                               ARS_INVHARVAL  Item was invalid or channel
 *                                              not supported on this device
 *                               ARS_INVBOARD   Board parameter value invalid or
 *                                              board has not been initialized
 * 
 * GLOBALS:     device[board]
 *
 * DESCRIPTION: This function assures the provided channel and configuration
 *              item are valid for the specified board configuration.
 *
 *================================================================================*/
CEI_INT16 a429_utl_validate_channel_request 
(
   CEI_INT16 board,                     /* board to reference (0 to 15)      */
   CEI_UINT32 channel,                  /* h/w channel select value          */
   CEI_INT16 item                       /* config register attribute to check*/
)
{
   CEI_INT16 status;                    /* function return status            */

   if ( (board >= MAX_BOARDS) || !(deviceGlobal[board]->boardInitialized) )
      return ARS_INVBOARD;
   else {
      status = ARS_NORMAL;

      switch (item) {
         /* Validate ARINC 429 specific receive channel attribute requests */
         case ARU_RX_PARITY:
            if ( ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate ARINC 429 and 717 specific receive channel attribute requests */
         case ARU_RX_BITRATE:
         case ARU_RECV_MODE:
         case ARU_RX_STORAGE_MODE:
         case ARU_RX_MERGED_MODE:
         case ARU_RX_FIFO_ENABLE:
         case ARU_RX_DISABLE:
         case ARU_RX_RESET:
         case ARU_RX_FIFO_COUNT:
         case ARU_ENHANCED_FILTER_LABEL:
         case ARU_ACCESS_SNAPSHOT_BUFFER:
            if ( ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ) 
                 &&
                 ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;

         /* Validate generic receive channel attribute requests */
         case ARU_GET_DATA:
            if ( ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]) ) 
                 &&
                 ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) 
                 &&
                   (channel != deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX])
               ) {
               status = ARS_INVHARVAL;
            }
            break;


         /* Validate ARINC 429 specific transmit channel attribute requests */
         case ARU_TX_BIT_ERROR:
         case ARU_TX_GAP_ERROR:
         case ARU_TX_PARITY:
         case ARU_TX_RESET:
            if ( ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate ARINC 429 and 717 specific transmit channel attribute requests */
         case ARU_TX_BITRATE:
         case ARU_PUT_DATA:
         case ARU_TX_DISABLE:
         case ARU_TX_FIFO_ENABLE:
         case ARU_DISCRETE_RAW:
         case ARU_FAST_SLEW_RATE:
         case ARU_TX_FIFO_COUNT:
         case ARU_ADD_SCHEDULED_MSG:
            if ( ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]) ) 
                 &&
                 ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ) 
               ) {
               status = ARS_INVHARVAL;
            }
            break;

         /* Validate ARINC 573 specific receive channel attribute requests */
         case ARU_573_RX_ENABLE:
         case ARU_573_RX_AUTO_DETECT:
            if ( (channel >= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) &&
                 (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) {
               status = ARS_INVHARVAL;
            }
            break;

         case ARU_573_SYNC_WORD1:
         case ARU_573_SYNC_WORD2:
         case ARU_573_SYNC_WORD3:
         case ARU_573_SYNC_WORD4:
            if ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) &&
                 (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) {
               status = ARS_INVHARVAL;
            }
            break;

         /* Validate ARINC 573 generic channel attribute requests */
         case ARU_573_TX_BPRZ_SELECT:
         case ARU_573_RX_BPRZ_SELECT:
            if ( ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ) &&
                 ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ||
                   (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX]) ) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         case ARU_573_TX_HBP_SELECT:
         case ARU_573_TX_SLEW_RATE:
            if ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC573_TX]) &&
                 (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Discrete Output channel attribute requests */
         case ARU_DIGITAL_OUT:
         case ARU_DISCRETE_OUT:
            if ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_DISC_OUT]) &&
                 (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DISC_OUT]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Discrete Output channel attribute requests */
         case ARU_DIGITAL_IN:
         case ARU_DISCRETE_IN:
            if ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_DISC_IN]) &&
                 (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DISC_IN]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Differential Input channel attribute requests */
         case ARU_DIFFERENTIAL_IN:
            if ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_DISC_IN]) &&
                 (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DISC_IN]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         /* Validate Differential Output channel attribute requests */
         case ARU_DIFFERENTIAL_OUT:
            if ( (channel > deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_DIFF_OUT]) &&
                 (channel < deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DIFF_OUT]) ) {
               status = ARS_INVHARVAL;
            }
            break;
   
         case ARU_573_RX_SUBFRAME_SIZE:
         case ARU_DISCRETE_OUT1:
         case ARU_DISCRETE_OUT2:
            break;
      
         default : status = ARS_INVHARVAL; 
            break;
      }
   }
   
   return status;
}


/*=========================================================================================================*
 * ENTRY POINT: A 4 2 9 _ U T L _V A L I D A T E _ S C H E D U L E D _ M E S S A G E _ P A R A M E T E R S
 *=========================================================================================================*
 *
 * FUNCTION:    Validate the parameters for a scheduled transmission request.
 *
 * PARAMETERS:  CEI_INT16 board      -- (input)  Board number (0 to 15)  
 *              CEI_INT16 channel    -- (input)  429 Tx channel selection,
 *                                               valid range is 0 to N, where N 
 *                                               refers to the ARINC 561 channel, 
 *                                               and N-1 refers to the last 429 
 *                                               Tx channel installed
 *              pCEI_INT32 hwChannel -- (output) resp. h/w-based channel number
 *
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid channel
 *                              ARS_INVHARVAL   Message scheduling is not
 *                                              supported on this device
 *
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: This routine validates the board and channel parameter
 *              values for a host-requested message scheduling function.
 *              It then returns the 429-based channel-index-referenced
 *              hardware-based channel number for use by the calling routine.
 * 
 *=======================================================================================================*/
CEI_INT32 a429_utl_validate_scheduled_message_parameters 
(
   CEI_UINT32 board,          /* (input)  Board number (0 to 15)             */
   CEI_UINT32 channel,        /* (input)  API-based 429 Tx channel selection */
   pCEI_UINT32 hwChannel      /* (output) resp. h/w-based channel number     */
)
{
   CEI_INT16 status = ARS_NORMAL;                /* function return status   */
   CEI_UINT32 localChannel = channel;            /* transmit channel select  */

   /* verify the channel parameter validity - must be of type 5 or 6 (enhanced 429 or 561) */
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* For current configurations, any message scheduler request is limited
    * to the number of enhanced ARINC 429 channels installed on the device
    */
   else if (channel > deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
      status = ARS_INVARG;
   }
   /* If the channel requested in one larger than the last ARINC 429 channel,
    * assume the request is to schedule a message on the enhanced ARINC 561 Receive Channel.
    */
   else if (channel == deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
      if (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC561_TX] != CEIDEV_MAX_CHANNELS)
         localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC561_TX];
      else
         status = ARS_INVARG;
   }
   else {
      localChannel = channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
   }

   if (status == ARS_NORMAL) {
      status = (CEI_INT16)a429_utl_validate_channel_request ((CEI_INT16)board, localChannel, ARU_ADD_SCHEDULED_MSG);
      if (status == ARS_NORMAL)
         *hwChannel = localChannel;
   }
   
   return status;
}


/*****************************************************************************
 *****************************************************************************
 *                    A P I   P U B L I C   F U N C T I O N S                *
 *****************************************************************************
 *****************************************************************************/


/*===========================================================================*
 * ENTRY POINT:          A R _ B O A R D _ T E S T
 *===========================================================================*
 *
 * FUNCTION:    This routine tests the functionality of the CEI device as part
 *              of the initialization process.
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 *              CEI_INT16 testType -- (input) type of test to execute
 *                                            INTERNAL_WRAP - perform an
 *                                             ARINC 429 internal wrap test 
 *                                            EXTERNAL_WRAP - perform an
 *                                             ARINC 429 external wrap test 
 *
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL          Success
 *                           ARS_INVBOARD        Invalid board parameter
 *                           ARS_BOARD_MUTEX     Access to the Board Lock timed-out/failed
 *                           ARS_MEMWRERR        Memory test failure
 *                           ARS_WRAP_DROP_FAIL  Wrap Data Missing
 *                           ARS_WRAP_DATA_FAIL  Wrap Data Pattern Mismatch
 *                           ARS_WRAP_FLUSH_FAIL Unknown External data received
 *                           ARS_XMITOVRFLO      Transmit buffer overrun has occurred
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine performs an internal wrap from each transmitter
 *              to the respective matched receiver.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_board_test 
(
   CEI_INT16 board,                  /* (input) board to reference (0 to 15) */
   CEI_INT16 testType                /* (input) type of test to execute      */
)
{
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT16 wrapState;                 /* local internal wrap state select  */
   
   if (VALID_BOARD_AND_INIT) {
      /*------------------------------------------------------------------*
       *  Perform a single message Wrap Test on all ARINC 429 channels
       *------------------------------------------------------------------*/
      wrapState = (CEI_INT16)((testType == INTERNAL_WRAP) ? AR_WRAP_ON : AR_WRAP_OFF);
      status = a429_utl_wrap_test (board, wrapState, 1);
   
      /* Initialize the device following the test */
      if (status == ARS_NORMAL)
         status = ar_initialize_device (board);
   }
   else
      status = ARS_INVBOARD;

   return (status);
}


/*===========================================================================*
 * ENTRY POINT:          A R _ B Y P A S S _ W R A P _ T E S T
 *===========================================================================*
 *
 * FUNCTION:    This routine saves the application's request to bypass
 *              the internal wrap test during board initialization
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *              CEI_INT16 bypass -- (input) AR_ON  bypass internal wrap test
 *                                          AR_OFF execute internal wrap test
 *
 * RETURN VAL:  CEI_INT16        -- ARS_NORMAL       Success
 *                                  ARS_INVBOARD     Invalid board parameter
 *
 * GLOBALS:     bypassWrapTest[]
 * 
 * DESCRIPTION: This routine saves the app's request for internal wrap
 *              execution during board initialization.  This is supplied
 *              in case the board is initialized while ARINC data is
 *              active on any receiver channel, which would cause the
 *              internal wrap test invoked via ar_loadslv() to fail.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_bypass_wrap_test 
(
   CEI_INT16 board,            /* (input) board to reference (0 to 15)       */
   CEI_INT16 bypass            /* (input) bypass (or not) internal wrap test */
)
{
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   
   if (board < MAX_BOARDS)
      bypassWrapTest[board] = (bypass == AR_ON) ? TRUE : FALSE;
   else
      status = ARS_INVBOARD;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:                 A R _ C L O S E 
 *===========================================================================*
 *
 * FUNCTION:    Release all resources for the specified device.
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *
 * RETURN:      CEI_INT16 -- ARS_NORMAL      - Succesful completion of routine
 *                           ARS_INVBOARD    - Invalid board number specified
 *                                             or board never initialized
 *                           ARS_BOARD_MUTEX - Access to the Board Lock timed-out/failed
 *                           ARS_INT_ISR     - Interrupt was not terminated successfully
 *                           ARS_WINRTFAIL   - Windows Driver Close Device failed
 *                           ARS_FAILURE     - Failed to close driver or terminate the ISR
 * GLOBALS:     device[board]
 * 
 * DESCRIPTION: Call into the low-level library and free the resources 
 *              associated with this board.
 *
 *===========================================================================*/                              
EXPORT32 CEI_INT16 DLL_EXPORTED ar_close (CEI_INT16 board) 
{    
   CEI_INT16 status = ARS_NORMAL;
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_UINT32 globalEnableRegData;      /* contents of the Global Enable reg */
   pCEI_UINT32 intCtrlRegPtr = 0;       /* ptr to the PLX Interrupt Ctrl Reg */
   CEI_UINT32 interruptBit;             /* the respective PLX Int Enable bit */

   if ((board < MAX_BOARDS) && (deviceGlobal[board] != NULL)) {
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         if ((deviceGlobal[board]->attachedProcessCount == 1) || (device[board].multiThreadProtect == AR_CONC_NONE)) {
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
            globalEnableRegData &= CEIDEV_GLOBAL_ENABLE_MASK;
            globalEnableRegData &= CEIDEV_INTERRUPTS_DISABLED;
            /* disable "global enable"  and "interrupt enable" */
            if (accessStatus == ARS_NORMAL)
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegData);
            
            if (deviceGlobal[board]->ceiPciInterruptEnabled) {
               if ((deviceGlobal[board]->configuration == CEIDEV_CONFIG_CEI430) || 
                   (deviceGlobal[board]->configuration == CEIDEV_CONFIG_RAR_EC) ||
                   (deviceGlobal[board]->configuration == CEIDEV_CONFIG_CEI430A)) {
                  intCtrlRegPtr = device[board].baseAddr + PLX9030_CFG_INT_CTRL_STAT_REG;
                  interruptBit = PLX9030_LOCAL_INT_INP_ENABLE;
               }
               else if (deviceGlobal[board]->configuration != CEIDEV_CONFIG_RAR_PCIE) {
                  intCtrlRegPtr = device[board].baseAddr + PLX_CONFIG_INT_CTRL_STAT_REG;
                  interruptBit = PLX9056_LOCAL_INT_INP_ENABLE;
               }
               else {
                  interruptBit = 0;
               }
               accessStatus = cei_terminate_interrupt(board, interruptBit, intCtrlRegPtr);
            }
            status = (CEI_INT16)cei_release_device(board, &device[board]);
   
            deviceGlobal[board]->boardInitialized = FALSE;
            deviceGlobal[board]->ceiPciInterruptEnabled = FALSE;
            deviceGlobal[board]->attachedProcessCount = 0;
            device[board].devicePtr = 0;
            device[board].baseAddr = 0;
            deviceGlobal[board] = NULL;
         }
         else if (deviceGlobal[board]->attachedProcessCount > 0) {
            deviceGlobal[board]->attachedProcessCount--;
         }
         
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
         
      }
   }
   else
      status = (CEI_INT16)ARS_INVBOARD;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:            A R _ D E F I N E _ M S G
 *===========================================================================*
 *
 * FUNCTION:    Define a message on the device for scheduled transmission.
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) Board number (0 to 15)  
 *              CEI_INT16 channel  -- (input) 429 Tx channel selection,
 *                                            valid range is 0 to N, where 
 *                                            N-1 refers to the last 429 
 *                                            Tx channel installed
 *              CEI_INT16 rate     -- (input) transmission rate (in msec)
 *              CEI_UINT16 start   -- (input) initial delay from the current 
 *                                            time until the message will be 
 *                                            transmitted for the first time  
 *                                            (subsequent transmissions at "rate")
 *              CEI_INT32 data     -- (input) the ARINC word to transmit
 *
 * RETURN VAL:  ARS_FAILURE    One of the following errors occurred:
 *                                board number out of range
 *                                board not loaded and initialized
 *                                an invalid channel was specified 
 *                                the maximum number of messages has already 
 *                                    been defined
 *                                feature is not supported by the hardware
 *
 *              Message Index  a return value in the range of 0 to 1023
 *                             indicates success, used as the unique
 *                             message number assigned to this message
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine defines an ARINC 429 or 561 message for
 *              periodic retransmission. It causes a specific data word to go
 *              out at a specific rate. Once a message has been defined, it 
 *              can be modified by calling AR_MODIFY_MSG. 
 *
 *              Messages can be defined anytime after AR_LOADSLV() has been 
 *              called.
 *
 *              The message rate is defined in milliseconds.  Based on h/w 
 *              design considerations, it is recommended that messages with  
 *              faster rates be defined before the slower ones. If the rate is
 *              set to zero, the message is considered to be removed and 
 *              eligable to be reused.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_define_msg 
(
   CEI_INT16 board,                /* (input) Board number (0 to 15)         */
   CEI_INT16 channel,              /* (input) 429 Tx channel selection       */
   CEI_INT16 rate,                 /* (input) transmission rate (in msec)    */
   CEI_UINT16 start,               /* (input) initial delay from the current */
   CEI_INT32 data                  /* (input) the ARINC word to transmit     */
)
{
   CEI_UINT32 localChannel;                     /* transmit channel select   */
   CEI_INT16 status = ARS_NORMAL;               /* function return status    */
   CEI_INT32 tempStatus = ARS_NORMAL;           /* local temp return status  */
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;        /* message content structure */
   msgEntry.messageIndex = (CEI_UINT32)0;

#ifdef TARGET_EMBEDDED
   localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
   /* verify calling parameter validity for h/w message scheduling */
   tempStatus = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, (CEI_UINT32)channel, &localChannel);

   if (tempStatus == ARS_NORMAL) {
#endif
      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      }
         
      if (status == ARS_NORMAL) {
         msgEntry.board        = (CEI_UINT32)board;
         msgEntry.channel      = localChannel;
         msgEntry.rate         = (CEI_UINT32)rate;
         msgEntry.start        = (CEI_UINT32)start;
         msgEntry.txCount      = (CEI_UINT32)CEIDEV_MSG_SCHEDULER_TX_ALWAYS;
         msgEntry.data         = (CEI_UINT32)data;
         
         /* define a single scheduled message entry */
         status = (CEI_INT16)a429_utl_store_message_scheduler_entry (FALSE, 1, &msgEntry);
   
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            if (status == ARS_NORMAL)
               status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
            else
               tempStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   
   if ((status == ARS_NORMAL) && (tempStatus == ARS_NORMAL))
      status = (CEI_INT16)msgEntry.messageIndex;
   else
      status = ARS_FAILURE;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ D E F I N E _ M S G _ B L O C K
 *===========================================================================*
 *
 * FUNCTION:    Define a block of messages for scheduled transmission.
 *
 * PARAMETERS:  CEI_INT32 numberOfEntries -- (input) the number of entries to
 *                                                define, used as a reference
 *                                                for the number of structure
 *                                                elements defined via the
 *                                                messageEntry pointer
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE
 *                  messageEntry     -- (input) array of structures of message 
 *                                              content, each element defined 
 *                                              as follows:
 *                                                 struct {
 *                                                    CEI_INT16 messageIndex;                  
 *                                                    CEI_INT16 board;                         
 *                                                    CEI_INT16 channel;                       
 *                                                    CEI_INT16 rate;                          
 *                                                    CEI_INT16 start;                         
 *                                                    CEI_UINT32 txCount;               
 *                                                    CEI_UINT32 data; } 
 *                                                    
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid number of entries argument
 *                              ARS_INVHARVAL   Invalid transmit channel entry
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_FAILURE     The number of entries requested
 *                                              exceeds the available number
 *                                              of remaining table entries
 * 
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine defines the supplied message(s) in the device
 *              message scheduler table.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_define_msg_block
(
   CEI_INT32 numberOfEntries,                /* (input) number of entries    */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry /* (input) ptr to struct of msgs*/
)
{
   CEI_INT16 board;                     /* board id extracted from structure */
   CEI_INT32 arrayIndex;                     /* scheduler access array index */
   CEI_UINT32 localChannel;                    /* transmit channel select    */
   CEI_INT16 status = ARS_NORMAL;              /* function return status     */
   CEI_INT32 tempStatus = ARS_NORMAL;          /* local temp return status   */
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;       /* message content structure  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE msgPtr;                /* local ptr to array */

   msgEntry.messageIndex = (CEI_UINT32)0;
   if ((numberOfEntries > 0) && (numberOfEntries < MAX_SCHEDULED_MSGS)) {
      msgPtr = messageEntry;
   }
   else {
      return ARS_INVARG;
   }

   for (arrayIndex = 0; (arrayIndex < numberOfEntries) && (status == ARS_NORMAL); arrayIndex++) {
      board = (CEI_INT16)msgPtr->board;

#ifdef TARGET_EMBEDDED
      localChannel = msgPtr->channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
      /* verify calling parameter validity for h/w message scheduling */
      tempStatus = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, msgPtr->channel, &localChannel);

      if (tempStatus == ARS_NORMAL) {
#endif
         /* if multi-thread protection requested, semaphore/mutex-lock device access */
         if (device[board].multiThreadProtect) {
            status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
         }
            
         if (status == ARS_NORMAL) {
            msgEntry.messageIndex = 0;
            msgEntry.board        = board;
            msgEntry.channel      = localChannel;
            msgEntry.rate         = msgPtr->rate;
            msgEntry.start        = msgPtr->start;
            msgEntry.txCount      = msgPtr->txCount;
            msgEntry.data         = msgPtr->data;
            
            /* define a single scheduled message entry */
            status = (CEI_INT16)a429_utl_store_message_scheduler_entry (FALSE, 1, &msgEntry);
            
            /* assign the message index in the app-supplied structure array entry */
            msgPtr->messageIndex = msgEntry.messageIndex;
            msgPtr++;
      
            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect) {
               if (status == ARS_NORMAL)
                  status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
               else
                  tempStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
      }
#ifndef TARGET_EMBEDDED
      else
         status = (CEI_INT16)tempStatus;
   }
#endif
   return status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ D E F I N E _ M S G _ B L O C K _ L V
 *===========================================================================*
 *
 * FUNCTION:    Define a block of messages for scheduled transmission,
 *              supporting LabVIEW structure array formatting
 *
 * PARAMETERS:  CEI_INT32 numberOfEntries -- (input) the number of entries to
 *                                                define, used as a reference
 *                                                for the number of structure
 *                                                elements defined via the
 *                                                messageEntry pointer
 *              pLV_AR_SCHEDULED_MSG_ENTRY_TYPE
 *                  messageEntry     -- (input) structure containing the size
 *                                              of the array along with the 
 *                                              array of structures of message 
 *                                              content, each element defined 
 *                                              as follows:
 *                                                 struct {
 *                                                    CEI_UINT32 messageIndex;                  
 *                                                    CEI_UINT32 board;                         
 *                                                    CEI_UINT32 channel;                       
 *                                                    CEI_UINT32 rate;                          
 *                                                    CEI_UINT32 start;                         
 *                                                    CEI_UINT32 txCount;               
 *                                                    CEI_UINT32 data; } 
 *                                                    
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid number of entries argument
 *                              ARS_INVHARVAL   Invalid transmit channel entry
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_FAILURE     The number of entries requested
 *                                              exceeds the available number
 *                                              of remaining table entries
 * 
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine defines the supplied message(s) in the device
 *              message scheduler table.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_define_msg_block_lv
(
   CEI_INT32 numberOfEntries,                /* (input) number of entries    */
   pLV_AR_SCHEDULED_MSG_ENTRY_TYPE messageEntry /* (input) ptr to msg array  */
)
{
   CEI_INT16 board;                     /* board id extracted from structure */
   CEI_INT32 arrayIndex;                     /* scheduler access array index */
   CEI_UINT32 localChannel;                    /* transmit channel select    */
   CEI_INT16 status = ARS_NORMAL;              /* function return status     */
   CEI_INT32 tempStatus = ARS_NORMAL;          /* local temp return status   */
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;       /* message content structure  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE msgPtr;     /* local pointer to struct array */

   msgEntry.messageIndex = (CEI_UINT32)0;
   if ((numberOfEntries > 0) && (numberOfEntries < MAX_SCHEDULED_MSGS)) {
      msgPtr = messageEntry->msgEntries;
   }
   else {
      return ARS_INVARG;
   }

   for (arrayIndex = 0; (arrayIndex < numberOfEntries) && (status == ARS_NORMAL); arrayIndex++) {
      board = (CEI_INT16)msgPtr->board;

#ifdef TARGET_EMBEDDED
      localChannel = msgPtr->channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
      /* verify calling parameter validity for h/w message scheduling */
      tempStatus = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, msgPtr->channel, &localChannel);

      if (tempStatus == ARS_NORMAL) {
#endif
         /* if multi-thread protection requested, semaphore/mutex-lock device access */
         if (device[board].multiThreadProtect) {
            status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
            msgEntry.messageIndex = 0;
            msgEntry.board        = board;
            msgEntry.channel      = localChannel;
            msgEntry.rate         = msgPtr->rate;
            msgEntry.start        = msgPtr->start;
            msgEntry.txCount      = msgPtr->txCount;
            msgEntry.data         = msgPtr->data;
            
            /* define a single scheduled message entry */
            status = (CEI_INT16)a429_utl_store_message_scheduler_entry (FALSE, 1, &msgEntry);
            
            /* assign the message index in the app-supplied structure array entry */
            msgPtr->messageIndex = msgEntry.messageIndex;
            msgPtr++;
   
            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect) {
               if (status == ARS_NORMAL)
                  status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
               else
                  tempStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
#ifndef TARGET_EMBEDDED
      }
      else
         status = (CEI_INT16)tempStatus;
#endif
   }

   return status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ E N H _ L A B E L _ F I L T E R
 *===========================================================================*
 *
 * FUNCTION:    Function to enable/disable enhanced label filtering on device.
 * 
 * PARAMETERS:  CEI_INT16 board     -- (input) Board number (0 to 15)  
 *              CEI_INT16 channel   -- (input) 429 receive channel selection,
 *                                     valid range is 0 to N, where N refers
 *                                     to the ARINC 561 channel, and N-1
 *                                     refers to the last 429 channel installed
 *              CEI_UINT16 label    -- (input) Label of interest.  Valid range
 *                                     is 0-255.  Also valid is ARU_ALL_LABELS
 *                                     which takes the action for all labels. 
 *              CEI_UINT16 sdi      -- (input) SDI of interest.  Valid range
 *                                     is 0-3.  Also valid is ARU_ALL_SDI
 *                                     which takes the action for all SDI values. 
 *              CEI_UINT16 essm     -- (input) ESSM of interest.  Valid range
 *                                     is 0-7.  Also valid is ARU_ALL_ESSM
 *                                     which takes the action for all ESSM values. 
 *              CEI_INT16 action    -- (input) Enable or disable filtering for
 *                                     this board/channel/label.  Valid values
 *                                     are: 
 * 
 *                 FILTER_SEQUENTIAL  0x10  if CLEAR add label to sequential
 *                                          receive buffer
 *                 FILTER_SNAPSHOT    0x20  If CLEAR add label to snapshot
 *                                          receive buffer
 *                 FILTER_INTERRUPT   0x40  If SET on reception insert channel #
 *                                          in the interrupt queue and if 
 *                                          enabled generate a PCI interrupt
 * 
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL      Success
 *                        -- ARS_INVBOARD    Invalid board argument
 *                        -- ARS_INVHARVAL   Invalid receive channel argument
 *                        -- ARS_INVARG      Invalid label, sdi, essm, or action argument
 *                        -- ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: The device supports hardware label filtering based on the
 *              label value for standard label filtering, or the combination
 *              of the label, SDI, and ESSM field values for enhanced label
 *              filtering.  This routine is designed to support enhanced
 *              label filtering for channel types 5 and 6 (enhanced ARINC 
 *              429).  This routine assigns the specified filter action to 
 *              the respective enhanced label filter table location for the 
 *              device receive channel.
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_enh_label_filter 
(
   CEI_INT16 board,                     /* (input) Board number (0 to 15)    */
   CEI_INT16 channel,                   /* (input) 429 Rx channel selection  */
   CEI_UINT16 label,                    /* (input) Label of interest         */
   CEI_UINT16 sdi,                      /* (input) SDI of interest           */
   CEI_UINT16 essm,                     /* (input) ESSM of interest          */
   CEI_INT16 action                     /* (input) Filtering Enable/Disable  */
)
{
   CEI_UINT32 labelFilterIndex;         /* label filter channel select index */
   CEI_UINT32 localChannel;             /* merged/individual channel select  */
   CEI_UINT32 filterActivation;         /* local translation of action parm  */
   CEI_UINT32 filterTableValue;         /* local translation of action parm  */
   CEI_UINT32 filterTableMask;          /* mask of the current table entry   */
   CEI_UINT32 essmEnd;                  /* local temp enh filter table index */
   CEI_UINT32 essmStart;                /* local temp enh filter table index */
   CEI_UINT32 essmIndex;                /* local temp enh filter table index */
   CEI_UINT32 sdiEnd;                   /* local temp enh filter table index */
   CEI_UINT32 sdiStart;                 /* local temp enh filter table index */
   CEI_UINT32 sdiIndex;                 /* local temp enh filter table index */
   CEI_UINT32 labelEnd;                 /* local temp enh filter table index */
   CEI_UINT32 labelStart;               /* local temp enh filter table index */
   CEI_UINT32 labelIndex;               /* local temp enh filter table index */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_INT32 tempStatus = ARS_NORMAL;           /* local temp return status  */

   localChannel = (CEI_UINT32)channel;
   filterActivation = FILTER_SNAPSHOT;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   
   /* For current configurations, any label filter channel request is limited
    * to the number of ARINC 429 channels installed on the device
    */
   else if (channel > (CEI_INT16)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
#ifndef TARGET_EMBEDDED
   }
   
   /* check the values passed in for all remaining parameters */
   if (status == ARS_NORMAL) {
      if ( (label > ARU_ALL_LABELS) || (sdi > ARU_ALL_SDI) || (essm > ARU_ALL_ESSM) )
         status = ARS_INVARG;
      else
         status = a429_utl_validate_channel_request (board, localChannel, ARU_ENHANCED_FILTER_LABEL);
      
      if (status == ARS_NORMAL) {
#endif
         if (action == ARU_FILTER_ON)
            filterActivation = FILTER_SEQUENTIAL + FILTER_SNAPSHOT;
         else if (action == ARU_FILTER_OFF)
            filterActivation = 0;
         else if ( (action < FILTER_SEQUENTIAL) || (action > (FILTER_SEQUENTIAL + FILTER_SNAPSHOT + FILTER_INTERRUPT)) )
            status = ARS_INVARG;
         else
            filterActivation = (CEI_INT32)action;
         
         filterActivation = filterActivation >> 4;    /* API definitions provide legacy support with x20 */
#ifndef TARGET_EMBEDDED
      }
   }
   
   if (status == ARS_NORMAL) {
#endif
      
      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
         if (status != ARS_NORMAL) {
            return (status);
         }
      }
      
      filterTableValue = 0;
      filterTableMask = 0;

      /* if the same action applies to all essm/sdi/label combinations, loop through all
       * and assign the specified action
       */
      if ( (label == ARU_ALL_LABELS) && (sdi == ARU_ALL_SDI) && (essm == ARU_ALL_ESSM) ) {

         for (essmIndex = 0; essmIndex < ARU_ALL_ESSM; essmIndex++) {
            filterTableValue |= filterActivation << (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT);
         }

         for ( labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
               (labelFilterIndex <= (CEIDEV_ENH_LABEL_FILTER_START + ((localChannel+1) * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE)))
                   && (accessStatus == ARS_NORMAL);
               labelFilterIndex++ ) {
            /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
            /* write the filter table entry data pattern to this location */
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], filterTableValue);
         }
      }
      else {   /* otherwise, apply the specified action to the specified essm/sdi/label location(s) */
         if (label == ARU_ALL_LABELS) {
            labelStart = 0;
            labelEnd = 256;
         }
         else {
            labelStart = label;
            labelEnd = label+1;
         }
         if (sdi == ARU_ALL_SDI) {
            sdiStart = 0;
            sdiEnd = ARU_ALL_SDI;
         }
         else {
            sdiStart = sdi;
            sdiEnd = sdi+1;
         }
         if (essm == ARU_ALL_ESSM) {
            essmStart = 0;
            essmEnd = ARU_ALL_ESSM;
         }
         else {
            essmStart = essm;
            essmEnd = essm+1;
         }
         /* create the 32-bit filter activation definition for each 4-bit ESSM field 
          * along with a mask for the current content of the respective table entry.
          */
         for (essmIndex = essmStart; essmIndex < essmEnd; essmIndex++) {
            filterTableValue |= filterActivation << (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT);
            filterTableMask  |= CEIDEV_ENH_FLTR_BIT_MASK  << (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT);
         }
         filterTableMask = ~filterTableMask;

         /* now assign the action to the specified locations in the table */
         for (labelIndex = labelStart; labelIndex < labelEnd; labelIndex++) {
            for (sdiIndex = sdiStart; sdiIndex < sdiEnd; sdiIndex++) {
               labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
               labelFilterIndex += (labelIndex << CEIDEV_ENH_LABEL_FILTER_LB_SHIFT) + sdiIndex;
               
               /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
               /* read the current contents of the label filter table location(s) to be updated */
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], &filterActivation);
               /* mask out only the selected ESSM field in this entry for the requested filter update, then add the new setting(s) */
               filterActivation &= filterTableMask;
               filterActivation |= filterTableValue;
               /* write the updated filter table entry back to this location */
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], filterActivation);
            }
         }
      }
      
      /* if multi-thread protection requested, free semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         if (status == ARS_NORMAL)
            status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
         else
            tempStatus = cei_release_thread_lock(LOCK_IDENT);
      }
#ifndef TARGET_EMBEDDED
   }
   
   if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
      status = (CEI_INT16)accessStatus;
#endif
   #ifdef INTEGRITY_PCI_PPC
      if (tempStatus) ;
   #endif
      
   return status;
}


/*===========================================================================*
 * ENTRY POINT:              A R _ E X E C U T E _ B I T
 *===========================================================================*
 *
 * FUNCTION:    This routine performs BIT functionality testing of the 
 *              device as specified by the host application.
 *
 * PARAMETERS:  CEI_INT16 board    -- (input) board to reference (0 to 15)
 *              CEI_INT16 testType -- (input) type of BIT to execute
 *                                        AR_BIT_BASIC_STARTUP
 *                                        AR_BIT_FULL_STARTUP
 *                                        AR_BIT_PERIODIC
 *                                        AR_BIT_INT_LOOPBACK
 *                                        AR_BIT_EXT_LOOPBACK
 *                                        AR_BIT_PARTIAL_SRAM   
 *                                        AR_BIT_FULL_SRAM      
 *                                        AR_BIT_SELECT_SRAM_MIN
 *                                        AR_BIT_SELECT_SRAM_MAX
 *
 * RETURN VAL:  CEI_INT16  ARS_NORMAL          Success
 *                         ARS_BOARD_MUTEX     Access to the Board Lock timed-out/failed
 *                         ARS_MEMWRERR        Memory test failure
 *                         ARS_WRAP_DROP_FAIL  Wrap Data Missing
 *                         ARS_WRAP_DATA_FAIL  Wrap Data Pattern Mismatch
 *                         ARS_WRAP_FLUSH_FAIL Unknown External data received
 *                         ARS_INVARG          Invalid testType parameter
 *                         ARS_INVBOARD        Invalid board parameter
 *                         ARS_FAILURE         Timer-deviation test failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine performs various tests of the hardware device.
 *              While none of these tests are "built-in" per se, they do
 *              provide basic health-status of the hardware.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_execute_bit
(
   CEI_INT16 board,                  /* (input) board to reference (0 to 15) */
   CEI_INT16 testType                /* (input) type of test to execute      */
)
{
   CEI_INT16 status = ARS_NORMAL;              /* function return status     */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_UINT32 firstTimerValue;                 /* timer deviation test value */
   CEI_UINT32 secondTimerValue;                /* timer deviation test value */
   CEI_UINT32 differenceTimerValue;            /* timer deviation test value */
   CEI_UINT32 hostFirstTimerValue;             /* timer deviation test value */
   CEI_UINT32 hostSecondTimerValue;            /* timer deviation test value */
   CEI_UINT32 hostDifferenceTimerValue;        /* timer deviation test value */
   CEI_UINT32 globalEnableRegData;      /* contents of the Global Enable reg */
   CEI_UINT32 enableDevice;         /* enable required via Global Enable reg */
   AR_TIMETAG_TYPE timeTag;                    /* temporary time storage     */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      if ( (device[board].multiThreadProtect == AR_CONC_MULTIPROC) && (deviceGlobal[board]->attachedProcessCount > 1) ) {
         status = ARS_FAILURE;
      }
      else {
         switch (testType) {
            case AR_BIT_BASIC_STARTUP:
               /* Initialize the device */
               status = ar_initialize_device (board);
               break;
   
            case AR_BIT_FULL_STARTUP:
               /* Initialize the device */
               status = ar_initialize_device (board);
               /* Perform a 1 msg Internal Wrap Test on each ARINC 429 channel */
               if (status == ARS_NORMAL)
                  status = a429_utl_wrap_test (board, AR_WRAP_ON, 1);
               /* Perform a Full SRAM Memory Test (6 seconds) */
               if (status == ARS_NORMAL)
                  status = a429_utl_memory_test (board, FULL_SRAM, 0);
               /* Initialize the device following the test */
               if (status == ARS_NORMAL)
                  status = ar_initialize_device (board);
               break;
   
            case AR_BIT_PERIODIC:
               /* save the existing channel global enable register state, enable data processing temporarily */
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
               enableDevice = globalEnableRegData & CEIDEV_GLOBAL_ENABLE_ON;
               if (enableDevice) {
                  status = ar_go (board);
                  ar_sleep (1);
               }
               status = a429_utl_memory_test (board, UNUSED_SRAM, 0);
               if (status == ARS_NORMAL) {
                  /* check consecutive reads of the h/w timer to check if "alive" */
                  status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &timeTag);
                  firstTimerValue = (CEI_UINT32)timeTag.timeTag;
                  cei_get_host_system_time(&hostFirstTimerValue);
      
                  ar_sleep (2);
                  
                  status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &timeTag);
                  secondTimerValue = (CEI_UINT32)timeTag.timeTag;
                  
                  cei_get_host_system_time(&hostSecondTimerValue);
                  if (hostSecondTimerValue > hostFirstTimerValue)
                     hostDifferenceTimerValue = hostSecondTimerValue - hostFirstTimerValue;
                  else
                     hostDifferenceTimerValue = hostSecondTimerValue + 1 + (0xFFFFFFFF - hostFirstTimerValue);
                  /* compare to host, host is in milliseconds so convert to microseconds */
                  if (hostDifferenceTimerValue > 0)
                     hostDifferenceTimerValue *= 1000;
                  else
                     hostDifferenceTimerValue = 1001;
      
                  if (secondTimerValue > firstTimerValue)
                     differenceTimerValue = secondTimerValue - firstTimerValue;
                  else
                     differenceTimerValue = secondTimerValue + 1 + (0xFFFFFFFF - firstTimerValue);
                  /* timer deviation should be between 1000 and 3000 microseconds */
                  if ( (differenceTimerValue < (hostDifferenceTimerValue - AR_BIT_TIMER_DEVIATION)) || (differenceTimerValue > (hostDifferenceTimerValue + AR_BIT_TIMER_DEVIATION)) )
                     status = ARS_FAILURE;
   
               }
               if (enableDevice) {
                  status = ar_stop(board);
               }
               break;
   
            case AR_BIT_INT_LOOPBACK:
               /* Perform an Internal Wrap Test on the first ARINC 429 channel */
               status = a429_utl_wrap_test (board, AR_WRAP_ON, AR_WRAP_TEST_MSG_COUNT);
               break;
   
            case AR_BIT_EXT_LOOPBACK:
               /* Perform an External Wrap Test on the first ARINC 429 channel */
               status = a429_utl_wrap_test (board, AR_WRAP_OFF, AR_WRAP_TEST_MSG_COUNT);
               break;
   
            case AR_BIT_PARTIAL_SRAM:
               /* Perform an Unused SRAM Memory Test */
               status = a429_utl_memory_test (board, UNUSED_SRAM, 0);
               break;
   
            case AR_BIT_FULL_SRAM:
               /* Perform a Full SRAM Memory Test (6 seconds) */
               status = a429_utl_memory_test (board, FULL_SRAM, 0);
               if (status == ARS_NORMAL)
                  status = ar_initialize_device (board);
               break;
   
            default:
               if ( (testType >= AR_BIT_SELECT_SRAM_MIN) && (testType <= AR_BIT_SELECT_SRAM_MAX) )
                  status = a429_utl_memory_test (board, SELECT_SRAM, testType);
               else
                  status = ARS_INVARG;
         }
      }
#ifndef TARGET_EMBEDDED
   }
   else
      status = ARS_INVBOARD;
#endif

   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:     A R _ G E T _ B A S E _ A D D R 
 *===========================================================================*
 *
 * FUNCTION:    This routine returns a pointer to the base address assigned
 *              to the device from the (plug and play) host.
 *
 * PARAMETERS:  CEI_INT16 board       -- (input) board to reference (0 to 15)
 *
 * RETURN VAL:  pCEI_UINT32 devicePtr -- pointer to the respective device host 
 *                                       interface location in virtual memory.
 * 
 * GLOBALS:     device[board].devicePtr
 * 
 * DESCRIPTION: Return the base address of the device.  Using the address 
 *              returned, any register on the device can be accessed directly.
 * 
 *===========================================================================*/
EXPORT32 pCEI_UINT32 DLL_EXPORTED ar_get_base_addr (CEI_INT16 board)
{
   if (board < MAX_BOARDS)
      return device[board].devicePtr;
   else
      return device[0].devicePtr;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ G E T _ B O A R D N A M E
 *===========================================================================*
 *
 * FUNCTION:    Return the respective CEI-x30 board name
 *                       
 * PARAMETERS:  CEI_UINT32 board    The board number of interest
 *              pCEI_CHAR boardName Reference to char array to store the name
 *
 * RETURN VAL:  NULL                Board parameter value invalid or
 *                                    board has not been initialized
 *              pointer to char     Ptr to string for the board type detected
 *
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: returns this API board type for the specified board index
 * 
 *===========================================================================*/ 
EXPORT32 pCEI_CHAR DLL_EXPORTED ar_get_boardname 
(
   CEI_UINT32 board, 
   pCEI_CHAR  boardName
) 
{
   CEI_UINT32 configuration;

   if (VALID_BOARD_AND_INIT) {
      configuration = deviceGlobal[board]->configuration;
      if ( (configuration < CEIDEV_FIRST_CEIX30_CONFIG) || (configuration > CEIDEV_LAST_CEIX30_CONFIG) ) {
         return NULL;
      }
      else {
         if (boardName != NULL) {
            strcpy (boardName, boardNames[configuration - CEIDEV_FIRST_CEIX30_CONFIG]); 
         }
         return boardNames[configuration - CEIDEV_FIRST_CEIX30_CONFIG];
      }
   }
   else {
      return NULL;
   }
}


/*===========================================================================*
 * ENTRY POINT:         A R _ G E T _ B O A R D N A M E L V
 *===========================================================================*
 *
 * FUNCTION:    Return a reference to a string describing the board type 
 *                       
 * PARAMETERS:  CEI_UINT32 board     -- The board number of interest
 *              pCEI_CHAR  boardName -- reference to a board type string
 *
 * RETURN VAL:  CEI_INT32 -- ARS_NORMAL   Normal succesful completion.
 *                           ARS_INVBOARD Board parameter value invalid or
 *                                        board has not been initialized
 *                           ARS_INVARG   boardName parameter not valid memory
 * 
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: returns the name string representing the current board
 * 
 *===========================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_boardnameLV 
(
   CEI_UINT32 board, 
   pCEI_CHAR  boardName
) 
{
   CEI_UINT32 configuration;
   CEI_INT32  status;

   if (VALID_BOARD_AND_INIT) {
      if (boardName != NULL) {
         configuration = deviceGlobal[board]->configuration;
         if ( (configuration < CEIDEV_FIRST_CEIX30_CONFIG) || (configuration > CEIDEV_LAST_CEIX30_CONFIG) ) {
            strcpy (boardName, "");
            status = ARS_INVBOARD;
         }
         else {
            strcpy (boardName, boardNames[configuration - CEIDEV_FIRST_CEIX30_CONFIG]); 
            status = ARS_NORMAL;
         }
      }
      else
         status = ARS_INVARG;
   }
   else
      status = ARS_INVBOARD;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ G E T _ B O A R D T Y P E
 *===========================================================================*
 *
 * FUNCTION:    Return the board type 
 *                       
 * PARAMETERS:  CEI_INT16 board -- The board number of interest
 * 
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL     Normal succesful completion.
 *                           ARS_INVBOARD   Board parameter value invalid or
 *                                          board has not been initialized
 *
 * GLOBALS:     deviceGlobal[]
 * 
 * DESCRIPTION: returns this API board type for the specified board index
 * 
 *===========================================================================*/ 
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_boardtype (CEI_INT16 board) 
{
   CEI_UINT32 configuration;

   if (VALID_BOARD_AND_INIT) {
      switch (deviceGlobal[board]->configuration) {
         case CEIDEV_CONFIG_CEI830:
            configuration = CEI830;
            break;
         case CEIDEV_CONFIG_CEI430:
            configuration = CEI430;
            break;
         case CEIDEV_CONFIG_CEI430A:
            configuration = CEI430A;
            break;
         case CEIDEV_CONFIG_AMCA30:
            configuration = AMCA30;
            break;
         case CEIDEV_CONFIG_CEI530:
            configuration = CEI530;
            break;
         case CEIDEV_CONFIG_R830RX:
            configuration = R830RX;
            break;
         case CEIDEV_CONFIG_RAR_CPCI:
            configuration = RARCPCI;
            break;
         case CEIDEV_CONFIG_RAR_EC:
            configuration = RAREC;
            break;
         case CEIDEV_CONFIG_RAR_PCIE:
            configuration = RARPCIE;
            break;
         default:
            configuration = CEIDEV_CONFIGURATION_NULL;
      }

      return (CEI_INT16)configuration;
   }
   else
      return (CEI_INT16)ARS_INVBOARD;
}             


/*===========================================================================*
 * ENTRY POINT:             A R _ G E T _ C O N F I G
 *===========================================================================*
 *
 * FUNCTION:    Retrieve board and channel configuration attributes (legacy)
 *
 * PARAMETERS:  CEI_INT16 board -- (input) board to reference (0 to 15)
 *              CEI_INT16 item  -- (input) item for which to return the 
 *                                     current state:
 * 
 *                     ARU_RX_CH0*_BIT_RATE        ARU_CONFIGURATION
 *                     ARU_RX_CH1*_BIT_RATE        ARU_DISCRETE_IN1          
 *                     ARU_TX_CH0*_BIT_RATE        ARU_DISCRETE_IN2          
 *                     ARU_TX_CH1*_BIT_RATE        ARU_DISCRETE_IN3          
 *                     ARU_RX_CH0*_PARITY          ARU_DISCRETE_IN4          
 *                     ARU_RX_CH1*_PARITY          ARU_DISCRETE_OUT1         
 *                     ARU_TX_CH0*_PARITY          ARU_DISCRETE_OUT2         
 *                     ARU_TX_CH1*_PARITY          ARU_DISCRETE_OUT2         
 *                     ARU_TX_CH0*_SHUT_OFF        ARU_DISCRETE_OUT3         
 *                     ARU_TX_CH1*_SHUT_OFF        ARU_ACCESS_SNAPSHOT_BUFFER
 *                     ARU_TX_CH**_HB_INJ          ARU_IRIG_WRAP_ENABLE      
 *                     ARU_TX_CH**_LB_INJ          ARU_IRIG_AVAILABLE        
 *                     ARU_TX_CH**_GAP_INJ         ARU_IRIG_CALIBRATED       
 *                     ARU_FW_VERSION              ARU_IRIG_INPUT_TIME
 *                     ARU_SERIAL_NUMBER
 *                     ARU_RX_TIMETAG_MODE w/values of:                          
 *
 *                              AR_TIMETAG_EXT_IRIG_64BIT                         
 *                              AR_TIMETAG_INT_USEC_64BIT                         
 *                              AR_TIMETAG_INT_20USEC_32BIT                       
 *                              AR_TIMETAG_INT_MSEC_32BIT                         
 *                              AR_TIMER_X20_COMPAT_32BIT
 *
 * RETURN VAL:  CEI_INT32 state -- the state of the item specified
 *                             ARS_INVARG     "item" parameter value is not 
 *                                            supported by this API function
 *                             ARS_INVHARVAL  Channel not supported on device
 *                             ARS_INVBOARD   Board parameter value invalid or
 *                                            board has not been initialized
 *                             ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * 
 * GLOBALS:     device[]
 * 
 * DESCRIPTION: Supplies the current state of the "item" requested.
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_config 
(
   CEI_INT16 board,          /* board to reference (0 to 15)                 */
   CEI_INT16 item            /* item for which to return the current state   */
)
{
   CEI_INT32 accessStatus;              /* status of device read/write access*/
   CEI_INT16 value;                     /* function return - item value      */
   CEI_INT32 channel = 0;               /* local item-derived channel index  */
   CEI_INT32 state = 0;                 /* function return - item value      */
   CEI_UINT32 bitFieldValue;         /* temp for device register read content*/
   CEI_INT16 status = ARS_NORMAL;       /* local invocation return status    */

   CEI_UINT32 deviceChannel;         /* current channel to select            */
   CEI_UINT32 offsetIntoChannelWindow = 0;  /* offset into chan mem region   */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /*-----------------------------------------------------------------*
       *  CEI-x20 ARINC 429 Set/Get Config Support
       *-----------------------------------------------------------------*/
      /* Report receive channel bit rate */
      if ( ((item >= ARU_RX_CH01_BIT_RATE) && (item <= ARU_RX_CH32_BIT_RATE)) || 
           ((item >= ARU_OLD_RX_CH01_BIT_RATE) && (item <= ARU_OLD_RX_CH16_BIT_RATE)) ) {

         if (item >= ARU_RX_CH01_BIT_RATE)
            channel = (CEI_INT32)item - ARU_RX_CH01_BIT_RATE;
         else
            channel = (CEI_INT32)item - ARU_OLD_RX_CH01_BIT_RATE;
            
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)ARU_RX_BITRATE, &value);
         
         if (status == ARS_NORMAL) {
            if (value > ARU_SPEED_LOW)
               state = AR_SIXTEEN_MHZ / ((CEI_INT32)value + 2);
            else
               state = (CEI_INT32)value;
         }
         else
            state = (CEI_INT32)status;
      }
      /* Report transmit channel bit rate */
      else if ( ((item >= ARU_TX_CH01_BIT_RATE) && (item <= ARU_TX_CH32_BIT_RATE)) || 
                ((item >= ARU_OLD_TX_CH01_BIT_RATE) && (item <= ARU_OLD_TX_CH16_BIT_RATE)) ) {
         
         if (item >= ARU_TX_CH01_BIT_RATE)
            channel = (CEI_INT32)item - ARU_TX_CH01_BIT_RATE;
         else
            channel = (CEI_INT32)item - ARU_OLD_TX_CH01_BIT_RATE;
            
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)ARU_TX_BITRATE, &value);
         
         if (status == ARS_NORMAL) {
            if (value > ARU_SPEED_LOW)
               state = AR_SIXTEEN_MHZ / ((CEI_INT32)value + 2);
            else
               state = (CEI_INT32)value;
         }
         else
            state = (CEI_INT32)status;
      }                                    
      /* Report receive channel parity */
      else if ( ((item >= ARU_RX_CH01_PARITY) && (item <= ARU_RX_CH32_PARITY)) ||
                ((item >= ARU_OLD_RX_CH01_PARITY) && (item <= ARU_OLD_RX_CH16_PARITY)) ) {    
         
         if (item >= ARU_RX_CH01_PARITY)
            channel = (CEI_INT32)item - ARU_RX_CH01_PARITY;
         else
            channel = (CEI_INT32)item - ARU_OLD_RX_CH01_PARITY;
            
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)ARU_RX_PARITY, &value);

         if (status == ARS_NORMAL)
            state = value == AR_ON ? AR_ODD : AR_OFF;
         else
            state = (CEI_INT32)status;
      }                                    
      /* Report transmit channel parity x20 style - requires accessing h/w directly */
      else if ( ((item >= ARU_TX_CH01_PARITY) && (item <= ARU_TX_CH32_PARITY)) ||
                ((item >= ARU_OLD_TX_CH01_PARITY) && (item <= ARU_OLD_TX_CH16_PARITY)) ) {    
         
         if (item >= ARU_TX_CH01_PARITY)
            channel = (CEI_INT32)item - ARU_TX_CH01_PARITY;
         else
            channel = (CEI_INT32)item - ARU_OLD_TX_CH01_PARITY;
            
         offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
         deviceChannel = (CEI_UINT32)channel + (CEI_UINT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
         status = a429_utl_validate_channel_request (board, deviceChannel, ARU_TX_PARITY);
         if (status == ARS_NORMAL) {
            /* read the device register with respect to the attribute specified */
            boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                         offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE));
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
            state = (bitFieldValue & CEIDEV_PARITY_SELECT_EVEN) ? AR_EVEN : AR_ODD;
         }
         else
            state = (CEI_INT32)status;
      }                                    
      /* Report transmit channel enable state */
      else if ( ((item >= ARU_TX_CH01_SHUT_OFF) && (item <= ARU_TX_CH32_SHUT_OFF)) ||
                ((item >= ARU_OLD_TX_CH01_SHUT_OFF) && (item <= ARU_OLD_TX_CH16_SHUT_OFF)) ) {    
         
         if (item >= ARU_TX_CH01_SHUT_OFF)
            channel = (CEI_INT32)item - ARU_TX_CH01_SHUT_OFF;
         else
            channel = (CEI_INT32)item - ARU_OLD_TX_CH01_SHUT_OFF;
            
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)ARU_TX_DISABLE, &value);

         if (status == ARS_NORMAL)
            state = (CEI_INT32)value;
         else
            state = (CEI_INT32)status;
            
      }                                    
      /* Report transmit channel bit count error injection */
      else if ( ((item >= ARU_TX_CH01_HB_INJ) && (item <= ARU_TX_CH32_HB_INJ)) ||
                ((item >= ARU_OLD_TX_CH01_HB_INJ) && (item <= ARU_OLD_TX_CH16_HB_INJ)) ) {    
         
         if (item >= ARU_TX_CH01_HB_INJ)
            channel = (CEI_INT32)item - ARU_TX_CH01_HB_INJ;
         else
            channel = (CEI_INT32)item - ARU_OLD_TX_CH01_HB_INJ;
            
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)ARU_TX_BIT_ERROR, &value);
         if (status != ARS_NORMAL)
            state = (CEI_INT32)status;
         else
            state = (value == AR_HI) ? AR_ON : AR_OFF;
      }                                    
      /* Report transmit channel bit count error injection */   
      else if ( ((item >= ARU_TX_CH01_LB_INJ) && (item <= ARU_TX_CH32_LB_INJ)) ||
                ((item >= ARU_OLD_TX_CH01_LB_INJ) && (item <= ARU_OLD_TX_CH16_LB_INJ)) ) {    
         
         if (item >= ARU_TX_CH01_LB_INJ)
            channel = (CEI_INT32)item - ARU_TX_CH01_LB_INJ;
         else
            channel = (CEI_INT32)item - ARU_OLD_TX_CH01_LB_INJ;
            
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)ARU_TX_BIT_ERROR, &value);
         if (status != ARS_NORMAL)
            state = (CEI_INT32)status;
         else
            state = (value == AR_LO) ? AR_ON : AR_OFF;
      }                                    
      /* Report transmit channel message gap error injection */   
      else if ( ((item >= ARU_TX_CH01_GAP_INJ) && (item <= ARU_TX_CH32_GAP_INJ)) ||
                ((item >= ARU_OLD_TX_CH01_GAP_INJ) && (item <= ARU_OLD_TX_CH16_GAP_INJ)) ) {    
         
         if (item >= ARU_TX_CH01_GAP_INJ)
            channel = (CEI_INT32)item - ARU_TX_CH01_GAP_INJ;
         else
            channel = (CEI_INT32)item - ARU_OLD_TX_CH01_GAP_INJ;
            
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)ARU_TX_GAP_ERROR, &value);
         if (status != ARS_NORMAL)
            state = (CEI_INT32)status;
         else
            state = (CEI_INT32)value;

      }
      /* return IRIG Receiver Raw Time-code Value */
      else if (item == ARU_IRIG_INPUT_TIME) {
         /* if multi-thread protection requested, semaphore/mutex-lock IRIG sample time register access */
         if (device[board].multiThreadProtect) {
            status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
            if (status != ARS_NORMAL)
               return (status);
         }
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME], (pCEI_UINT32)&bitFieldValue);  

         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect)
            accessStatus = cei_release_thread_lock(LOCK_IDENT);

         bitFieldValue &= 0x3FFFFFFF;
         state = (CEI_INT32)bitFieldValue;
      }
      /* return the board serial number (currently applies only to RAR-EC) */
      else if (item == ARU_SERIAL_NUMBER) {
         state = (CEI_UINT32)deviceGlobal[board]->serialNumber;
      }
      /* return the board serial number (currently applies only to RAR-EC) */
      else if (item == ARU_PARAMETRIC_SUPPORT) {
         state = AR_OFF;
      }
      else { /* all remaining non-x20 legacy items are handled in the routine ar_get_device_config */
         status = ar_get_device_config (board, (CEI_INT16)channel, (CEI_INT16)item, &value);
         if (status != ARS_NORMAL)
            state = (CEI_INT32)status;
         else
            state = (CEI_INT32)value;
      }
#ifndef TARGET_EMBEDDED
   }
   else
      state = ARS_INVBOARD;
#endif
   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
   
   return state;
}
 

/*===========================================================================*
 * ENTRY POINT:        A R _ G E T _ D E V I C E _ C O N F I G          
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the current state of protocol-generic device 
 *              configuration register and board-level/API attributes 
 * 
 * PARAMETERS:  CEI_INT16  board    -- (input)  board to reference (0 to 15)
 *              CEI_INT16  channel  -- (input)  channel number (0 to 31)
 *              CEI_INT16  item     -- (input)  selected channel config item
 *              pCEI_INT16 value    -- (output) current bit field state
 * 
 *              Valid combinations of item / value parameters are:
 * 
 *              for this "item"        valid options for "value" returned are:
 * 
 *              ARU_RX_FIFO_ENABLE     AR_ON           AR_OFF
 *              ARU_TX_FIFO_ENABLE     AR_ON           AR_OFF
 *              ARU_TX_DISABLE         AR_ON           AR_OFF
 *
 *              ARU_RX_BITRATE         ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom
 *              ARU_RX_PARITY          AR_ON           AR_OFF
 *              ARU_RX_MERGED_MODE     AR_ON           AR_OFF
 *              ARU_RECV_MODE          AR_WRAP_OFF     AR_WRAP_ON
 * 
 *              ARU_TX_BITRATE         ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom
 *              ARU_TX_PARITY          ARU_PARITY_NONE ARU_PARITY_ODD  ARU_PARITY_EVEN
 *              ARU_TX_GAP_ERROR       AR_ON           AR_OFF
 *              ARU_TX_BIT_ERROR       AR_HI   AR_LO   AR_OFF
 *              ARU_FAST_SLEW_RATE     AR_ON           AR_OFF
 *
 *              ARU_DISCRETE_VALUES    $0000           $FFFF
 *              ARU_DISCRETE_INPUTS    $0000           $FFFF
 *              ARU_DISCRETE_IN        AR_HI           AR_LO
 *              ARU_DIGITAL_IN         AR_HI           AR_LO
 *              ARU_DISCRETE_OUT       AR_HI           AR_LO
 *              ARU_DIGITAL_OUT        AR_HI           AR_LO
 *              ARU_DIFFERENTIAL_IN    AR_HI           AR_LO
 *              ARU_DIFFERENTIAL_OUT   AR_ON           AR_OFF
 *              ARU_HW_INTERRUPT_ENABLE    AR_ON       AR_OFF       
 *
 *              ARU_IRIG_AVAILABLE     TRUE            FALSE
 *              ARU_IRIG_WRAP_ENABLE   AR_ON           AR_OFF
 *              ARU_IRIG_OUTPUT_ENABLE AR_ON           AR_OFF
 *              ARU_IRIG_CALIBRATED
 *              ARU_IRIG_INPUT_TIME
 *              ARU_ACCESS_SNAPSHOT_BUFFER     ARU_LABEL_ONLY  ARU_LABEL_WITH_SDI
 *              ARU_RX_FIFO_COUNT      0 to 8191
 *              ARU_TX_FIFO_COUNT      0 to 8191
 *              ARU_RX_MSG_COUNT       0 to 65535
 *              ARU_TX_MSG_COUNT       0 to 65535
 *              ARU_RX_TIMETAG_MODE w/values of:                          
 *
 *                 AR_TIMETAG_EXT_IRIG_64BIT                         
 *                 AR_TIMETAG_INT_USEC_64BIT                         
 *                 AR_TIMETAG_INT_20USEC_32BIT                       
 *                 AR_TIMETAG_INT_MSEC_32BIT                         
 *                 AR_TIMER_X20_COMPAT_32BIT
 *
 *              ARU_CHAN_COUNT_429
 *              ARU_CHAN_COUNT_573
 *              ARU_CHAN_COUNT_DISC
 *              ARU_CHAN_COUNT_DIFF
 *              ARU_CONFIGURATION
 *              ARU_DEVICE_DISABLE
 *              ARU_FW_VERSION
 *              ARU_HW_ENHANCE_CHECK
 *              ARU_HW_FPGA_TEMPERATURE
 *              ARU_HW_1PT0V_PWR_SUPPLY
 *              ARU_HW_2PT5V_PWR_SUPPLY
 *
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL      Normal succesful completion.
 *                           ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                           ARS_INVARG      "item" parameter value is not
 *                                              supported by this function
 *                           ARS_INVHARVAL   Item was invalid or channel
 *                                              not supported on this device
 *                           ARS_INVBOARD    Board parameter value invalid or
 *                                              board has not been initialized
 * 
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This function returns the state of the device configuration
 *              register attribute based on the combined item/value
 *              parameter selections.  It is designed to support the 
 *              ARINC 429 and all discrete and digital I/O configuration  
 *              register attributes available to the device.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_device_config 
( 
   CEI_INT16 board,         /* (input)  board to reference (0 to 15)         */
   CEI_INT16 channel,       /* (input)  channel number (0 to 31)             */
   CEI_INT16 item,          /* (input)  selected channel configuration item  */
   pCEI_INT16 value         /* (output) current bit field state              */
)
{
   CEI_INT32 accessStatus;           /* status of device read/write access   */
   CEI_INT32 itemValue = 0;          /* temp placeholder for return value    */
   CEI_UINT32 deviceChannel;         /* current channel to select            */
   CEI_UINT32 offsetIntoChannelWindow = 0;  /* offset into chan mem region   */
   CEI_UINT32 bitFieldValue;         /* temp for device register read content*/
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   CEI_INT16 status = ARS_NORMAL;    /* function return status               */
   CEI_FLOAT parametricValue;        /* FPGA parametric working variable     */

   if ((item == ARU_HW_ENHANCE_CHECK) && (board < MAX_BOARDS)) {
      status = a429_utl_enh (board, PLX_ENHANCED_CHECK);
   }
   else if (VALID_BOARD_AND_INIT) {
      deviceChannel = (CEI_UINT32)channel;
   
      /* Validate the channel parameter for the specific item parameter */
      switch (item) {
         case ARU_RECV_MODE:
         case ARU_RX_BITRATE:
         case ARU_RX_PARITY:
         case ARU_RX_FIFO_ENABLE:
         case ARU_RX_MERGED_MODE:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            deviceChannel += (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
            break;
      
         case ARU_TX_BIT_ERROR:
         case ARU_TX_GAP_ERROR:
         case ARU_TX_BITRATE:
         case ARU_TX_PARITY:
         case ARU_TX_FIFO_ENABLE:
         case ARU_TX_DISABLE:
         case ARU_FAST_SLEW_RATE:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            deviceChannel += (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
            break;
   
         case ARU_DISCRETE_OUT:
            deviceChannel += (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DISC_OUT];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
      
         case ARU_DIGITAL_OUT:
            deviceChannel += (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DISC_OUT];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
      
         case ARU_DISCRETE_VALUES:
         case ARU_DISCRETE_INPUTS:
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_DISCRETE_INPUT_0], &bitFieldValue);
            *value = (CEI_INT16)(bitFieldValue & 0xFFFF);
            status = EXIT_ROUTINE;
            break;
      
         case ARU_DISCRETE_IN:
#ifndef TARGET_EMBEDDED
            status = a429_utl_validate_channel_request (board, deviceChannel, item);
            if ( status == ARS_NORMAL ) {
#endif
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_DISCRETE_INPUT_0], &bitFieldValue);
               *value = (CEI_INT16)(((bitFieldValue >> deviceChannel) & 1) ? AR_HI : AR_LO);
               status = EXIT_ROUTINE;
#ifndef TARGET_EMBEDDED
            }
#endif
            break;
      
         case ARU_DIGITAL_IN:
#ifndef TARGET_EMBEDDED
            status = a429_utl_validate_channel_request (board, deviceChannel, item);
            if ( status == ARS_NORMAL ) {
#endif
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_DISCRETE_INPUT_0], &bitFieldValue);
               *value = (CEI_INT16)(((bitFieldValue >> deviceChannel) & 1) ? AR_HI : AR_LO);
               status = EXIT_ROUTINE;
#ifndef TARGET_EMBEDDED
            }
#endif
            break;

         case ARU_DIFFERENTIAL_IN:
#ifndef TARGET_EMBEDDED
            status = a429_utl_validate_channel_request (board, deviceChannel, item);
            if ( status == ARS_NORMAL ) {
#endif
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_DISCRETE_INPUT_1], &bitFieldValue);
               *value = (CEI_INT16)(((bitFieldValue >> deviceChannel) & 1) ? AR_HI : AR_LO);
               status = EXIT_ROUTINE;
#ifndef TARGET_EMBEDDED
            }
#endif
            break;

         case ARU_DIFFERENTIAL_OUT:
            deviceChannel += (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DIFF_OUT];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;

         case ARU_DISCRETE_RAW:
            return ARS_NORMAL;

         case ARU_ACCESS_SNAPSHOT_BUFFER:
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
            *value = (CEI_INT16)((bitFieldValue & CEIDEV_SNAPSHOT_STORAGE_ENHANCED) ? ARU_LABEL_ONLY : ARU_LABEL_WITH_SDI);
            status = EXIT_ROUTINE;
            break;

         case ARU_HW_INTERRUPT_ENABLE:
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
            *value = (CEI_INT16)((bitFieldValue & CEIDEV_INTERRUPTS_ENABLED) ? AR_ON : AR_OFF);
            status = EXIT_ROUTINE;
            break;

         case ARU_TX_FIFO_COUNT:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_STATUS;
            deviceChannel += (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
            break;

         case ARU_RX_FIFO_COUNT:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_STATUS;
            deviceChannel += (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
            break;

         case ARU_CHAN_COUNT_429:
            *value = (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_CHAN_COUNT_573:
            *value = (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_TX]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_CHAN_COUNT_DISC:
            *value = (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_DISC_OUT]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_CHAN_COUNT_DIFF:
            *value = (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_DIFF_OUT]; 
            status = EXIT_ROUTINE;
            break;

         case ARU_DEVICE_DISABLE:
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
            *value = (CEI_INT16)( (bitFieldValue & CEIDEV_DEVICE_DISABLED) >> 31 );
            status = EXIT_ROUTINE;
            break;

         case ARU_IRIG_OUTPUT_ENABLE:
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], (pCEI_UINT32)&bitFieldValue);
            *value = (CEI_INT16)( (bitFieldValue & CEIDEV_IRIG_830RX_IRIG_ENABLE) ? AR_ON : AR_OFF);
            status = EXIT_ROUTINE;
            break;
   
         case ARU_FW_VERSION:
            *value = (CEI_INT16)deviceGlobal[board]->firmwareVersion;
            status = EXIT_ROUTINE;
            break;
      
         case ARU_CONFIGURATION:
            *value = (CEI_INT16)(deviceGlobal[board]->configuration + deviceGlobal[board]->alternateJconfiguration);
            status = EXIT_ROUTINE;
            break;
      
         case ARU_RX_TIMETAG_MODE:
            *value = (CEI_INT16)deviceGlobal[board]->deviceTimeTagReference;
            status = EXIT_ROUTINE;
            break;

         /* return IRIG Internal Wrap State */
         case 0: *value=(CEI_INT16)baseConfig; status = EXIT_ROUTINE; break;
         case ARU_IRIG_WRAP_ENABLE:
            if (deviceGlobal[board]->irigInstalled) {
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], (pCEI_UINT32)&bitFieldValue);
      
               if (bitFieldValue & CEIDEV_IRIG_INTERNAL_ENABLE)
                  *value = AR_ON;
               else
                  *value = AR_OFF;
               status = EXIT_ROUTINE;
            }
            else
               status = ARS_INVHARVAL;
            break;
      
         case ARU_IRIG_AVAILABLE:
            #if defined DEBUG_PROGRAMMING
               #include "irigdefs.h"
            #else
               *value = (CEI_UINT16)deviceGlobal[board]->irigInstalled;
            #endif
               status = EXIT_ROUTINE;
            break;
     
         case ARU_IRIG_CALIBRATED:
            *value = (CEI_UINT16)a429_utl_is_irig_calibrated(board);
            status = EXIT_ROUTINE;
            break;

         /* return IRIG Receiver Raw Time-code Value */
         case ARU_IRIG_INPUT_TIME:
            /* if multi-thread protection requested, semaphore/mutex-lock IRIG sample time register access */
            if (device[board].multiThreadProtect) {
               status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
               if (status == ARS_NORMAL) {
                  accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_INPUT_TIME], (pCEI_UINT32)&bitFieldValue);  
            
                  /* if multi-thread protection requested, free semaphore/mutex-locking */
                  if (device[board].multiThreadProtect)
                     accessStatus = cei_release_thread_lock(LOCK_IDENT);

                  bitFieldValue &= 0x3FFFFFFF;
                  *value = (CEI_INT16)bitFieldValue;
                  status = EXIT_ROUTINE;
               }
            }
            break;
            
         case ARU_RX_MSG_COUNT:
            if (deviceChannel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
               offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + deviceChannel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[offsetIntoChannelWindow], &bitFieldValue);
               *value = (CEI_INT16)bitFieldValue;
               status = EXIT_ROUTINE;
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;

         case ARU_TX_MSG_COUNT:
            if (deviceChannel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
               offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + deviceChannel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[offsetIntoChannelWindow], &bitFieldValue);
               *value = (CEI_INT16)bitFieldValue;
               status = EXIT_ROUTINE;
            }
            else {
               status = ARS_INVHARVAL;
            }
            break;

         case ARU_HW_FPGA_TEMPERATURE:
            if (deviceGlobal[board]->configuration == CEIDEV_CONFIG_CEI430A) {
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               ar_sleep (100);
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               ar_sleep (100);
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               if (bitFieldValue & 0x80) {
                  bitFieldValue &= 0x7F;
                  bitFieldValue = 127 - bitFieldValue;
                  bitFieldValue *= -1;
               }
               *value = (CEI_INT16)bitFieldValue;
            }
            else {
               accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_FPGA_FPGA_TEMPERATURE], &bitFieldValue);
               parametricValue = ((CEI_FLOAT)(bitFieldValue & 0xFFF) * ((CEI_FLOAT)503.975 / (CEI_FLOAT)1024.0)) - (CEI_FLOAT)273.15;
               *value = (CEI_INT16)((CEI_FLOAT)1000 * parametricValue); /* return temperature in millidegrees C */
            }
            status = EXIT_ROUTINE;
            break;

         case ARU_HW_1PT0V_PWR_SUPPLY:
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_FPGA_1PT0V_PWR_SUPPLY], &bitFieldValue);
            parametricValue = (CEI_FLOAT)(bitFieldValue & 0xFFF) * (CEI_FLOAT)0.00293;
            *value = (CEI_INT16)((CEI_FLOAT)1000 * parametricValue);  /* return 1.0V supply level in millivolts */
            status = EXIT_ROUTINE;
            break;

         case ARU_HW_2PT5V_PWR_SUPPLY:
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_FPGA_2PT5V_PWR_SUPPLY], &bitFieldValue);
            parametricValue = (CEI_FLOAT)(bitFieldValue & 0xFFF) * (CEI_FLOAT)0.00293;
            *value = (CEI_INT16)((CEI_FLOAT)1000 * parametricValue);  /* return 2.5V supply level in millivolts */
            status = EXIT_ROUTINE;
            break;

         default : status = ARS_INVARG; 
            break;
      }
      
      /* validate the specified channel */
#ifndef TARGET_EMBEDDED
      if ( status == ARS_NORMAL )
         status = a429_utl_validate_channel_request (board, deviceChannel, item);
#endif

      if (status == ARS_NORMAL) {
         /* read the device register with respect to the attribute specified */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE));

         accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
         if (accessStatus == ARS_NORMAL) {
   
            /* test the specified bit field and return the respective state */
            switch (item) {
               case ARU_RECV_MODE:
                  itemValue = (bitFieldValue & CEIDEV_INTERNAL_WRAP_ENABLE) ? AR_WRAP_ON : AR_WRAP_OFF;
                  break;
      
               case ARU_RX_BITRATE:
                  bitFieldValue &= CEIDEV_BAUD_RATE_FIELD;
                  bitFieldValue = bitFieldValue >> CEIDEV_BAUD_RATE_SHIFT;
      
                  /* flag default ARINC 429 bus speeds with API constant values, otherwise it's custom */
                  if (bitFieldValue == CEIDEV_BAUD_RATE_100K)
                     itemValue = ARU_SPEED_HIGH;
                  else if (bitFieldValue == CEIDEV_BAUD_RATE_12pt5K)
                     itemValue = ARU_SPEED_LOW;
                  else 
                     itemValue = (CEI_INT16)bitFieldValue;
                  break;
      
               case ARU_RX_FIFO_ENABLE:
                  itemValue = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_ON : AR_OFF;
                  break;
         
               case ARU_RX_MERGED_MODE:
                  itemValue = (bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) ? AR_ON : AR_OFF;
                  break;
      
               case ARU_RX_PARITY:
                  itemValue = (bitFieldValue & CEIDEV_PARITY_ENABLE) ? AR_ON : AR_OFF;
                  break;
      
               case ARU_TX_BITRATE:
                  bitFieldValue &= CEIDEV_BAUD_RATE_FIELD;
                  bitFieldValue = bitFieldValue >> CEIDEV_BAUD_RATE_SHIFT;
      
                  /* flag default ARINC 429 bus speeds with API constant values, otherwise it's custom */
                  if (bitFieldValue == CEIDEV_BAUD_RATE_100K)
                     itemValue = ARU_SPEED_HIGH;
                  else if (bitFieldValue == CEIDEV_BAUD_RATE_12pt5K)
                     itemValue = ARU_SPEED_LOW;
                  else 
                     itemValue = (CEI_INT16)bitFieldValue;
                  break;
                  
               case ARU_TX_BIT_ERROR:
                  if (bitFieldValue & CEIDEV_429_TXCFG_BIT_COUNT_LO)
                     itemValue = AR_LO;
                  else if (bitFieldValue & CEIDEV_429_TXCFG_BIT_COUNT_HI)
                     itemValue = AR_HI;
                  else
                     itemValue = AR_OFF;
                  break;
      
               case ARU_TX_GAP_ERROR:
                  itemValue = (bitFieldValue & CEIDEV_429_TXCFG_GAP_ERROR) ? AR_ON : AR_OFF;
                  break;
      
               case ARU_TX_PARITY:
                  if (bitFieldValue & CEIDEV_PARITY_ENABLE)
                     itemValue = (bitFieldValue & CEIDEV_PARITY_SELECT_EVEN) ? ARU_PARITY_EVEN : ARU_PARITY_ODD;
                  else
                     itemValue = ARU_PARITY_NONE;
                  break;
      
               case ARU_TX_FIFO_ENABLE:
                  itemValue = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_ON : AR_OFF;
                  break;
            
               case ARU_TX_DISABLE:
                  itemValue = (bitFieldValue & CEIDEV_TRANSMIT_DISABLE) ? AR_ON : AR_OFF;
                  break;
            
               case ARU_FAST_SLEW_RATE:
                  itemValue = (bitFieldValue & CEIDEV_FAST_SLEW_RATE) ? AR_ON : AR_OFF;
                  break;
            
               case ARU_DIGITAL_OUT:
               case ARU_DISCRETE_OUT:
                  itemValue = (bitFieldValue & 1) ? AR_HI : AR_LO;
                  break;
            
               case ARU_DIFFERENTIAL_OUT:
                  if (bitFieldValue & CEIDEV_CHANNEL_ENABLE)
                     itemValue = AR_ON;
                  else
                     itemValue = AR_OFF;
                  break;
            
               case ARU_TX_FIFO_COUNT:
                  itemValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  break;

               case ARU_RX_FIFO_COUNT:
                  itemValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  break;

               default : status = ARS_INVARG; 
            }

            *value = (CEI_INT16)itemValue;
         }
         else
            status = (CEI_INT16)accessStatus;
      }
   }
   else
      status = ARS_INVBOARD;

   if (status == EXIT_ROUTINE)
      status = ARS_NORMAL;

   return (status);
}


/*===========================================================================*
 * ENTRY POINT:         A R _ G E T _ E R R O R
 *===========================================================================*
 *
 * FUNCTION:    Function to return a message string associated with an error.
 * 
 * PARAMETERS:  CEI_INT16 error_num -- (input) A status value returned by  
 *                                             any API routine
 * 
 * RETURN VAL:  pCEI_CHAR -- string containing associated error 
 *
 * GLOBALS:     ar_errors_1[]
 *              ar_errors_2[]
 *              ar_intfc_errors[]
 *              arw_errors[]
 * 
 * DESCRIPTION: Many of the ARINC functions described here return status 
 *              values which may indicate an error condition.  This function 
 *              is passed such an error value and returns a pointer to a  
 *              message string describing the error. 
 *
 *              NOTE: This routine needs to be fixed to handle a -1 status.
 *
 *===========================================================================*/  
#ifdef NIVISA_PXI
EXPORT32 void DLL_EXPORTED ar_get_error
(
   CEI_INT16 error_num,
   pCEI_CHAR errorStringOut
)
#else
EXPORT32 pCEI_CHAR DLL_EXPORTED ar_get_error
(
   CEI_INT16 error_num
)
#endif
{   
   pCEI_CHAR error_string = NULL;

   if ( error_num == ARS_FAILURE)
      error_string = "API Call Failure";
   else if ((error_num == ARS_WINRTFAIL) || (error_num == ARS_ERR_SH_MEM_OBJ))
      error_string = cei_get_open_error();
   else if ( (error_num < 1000) && (error_num < MAX_830_1ST_ERRORS) )
      error_string = ar_830_errors_1[error_num];
   else if ( (error_num > 1000) && (error_num < 2000) 
         && ((error_num - 1000) < MAX_830_2ND_ERRORS) )
      error_string = ar_830_errors_2[error_num - 1000];
   else if ( (error_num > 1899) && (error_num < 1999) 
         && ((error_num - 1900) < MAX_830_INTFC_ERRORS) )
      error_string = ar_830_intfc_errors[error_num - 1900];
   else if ( (error_num > 2000) && (error_num < 3000) 
         && ((error_num - 2000) < MAX_830_WMS_ERRORS) )
      error_string = arw_830_errors[error_num - 2000];
   else
      error_string = "Invalid error passed to AR_GET_ERROR";

#ifdef NIVISA_PXI
   strcpy (errorStringOut, error_string);
#else
   return error_string;
#endif
}   


/*===========================================================================*
 * ENTRY POINT:                 A R _ G E T F I L T E R
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the label interrupt and filter buffer definition used 
 *              by the device enhanced label filter feature.
 * 
 * PARAMETERS:  CEI_UINT32 board      -- (input) Board number (0 to 15)  
 *              CEI_UINT32 channel    -- (input) 429 Rx channel selection,
 *                                               valid range is 0 to one less
 *                                               than the number of enhanced
 *                                               ARINC 429 channels installed
 *              pCEI_CHAR filterTable -- (input) pointer to the host filter table
 *                                               at least 8Kb in size.
 * 
 * RETURN VAL:  ARS_NORMAL        operation completed successfully
 *              ARS_INVBOARD      invalid board number or board not initialized
 *              ARS_INVARG        invalid argument
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine copies the specified label filter buffer entry
 *              from the enhanced label filter table on the device to 
 *              the caller's memory space.  Each element of the enhanced label 
 *              filter buffer consists of a bit field defined as follows:
 *
 *                 FILTER_SEQUENTIAL  0x10   if CLEAR add label to sequential
 *                                              receive buffer
 *                 FILTER_SNAPSHOT    0x20   If CLEAR add label to snapshot
 *                                              receive buffer
 *                 FILTER_INTERRUPT   0x40   If SET on reception insert channel #
 *                                              in the interrupt queue and if 
 *                                              enabled generate a PCI interrupt
 *
 *              The filter buffer for a single channel is defined as follows:
 *
 *                 CEI_UCHAR filterTable[MAX_ESSM][MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *              field   ---eSSM---  -SDI-  --------label---------
 *               bits   30  29  28   9 8   7  6  5  4  3  2  1  0
 *                    
 *              On the device each 32-bit entry in the filter table is 
 *              allocated once for each label/SDI combination, divided 
 *              into individual filter activation fields by eSSM:
 *
 *                 Label Filter Table [MAX_LABEL][MAX_SDI]
 *
 *               bits   31 30 29 28 27 ... 8  7  6  5  4  3  2  1  0
 *              field  | eSSM = 7  |   ...   | eSSM = 1  | eSSM = 0 |
 *                    
 *===========================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_getfilter 
(
   CEI_UINT32 board,             /* (input) Board number (0 to 15)           */
   CEI_UINT32 channel,           /* (input) 429 Rx channel selection         */
   pCEI_CHAR filterTable         /* (input) pointer to the host filter table */
)
{
   CEI_INT32  localChannel;    /* channel offset into the label filter table */
   CEI_INT32  labelFilterIndex;         /* label filter channel select index */
   CEI_UINT32 filterActivation;         /* local translation of action parm  */
   CEI_UINT32 essmIndex;                /* local temp enh filter table index */
   CEI_UINT32 labelIndex;               /* local temp enh filter table index */
   CEI_UINT32 sdiIndex;                 /* local temp enh filter table index */
   pCEI_CHAR  hostFilterTablePtr;       /* local pointer to host filterTable */
   CEI_INT16  status = ARS_NORMAL;      /* function return status            */
   CEI_INT32  accessStatus = ARS_NORMAL;/* status of device read/write access*/

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   
   /* Label Filter Table access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel >= deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) ||
             (filterTable == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_INT32)channel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
   
      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      }
      
      if (status == ARS_NORMAL) {
         /* copy the specified label filter table entry states from the device table to the host table */
         labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
         for (labelIndex = 0; labelIndex < MAX_LABEL; labelIndex++) {
            for (sdiIndex = 0; sdiIndex < MAX_SDI; sdiIndex++) {
               /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
               accessStatus = cei_write_32bit ((CEI_INT16)board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
               /* read the filter table entry data pattern from this location */
               accessStatus = cei_read_32bit ((CEI_INT16)board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], &filterActivation);
               /* read eight ESSM entries for each combination of SDI/label entry */
               for (essmIndex = 0; essmIndex < ARU_ALL_ESSM; essmIndex++) {
                  hostFilterTablePtr = filterTable  + (essmIndex << 10) + (sdiIndex << 8) + labelIndex;
                  *hostFilterTablePtr = (CEI_CHAR)(((filterActivation >> (essmIndex * CEIDEV_ENH_LABEL_ESSM_FILTER_SHIFT)) & 0x0000000F) << 4);
               }
               labelFilterIndex++;  
            }
         }

         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
         }
         if (accessStatus != ARS_NORMAL) {
            status = (CEI_INT16)accessStatus;
         }
      }
#ifndef TARGET_EMBEDDED
   }
#endif

   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:        A R _ G E T _ L A B E L _ F I L T E R
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the active state of filtering for all receive
 *              channels for the specified label
 * 
 * PARAMETERS:  CEI_INT16 board       -- (input) Board number (0 to 15)  
 *              CEI_UINT16 label      -- (input) the label to query the filter state
 * 
 * RETURN VAL:  Label Filter State    -- each bit indicates the label state on
 *                                       the respective receive channel 0-15
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine indicates the active state of label filtering
 *              for the specified label on each receive channel.  The state
 *              indicates active filtering with a "1" bit state for the bit 
 *              respective to the receive channel.  Active filtering is 
 *              defined as any label filter table entry for any Enhanced SSM
 *              and SDI value combination where the following filtering is
 *              enabled
 *                 FILTER_SEQUENTIAL  0x01   if CLEAR add label to sequential
 *                                              receive buffer
 *                 FILTER_SNAPSHOT    0x02   If CLEAR add label to snapshot
 *                                              receive buffer
 *
 *===========================================================================*/ 
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_label_filter 
(
   CEI_INT16  board, 
   CEI_UINT16 label
) 
{
   CEI_UINT32 status = 0;               /* function return status            */
   CEI_UINT32 sdiIndex;                 /* local temp enh filter table index */
   CEI_UINT32 localChannel;          /* channel reference for ar_get_data_xt */
   CEI_UINT32 labelFilterIndex;         /* label filter channel select index */
   CEI_UINT32 filterActivation;         /* local translation of action parm  */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   
   if (VALID_BOARD_AND_INIT) {
      /* read the label filter table for the specified filter on all channels, 
       * and if any eSSM enties are non-Zero, this label is considered "filtered" */
      for (localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
           localChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
           localChannel++) {
         for (sdiIndex = 0; sdiIndex < ARU_ALL_SDI; sdiIndex++) {
            labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
            labelFilterIndex += ((CEI_UINT32)label << CEIDEV_ENH_LABEL_FILTER_LB_SHIFT) + sdiIndex;
            
            /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
            /* read the current contents of the label filter table location(s) to be updated */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], &filterActivation);
            /* mask out only the selected ESSM field in this entry for the requested filter update, then add the new setting(s) */
            if (filterActivation & 0x33333333)
               status |= 1 << (localChannel - deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]);
         }
      }
   }
   
   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:             A R _ G E T _ R X _ C O U N T
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the count of received messages on this channel
 *
 * PARAMETERS:  CEI_INT16 board   -- (input) Board to reference (0 to 15)
 *              CEI_INT16 channel -- (input) Receive channel to use (0 to 31)
 * 
 * RETURN VAL   CEI_UINT32                   Count of ARINC messages received
 * 
 * GLOBALS:     deviceGlobal[board]
 *              device[board]
 * 
 * DESCRIPTION: The device maintains a count of the number of ARINC data 
 *              messages received over the interface for each channel since 
 *              the device was last initialized (see AR_LOADSLV). 
 *              This routine returns that number.
 *
 *===========================================================================*/
EXPORT32 CEI_UINT32 DLL_EXPORTED ar_get_rx_count
( 
   CEI_INT16 board,              /* (input)  board to reference (0 to 15)    */
   CEI_INT16 channel             /* (input)  recv channel to ref (0 to 31)   */
)
{
   CEI_INT32  accessStatus = ARS_NORMAL; /* status of dev read/write access  */
   CEI_INT32  offsetChannelValue;        /* actual channel on the device     */
   CEI_UINT32 offsetIntoChannelWindow;   /* offset into chan mem region      */
   CEI_UINT32 bitFieldValue = 0;         /* temp for device reg read content */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if ((CEI_UINT32)channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
#endif
         offsetChannelValue = (CEI_INT32)channel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + (CEI_UINT32)offsetChannelValue;
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[offsetIntoChannelWindow], &bitFieldValue);
         if (bitFieldValue >= deviceGlobal[board]->receiveChannelMessageCount[channel]) {
            bitFieldValue -= deviceGlobal[board]->receiveChannelMessageCount[channel];
         }
         else {
            bitFieldValue += (~deviceGlobal[board]->receiveChannelMessageCount[channel]) + 1;
         }
#ifndef TARGET_EMBEDDED
      }
   }
#endif
   
   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
   return bitFieldValue;
}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_clr_rx_count (CEI_INT16 board, CEI_INT16 channel) 
{
   CEI_INT32  accessStatus = ARS_NORMAL; /* status of dev read/write access  */
   CEI_INT32  offsetChannelValue;        /* actual channel on the device     */
   CEI_UINT32 offsetIntoChannelWindow;   /* offset into chan mem region      */
   CEI_UINT32 bitFieldValue = 0;         /* temp for device reg read content */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if ((CEI_UINT32)channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
#endif
         offsetChannelValue = (CEI_INT32)channel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         offsetIntoChannelWindow = CEIDEV_CHANNEL_STATS_START + (CEI_UINT32)offsetChannelValue;
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[offsetIntoChannelWindow], &bitFieldValue);
         deviceGlobal[board]->receiveChannelMessageCount[channel] = bitFieldValue;
#ifndef TARGET_EMBEDDED
      }
   }
#endif
   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
}
 

/*===========================================================================*
 * ENTRY POINT:             A R _ G E T _ S T A T U S
 *===========================================================================*
 *
 * FUNCTION:    Retrieve FIFO Data Available bit status for all receivers
 *
 * PARAMETERS:  CEI_INT16  board -- (input)  board to reference (0 to 15)
 *              pCEI_INT16 state -- (output) value if interrupt status reg.
 * 
 * RETURN VAL   CEI_INT16        -- ARS_NORMAL       Success
 *                                  ARS_INVBOARD      Invalid board parameter
 * 
 * GLOBALS:     deviceGlobal[board]
 *              device[board]
 * 
 * DESCRIPTION: Supplies the combined state of each Rx FIFO status register
 *              FIFO Data Available bit.  Supports up to 16 multi-protocol 
 *              receivers.
 *
 *              For the CEI-830, this routine returns the combined state of 
 *              the Rx FIFO Status Register Data Available Bit for each of
 *              the individual receiver FIFOs. 
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_status 
( 
   CEI_INT16 board,         /* (input)  board to reference (0 to 15)         */
   pCEI_INT16 state         /* (output) bit value from each FIFO status reg. */
)
{
   CEI_UINT32 channelIndex;             /* local channel index               */
   CEI_UINT32 channelStatus;            /* temp FIFO status by channel       */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_INT32 combinedStatus = 0;        /* accum. FIFO status for all chan's */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   else {
#endif
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS +
                                   (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE));
      for ( channelIndex = 0; 
            (channelIndex <= deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) && (accessStatus == ARS_NORMAL); 
            channelIndex++ ) {
         accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
         combinedStatus |= ( (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) << channelIndex);
         boardPointer += CEIDEV_CHANNEL_REGION_SIZE;                        /* setup for next channel */
      }
      
      *state = (CEI_INT16)combinedStatus;
      
#ifndef TARGET_EMBEDDED
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
         status = (CEI_INT16)accessStatus;
   }
#endif
   
   return status;
}
 

/*===========================================================================*
 * ENTRY POINT:      A R _ G E T _ S T O R A G E _ M O D E
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the board-level receive data storage mode
 *
 * PARAMETERS:  CEI_INT16  board -- (input)  board index (0 to 15)
 *              pCEI_INT16 mode  -- (output) valid values are:
 *                                  ARU_BUFFERED individual buffer per channel
 *                                  ARU_MERGED   one merged circular buffer
 * 
 * RETURN VAL:  ARS_NORMAL        success
 *              ARS_INVBOARD      invalid board parameter or board not 
 *                                initialized
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine returns the current "global" storage mode to the
 *              application .
 *
 *              This routine is provided for backward compatibility with
 *              the CEI-710 API.  The proper method to retrieve the Merged
 *              Receive Mode is provided for individual channels via the 
 *              respective channel get-config routine using the item
 *              parameter value ARU_RX_MERGED_MODE and its ON/OFF selection.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_storage_mode 
(
   CEI_INT16 board,   /* (input)  board index (0 to 15)                      */
   pCEI_INT16 mode    /* (output) storage mode (ARU_BUFFERED or ARU_MERGED)  */
)
{
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT)
#endif
      *mode = (CEI_INT16)deviceGlobal[board]->receiveMode;
#ifndef TARGET_EMBEDDED
   else
      status = ARS_INVBOARD;
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ G E T _ T I M E R C N T L
 *===========================================================================*
 *
 * FUNCTION:    This routine returns the value of the device timer in a
 *              32-bit format.  If the current selected system time format
 *              is 64-bit or IRIG, the 1 millisecond 32-bit format is used.
 *
 * PARAMETERS:  CEI_INT16 board   -- (input) Board number (0 to 15)  
 * 
 * RETURN VAL:  CEI_UINT32 -- the value of the device timer
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 *===========================================================================*/
EXPORT32 CEI_UINT32 DLL_EXPORTED ar_get_timercntl (CEI_INT16 board) 
{
   CEI_UINT32 currentTime;           /* temp storage for current timer value */
   AR_TIMETAG_TYPE timeTag;          /* temp struct storage for current time */
   CEI_INT32 timeTagReference;
   CEI_INT32 status;
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      return ARS_INVBOARD;
#endif
   /* if a 32-bit time-tag ref. isn't already selected, force 32-bit default */
   timeTagReference = (CEI_INT32)deviceGlobal[board]->deviceTimeTagReference;
   if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) &&
        (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) ) {
      timeTagReference = AR_TIMETAG_INT_MSEC_32BIT;
   }
   
   boardPointer = (pCEI_UINT32)(device[board].devicePtr);
   status = a429_utl_acquire_formatted_time_tag ( board,
                                                  boardPointer,
                                                  AR_TIMER_REGISTER_SET, 
                                                  timeTagReference, 
                                                  AR_OFF, 
                                                  &timeTag );
   
   currentTime = (CEI_UINT32)timeTag.timeTag;
   
   #ifdef INTEGRITY_PCI_PPC
      if (status) ;
   #endif
   return ( currentTime );
}
EXPORT32 CEI_UINT16 DLL_EXPORTED ar_get_timercnt (CEI_INT16 board) 
{
   return (CEI_UINT16)ar_get_timercntl(board);
}


/*===========================================================================*
 * ENTRY POINT:             A R _ G E T _ T I M E
 *===========================================================================*
 *
 * FUNCTION:    This routine returns the current time reference value scaled 
 *              from either the CEI-830 internal 64-bit timer or the most 
 *              recently received IRIG timer reference, as specified via 
 *              the format parameter.  If the requested timer reference is 
 *              IRIG, the syncTime parameter will contain the timer-
 *              referenced time-tag assigned when the IRIG time reference 
 *              was received by the CEI-830.
 *
 * PARAMETERS:  CEI_INT16 board  -- (input)  Board number (0 to 15)  
 * 
 *              CEI_INT16 format -- (input)  Format of time value to be returned     
 * 
 *              pAR_TIMETAG_TYPE timeTagRef    -- (output)
 *                             ->timeTagFormat -- format of timeTag member
 *                                              AR_TIMETAG_EXT_IRIG_64BIT
 *                                              AR_TIMETAG_INT_USEC_64BIT
 *                                              AR_TIMETAG_HOST_USEC_64BIT
 *                                              AR_TIMETAG_INT_20USEC_32BIT
 *                                              AR_TIMETAG_INT_MSEC_32BIT
 *                                              AR_TIMER_X20_COMPAT_32BIT
 * 
 *                             ->timeTag -- timer value in the resolution and 
 *                                          size selected via "format" parameter
 * 
 *                             ->referenceTimeTag -- a 64-bit 1 microsecond
 *                                          indicating one of the following:
 * 
 *                                  1) The current device timer value, or
 *                                  2) For IRIG, the device timer value 
 *                                     when the last IRIG update was received.
 * 
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid timeTagFormat structure member
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     boardInitialized
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_time 
(
   CEI_INT16 board,                 /* (input)  board to reference (0 to 15) */
   CEI_INT16 format,                /* (input)  requested format of time     */
   pAR_TIMETAG_TYPE timeTag         /* (output) formatted time-tag value     */
)
{
   CEI_INT16 status;
   TIME_TAG_TYPE hostTime;
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   #if defined (NO_INT64)
   else if ( (format == AR_TIMETAG_HOST_USEC_64BIT) || (format == AR_TIMETAG_HOST_SEC_32BIT) ) {
   #else   
   else if ( format == AR_TIMETAG_HOST_USEC_64BIT ) {
   #endif
#else
   #if defined (NO_INT64)
   if ( (format == AR_TIMETAG_HOST_USEC_64BIT) || (format == AR_TIMETAG_HOST_SEC_32BIT) ) {
   #else   
   if ( format == AR_TIMETAG_HOST_USEC_64BIT ) {
   #endif
#endif
   #ifndef NO_INT64
      boardPointer = (pCEI_UINT32)(device[board].devicePtr);
      status = (CEI_INT16)a429_utl_acquire_formatted_time_tag ( board,
                                                                boardPointer,
                                                                AR_TIMER_REGISTER_SET, 
                                                                AR_TIMETAG_INT_USEC_64BIT, 
                                                                AR_OFF, 
                                                                timeTag );
      timeTag->timeTagFormat = AR_TIMETAG_HOST_USEC_64BIT;
   #else   
      timeTag->timeTagFormat = AR_TIMETAG_HOST_SEC_32BIT;
   #endif
      cei_get_host_system_time_x (&hostTime);
      timeTag->timeTag = hostTime;
   }
   else {
      boardPointer = (pCEI_UINT32)(device[board].devicePtr);
      status = (CEI_INT16)a429_utl_acquire_formatted_time_tag ( board,
                                                                boardPointer,
                                                                AR_TIMER_REGISTER_SET, 
                                                                (CEI_INT32)format, 
                                                                AR_OFF, 
                                                                timeTag );
   }
   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:      A R _ G E T _ T R A N S M I T T E R _ M O D E
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the ARINC 573 multi-protocol transmitter mode
 *
 * PARAMETERS:  CEI_INT16 board -- (input) board index (0 to 15)
 *              pCEI_INT16 mode -- (input) valid values are:
 *                       ARU_XMIT_573_HBP    P3 Pins 9-11 are 573-HBP output
 *                       ARU_XMIT_573_BPRZ   P3 Pins 9-11 are 573-BPRZ output
 *
 * RETURN VAL:  ARS_NORMAL        success
 *              ARS_INVBOARD      invalid board parameter or board not 
 *                                initialized
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine returns the currently enabled ARINC 573/717/CSDB  
 *              transmitter configuration to the application
 *
 *              This routine is provided for backward compatibility with
 *              the CEI-710 API.  The proper method to retrieve the ARINC 573
 *              Encoding as BPRZ or HBP is provided for individual channel(s) 
 *              via the ar_get_573_config routine using the item parameter
 *              values ARU_573_TX_HBP_SELECT and ARU_573_TX_BPRZ_SELECT.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_transmitter_mode 
(
   CEI_INT16 board,   /* (input)  board index (0 to 15)                      */
   pCEI_INT16 mode    /* (output) transmitter mode (CSDB 717-HBP or 717-BPRZ */
)
{
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT)
#endif
      *mode = (CEI_INT16)deviceGlobal[board]->transmitMode;
#ifndef TARGET_EMBEDDED
   else
      status = ARS_INVBOARD;
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:              A R _ G E T _ 4 2 9 _ M E S S A G E
 *===========================================================================*
 *
 * FUNCTION:    Read a 429 message from the ARINC 429 receiver FIFO 
 *              buffer specified via the channel parameter.
 *
 * PARAMETERS:  CEI_INT16 board     -- (input)  board to reference (0 to 15)
 *              CEI_INT16 channel   -- (input)  429 recvr channel to use (0 to 31)
 *              CEI_INT16 waitState -- (input)  to wait for 1/2 second or not
 *                                              designated as AR_ON or AR_OFF
 *              pCEI_VOID data      -- (output) location to store received data
 *              pCEI_VOID timeTag   -- (output) location to store data timeTag
 *
 * RETURN VAL   CEI_INT16 status -- ARS_GOTDATA       Success
 *                                  ARS_NODATA        No data available (no wait)
 *                                  ARS_CHAN_TIMEOUT  No data available (w/wait)
 *                                  ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *                                  ARS_INVBOARD      Invalid board parameter
 *                                                       or board not initialized
 *                                  ARS_INVHARVAL     Unsupported channel parameter
 *                                  ARS_INVARG        NULL data parameter value
 *                                  ARS_BAD_MESSAGE   An invalid length msg was received
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine invokes one of the two generic get-data  
 *              routines based on whether or not the waitState indicates to  
 *              delay 1/2 second.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_429_message  
( 
   CEI_INT16 board,              /* (input)  board to reference (0 to 15)    */
   CEI_INT16 channel,            /* (input)  channel number to ref (0 to 31) */
   CEI_INT16 waitState,          /* (input)  to wait for 1/2 second or not   */
   pCEI_VOID data,               /* (output) location to store received data */
   pCEI_VOID timeTag             /* (output) location to store data timeTag  */
)
{
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
      
   if (waitState == AR_ON)
      status = ar_getnextt ( board, channel, data, timeTag );
   else
      status = ar_getwordt ( board, channel, data, timeTag );

   return (status);
}
           

/*===========================================================================*
 * ENTRY POINT:              A R _ G E T N E X T
 *===========================================================================*
 *
 * FUNCTION:    Read a message from the specified Receive FIFO.  Give the 
 *              message a maximum of 1/2 second to arrive before timing out.
 *
 * PARAMETERS:  CEI_INT16 board        -- (input)  board to reference
 *              CEI_INT16 channel      -- (input)  recvr channel to use (0 to 31)
 *              pCEI_VOID destination  -- (output) location to receive data
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA       Success
 *                              ARS_INVHARVAL     Unsupported channel #
 *                              ARS_INVARG        NULL data parameter value
 *                              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *                              ARS_CHAN_TIMEOUT  No data available
 *                              ARS_BAD_MESSAGE   An invalid length msg was received
 *                              ARS_INVBOARD      Invalid board parameter
 *                                                or board not initialized
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC 429 data.
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 * 
 *              This routine polls the receiver continuously, incorporating a
 *              1 millisecond sleep/task delay until:
 *                 1) An ARINC word is received or
 *                 2) The timeout threshold of 1/2 second is reached.
 * 
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_getnext 
( 
   CEI_INT16 board,              /* (input)  board to reference (0 to 15)    */
   CEI_INT16 channel,            /* (input)  recv channel to ref (0 to 31)   */
   pCEI_VOID destination         /* (output) location to store received data */
)
{
   CEI_INT16 status;                 /* function return status               */
   CEI_UINT32 goal;                  /* local time-out target value          */
   CEI_INT32 passes = 50;            /* 50 10ms retries to receive data      */
   CEI_INT16 localChannel;           /* channel reference for ar_get_data_xt */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getword channel is 0-referenced ARINC 429 receive channel index;
       * ar_get_data_xt uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
#endif
         localChannel = channel + (CEI_INT16)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];

         goal = ar_get_timercntl(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);

         /*-----------------------------------------------------------------------*
          *  Wait for a message to arrive in the buffer, or exit if timeout.
          *-----------------------------------------------------------------------*/
         while ((status = ar_get_data_xt (board, &localChannel, (pCEI_UINT32)destination, NULL)) == ARS_NODATA) {
            /*-------------------------------------------------------------------*
             *  Initialize variable for time to wait until if this is first pass.
             *
             *  NOTE: this logic is inside the loop for efficiency so as to
             *        possibly avoid an extra, time-consuming call to clock()
             *        if it is not needed (i.e. if data is waiting at the first
             *        call to ip_getword().)
             *-------------------------------------------------------------------*/
            if ( (passes > 0) && (ar_get_timercntl(board) > goal) ) {
               passes--;
               goal = ar_get_timercntl(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
            }
      
            /*-------------------------------------------------------------------*
             *  If more than 1/2 second of processing time has passed, exit.
             *-------------------------------------------------------------------*/
            else if ( (passes == 0) && (ar_get_timercntl(board) > goal) ) {
               status = ARS_CHAN_TIMEOUT;
               break;
            }
            else { /* release this thread from the o/s for 1 millisecond */
               ar_sleep(1);
            }
         }
         
         /* if we timed-out, make one last attempt to get some data */
         if (status == ARS_CHAN_TIMEOUT) {
            status = ar_get_data_xt ( board, &localChannel, (pCEI_UINT32)destination, NULL );
            if (status == ARS_NODATA)
               status = ARS_CHAN_TIMEOUT;
         }
#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return status;
}
                                                 

/*===========================================================================*
 * ENTRY POINT:              A R _ G E T N E X T T
 *===========================================================================*
 *
 * FUNCTION:    Read the next incoming ARINC data word and its timetag.  Give
 *              the word a maximum of 1/2 second to arrive before timing out.
 *
 * PARAMETERS:  CEI_INT16 board       -- (input)  board to reference (0 to 15)
 *              CEI_INT16 channel     -- (input)  recvr channel to use (0 to 31)
 *              pCEI_VOID destination -- (output) location to receive data
 *              pCEI_VOID timeTag     -- (output) location to place timetag
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA       Success
 *                              ARS_CHAN_TIMEOUT  No data available
 *                              ARS_INVHARVAL     Unsupported channel parameter
 *                              ARS_INVARG        NULL data parameter value
 *                              ARS_INVBOARD      Invalid board parameter
 *                              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *                              ARS_BAD_MESSAGE   An invalid length msg was received
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC 429 data.
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 * 
 *              This routine polls the receiveer continuously until:
 *                 1) An ARINC word is received or
 *                 2) The timeout threshold (1/2 second) is reached.
 * 
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_getnextt 
( 
   CEI_INT16 board,              /* (input)  board to reference (0 to 15)    */
   CEI_INT16 channel,            /* (input)  recv channel to ref (0 to 31)   */
   pCEI_VOID destination,        /* (output) location to store received data */
   pCEI_VOID timeTag             /* (output) location to store data time-tag */
)
{
   CEI_INT16 status;                   /* function return status             */
   CEI_UINT32 goal;                    /* local time-out target value        */
   CEI_INT32 passes = 50;              /* 50 10ms retries to receive data    */
   AR_TIMETAG_TYPE tempTimeTag;        /* local time-tag reference structure */
   CEI_INT32 timeTagReference;         /* local Time-Tag Mode reference      */
   CEI_INT16 localChannel;             /* merged/individual channel select   */
   pCEI_UINT32 timeTagPtr = (pCEI_UINT32)timeTag;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getword channel is 0-referenced ARINC 429 receive channel index;
       * ar_get_data_xt uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CEI_INT16)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         /* if a 32-bit time-tag ref. isn't already selected, force 32-bit as a default */
         timeTagReference = deviceGlobal[board]->deviceTimeTagReference;
         if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) && (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) ) {
            timeTagReference = AR_DEFAULT_RECEIVE_DATA_TIMETAG_FORMAT;
         }
         localChannel = channel + (CEI_INT16)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];

         goal = ar_get_timercntl(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
      
         /*-----------------------------------------------------------------------*
          *  Wait for something, or exit if timeout.
          *-----------------------------------------------------------------------*/
         tempTimeTag.timeTagFormat = timeTagReference;
      
         while ( (status = ar_get_data_xt (board, &localChannel, (pCEI_UINT32)destination, &tempTimeTag)) == ARS_NODATA ) {
      
            /*-------------------------------------------------------------------*
             *  Initialize variable for time to wait until if this is first pass.
             *
             *  NOTE: this logic is inside the loop for efficiency so as to
             *        possibly avoid an extra, time-consuming call to clock()
             *        if it is not needed (i.e. if data is waiting at the first
             *        call to ip_getword().)
             *-------------------------------------------------------------------*/
      
            if ( (passes > 0) && (ar_get_timercntl(board) > goal) ) {
               passes--;
               goal = ar_get_timercntl(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
            }
      
            /*-------------------------------------------------------------------*
             *  If more than 1/2 second of processing time has passed, exit.
             *-------------------------------------------------------------------*/
            else if ( (passes == 0) && (ar_get_timercntl(board) > goal) ) {
               status = ARS_CHAN_TIMEOUT;
               break;
            }
            else { /* release this thread from the o/s for 1 millisecond */
               ar_sleep(1);
            }
         }
         
         /* if we timed-out, make one last attempt to get some data */
         if (status == ARS_CHAN_TIMEOUT) {
            status = ar_get_data_xt (board, &localChannel, (pCEI_UINT32)destination, &tempTimeTag);
            if (status == ARS_NODATA)
               status = ARS_CHAN_TIMEOUT;
         }

         if ((status == ARS_GOTDATA) & (timeTag != NULL)) {
            *timeTagPtr = (CEI_UINT32)(tempTimeTag.timeTag & 0xFFFFFFFF);
         }

#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:              A R _ G E T N E X T _ X T
 *===========================================================================*
 *
 * FUNCTION:    Read the next incoming ARINC data word and its timetag.  Give
 *              the word a maximum of 1/2 second to arrive before timing out.
 *              Return the extended time-tag structure supporting IRIG.
 *
 * PARAMETERS:  CEI_INT16 board    -- (input)  board to reference (0 to 15)
 *              CEI_INT16 channel  -- (input)  recvr channel to use (0 to 7)
 *              pCEI_UINT32 data   -- (output) location for receive data
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                 -- (output) location for time-tag struct
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA       Success
 *                              ARS_CHAN_TIMEOUT  No data available
 *                              ARS_INVHARVAL     Unsupported channel parameter
 *                              ARS_INVARG        NULL data parameter value
 *                              ARS_INVBOARD      Invalid board parameter
 *                              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *                              ARS_BAD_MESSAGE   An invalid length msg was received
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC 429 data.
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 * 
 *              This routine polls the receiveer continuously until:
 *                 1) An ARINC word is received or
 *                 2) The timeout threshold (1/2 second) is reached.
 * 
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_getnext_xt 
( 
   CEI_INT16 board,             /* (input)  board to reference (0 to 15)     */
   CEI_INT16 channel,           /* (input)  recv channel to ref (0 to 31)    */
   pCEI_UINT32 data,            /* (output) location to store received data  */
   pAR_TIMETAG_TYPE timeTagRef  /* (output) location to store data time-tags */
)
{
   CEI_INT16 localChannel;              /* merged/individual channel select  */
   CEI_INT16 status;                    /* function return status            */
   CEI_UINT32 goal;                     /* local time-out target value       */
   CEI_INT32 passes = 50;               /* 50 10ms retries to receive data   */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getword channel is 0-referenced ARINC 429 receive channel index;
       * ar_get_data_xt uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CEI_INT16)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         localChannel = channel + (CEI_INT16)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         
         goal = ar_get_timercntl(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
      
         /*-----------------------------------------------------------------------*
          *  Wait for something, or exit if timeout.
          *-----------------------------------------------------------------------*/
      
         while ( (status = ar_get_data_xt (board, &localChannel, data, timeTagRef)) == ARS_NODATA ) {
      
            /*-------------------------------------------------------------------*
             *  Initialize variable for time to wait until if this is first pass.
             *
             *  NOTE: this logic is inside the loop for efficiency so as to
             *        possibly avoid an extra, time-consuming call to clock()
             *        if it is not needed (i.e. if data is waiting at the first
             *        call to ip_getword().)
             *-------------------------------------------------------------------*/
      
            if ( (passes > 0) && (ar_get_timercntl(board) > goal) ) {
               passes--;
               goal = ar_get_timercntl(board) + ((CEI_INT32)TIMER_CLOCK_TICKS_PER_SECOND / 100);
            }
      
            /*-------------------------------------------------------------------*
             *  If more than 1/2 second of processing time has passed, exit.
             *-------------------------------------------------------------------*/
            else if ( (passes == 0) && (ar_get_timercntl(board) > goal) ) {
               status = ARS_CHAN_TIMEOUT;
               break;
            }
         }
         
         /* if we timed-out, make one last attempt to get some data */
         if (status == ARS_CHAN_TIMEOUT) {
            status = ar_get_data_xt (board, &localChannel, data, timeTagRef);
            if (status == ARS_NODATA)
               status = ARS_CHAN_TIMEOUT;
         }

#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:              A R _ G E T W O R D
 *===========================================================================*
 *
 * FUNCTION:    Read the ARINC word from a receiver FIFO buffer.
 *
 * PARAMETERS:  CEI_INT16 board       -- (input)  board to reference (0 to 15)
 *              CEI_INT16 channel     -- (input)  recvr channel to use (0 to *)
 *              pCEI_VOID destination -- (output) location to receive data
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA     Success
 *                              ARS_NODATA      No data available
 *                              ARS_INVHARVAL   Protocol not available on device
 *                              ARS_INVARG      NULL data parameter value
 *                              ARS_INVBOARD    Invalid board parameter
 *                                              or board not initialized
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_BAD_MESSAGE An invalid length msg was received
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC data.
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 receive channel index, starting with zero.  If it *
 *              exceeds the 429 receive channel count and an ARINC 573/717
 *              receiver exists, it will be used as the designated receiver.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_getword 
( 
   CEI_INT16 board,              /* (input)  board to reference (0 to 15)    */
   CEI_INT16 channel,            /* (input)  recv channel to ref (0 to 31)   */
   pCEI_VOID destination         /* (output) location to store received data */
)
{
   CEI_INT16 status = ARS_NORMAL;    /* function return status               */
   CEI_INT16 localChannel;           /* channel reference for ar_get_data_xt */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* legacy ar_getword channel is 0-referenced ARINC 429 receive channel index;
       * ar_get_data_xt uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel >= (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
         if (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] != CEIDEV_MAX_CHANNELS) {
            localChannel = (CEI_INT16)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
         }
         else {
            status = ARS_INVHARVAL;
         }
      }
      else {
         localChannel = channel + (CEI_INT16)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      }
      if (status == ARS_NORMAL) {
         status = ar_get_data_xt ( board, &localChannel, (pCEI_UINT32)destination, NULL );
      }

#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}
           

/*===========================================================================*
 * ENTRY POINT:              A R _ G E T W O R D T
 *===========================================================================*
 *
 * FUNCTION:    Read the ARINC word and timetag from a receiver FIFO buffer.
 *
 * PARAMETERS:  CEI_INT16 board       -- (input)  board to reference (0 to 15)
 *              CEI_INT16 channel     -- (input)  recvr channel to use (0 to 255)
 *              pCEI_VOID destination -- (output) location to receive data
 *              pCEI_VOID timeTag     -- (output) location to receive timetag
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA     Success
 *                              ARS_NODATA      No data availalble
 *                              ARS_INVHARVAL   Protocol not available on device
 *                              ARS_INVARG      NULL data parameter value
 *                              ARS_INVBOARD    Invalid board parameter
 *                                              or board not initialized
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_BAD_MESSAGE An invalid length msg was received
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC word 
 *              and timetag from the specified channel buffer.  The 
 *              channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_getwordt 
( 
   CEI_INT16 board,             /* (input)  board to reference (0 to 15)     */
   CEI_INT16 channel,           /* (input)  recv channel to ref (0 to 31)    */
   pCEI_VOID destination,       /* (output) location to store received data  */
   pCEI_VOID timeTag            /* (output) location to store data time-tags */
)
{
   CEI_INT16 localChannel;             /* merged/individual channel select   */
   CEI_INT16 status;                   /* function return status             */
   AR_TIMETAG_TYPE tempTimeTag;        /* local time-tag reference structure */
   CEI_INT32 timeTagReference;         /* local Time-Tag Mode reference      */
   pCEI_UINT32 timeTagPtr = (pCEI_UINT32)timeTag;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getwordt channel is 0-referenced ARINC 429 receive channel index;
       * ar_get_data_xt uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CEI_INT16)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         /* if a 32-bit time-tag ref. isn't already selected, force 32-bit as a default */
         timeTagReference = deviceGlobal[board]->deviceTimeTagReference;
         if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) && (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) ) {
            timeTagReference = AR_DEFAULT_RECEIVE_DATA_TIMETAG_FORMAT;
         }
         localChannel = channel + (CEI_INT16)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         tempTimeTag.timeTagFormat = timeTagReference;
         status = ar_get_data_xt ( board, &localChannel, (pCEI_UINT32)destination, &tempTimeTag );
         if ((status == ARS_GOTDATA) & (timeTag != NULL)) {
            *timeTagPtr = (CEI_UINT32)(tempTimeTag.timeTag & 0xFFFFFFFF);
         }
#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:              A R _ G E T W O R D _ X T
 *===========================================================================*
 *
 * FUNCTION:    Read the ARINC word and time-tag from a receiver FIFO buffer,
 *              returning the extended time-tag structure supporting IRIG
 *
 * PARAMETERS:  CEI_INT16 board    -- (input)  board to reference (0 to 15)
 *              CEI_INT16 channel  -- (input)  recvr channel to ref (0 to 31)
 *              pCEI_UINT32 data   -- (output) location for receive data
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                 -- (output) location for time-tag struct
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA     Success
 *                              ARS_NODATA      No data availalble
 *                              ARS_INVHARVAL   Protocol not available on device
 *                              ARS_INVARG      NULL data parameter value
 *                              ARS_INVBOARD    Invalid board parameter
 *                                              or board not initialized
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_BAD_MESSAGE An invalid length msg was received
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine reads the most recently received ARINC word 
 *              and 64-bit timetag from the specified channel buffer.  The 
 *              channel number passed to this routine corresponds to the
 *              ARINC 429 channel number index, starting with zero.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_getword_xt
(
   CEI_INT16 board,             /* (input)  board to reference (0 to 15)     */
   CEI_INT16 channel,           /* (input)  recv channel to ref (0 to 31)    */
   pCEI_UINT32 data,            /* (output) location to store received data  */
   pAR_TIMETAG_TYPE timeTagRef  /* (output) location to store data time-tags */
)
{
   CEI_INT16 localChannel;              /* merged/individual channel select  */
   CEI_INT16 status;                    /* function return status            */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* legacy ar_getword_xt channel is 0-referenced ARINC 429 receive channel index;
       * ar_get_data_xt uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
      if (channel < (CEI_INT16)(deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX])) {
#endif
         localChannel = channel + (CEI_INT16)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         status = ar_get_data_xt (board, &localChannel, data, timeTagRef);

#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (status);
}


/*===========================================================================*
 * ENTRY POINT:                 A R _ G E T _ D A T A  
 *===========================================================================*
 *
 * FUNCTION:    Read the next data and 64-bit time-tag from a receive FIFO.
 *
 * PARAMETERS:  CEI_INT16 board    -- (input)  board to reference (0 to 15)
 *              pCEI_INT16 channel -- (input)  recv channel to reference (0 to 255)
 *                                    (output) in merged mode, the channel this
 *                                             data was actually received on
 *              pCEI_UINT32 
 *                  data           -- (output) location to store received data
 *
 *              pCEI_UINT32 
 *                  timeTagLo      -- (output) location to store timetag lsw
 *
 *              pCEI_UINT32 
 *                  timeTagHi      -- (output) location to store timetag msw
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA     Success
 *                              ARS_NODATA      No data availalble
 *                              ARS_INVHARVAL   Unsupported channel parameter
 *                              ARS_INVBOARD    Invalid board parameter
 *                                                 or board not initialized
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_BAD_MESSAGE An invalid length msg was received
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine reads the most recently received entry (data and
 *              timetag) from the specified receive channel FIFO.  The 
 *              channel number passed to this routine corresponds to the
 *              hardware channel assignment on the device.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_data 
( 
   CEI_INT16 board,            /* (input)  board to reference (0 to 15)      */
#if defined (LABVIEW_RT)
   CEI_INT16 channel,          /* (input)  channel number to receive from    */
#else   
   pCEI_INT16 channel,         /* (output) in merged mode, chan received on  */
#endif
   pCEI_UINT32 data,           /* (output) location to store received data   */
   pCEI_UINT32 timeTagLo,      /* (output) location to store time-tag low wd */
   pCEI_UINT32 timeTagHi       /* (output) location to store time-tag hi wrd */
)
{
#if defined (LABVIEW_RT)
   CEI_INT16 localChannel = channel;
#else   
   CEI_INT16 localChannel = *channel;
#endif
   CEI_INT16 status;
   AR_TIMETAG_TYPE timeTagStruct;
   status = ar_get_data_xt (board, &localChannel, data, &timeTagStruct);
   if (status == ARS_GOTDATA) {
#if defined (NO_INT64)
      *timeTagLo = (CEI_UINT32)(timeTagStruct.timeTag);
      *timeTagHi = (CEI_UINT32)(timeTagStruct.referenceTimeTag);
#else   
      *timeTagLo = (CEI_UINT32)(timeTagStruct.timeTag & 0xFFFFFFFFL);
      *timeTagHi = (CEI_UINT32)(timeTagStruct.timeTag >> 32);
#endif
#ifndef LABVIEW_RT
      *channel = localChannel;
#endif
   }
   return status;
}


/*===========================================================================*
 * ENTRY POINT:                A R _ G E T _ D A T A _ X T
 *===========================================================================*
 *
 * FUNCTION:    Read the next data and 64-bit time-tag from a receive FIFO,
 *              returns the extended time-tag structure supporting IRIG.
 *
 * PARAMETERS:  CEI_INT16  board   -- (input)  board to reference (0 to 15)
 *              pCEI_INT16 channel -- (input)  recv channel to reference (0 to 255)
 *                                    (output) in merged mode, the channel this
 *                                             data was actually received on
 *              pCEI_UINT32 data   -- (output) location to store received data
 *
 *              pAR_TIMETAG_TYPE timeTagRef
 *                                 -- (output) location for time-tag struct
 *
 * RETURN VAL   CEI_INT16    -- ARS_GOTDATA     Success
 *                              ARS_INVHARVAL   Unsupported channel parameter
 *                              ARS_INVARG      NULL data parameter value
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_NODATA      No data availalble
 *                              ARS_INVBOARD    Invalid board parameter
 *                                                 or board not initialized
 *                              ARS_BAD_MESSAGE An invalid length msg was received
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine reads the most recently received entry (data and
 *              timetag) from the specified receive channel FIFO.  The 
 *              channel number passed to this routine corresponds to the
 *              hardware channel assignment on the device.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_data_xt
(
   CEI_INT16   board,           /* (input)  board to reference (0 to 15)     */
   pCEI_INT16  channel,         /* (input)  recv channel to ref (0 to 255)   */
   pCEI_UINT32 data,            /* (output) location to store received data  */
   pAR_TIMETAG_TYPE timeTagRef  /* (output) location to store data time-tags */
)
{
   CEI_UINT32 localChannel;             /* merged/individual channel select  */
   CEI_UINT32 channelStatus;            /* Rx channel FIFO data avail status */
   CEI_INT16 mergedModeState = AR_OFF;  /* 429 Rx ch merged/individual state */
   CEI_INT16 recvStatus = ARS_NORMAL;   /* FIFO receive status               */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_UINT32 bitFieldValue;        /* temp for device register read content */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   CEI_UINT32 channelToRead = 0;        /* merged/individual channel select  */
   TIME_TAG_TYPE requestedTimeTagFormat; /* Format of the time-tag returned  */
   
   localChannel = (CEI_INT32)(*channel);

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if (data == NULL) {
         status = ARS_INVARG;
      }
      else {
#endif
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
            /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
            if (*channel == (CEI_INT16)deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX]) {
               channelToRead = *channel;
               mergedModeState = AR_ON;
            }
            else {
               boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                            CEIDEV_CHANNEL_CONFIG_1 + (*channel * CEIDEV_CHANNEL_REGION_SIZE));
      
               /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
               accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
               if (accessStatus == ARS_NORMAL) {
                  mergedModeState = (CEI_INT16)((bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) ? AR_ON : AR_OFF);
                  if (mergedModeState == AR_ON) {
                     localChannel = (CEI_INT32)deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
                     channelToRead = localChannel;
                  }
                  else {
                     localChannel = (CEI_INT32)*channel;
                     channelToRead = *channel;
                  }
               }
            }
      
#ifndef TARGET_EMBEDDED
            status = a429_utl_validate_channel_request (board, localChannel, ARU_GET_DATA);
      
            if (status == ARS_NORMAL) {
#endif
               
               /* check the receive channel status, if it indicates "NOT EMPTY"
                * then retrieve the data from the queue
                */
               boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                            CEIDEV_CHANNEL_STATUS + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE));
      
               accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
               /* check the receive channel status, if it indicates "NOT EMPTY", then retrieve the data
                * from the queue.  If it indicates "BAD MESSAGE RECEIVED", return status indicating
                * as such unless there was an invalid access to the board encountered on top of that.
                */
               if (channelStatus & CEIDEV_CSR_FIFO_INVALID_MSG) {
                  recvStatus = ARS_BAD_MESSAGE;
                  accessStatus = cei_write_32bit (board, boardPointer, CEIDEV_CSR_FIFO_INVALID_MSG);
               }
               if ( (channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL) ) {
                  /* if merged mode is enabled for this channel or the merged channel was specified,
                   * read the actual channel on which the message was received from the Rx buffer */
                  if (mergedModeState == AR_ON) {
                     boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                                  CEIDEV_CHANNEL_BUFFER_2 + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE));
                     accessStatus = cei_read_32bit (board, boardPointer, (pCEI_UINT32)&localChannel);
                     mergedModeState = (CEI_INT16)localChannel;
                  }
                  else
                     mergedModeState = CEIDEV_MAX_CHANNELS;
      
                  /* if the reference is valid, read the time-stamp */
                  if (timeTagRef != NULL) {
                     boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                                  CEIDEV_CHANNEL_BUFFER_0 + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE));
                     /* for any time-tag format less than 1 microsecond resolution, use the supplied format */
                     requestedTimeTagFormat = ((timeTagRef->timeTagFormat >= AR_TIMETAG_EXT_IRIG_64BIT) && (timeTagRef->timeTagFormat < AR_TIMETAG_EXT_IRIG_REF_64BIT)) ? timeTagRef->timeTagFormat : AR_USE_PROGRAMMED_TIMETAG_MODE;
         
                     accessStatus = a429_utl_acquire_formatted_time_tag ( board,
                                                                          boardPointer,
                                                                          AR_RECEIVER_FIFO_TIMETAG,
                                                                          (CEI_INT32)requestedTimeTagFormat, 
                                                                          mergedModeState, 
                                                                          timeTagRef );
                  }
                  
                  /* read the actual message last, triggering the fifo head ptr increment */
                  if (accessStatus == ARS_NORMAL) {
                     *channel = (CEI_INT16)localChannel;
                     boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                                  CEIDEV_CHANNEL_BUFFER_3 + (channelToRead * CEIDEV_CHANNEL_REGION_SIZE));
                     accessStatus = cei_read_32bit (board, boardPointer, data);
                     if (accessStatus == ARS_NORMAL)
                        status = ARS_GOTDATA;
                  }
               }
               else { /* no data available, exit with status = ARS_NODATA */
                  status = ARS_NODATA;
               }
            }
            if (device[board].multiThreadProtect) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
#ifndef TARGET_EMBEDDED
      }
      
      if ( ((status == ARS_NORMAL) || (status == ARS_NODATA)) && (accessStatus != ARS_NORMAL))
         status = (CEI_INT16)accessStatus;
#endif
      if ( (recvStatus == ARS_BAD_MESSAGE) && (accessStatus == ARS_NORMAL) ) {
         status = ARS_BAD_MESSAGE;
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (status);
}


/*===========================================================================*
 * ENTRY POINT:               A R _ G E T B L O C K
 *===========================================================================*
 *
 * FUNCTION:    Retrieve multiple messages from a specified receive buffer.
 *
 * PARAMETERS:  CEI_UINT32 board       -- (input)  board to reference (0 to 15)
 *              CEI_UINT32 channel     -- (input)  recvr channel to use (0 to 31)
 *              CEI_INT32  maxMessages -- (input)  number of messages to retrieve
 *              CEI_INT32  offset      -- (input)  unused parameter (legacy support)
 *              pCEI_INT32 actualCount -- (output) number of messages retrieved
 *              pCEI_INT32 data        -- (output) location to store data
 *              pCEI_INT32 timeTags    -- (output) location to store timetags
 *
 * RETURN VAL:  ARS_GOTDATA       one or more ARINC messages were retrieved
 *              ARS_NODATA        no ARINC messages were available in the buffer
 *              ARS_INVBOARD      invalid board number
 *              ARS_INVHARVAL     Item was invalid or channel
 *              ARS_INVARG        invalid parameter value
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *              ARS_BAD_MESSAGE   An invalid length msg was received
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine retrieves all of the available ARINC messages 
 *              from the requested receive channel buffer and copies them to 
 *              the desired destination.
 *
 *              If the TimeTag pointer is not NULL, the time tags associated
 *              with each retrieved message are also copied; however, only
 *              32-bit time-tagging options are supported.
 *
 *              If the buffer is empty this routine returns ARS_NODATA.
 *
 *              If the return value for actualCount is equal to maxMessages,
 *              additional unread data may be available in the receive buffer. 
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_getblock
(
   CEI_UINT32 board,           /* (input)  board to reference (0 to 15)      */
   CEI_UINT32 channel,         /* (input)  recvr channel to use (0 to 31)    */
   CEI_INT32  maxMessages,     /* (input)  number of messages to retrieve    */
   CEI_INT32  offset,          /* (input)  unused parameter (legacy support) */
   pCEI_INT32 actualCount,     /* (output) number of messages retrieved      */
   pCEI_INT32 data,            /* (output) location to store data            */
   pCEI_INT32 timeTags         /* (output) location to store timetags        */
)
{
   CEI_UINT32 channelStatus = 0;        /* Rx channel FIFO data avail status */
   CEI_UINT32 localChannel = 0;         /* merged/individual channel select  */
   pCEI_INT32 dataPtr;                  /* local pointer to store Rx data    */
   pCEI_INT32 timeTagPtr;               /* local pointer to store Rx timetag */
   CEI_INT32 messageCount;              /* receiver buffer message counter   */
   CEI_INT16 mergedModeState =  AR_OFF; /* 429 Rx ch merged/individual state */
   CEI_INT32 mergedModeChan = 0;        /* 429 Rx ch merged/individual state */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT16 recvStatus = ARS_NORMAL;   /* FIFO receive status               */
   CEI_INT32 retrieveTimeTag = FALSE;   /* indicates to pass time-tags back  */
   CEI_INT32 accessStatus = ARS_NORMAL;/* status of device read/write access */
   CEI_INT32 timeTagReference;         /* currently selected time-tag format */
   AR_TIMETAG_TYPE timeTagRef;         /* local time-tag reference structure */
   CEI_UINT32 bitFieldValue = 0;    /* temp for device register read content */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 chanPointer;         /* local pointer to channel memory area  */
   pCEI_UINT32 statusPointer;       /* local pointer to channel memory area  */
   pCEI_UINT32 dataPointer;         /* local pointer to channel memory area  */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((data == NULL) || (actualCount == NULL) || (maxMessages < 0)){
      status = ARS_INVARG;
   }
   else {
#endif
      timeTagReference = (CEI_INT32)deviceGlobal[board]->deviceTimeTagReference;
      timeTagPtr = timeTags;       /* reference supplied pointers locally */
      if (channel >= deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
         if (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] != CEIDEV_MAX_CHANNELS) {
            localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
         }
         else {
            status = ARS_INVHARVAL;
         }
      }
      else {
         localChannel = channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      }
      
      if (status == ARS_NORMAL) {
         /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      CEIDEV_CHANNEL_CONFIG_1 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
     
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_read_32bit ((CEI_INT16)board, boardPointer, &bitFieldValue);
            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect)
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      
         if (accessStatus == ARS_NORMAL) {
            mergedModeState = (CEI_INT16)((bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) ? AR_ON : AR_OFF);
            if (mergedModeState == AR_ON)
               localChannel = deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
            else
               mergedModeChan = CEIDEV_MAX_CHANNELS;
         }
         
#ifndef TARGET_EMBEDDED
         status = a429_utl_validate_channel_request ((CEI_INT16)board, localChannel, ARU_GET_DATA);
#endif
      }

      if (status == ARS_NORMAL) {
         /* check the receive channel status, if it indicates "NOT EMPTY"
          * then retrieve the data from the queue
          */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));

         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_read_32bit ((CEI_INT16)board, boardPointer, &channelStatus);
         }
         /* check the receive channel status, if it indicates "NOT EMPTY", then retrieve the data
          * from the queue.  If it indicates "BAD MESSAGE RECEIVED", return status indicating
          * as such unless there was an invalid access to the board encountered on top of that.
          */
         if (channelStatus & CEIDEV_CSR_FIFO_INVALID_MSG) {
            recvStatus = ARS_BAD_MESSAGE;
            accessStatus = cei_write_32bit ((CEI_INT16)board, boardPointer, CEIDEV_CSR_FIFO_INVALID_MSG);
         }
         if ((channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
            
            /* reference supplied pointers locally */
            dataPtr = data;
            if (timeTags != NULL) {
               retrieveTimeTag = TRUE;
            
               /* if a 32-bit time-tag ref. isn't already selected, force 32-bit as a default */
               if ( (timeTagReference != AR_TIMER_X20_COMPAT_32BIT) &&
                    (timeTagReference != AR_TIMETAG_INT_20USEC_32BIT) && 
                    (timeTagReference != AR_TIMETAG_INT_MSEC_32BIT) ) {
                  timeTagReference = AR_DEFAULT_RECEIVE_DATA_TIMETAG_FORMAT;
               }
            }

            /* setup all pointers that might be used in the message/time-stamp retreival loop */
            boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                         CEIDEV_CHANNEL_BUFFER_0 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
            chanPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                        CEIDEV_CHANNEL_BUFFER_2 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
            dataPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                        CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
            statusPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                          CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));

            /* attempt to retrieve as much data as requested */
            messageCount = 0;
            while ((messageCount < maxMessages) && (accessStatus == ARS_NORMAL)) {
               if (retrieveTimeTag) {
                  /* if merged mode is enabled for this channel read the actual 
                   * channel on which the message was received from the Rx buffer */
                  if (mergedModeState == AR_ON) {
                     accessStatus = cei_read_32bit ((CEI_INT16)board, chanPointer, (pCEI_UINT32)&mergedModeChan);
                  }

                  accessStatus = a429_utl_acquire_formatted_time_tag ((CEI_INT16)board, 
                                                                      boardPointer,
                                                                      AR_RECEIVER_FIFO_TIMETAG, 
                                                                      timeTagReference, 
                                                                      (CEI_INT32)mergedModeChan, 
                                                                      &timeTagRef );
                  if (accessStatus == ARS_NORMAL) {
                     /* return the modified 32-bit 20ms resolution time-tag */
                     *timeTagPtr = (CEI_UINT32)(timeTagRef.timeTag & 0xFFFFFFFF);
                     timeTagPtr++;
                  }
               }
      
               /* return the 32-bit data word */
               accessStatus = cei_read_32bit ((CEI_INT16)board, dataPointer, (pCEI_UINT32)dataPtr);
               dataPtr++;
               messageCount++;
               
               accessStatus = cei_read_32bit ((CEI_INT16)board, statusPointer, &channelStatus);
               if ( (accessStatus == ARS_NORMAL) && !(channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) )
                  break;
            }
            
            if (messageCount) {
               status = ARS_GOTDATA;
               *actualCount = messageCount;
            }
            else
               *actualCount = 0;
         }
         else { /* no data available, exit with status = ARS_NODATA */
            status = ARS_NODATA;
         }
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
      if ( ((status == ARS_NORMAL) || (status == ARS_NODATA)) && (accessStatus != ARS_NORMAL))
         status = (CEI_INT16)accessStatus;
#endif
      if ( (recvStatus == ARS_BAD_MESSAGE) && (accessStatus == ARS_NORMAL) ) {
         status = ARS_BAD_MESSAGE;
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:               A R _ G E T B L O C K _ T
 *===========================================================================*
 *
 * FUNCTION:    Retrieve multiple messages from a specified receive buffer,
 *              supporting 64-bit time-tag retrieval.
 *
 * PARAMETERS:  CEI_UINT32  board       -- (input)  board to reference (0 to 15)
 *              CEI_UINT32  channel     -- (input)  recvr channel to use (0 to 31)
 *              CEI_INT32   maxMessages -- (input)  number of messages to retrieve
 *              pCEI_INT32  actualCount -- (output) number of messages retrieved
 *              pCEI_UINT32 msgChan     -- (output) location to store receive channel
 *              pCEI_INT32  data        -- (output) location to store data
 *              pCEI_INT32  timeTagMsw  -- (output) location to store time-tag MSW
 *              pCEI_INT32  timeTagLsw  -- (output) location to store time-tag LSW
 *
 * RETURN VAL:  ARS_GOTDATA       one or more ARINC messages were retrieved
 *              ARS_NODATA        no ARINC messages were available in the buffer
 *              ARS_INVBOARD      invalid board number
 *              ARS_INVHARVAL     Unsupported channel parameter
 *              ARS_INVARG        invalid parameter value
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *              ARS_BAD_MESSAGE   An invalid length msg was received
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine retrieves all of the available ARINC messages 
 *              from the requested receive channel buffer and copies them to 
 *              the desired destination.
 *
 *              If the TimeTag pointers are not NULL, the 64-bit time-tag associated
 *              with each retrieved message are also copied, in separate 32-bit
 *              MSW and LSW arrays.
 *
 *              If the buffer is empty this routine returns ARS_NODATA.
 *
 *              If the return value for actualCount is equal to maxMessages,
 *              additional unread data may be available in the receive buffer. 
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_getblock_t
(
   CEI_UINT32  board,          /* (input)  board to reference (0 to 15)      */
   CEI_UINT32  channel,        /* (input)  recvr channel to use (0 to 31)    */
   CEI_INT32   maxMessages,    /* (input)  number of messages to retrieve    */
   pCEI_INT32  actualCount,    /* (output) number of messages retrieved      */
   pCEI_UINT32 msgChan,        /* (output) location to store rx chan (mrg md)*/
   pCEI_INT32  data,           /* (output) location to store data            */
   pCEI_INT32  timeTagMsw,     /* (output) location to store time-tag MSW    */
   pCEI_INT32  timeTagLsw      /* (output) location to store time-tag LSW    */
)
{
   CEI_UINT32 channelStatus = 0;        /* Rx channel FIFO data avail status */
   CEI_UINT32 localChannel = 0;         /* merged/individual channel select  */
   pCEI_UINT32 rxChanPtr = NULL;        /* local pointer to store Rx channel */
   pCEI_INT32 dataPtr;                  /* local pointer to store Rx data    */
   pCEI_INT32 timeTagPtr1 = NULL;       /* local pointer to store Rx timetag */
   pCEI_INT32 timeTagPtr2 = NULL;       /* local pointer to store Rx timetag */
   CEI_INT32 messageCount;              /* receiver buffer message counter   */
   CEI_INT16 mergedModeState = AR_OFF;  /* 429 Rx ch merged/individual state */
   CEI_UINT32 mergedModeChan = 0;        /* 429 Rx ch merged/individual state */
   AR_TIMETAG_TYPE timeTagRef;          /* local time-tag reference structure*/
   CEI_INT16 recvStatus = ARS_NORMAL;   /* FIFO receive status               */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 retrieveTimeTag = FALSE;   /* indicates to pass time-tags back  */
   CEI_INT32 retrieveChannel = FALSE;   /* indicates to pass channel back    */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_INT32 timeTagReference = 0;      /* local current time-tag reference  */
   CEI_UINT32 bitFieldValue = 0;    /* temp for device register read content */
   pCEI_UINT32 boardPointer;        /* local pointer to channel memory area  */
   pCEI_UINT32 chanPointer = 0;     /* local pointer to channel memory area  */
   pCEI_UINT32 statusPointer;       /* local pointer to channel memory area  */
   pCEI_UINT32 dataPointer;         /* local pointer to channel memory area  */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((data == NULL) || (actualCount == NULL) || (maxMessages < 0)) {
      status = ARS_INVARG;
   }
   else if (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) {
      /* legacy ar_getblock_t channel is 0-referenced ARINC 429 receive channel index;
       * ar_get_data_xt uses the board-referenced channel number 0-255, where
       * ARINC 429 receive channels can be located anywhere in that range.
       */
#endif
      /* if External IRIG time-tagging is selected, convert IRIG once per block */
      timeTagReference = (CEI_INT32)deviceGlobal[board]->deviceTimeTagReference;
      if (timeTagReference == AR_TIMETAG_EXT_IRIG_64BIT)
         timeTagReference = AR_TIMETAG_REF_IRIG_64BIT;
      
      localChannel = channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
      /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                   CEIDEV_CHANNEL_CONFIG_1 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
   
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
      if (accessStatus == ARS_NORMAL) {
         accessStatus = cei_read_32bit ((CEI_INT16)board, boardPointer, &bitFieldValue);
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect)
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         mergedModeState = (CEI_INT16)((bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) ? AR_ON : AR_OFF);
         if (mergedModeState == AR_ON)
            localChannel = (CEI_INT32)deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
         else
            mergedModeChan = CEIDEV_MAX_CHANNELS;
      }
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request ((CEI_INT16)board, localChannel, ARU_GET_DATA);
   }
   else {
      status = ARS_INVHARVAL;
   }

   if (status == ARS_NORMAL) {
#endif
      /* check the receive channel status, if it indicates "NOT EMPTY"
       * then retrieve the data from the queue
       */
      statusPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                    CEIDEV_CHANNEL_STATUS + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));

      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         accessStatus = cei_read_32bit ((CEI_INT16)board, statusPointer, &channelStatus);
      }
      /* check the receive channel status, if it indicates "NOT EMPTY", then retrieve the data
       * from the queue.  If it indicates "BAD MESSAGE RECEIVED", return status indicating
       * as such unless there was an invalid access to the board encountered on top of that.
       */
      if (channelStatus & CEIDEV_CSR_FIFO_INVALID_MSG) {
         recvStatus = ARS_BAD_MESSAGE;
         accessStatus = cei_write_32bit ((CEI_INT16)board, statusPointer, CEIDEV_CSR_FIFO_INVALID_MSG);
      }
      if ((channelStatus & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {

         /* To avoid having to poll the channel status register for each message read, read
          * how many unread entries exist in the respective buffer from the channel status register.
          */
         messageCount = (CEI_INT32)((channelStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT);
         if (maxMessages > messageCount) {
            maxMessages = messageCount;
         }

         /* reference supplied pointers locally */
         dataPtr = data;
         if ( (timeTagLsw != NULL) && (timeTagMsw != NULL) ) {
            retrieveTimeTag = TRUE;
            timeTagPtr1 = timeTagLsw;
            timeTagPtr2 = timeTagMsw;
      
         }
         if (msgChan != NULL) {
            retrieveChannel = TRUE;
            rxChanPtr = msgChan;
         }

         /* If the reference is valid and the active Time-Tag Mode is IRIG, 
          * read the time-stamp in a two-step process. First, read the current
          * IRIG sample time value, then extrapolate the individual received 
          * message time-stamps to IRIG-referenced time as they are read.
          * This saves time by not having to read and convert the IRIG sample
          * for each received message time-stamp extrapolation.
          */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      CEIDEV_CHANNEL_BUFFER_0 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));

         if ((retrieveTimeTag) && (timeTagReference == AR_TIMETAG_REF_IRIG_64BIT)) {
            accessStatus = a429_utl_acquire_formatted_time_tag ((CEI_INT16)board, 
                                                                 boardPointer,
                                                                 AR_BYPASS_TIMER_READ, 
                                                                 AR_LOG_EXT_IRIG_SAMPLE, 
                                                                 (CEI_INT32)mergedModeChan, 
                                                                 &timeTagRef );
         }
         
         /* if merged mode is enabled for this channel, setup a pointer to read the actual 
          * channel on which the message was received from the Rx buffer */
         if (mergedModeState == AR_ON) {
            chanPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                        CEIDEV_CHANNEL_BUFFER_2 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
         }
         /* setup a pointer to read the ARINC 429 message from the FIFO */
         dataPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      CEIDEV_CHANNEL_BUFFER_3 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));

         /* attempt to retrieve as much data as requested */
         messageCount = 0;
         while ((messageCount < maxMessages) && (accessStatus == ARS_NORMAL)) {
            /* if merged mode is enabled for this channel read the actual 
             * channel on which the message was received from the Rx buffer */
            if (mergedModeState == AR_ON) {
               accessStatus = cei_read_32bit ((CEI_INT16)board, chanPointer, (pCEI_UINT32)&mergedModeChan);
            }

            if (retrieveTimeTag) {
               accessStatus = a429_utl_acquire_formatted_time_tag ((CEI_INT16)board, 
                                                                   boardPointer,
                                                                   AR_RECEIVER_FIFO_TIMETAG,
                                                                   timeTagReference, 
                                                                   (CEI_INT32)mergedModeChan, 
                                                                   &timeTagRef );
               if (accessStatus == ARS_NORMAL) {
                     /* return the modified 32-bit 20ms resolution time-tag */
                     *timeTagPtr1 = (CEI_UINT32)(timeTagRef.timeTag & 0xFFFFFFFF);
#if defined (NO_INT64)
                     *timeTagPtr2 = (CEI_UINT32)(timeTagRef.referenceTimeTag & 0xFFFFFFFF);
#else
                     *timeTagPtr2 = (CEI_UINT32)((timeTagRef.timeTag >> 32) & 0xFFFFFFFF);
#endif
                     timeTagPtr1++;
                     timeTagPtr2++;
               }
            }
            if (retrieveChannel) {
               /* return the channel identifier from the FIFO */
               *rxChanPtr = (mergedModeState == AR_ON) ? mergedModeChan: channel;  
               rxChanPtr++;
            }

            /* return the 32-bit data word */
            accessStatus = cei_read_32bit ((CEI_INT16)board, dataPointer, (pCEI_UINT32)dataPtr);
            dataPtr++;
            messageCount++;
         }
         
         if (messageCount) {
            status = ARS_GOTDATA;
            *actualCount = messageCount;
         }
         else
            *actualCount = 0;
      }
      else { /* no data available, exit with status = ARS_NODATA */
         status = ARS_NODATA;
      }
      /* if multi-thread protection requested, free semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_release_thread_lock(LOCK_IDENT);
      }
#ifndef TARGET_EMBEDDED
   }
   if ( ((status == ARS_NORMAL) || (status == ARS_NODATA)) && (accessStatus != ARS_NORMAL))
      status = (CEI_INT16)accessStatus;
#endif
   if ( (recvStatus == ARS_BAD_MESSAGE) && (accessStatus == ARS_NORMAL) ) {
      status = ARS_BAD_MESSAGE;
   }
   
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:                 A R _ G E T _ L A T E S T
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the latest message received for the specified 
 *              channel/label combination from the snapshot buffer
 * 
 * PARAMETERS:  CEI_INT16 board    -- (input)  Board number (0 to 15)  
 *              CEI_INT16 channel  -- (input)  429 Rx channel selection,
 *                                             valid range is 0 to one less
 *                                             than the number of enhanced
 *                                             ARINC 429 channels installed
 *              CEI_UINT16 label   -- (input)  Label of interest
 *              pCEI_VOID data     -- (output) pointer to the host data storage
 *              pCEI_CHAR seq_num  -- (n/a)    legacy parameter, not supported
 * 
 * RETURN VAL:  NONE
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine copies the latest ARINC 492/561 message received 
 *              for the specified channel/label comb. from the snapshot buffer.
 *              When using this routine, the host application should set the
 *              snapshot storage mode to "label field only".  This will setup
 *              the device to store snapshot data based on the label field value
 *              only, ignoring the SDI bit field value.
 *
 *              The snapshot buffer access by this routine for a single label/channel
 *              combination is defined with the bits of the ARINC word as follows:
 *
 *                     field    --------label---------
 *                      bits    7  6  5  4  3  2  1  0
 *
 *              If the label requested is label 256 (an ARINC label that does 
 *              not exist) or the value ARU_ALL_LABELS, this function returns
 *              all 255 ARINC labels for the specified channel (000-255), in sequence, 
 *              in successive data array elements.  This function assumes the caller
 *              has allocated at least 256*4 bytes for data when used in this mode.
 *                    
 *===========================================================================*/ 
EXPORT32 CEI_VOID  DLL_EXPORTED ar_get_latest 
(
   CEI_INT16 board,             /* (input)  Board number (0 to 15)           */
   CEI_INT16 channel,           /* (input)  429 Rx channel selection         */
   CEI_UINT16 label,            /* (input)  Label of interest                */
   pCEI_VOID data,              /* (output) pointer to the host data storage */
   pCEI_CHAR seq_num            /* (n/a)    legacy parameter, not supported  */
)
{
   CEI_UINT32 snapshotBufferIndex = 0;      /* snapshot buffer offset label/sdi  */
   CEI_UINT32 snapStorageMode = 0;      /* local snapshot storage selection  */
   CEI_UINT32 labelIndex;               /* local temp enh filter table index */
   CEI_INT32 localChannel;              /* merged/individual channel select  */
   pCEI_UINT32 boardPointer;            /* local pointer to chan memory area */
   pCEI_UINT32 localDataPtr;            /* temp pointer to application data  */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* Snapshot Buffer access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel > deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] ) || (data == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_ACCESS_SNAPSHOT_BUFFER);
      if (status == ARS_NORMAL) {
#endif
         if (seq_num != NULL) {
            *seq_num = 0;
         }
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
            /* read the current snapshot storage mode selection: label only or label/SDI combined */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &snapStorageMode);
         }
         /* this routine must only be invoked when the snapshot storage mode selection is label only */
         if ((accessStatus == ARS_NORMAL) && (snapStorageMode & CEIDEV_SNAPSHOT_STORAGE_ENHANCED)) {
            if ((accessStatus == ARS_NORMAL) && (label < 256)) {
               /* access the specific location in the snapshot buffer for the specified channel and label */
               snapshotBufferIndex = (CEI_UINT32)(localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT) + 
                                     (CEI_UINT32)(label << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT) + CEIDEV_SNAPSHOT_BUFFER_DATA_OFF;
               boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_SNAPSHOT_BUFFER_START + snapshotBufferIndex);
               accessStatus = cei_read_32bit (board, boardPointer, (pCEI_UINT32)data);
            }
            else if ((accessStatus == ARS_NORMAL) && ((label == 256) || (label == ARU_ALL_LABELS)) ) {
               localDataPtr = (pCEI_UINT32)data;
               snapshotBufferIndex = (CEI_UINT32)(localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT) + CEIDEV_SNAPSHOT_BUFFER_DATA_OFF;
               boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_SNAPSHOT_BUFFER_START + snapshotBufferIndex);
               for (labelIndex = 0; labelIndex < 256; labelIndex++) {
                  accessStatus = cei_read_32bit (board, boardPointer, localDataPtr);
                  boardPointer += 4;
                  localDataPtr++;
               }
            }
         }
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
            status = (CEI_INT16)accessStatus;
#ifndef TARGET_EMBEDDED
      }
   }
#endif
}


/*===========================================================================*
 * ENTRY POINT:         A R _ G E T _ L A T E S T _ T
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the latest message & time-stamp for the specified 
 *              channel/label combination from the snapshot buffer
 * 
 * PARAMETERS:  CEI_INT16 board   -- (input)  Board number (0 to 15)  
 *              CEI_INT16 channel -- (input)  429 Rx channel selection,
 *                                               valid range is 0 to one less
 *                                               than the number of enhanced
 *                                               ARINC 429 channels installed
 *              CEI_UINT16 label  -- (input)  Label of interest
 *              pCEI_UINT32 data  -- (output) pointer to the host data storage
 *              pCEI_INT32 seq    -- (output) ptr to the sequence number storage
 *              TIME_TAG_TYPE timeTag -- (output) pointer to time-stamp storage
 * 
 * 
 * RETURN VAL:  ARS_NORMAL        Operation completed successfully
 *              ARS_INVBOARD      Invalid board number or board not initialized
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid label argument, null data ptr
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine copies the latest ARINC 492/561 message received 
 *              for the specified channel/label combination from the snapshot buffer.
 *              When using this routine, the host application should set the
 *              snapshot storage mode to "label/sdi storage".  This will setup
 *              the device to store snapshot data based on the label field value
 *              in combination with the SDI bit field value.
 *
 *              The snapshot buffer for a single channel is defined as follows:
 *
 *                 CEI_UCHAR snapshot[MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *                     field   -SDI-  --------label---------
 *                      bits    9 8   7  6  5  4  3  2  1  0
 *                    
 *===========================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_latest_t 
(
   CEI_INT16 board,            /* (input)  Board number (0 to 15)            */
   CEI_INT16 channel,          /* (input)  429 Rx channel selection          */
   CEI_UINT16 label,           /* (input)  Label of interest                 */
   pCEI_UINT32 data,           /* (output) pointer to the host data storage  */
   TIME_TAG_TYPE * timeTag     /* (output) pointer to time-stamp storage     */
)
{
   CEI_UINT32 snapStorageMode;          /* local snapshot storage selection  */
   CEI_UINT32 localChannel;             /* merged/individual channel select  */
   pCEI_UINT32 boardPointer;            /* local pointer to chan memory area */
   CEI_INT32 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   AR_TIMETAG_TYPE timeTagRef;          /* structure for the 64-bit time-tag */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* Snapshot Buffer access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel > deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) || (data == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_ACCESS_SNAPSHOT_BUFFER);
      if (status == ARS_NORMAL) {
#endif
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
            /* read the current snapshot storage mode selection: label only or label/SDI combined */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &snapStorageMode);
   
            /* this routine must only be invoked when the snapshot storage mode selection is label only */
            if ((snapStorageMode & CEIDEV_SNAPSHOT_STORAGE_ENHANCED) == 0) {
               status = ARS_INVARG;
            }
            else {

               boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_SNAPSHOT_BUFFER_START + 
                                            (CEI_UINT32)(localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT) + 
                                            (CEI_UINT32)(label << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT));
               if (timeTag != NULL) {
                  accessStatus = a429_utl_acquire_formatted_time_tag ( board,
                                                                       boardPointer,
                                                                       AR_RECEIVER_FIFO_TIMETAG,
                                                                       AR_USE_PROGRAMMED_TIMETAG_MODE, 
                                                                       CEIDEV_MAX_CHANNELS, 
                                                                       &timeTagRef );
                  *timeTag = timeTagRef.timeTag;
               }
               boardPointer += 3;
               accessStatus = cei_read_32bit (board, boardPointer, (pCEI_UINT32)data);
            }
            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
         
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
            status = accessStatus;
#ifndef TARGET_EMBEDDED
      }
   }
#endif
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:                 A R _ G E T _ S N A P _ D A T A
 *===========================================================================*
 *
 * FUNCTION:    Retrieve the latest message received for the specified 
 *              channel/label combination from the snapshot buffer
 * 
 * PARAMETERS:  CEI_INT16 board   -- (input)  Board number (0 to 15)  
 *              CEI_INT16 channel -- (input)  429 Rx channel selection,
 *                                               valid range is 0 to one less
 *                                               than the number of enhanced
 *                                               ARINC 429 channels installed
 *              CEI_UINT16 label  -- (input)  Label of interest
 *              CEI_UINT16 sdi    -- (input)  SDI Field Value of interest
 *              pCEI_INT32 data   -- (output) pointer to the host data storage
 * 
 * RETURN VAL:  ARS_NORMAL        Operation completed successfully
 *              ARS_INVBOARD      Invalid board number or board not initialized
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid label or SDI argument, null data ptr
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine copies the latest ARINC 492/561 message received 
 *              for the specified channel/label combination from the snapshot buffer.
 *              When using this routine, the host application should set the
 *              snapshot storage mode to "label/sdi storage".  This will setup
 *              the device to store snapshot data based on the label field value
 *              in combination with the SDI bit field value.
 *
 *              The snapshot buffer for a single channel is defined as follows:
 *
 *                 CEI_UCHAR snapshot[MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *                     field   -SDI-  --------label---------
 *                      bits    9 8   7  6  5  4  3  2  1  0
 *                    
 *===========================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_snap_data 
(
   CEI_INT16 board,            /* (input)  Board number (0 to 15)            */
   CEI_INT16 channel,          /* (input)  429 Rx channel selection          */
   CEI_UINT16 label,           /* (input)  Label of interest                 */
   CEI_UINT16 sdi,             /* (input)  SDI Field Value of interest       */
   pCEI_INT32 data             /* (output) pointer to the host data storage  */
)
{
   CEI_UINT32 snapshotBufferIndex;      /* snapshot buffer offset label/sdi  */
   CEI_UINT32 snapStorageMode;          /* local snapshot storage selection  */
   CEI_UINT32 localChannel;             /* merged/individual channel select  */
   pCEI_UINT32 boardPointer;            /* local pointer to chan memory area */
   CEI_INT32 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   /* Snapshot Buffer access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel > deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) || (data == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_ACCESS_SNAPSHOT_BUFFER);
      if (status == ARS_NORMAL) {
#endif
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
            /* read the current snapshot storage mode selection: label only or label/SDI combined */
            accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &snapStorageMode);
   
            if (snapStorageMode & CEIDEV_SNAPSHOT_STORAGE_ENHANCED) 
               snapshotBufferIndex = (label << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT) + CEIDEV_SNAPSHOT_BUFFER_DATA_OFF;
            else
               snapshotBufferIndex = (label << CEIDEV_SNAPSHOT_BUFFER_LB_SHIFT) + sdi;

            snapshotBufferIndex += localChannel << CEIDEV_SNAPSHOT_BUFFER_CH_SHIFT;
            boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_SNAPSHOT_BUFFER_START + snapshotBufferIndex);
            accessStatus = cei_read_32bit (board, boardPointer, (pCEI_UINT32)data);
            
            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
         
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
            status = accessStatus;
#ifndef TARGET_EMBEDDED
      }
   }
#endif
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:             A R _ G E T _ 5 7 3 _ C O N F I G
 *===========================================================================*
 *
 * FUNCTION:    Retrieve an ARINC 573 channel configuration register bit field def.
 *
 * PARAMETERS:  CEI_INT16 board   -- (input) board to reference (0 to 15)
 *              CEI_INT16 item    -- (input) item to process, valid items are:
 *                                          ARU_RECV_MODE           
 *                                          ARU_RX_BITRATE          
 *                                          ARU_TX_BITRATE          
 *                                          ARU_RX_FIFO_ENABLE
 *                                          ARU_TX_FIFO_ENABLE
 *                                          ARU_573_RX_ENABLE
 *                                          ARU_573_RX_SUBFRAME_SIZE       
 *                                          ARU_573_RX_AUTO_DETECT  
 *                                          ARU_573_RX_BPRZ_SELECT  
 *                                          ARU_573_TX_BPRZ_SELECT
 *                                          ARU_573_TX_HBP_SELECT
 *                                          ARU_573_TX_SLEW_RATE    
 *                                          ARU_573_SYNC_WORD1
 *                                          ARU_573_SYNC_WORD2
 *                                          ARU_573_SYNC_WORD3
 *                                          ARU_573_SYNC_WORD4
 * 
 *              pCEI_INT32 value  -- (output) the state of the item specified
 *                                        with valid item/value combinations:
 *
 *                  item                     valid values
 * 
 *                 ARU_RECV_MODE            AR_WRAP_OFF, AR_WRAP_ON
 *                 ARU_RX_FIFO_ENABLE       AR_OFF, AR_ON
 *                 ARU_573_RX_ENABLE        AR_OFF, AR_ON (same as ARU_RX_FIFO_ENABLE)
 *                 ARU_573_RX_AUTO_DETECT   AR_OFF, AR_ON
 *                 ARU_573_RX_BPRZ_SELECT   AR_OFF, AR_ON
 *                 ARU_573_TX_BPRZ_SELECT   AR_OFF, AR_ON
 *                 ARU_573_TX_HBP_SELECT    AR_OFF, AR_ON
 *                 ARU_RX_BITRATE &         ARU_573_RATE_SIZE_384_32    
 *                 ARU_TX_BITRATE           ARU_573_RATE_SIZE_768_64    
 *                                          ARU_573_RATE_SIZE_1536_128  
 *                                          ARU_573_RATE_SIZE_3072_256  
 *                                          ARU_573_RATE_SIZE_6144_512  
 *                                          ARU_573_RATE_SIZE_12288_1024
 *                                          ARU_573_RATE_SIZE_24576_2048
 *                                          ARU_573_RATE_SIZE_49152_4096
 *                 ARU_573_RX_SUBFRAME_SIZE ARU_573_SUB_FRAME_64  
 *                                          ARU_573_SUB_FRAME_128 
 *                                          ARU_573_SUB_FRAME_256 
 *                                          ARU_573_SUB_FRAME_512 
 *                                          ARU_573_SUB_FRAME_1024
 *                                          ARU_573_SUB_FRAME_2048
 *                                          ARU_573_SUB_FRAME_4096
 *                 ARU_TX_FIFO_ENABLE       AR_OFF, AR_ON
 *                 ARU_TX_DISABLE           AR_OFF, AR_ON (inverse of ARU_TX_FIFO_ENABLE)
 *                 ARU_573_TX_SLEW_RATE     ARU_573_TX_SLEW_1PT5
 *                                          ARU_573_TX_SLEW_10PT0
 *                 ARU_RX_MERGED_MODE       AR_ON, AR_OFF
 *                 ARU_573_SYNC_WORD1       value from 0x0000 to 0x0fff
 *                 ARU_573_SYNC_WORD2       value from 0x0000 to 0x0fff
 *                 ARU_573_SYNC_WORD3       value from 0x0000 to 0x0fff
 *                 ARU_573_SYNC_WORD4       value from 0x0000 to 0x0fff
 *                 ARU_TX_FIFO_COUNT        value from 0 to 8191
 *      
 * RETURN VAL:  CEI_INT16     -- ARS_NORMAL      Normal succesful completion.
 *                               ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                               ARS_INVARG      "item" parameter value is not
 *                                                  supported by this function
 *                               ARS_INVHARVAL   Item was invalid or channel
 *                                                  not supported on this device
 *                               ARS_INVBOARD    Board parameter value invalid or
 *                                                  board has not been initialized
 * 
 * GLOBALS:     deviceGlobal[board]
 *              
 * DESCRIPTION: Supplies the current state of the "item" requested from the device.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_573_config 
( 
   CEI_INT16 board,                 /* (input)  board to reference (0 to 15) */
   CEI_INT16 item,                  /* (input)  item to process              */
   pCEI_INT32 value                 /* (output) state of the item            */
)
{
   CEI_INT32 accessStatus = ARS_NORMAL;/* status of device read/write access */
   CEI_INT32 state = 0;              /* temporary config reg bit field state */
   CEI_UINT32 deviceChannel;         /* temp device-specific channel number  */
   CEI_UINT32 bitFieldValue = 0;     /* temp for device register read content*/
   CEI_UINT32 offsetIntoChannelWindow = 0;  /* offset into chan mem region   */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   CEI_INT16 status = ARS_NORMAL;    /* function return status               */

   /* return state is based on the value of the respective shadow 
    * register's bit field 
    */
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   else {
      deviceChannel = (CEI_INT32)0;  /* here for future growth w/channel parm */ 

      /* offset the first installed CSDB channel by the specified channel no. */
      switch (item) {
         case ARU_RX_BITRATE:
         case ARU_RX_PARITY:
         case ARU_RECV_MODE:
         case ARU_RX_DISABLE:
         case ARU_RX_FIFO_ENABLE:
         case ARU_RX_MERGED_MODE:
         case ARU_573_RX_ENABLE:
         case ARU_573_RX_SUBFRAME_SIZE:
         case ARU_573_RX_AUTO_DETECT:
         case ARU_573_RX_BPRZ_SELECT:
            deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
   
         case ARU_TX_BITRATE:
         case ARU_TX_DISABLE:
         case ARU_TX_FIFO_ENABLE:
         case ARU_573_TX_SLEW_RATE:
         case ARU_573_TX_BPRZ_SELECT:
         case ARU_573_TX_HBP_SELECT:
            deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;

         case ARU_TX_FIFO_COUNT:
            deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_STATUS;
            break;
   
         case ARU_573_SYNC_WORD1:
         case ARU_573_SYNC_WORD2:
         case ARU_573_SYNC_WORD3:
         case ARU_573_SYNC_WORD4:
            *value = (CEI_INT32)a717syncWords[board][item-ARU_573_SYNC_WORD1];
            return status;
   
         default:
            status = ARS_INVARG;
      }

#ifndef TARGET_EMBEDDED
      /* validate the specified channel */
      status = a429_utl_validate_channel_request (board, deviceChannel, item);

      if (status == ARS_NORMAL) {
#endif
         /* read the device register with respect to the attribute specified */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE));
         
         /* if multi-thread protection requested, semaphore/mutex-lock IRIG sample time register access */
         if (device[board].multiThreadProtect) {
            accessStatus = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
         }
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {

            switch (item) {
               case ARU_RECV_MODE:
                  state = (bitFieldValue & CEIDEV_INTERNAL_WRAP_ENABLE) ? AR_WRAP_ON : AR_WRAP_OFF;
                  break;
         
               case ARU_573_RX_ENABLE:
               case ARU_RX_FIFO_ENABLE:
               case ARU_TX_FIFO_ENABLE:
                  state = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_ON : AR_OFF;
                  break;

               case ARU_RX_DISABLE:
                  /* invert the state for CEI-710 legacy ARU_RX_DISABLE option support */
                  state = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_OFF : AR_ON;
                  break;
         
               case ARU_573_RX_AUTO_DETECT:  
                  state = (bitFieldValue & CEIDEV_573_RX_MODE_RAW) ? AR_OFF : AR_ON;
                  break;
         
               case ARU_573_RX_BPRZ_SELECT:  
                  state = (bitFieldValue & CEIDEV_573_RX_BPRZ_SELECT) ? AR_ON : AR_OFF;
                  break;
         
               case ARU_RX_MERGED_MODE:
                  state = (bitFieldValue & CEIDEV_MERGE_MODE_ENABLE) ? AR_ON : AR_OFF;
                  break;
   
               case ARU_RX_BITRATE:
                  state = bitFieldValue & CEIDEV_573_FRAME_49152_4096;
                  state = state >> CEIDEV_573_FRAME_SHIFT;
                  state = a717busSpeedOptions[state];
                  break;
         
               /* following case only provided for CEI-710 backward compatibility */
               case ARU_573_RX_SUBFRAME_SIZE:
                  state = bitFieldValue & CEIDEV_573_FRAME_49152_4096;
                  state = state >> CEIDEV_573_FRAME_SHIFT;
                  state = a717frameSizeOptions[state];
                  break;
         
               case ARU_TX_BITRATE:
                  state = bitFieldValue & CEIDEV_573_FRAME_49152_4096;
                  state = state >> CEIDEV_573_FRAME_SHIFT;
                  state = a717busSpeedOptions[state];
                  break;
         
               case ARU_573_TX_SLEW_RATE:    
                  state = (bitFieldValue & CEIDEV_573_SLEW_1PT5_SELECT) ? ARU_573_TX_SLEW_1PT5 : ARU_573_TX_SLEW_10PT0;
                  break;
         
               case ARU_573_TX_BPRZ_SELECT:
                  state = (bitFieldValue & CEIDEV_573_TX_BPRZ_SELECT) ? AR_ON : AR_OFF;
                  break;
         
               case ARU_573_TX_HBP_SELECT:
                  state = (bitFieldValue & CEIDEV_573_TX_HBP_SELECT) ? AR_ON : AR_OFF;
                  break;
         
               case ARU_TX_DISABLE:
                  /* invert the state for CEI-710 legacy ARU_TX_DISABLE option support */
                  state = (bitFieldValue & CEIDEV_CHANNEL_ENABLE) ? AR_OFF : AR_ON;
                  break;
         
               case ARU_TX_FIFO_COUNT:
                  state = (CEI_INT32)(bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
                  break;
   
               default:
                  status = ARS_INVARG;
            }
#ifndef TARGET_EMBEDDED
         }
         else
            status = (CEI_INT16)accessStatus;
#endif

         *value = (CEI_INT32)state;
      }
   }
   
   return status;
}
 

/*===========================================================================*
 * ENTRY POINT:             A R _ G E T _ 5 7 3 _ F R A M E
 *===========================================================================*
 *
 * FUNCTION:    Receive an ARINC 573 frame from the ARINC 573 FIFO. 
 *
 * PARAMETERS:  CEI_INT16 board         -- (input) Board number of interest (0-8)                          
 *              pCEI_UINT32 numberWords -- (input/output) Number of ARINC 573 
 *                                          words to read / number actually read
 *              pCEI_UINT16 arincData   -- (output) Pointer to the ARINC 573 buffer
 * 
 * RETURN VAL:  CEI_INT16    -- ARS_NODATA      No 573 frame data was available
 *                              ARS_GOTDATA     Some 573 frame data was retrieved
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_INVBOARD    Invalid board parameter or board not initialized
 *                              ARS_INVHARVAL   Protocol not available on device
 *                              ARS_INVARG      Invalid numberWords or arincData parameter
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: When the ARINC 573 receiver is configured for auto-sync, 
 *      this routine will read an entire ARINC 573 frame based on the 
 *      specified sub-frame size (from speed selection).  If the receiver
 *      is configured for raw reception, all available data will be retrieved
 *      from the receiver channel FIFO buffer.  If the receive buffer doesn't  
 *      contain any data, it will return a zero; otherwise, it will
 *      read and return what is available.
 * 
 *      The data is returned as an array of 16-bit unsigned values.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_573_frame 
( 
   CEI_INT16 board,              /* (input)  board to reference (0 to 15)    */
   pCEI_UINT32 numberWords,      /* (in/output) number words to read/actual  */
   pCEI_UINT16 arincData         /* (output) data read from 717 FIFO         */
)
{
   CEI_INT32 mergedModeState;           /* 573 Rx ch merged/individual state */
   CEI_INT32 timeOutLimit = 0;          /* local time-out duration selection */
   pCEI_UINT16 arincDataPtr = NULL;     /* local ptr to applic. data array   */
   CEI_UINT32 localChannel = 0;         /* merged/individual channel select  */
   CEI_UINT32 numberToRead = 0;         /* local value for # of words to read*/
   CEI_UINT32 localData;                /* temp storage for received 717 data*/
   CEI_UINT32 index = 0;                /* local # of words read counter     */
   pCEI_UINT32 boardPointer = NULL;     /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer = NULL;    /* local pointer to channel memory area */
   CEI_UINT32 autoSyncDisable = 0;      /* current 573 Rx autosync setting   */
   CEI_INT32 delayDuration = 0;         /* receive word time-out duration    */
   CEI_INT32 timeOutCount = 0;          /* receive frame time-out counter    */
   CEI_INT32 subframeCount = 0;         /* local sub-frame count index       */
   CEI_INT32 nSync = FALSE;             /* local sync-detect flag            */
   CEI_INT16 status = ARS_NODATA;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) {
      status = ARS_INVBOARD;
   }
   else if ((arincData == NULL) || (numberWords == NULL)) {
      status = ARS_INVARG;
   }
   else {
#endif
      /* future support: translate the channel using the device channel 
       * allocation instead a the fixed single ARINC 573 channel */
      localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
      
      /* If the Merged Receive Mode is active for this channel, use the Merged Rx FIFO */
      status = ar_get_573_config (board, ARU_RX_MERGED_MODE, &mergedModeState);

      if (mergedModeState == AR_ON) {
         localChannel = deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX];
      }
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_GET_DATA);
   }

   if (status == ARS_NORMAL) {
#endif
      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         accessStatus = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      }
      /* setup pointers to the ARINC 717 channel status and data buffer */
      if (accessStatus == ARS_NORMAL) {
         /* If auto-sync is selected, setup the time-out logic to wait for an 
          * entire ARINC 573/717 frame to be received.  In the raw mode, return
          * only what is available in the FIFO.
          */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + 
                                      CEIDEV_CHANNEL_CONFIG_1 + (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
         accessStatus = cei_read_32bit (board, boardPointer, &autoSyncDisable);
         
         if (autoSyncDisable & CEIDEV_573_RX_MODE_RAW) {
            timeOutLimit = A573_RAW_TIMEOUT;
         }
         else {
            /* extract the current subframe size value from the configuraiton register and vary delay and count respectively */
            autoSyncDisable = (autoSyncDisable & CEIDEV_573_FRAME_49152_4096) >> CEIDEV_573_FRAME_SHIFT;
            timeOutLimit = a717frameSizeRxTimeOutCount[autoSyncDisable];
            delayDuration = a717frameSizeRxDelayDuration[autoSyncDisable];
         }
         
         status = ARS_NODATA;
         numberToRead = *numberWords;
         arincDataPtr = arincData;
         /* setup pointers to the ARINC 717 channel status and data buffer */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + 
                                      (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
         bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
         boardPointer += CEIDEV_CHANNEL_STATUS;
      }

      while ((index < numberToRead) && (accessStatus == ARS_NORMAL)) {

         if (timeOutCount > timeOutLimit) { /* if we get "no data" too many times */
            if (!autoSyncDisable)           /* for raw mode return active status setting */
               status = ARS_NODATA;         /* for autosync return NO DATA when frame is incomplete */
            break;
         }
         accessStatus = cei_read_32bit (board, boardPointer, &localData);
         if ((localData & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) && (accessStatus == ARS_NORMAL)) {
            accessStatus = cei_read_32bit (board, bufferPointer, &localData);
   
            if (accessStatus == ARS_NORMAL) {
               if (!autoSyncDisable) {

                  /* if h/w indicates this is a subframe-sync word, 
                   * check it against what is programmed by the API */
                  if ( (localData & 0x8000) == 0x8000 ) {
                     if ( (localData & 0x0FFF) == a717syncWords[board][subframeCount] ) {
                        nSync = TRUE;
                        subframeCount++;
                        if (subframeCount == 4) {
                           subframeCount = 0;
                        }
                        cei_delay_milliseconds(1000);
                     }
                     else {   /* subframe sync word doesn't match expected order */
                        nSync = FALSE;
                     }
                  }

                  if (nSync) {    /* still in valid sync, log data and continue */
                     *arincDataPtr = (CEI_UINT16)localData;
                     arincDataPtr++;
                     index++;
                     status = ARS_GOTDATA;
                  }
               }
               else {    /* autoSyncDisable = AR_OFF, raw mode recording only */
                  *arincDataPtr = (CEI_UINT16)localData;
                  arincDataPtr++;
                  index++;
                  status = ARS_GOTDATA;
               }
            }
         }
         else {    /* if the FIFO status indicates no data, delay and increment the time-out count */
            ar_sleep(delayDuration);
            timeOutCount++;
         }
      }
      /* if multi-thread protection requested, free semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
      }
      /* return the number of words read */
      *numberWords = index;

#ifndef TARGET_EMBEDDED
   }

   if ( ((status == ARS_NORMAL) || (status == ARS_NODATA)) && (accessStatus != ARS_NORMAL) )
      status = (CEI_INT16)accessStatus;
#endif
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:             A R _ G O
 *===========================================================================*
 *
 * FUNCTION:    Enable the transmitter and receiver FIFOs on the device
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL     Success
 *                              ARS_INVBOARD   Invalid board parameter
 *                                             or board not initialized
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Sets the global enable register global enable bit to ENABLE
 *              and enables the CEI-830 Message Scheduler
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_go (CEI_INT16 board)
{
   CEI_UINT32 globalEnableRegData;      /* contents of the Global Enable reg */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
         
         /* enable "global enable" */
         if (accessStatus == ARS_NORMAL) {
            globalEnableRegData |= CEIDEV_GLOBAL_ENABLE_ON;
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegData);
         }
         
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
      status = (CEI_INT16)accessStatus;
   }
   else
      status = ARS_INVBOARD;
#endif

   return (status);
}


/*===========================================================================*
 * ENTRY POINT:            A R _ I N I T I A L I Z E _ A P I
 *===========================================================================*
 *
 * FUNCTION:    Initialize the API local variables
 * 
 * PARAMETERS:  CEI_INT32 board   -- (input) WinRT device number to open (0-15)
 *
 * RETURN VAL:  CEI_INT16   -- ARS_NORMAL      Success
 *                             ARS_WINRTFAIL   Board not in WIN Registry
 *                                             or conflict mapping device
 *                             ARS_INVBOARD    Invalid board parameter
 *                             ARS_BADLOAD     FPGA load process failed
 *                             ARS_HW_DETECT   Supported device not detected
 *
 * GLOBALS:     deviceGlobal[board]->boardInitialized
 *              device[board].devicePtr
 *              deviceGlobal[board]->configuration
 *                                    
 * DESCRIPTION: This routine opens a WinRT session to the device and
 *              defines the API I/O register pointers.  This routine must be
 *              called first, before any other in the API (with the exception
 *              of ar_get_version and ar_loadslv - where it is invoked).
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_initialize_api ( CEI_INT16 board )
{
   CEI_UINT32 configuration;           /* device channel configuration       */
   CEI_UINT32 firmwareVersion;         /* device firmware configuration      */
   CEI_INT32  vendorIdent;             /* PCI Vendor ID for opened device    */
   CEI_INT32  deviceIdent;             /* PCI Device ID for opened device    */
   CEI_INT32  status = ARS_NORMAL;     /* OS-specific function return status */
   CEI_INT32  altStatus = ARS_NORMAL;  /* OS-specific function return status */
   CEI_UINT32 interruptBit;             /* the respective PLX Int Enable bit */

#ifndef TARGET_EMBEDDED
   if ( board < MAX_BOARDS ) {
#endif
      /*-------------------------------------------------------------------*
       * We are now ready to identify and map the device.
       *-------------------------------------------------------------------*/
      status = cei_acquire_device (board, &device[board]);
      
      /* validate the device configuration */
      if (status == ARS_NORMAL) { 
         status = cei_acquire_device_identity ((CEI_INT32)board, &vendorIdent, &deviceIdent);
         if (status != ARS_FAILURE) {
            if ((deviceGlobal[board]->attachedProcessCount == 0) || (device[board].multiThreadProtect == AR_CONC_NONE)) {
               /*------------------------------------------------------------------*
                * In not already initialized, load the FPGA and check configuration.
                *------------------------------------------------------------------*/
               if (vendorIdent == PCI_X30_MANUF_ID) {
#if defined (_WIN32) || defined (_LINUX_X86_) || defined (_LINUX_PPC_)
                  if (deviceIdent != PCI_DEVICE_ID_RAR_PCIE) {
                     status = a429_utl_enh_check (board, deviceIdent);
                  }
#endif
                  if (status == ARS_NORMAL) { 
                     switch (deviceIdent) { /* load routine is dependent on PCI intfc */
                        case PCI_DEVICE_ID_830:
                           status = a429_utl_program_xilinx_9056 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_830), 
                                                                   (pCEI_UCHAR)fpga_830 );
                           break;
         
                        case PCI_DEVICE_ID_830RX:
                           status = a429_utl_program_xilinx_9056 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_830_rx), 
                                                                   (pCEI_UCHAR)fpga_830_rx );
                           break;
         
                        case PCI_DEVICE_ID_630:
                           status = a429_utl_program_xilinx_9056_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)((sizeof(fpga_630))/4), 
                                                                         (pCEI_UINT32)fpga_630 );
                           break;
         
                        case PCI_DEVICE_ID_530:
                           status = a429_utl_program_xilinx_9056_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)((sizeof(fpga_530))/4), 
                                                                         (pCEI_UINT32)fpga_530 );
                           break;
   #ifndef NIVISA_PXI
                        case PCI_DEVICE_ID_A30:
                           status = a429_utl_program_xilinx_9056_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)((sizeof(fpga_A30))/4), 
                                                                         (pCEI_UINT32)fpga_A30 );
                           break;
         
                        case PCI_DEVICE_ID_RAREC:
                           status = cei_read_32bit (board, &device[board].baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], &interruptBit);
                           interruptBit &= ~PLX9030_LOCAL_INT_INP_ENABLE;
                           status = cei_write_32bit (board, &device[board].baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], interruptBit);
                           status = a429_utl_program_xilinx_9030_32bit ( board,
                                                                         deviceIdent, 
                                                                         (CEI_UINT32)(sizeof(fpga_ec)/4), 
                                                                         (pCEI_UINT32)fpga_ec );
                           break;
         
                        case PCI_DEVICE_ID_430:
                           status = cei_read_32bit (board, &device[board].baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], &interruptBit);
                           interruptBit &= ~PLX9030_LOCAL_INT_INP_ENABLE;
                           status = cei_write_32bit (board, &device[board].baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], interruptBit);
                           status = a429_utl_program_xilinx_9030 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_430), 
                                                                   (pCEI_UCHAR)fpga_430 );
                           break;
         
                        case PCI_DEVICE_ID_430A:
                           status = cei_read_32bit (board, &device[board].baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], &interruptBit);
                           interruptBit &= ~PLX9030_LOCAL_INT_INP_ENABLE;
                           status = cei_write_32bit (board, &device[board].baseAddr[PLX9030_CFG_INT_CTRL_STAT_REG], interruptBit);
                           status = a429_utl_program_xilinx_9030 ( board,
                                                                   deviceIdent, 
                                                                   (CEI_UINT32)sizeof(fpga_430A), 
                                                                   (pCEI_UCHAR)fpga_430A );
                           break;
   #endif
                        case PCI_DEVICE_ID_RAR_PCIE:
                           /* the RAR-PCIE firmware interface is implemented in BAR0, so fake it */
                           device[board].devicePtr = device[board].baseAddr;
                           break;
         
                        default:
                           status = ARS_HW_DETECT;
                           break;
                     }
                  }
               }
            }
            else if ((vendorIdent == PCI_X30_MANUF_ID) && (deviceIdent == PCI_DEVICE_ID_RAR_PCIE)) {
               /* for any multi-process instance, the RAR-PCIE host interface is implemented in BAR0, fix board pointer */
               device[board].devicePtr = device[board].baseAddr;
            }
         }

         if (status == ARS_NORMAL) {
            /* read the board configuration field */
            status = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &firmwareVersion);
            configuration = (firmwareVersion >> 8) & 0x000000FF;
            
            if ( (configuration >= CEIDEV_FIRST_CEIX30_CONFIG) && (configuration <= CEIDEV_LAST_CEIX30_CONFIG) ) {  
               deviceGlobal[board]->configuration = (CEI_UINT16)configuration;

               if (firmwareVersion & CEIDEV_IRIG_AVAILABLE)
                  deviceGlobal[board]->irigInstalled = TRUE;

               deviceGlobal[board]->firmwareVersion = (firmwareVersion >> CEIDEV_FIRMWARE_VERSION_SHIFT) & CEIDEV_FIRMWARE_VERSION_MASK;
               a429_utl_determine_channel_support (board);
            }
            else {
               status = ARS_HW_DETECT;
            }
         }
         else {  /* otherwise, its an invalid configuration... close and exit */
            altStatus = cei_release_device(board, &device[board]);
         }
      }
#ifndef TARGET_EMBEDDED
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   #ifdef INTEGRITY_PCI_PPC
      if (altStatus) ;
   #endif
   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:            A R _ I N I T I A L I Z E _ D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Initialize the device to a ready state
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) device number to reference (0-15)
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL       Success
 *                              ARS_MEMWRERR     Memory test failure
 *                              ARS_INVBOARD     Invalid board parameter
 *                              ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine initializes the device I/O configuration 
 *              registers to a state ready for use by the host application.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_initialize_device (CEI_INT16 board)
{
   CEI_UINT32 labelIndex;               /* local label filter table index    */
   CEI_UINT32 initialIrigTime;          /* initial IRIG BCD time value       */
   CEI_UINT32 bitFieldValue;            /* temp var for device register data */
   CEI_UINT32 configRegValue;           /* temp var for device register data */
   CEI_UINT32 deviceChannel;            /* temp var for device register data */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 configPointer;        /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer;        /* local pointer to channel memory area */
   CEI_INT16  status = ARS_NORMAL;      /* function return status            */
   CEI_INT32  accessStatus = ARS_NORMAL;/* status of device read/write access*/

#ifndef TARGET_EMBEDDED
   if ( (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) && (deviceGlobal[board]->attachedProcessCount > 0) ) {
#endif
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         if (accessStatus != ARS_NORMAL) {
            return (CEI_INT16)accessStatus;
         }
      }

      /*--------------------------------------------------------------------*
       *  Disable all data processing
       *--------------------------------------------------------------------*/
      bitFieldValue = 0;
      accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
      cei_delay_milliseconds(20);

      /*--------------------------------------------------------------------*
       *  Execute a quick memory test on a few unused SRAM locations
       *--------------------------------------------------------------------*/
      status = a429_utl_memory_test(board, UNUSED_SRAM, 0);
      if (status != ARS_NORMAL) {
         return status;
      }

      /*----------------------------------------------------------------------*
       *  Initialize local timer reference to 64-bit integer
       *  x20 compatibility to a 5ms resolution
       *  host-programmable IRIG offset to zero
       *----------------------------------------------------------------------*/
      deviceGlobal[board]->deviceTimeTagReference = AR_TIMETAG_INT_USEC_64BIT;
      deviceGlobal[board]->deviceX20CompatibleResolution = 20000;
      deviceGlobal[board]->userProgrammableIrigOffset = (TIME_TAG_TYPE)0;
      deviceGlobal[board]->receiveMode = ARU_BUFFERED;
      deviceGlobal[board]->transmitMode = ARU_XMIT_573_BPRZ;
      
      /*--------------------------------------------------------------------*
       *  Initialize the ARINC 429 receiver configuration registers
       *    Bus Speed of 100Kbps with Fast Slew Rate
       *    Odd Parity
       *    Buffered Rx mode
       *    Internal Wrap OFF
       *    Receiver FIFO Disabled
       *--------------------------------------------------------------------*/
      if (deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0) {
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         bitFieldValue = CEIDEV_PARITY_ENABLE + CEIDEV_FAST_SLEW_RATE + (CEIDEV_BAUD_RATE_100K << CEIDEV_BAUD_RATE_SHIFT);
         for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
              deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
              deviceChannel++) {
            accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
            /* Test for access conflict with register set */
            accessStatus = cei_read_32bit (board, boardPointer, &configRegValue);
            if (configRegValue != bitFieldValue) {
               return ARS_BAD_STATIC;
            }
            boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
         }
      }

      /*--------------------------------------------------------------------*
       *  Since the RAR-PCIE retains transmitter buffer content,
       *  flush it here.
       *--------------------------------------------------------------------*/
      if (deviceGlobal[board]->configuration == CEIDEV_CONFIG_RAR_PCIE) {
         configRegValue = CEIDEV_FAST_SLEW_RATE + CEIDEV_TRANSMIT_DISABLE + CEIDEV_CHANNEL_ENABLE + (8 << CEIDEV_BAUD_RATE_SHIFT);
         for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
              deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
              deviceChannel++) {
            boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS +
                                         (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE));
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
            bitFieldValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
            if (bitFieldValue) {
               configPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                            (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE));
               accessStatus = cei_write_32bit (board, configPointer, configRegValue);
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], CEIDEV_GLOBAL_ENABLE_ON);
               while (bitFieldValue) {
                  accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
                  bitFieldValue = (bitFieldValue & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
               }
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], 0);
            }
         }
      }
      
      /*--------------------------------------------------------------------*
       *  Initialize the ARINC 429 transmitter configuration registers
       *    Transmitter FIFO Enabled, Bus Speed of 100Kbps, Odd Parity
       *--------------------------------------------------------------------*/
      if (deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX] > 0) {
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE));
         bitFieldValue = CEIDEV_FAST_SLEW_RATE + CEIDEV_PARITY_ENABLE + CEIDEV_CHANNEL_ENABLE + (CEIDEV_BAUD_RATE_100K << CEIDEV_BAUD_RATE_SHIFT);
         for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
              deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
              deviceChannel++) {
            accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
            /* Test for access conflict with register set */
            accessStatus = cei_read_32bit (board, boardPointer, &configRegValue);
            if (configRegValue != bitFieldValue) {
               return ARS_BAD_STATIC;
            }
            boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
         }
      }
   
      /* Wait 10 Msec for signal    */
      cei_delay_milliseconds(10);
      
      /*---------------------------------------------------------------------------*
       * disable ARINC 429 enhanced label filtering, setting each table entry to 0 
       *---------------------------------------------------------------------------*/
      if (deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0) {
         for (labelIndex = CEIDEV_ENH_LABEL_FILTER_START; 
              (labelIndex < CEIDEV_ENH_LABEL_FILTER_END) && (accessStatus == ARS_NORMAL); 
              labelIndex++ ) {
            /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelIndex);
            /* write the filter table entry data pattern to this location */
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], 0);
         }
      }
   
      /*--------------------------------------------------------------------*
       * Initialize the message scheduler table to zero
       *--------------------------------------------------------------------*/
      if (deviceGlobal[board]->configuration != CEIDEV_CONFIG_R830RX) {
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_MSG_SCHED_TABLE_START);
         for (labelIndex = CEIDEV_MSG_SCHED_TABLE_START; 
              (labelIndex < CEIDEV_MSG_SCHED_TABLE_END) && (accessStatus == ARS_NORMAL); 
              labelIndex++ ) {
            accessStatus = cei_write_32bit (board, boardPointer, (CEI_UINT32)0);
            /* Test for access conflict with the message scheduler table */
            accessStatus = cei_read_32bit (board, boardPointer, &configRegValue);
            if (configRegValue != 0) {
               return ARS_BAD_STATIC;
            }
            boardPointer++;
         }
      }
   
      /*--------------------------------------------------------------------*
       * Initialize the snapshot buffer to zero
       *--------------------------------------------------------------------*/
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_SNAPSHOT_BUFFER_START);
      for (labelIndex = CEIDEV_SNAPSHOT_BUFFER_START; 
           (labelIndex < CEIDEV_SNAPSHOT_BUFFER_END) && (accessStatus == ARS_NORMAL); 
           labelIndex++ ) {
            accessStatus = cei_write_32bit (board, boardPointer, (CEI_UINT32)0);
            /* Test for access conflict with the message scheduler table */
            accessStatus = cei_read_32bit (board, boardPointer, &configRegValue);
            if (configRegValue != 0) {
               return ARS_BAD_STATIC;
            }
            boardPointer++;
      }
   
      /*--------------------------------------------------------------------*
       *  Initialize ARINC 573/717 receive and transmit channels
       *  Rx & Tx set to 768bps-64 wd subframe, FIFO Disabled, BPRZ encoding
       *  Receive API and h/w Sync Words set to 0
       *  Receive set to internal wrap OFF, individual buffered receive mode
       *     and auto-detect enabled after the FIFO is flushed
       *  Transmit set to External transmiter disabled, low slew rate
       *--------------------------------------------------------------------*/
      if ( deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_RX] > 0 ) {
         a717syncWords[board][0] = 0;
         a717syncWords[board][1] = 0;
         a717syncWords[board][2] = 0;
         a717syncWords[board][3] = 0;
         bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_RX_BPRZ_SELECT + CEIDEV_573_RX_MODE_RAW;
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         accessStatus = cei_write_32bit (board, boardPointer++, bitFieldValue);
         accessStatus = cei_write_32bit (board, boardPointer, (CEI_UINT32)0);
         accessStatus = cei_write_32bit (board, boardPointer, (CEI_UINT32)0);
      }
      if ( deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_TX] > 0 ) {
         bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_TX_BPRZ_SELECT + CEIDEV_TRANSMIT_DISABLE_MASK;
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] * CEIDEV_CHANNEL_REGION_SIZE));
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
      }

      /*--------------------------------------------------------------------*
       * if present, flush the 717 receiver
       *--------------------------------------------------------------------*/
      if ( deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_RX] > 0 ) {
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
         boardPointer += CEIDEV_CHANNEL_STATUS;
         /* clear any outstanding error status bits */
         bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_FIFO_INVALID_MSG;
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
         while (bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
            accessStatus = cei_read_32bit (board, bufferPointer, &labelIndex);
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
         }
         /* now enable auto-sync */
         bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_RX_BPRZ_SELECT;
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
      }

      /*------------------------------------------------------------------------------------------*
       *  Enable data processing and flush all of the 429 receive FIFOs
       *  Initialize ARINC 429/561 enhanced receiver snapshot storage to be on a label-only basis
       *------------------------------------------------------------------------------------------*/
      bitFieldValue = CEIDEV_GLOBAL_ENABLE_ON;
      accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
      
      /*--------------------------------------------------------------------*
       *  Enable and Flush the ARINC 429 receivers
       *--------------------------------------------------------------------*/
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                   (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE));
      configPointer = boardPointer + CEIDEV_CHANNEL_CONFIG_1;
      bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
      configRegValue = CEIDEV_PARITY_ENABLE + CEIDEV_FAST_SLEW_RATE + CEIDEV_CHANNEL_ENABLE + (CEIDEV_BAUD_RATE_100K << CEIDEV_BAUD_RATE_SHIFT);
      if (deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0) {
         for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
              deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
              deviceChannel++) {
            accessStatus = cei_write_32bit (board, configPointer, configRegValue);  /* write the config register - enabling reception */
            /* clear any outstanding error status bits */
            bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_FIFO_INVALID_MSG;
            accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);    /* read the status register - check for presense of data */
            while (bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
               accessStatus = cei_read_32bit (board, bufferPointer, &labelIndex);   /* read the data register to reset status for this message */
               accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue); /* read the status register - check for presense of more data */
            }
            boardPointer  += CEIDEV_CHANNEL_REGION_SIZE;                            /* setup for next channel */
            configPointer += CEIDEV_CHANNEL_REGION_SIZE;
            bufferPointer += CEIDEV_CHANNEL_REGION_SIZE;
         }
      
         /*--------------------------------------------------------------------*
          * Flush the merged mode receiver FIFO
          *--------------------------------------------------------------------*/
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      (deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
         /* clear any outstanding error status bits */
         bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_FIFO_INVALID_MSG;
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
         accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);    /* read the status register - check for presense of data */
         while (bitFieldValue == CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
            accessStatus = cei_read_32bit (board, bufferPointer, &labelIndex);   /* read the data register to reset status for this message */
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue); /* read the status register - check for presense of more data */
         }
      }

      /*--------------------------------------------------------------------*
       *  Disable all data processing
       *--------------------------------------------------------------------*/
      bitFieldValue = 0;
      accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);

      /* initialize the Discrete DAC to a mid-level value */
      accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_DAC_CONTROL], 0x0180);
      
      /* if present, initialize the IRIG DAC to the last adjusted value */
      if (deviceGlobal[board]->irigInstalled) {
         accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_DAC_CONTROL], irigDacThreshold[board]);
      
         /* initialize the IRIG timer to host current time-of-year */
         if (accessStatus == ARS_NORMAL) {
            cei_get_irig_initial_time (&initialIrigTime);
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_UPDATE], initialIrigTime);
         }
      }
      
      /* Acquire the current interrupt queue head pointer */
      accessStatus = cei_read_32bit(board, &device[board].devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], &lastReadIntQueueLocation[board]);
      lastApplicationReadIntQueueLocation[board] = lastReadIntQueueLocation[board];
      
      /* if multi-thread protection requested, free semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_release_thread_lock(LOCK_IDENT);
      }
#ifndef TARGET_EMBEDDED
   }
   else
      status = ARS_INVBOARD;
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:     A R _ H W _ I N T E R R U P T _ B U F F E R _ R E A D
 *===========================================================================*
 *
 * FUNCTION:    Read from the local API copy of the device h/w interrupt queue
 *
 * PARAMETERS:  board         -- (input)  device number to reference (0-15)
 *              numberOfWords -- (in/output) number of words to read & no. read
 *              data          -- (output) the queue entry value(s) read
 *
 * RETURN VAL:  ARS_GOTDATA   -- Data was successfully retrieved
 *              ARS_NODATA    -- No data was retrieved
 *              ARS_INVBOARD  -- Invalid device ID parameter
 *              ARS_INVARG    -- Invalid data buffer parameter
 *
 * GLOBALS:     deviceConfig[].deviceAcquired
 *              device[].devicePtr
 *              lastApplicationReadIntQueueLocation[]
 *
 * DESCRIPTION: This routine provides read access to the local API copy of
 *              the CEI-x30 device interrupt queue.  The local API copy is
 *              filled via h/w interrupt processing within the default API ISR.
 *              If the host application replaces the default API ISR with a
 *              custom ISR, this routine will not be usable.
 *
 *              For each invocation of this routine, the specified number of 
 *              queue entries will be read from the buffer region starting at
 *              the location last referenced by the API and ending at the 
 *              location referenced by the interrupt queue pointer.
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_hw_interrupt_buffer_read 
(
   CEI_INT32   board,                        /* device number (0-15)         */
   pCEI_UINT32 numberOfWords,                /* number of words to read      */
   pCEI_UINT32 data                          /* data read from the buffer    */
)
{
   CEI_UINT32 tempTailPtr;                      /* API int buffer tail ptr   */
   CEI_UINT32 tempHeadPtr;                      /* API int buffer head ptr   */
   CEI_UINT32 wordsRead;                        /* words to read index       */
   CEI_UINT32 unreadBufferSize = 0;          /* unused space in interrupt q  */
   CEI_INT32  status = ARS_NODATA;              /* local API routine status  */
   pCEI_UINT32 dataPtr;                      /* pointer to host data buffer  */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if (data == NULL) {
         status = ARS_INVARG;
      }
      else {
#endif
         tempTailPtr = interruptQueue[board].intQueueTailPtr;
         tempHeadPtr = interruptQueue[board].intQueueHeadPtr;

         if (tempTailPtr != tempHeadPtr) {
            if (tempTailPtr < tempHeadPtr)
               unreadBufferSize = tempHeadPtr - tempTailPtr;
            else
               unreadBufferSize = (INTERRUPT_QUEUE_SIZE - tempTailPtr) + tempHeadPtr + 1;
      
            if (unreadBufferSize > 0) {
               status = ARS_GOTDATA;
               dataPtr = data;
               for (wordsRead = 0; (wordsRead < *numberOfWords) && (wordsRead < unreadBufferSize); wordsRead++) {
                  *dataPtr = interruptQueue[board].intQueueEntries[tempTailPtr++];
                  if (tempTailPtr == INTERRUPT_QUEUE_SIZE) {
                     tempTailPtr = 0;
                  }
                  dataPtr++;
               }
               *numberOfWords = wordsRead;
               interruptQueue[board].intQueueTailPtr = tempTailPtr;
            }
         }
#ifndef TARGET_EMBEDDED
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return (status);
}

/*===========================================================================*
 * ENTRY POINT:     A R _ I N T E R R U P T _ Q U E U E _ R E A D
 *===========================================================================*
 *
 * FUNCTION:    Provides application read access to the h/w interrupt queue
 *
 * PARAMETERS:  board         -- (input)  device number to reference (0-15)
 *              numberOfWords -- (in/output) number of words to read & no. read
 *              data          -- (output) the queue entry value(s) read
 *
 * RETURN VAL:  ARS_GOTDATA     -- Data was successfully retrieved
 *              ARS_NODATA      -- No data was retrieved
 *              ARS_INVBOARD    -- Invalid device ID parameter
 *              ARS_INVARG      -- Invalid data buffer parameter
 *              ARS_BOARD_MUTEX -- Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceConfig[].deviceAcquired
 *              device[].devicePtr
 *              lastApplicationReadIntQueueLocation[]
 *
 * DESCRIPTION: This routine provides read access to the h/w interrupt queue.
 *              For each invocation, the specified number of queue entries
 *              will be read from the buffer region starting at the location
 *              last referenced by the API and ending at the location
 *              referenced by the interrupt queue pointer.
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_interrupt_queue_read 
(
   CEI_INT32   board,                        /* device number (0-15)         */
   pCEI_UINT32 numberOfWords,                /* number of words to read      */
   pCEI_UINT32 data                          /* data read from the buffer    */
)
{
   CEI_UINT32  bufferHead = 0;               /* temp register data parm      */
   CEI_UINT32  bufferTail = 0;               /* temp API tail ptr tracker    */
   CEI_UINT32  unreadBufferSize = 0;         /* unused space in interrupt q  */
   CEI_UINT32  wordsRead = 0;                /* words to write index         */
   CEI_INT32   rxStatus = ARS_NODATA;        /* API routine status           */
   CEI_INT32   status = ARS_NORMAL;          /* local API routine status     */
   pCEI_UINT32 appDataPtr;                   /* ptr to the app data array    */
   pCEI_UINT32 bufferDataPtr;                /* general purpose pointer var  */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   else if (data == NULL) {
      status = ARS_INVARG;
   }
   else {
#endif
      /* Acquire the current interrupt queue head pointer */
      status = cei_read_32bit((CEI_INT16)board, &device[board].devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], &bufferHead);
      bufferTail = lastApplicationReadIntQueueLocation[board];
   
      /* Determine how many unread queue entries are available in the h/w interrupt queue */
      if (status == ARS_NORMAL) {
         if (bufferTail != bufferHead) {
            if (bufferTail < bufferHead)
               unreadBufferSize = bufferHead - bufferTail;
            else
               unreadBufferSize = (CEIDEV_INT_QUEUE_SIZE - bufferTail) + bufferHead + 1;
   
            if (unreadBufferSize > 0)
               rxStatus = ARS_GOTDATA;
         }
      }
   
      if (rxStatus == ARS_GOTDATA) {
         appDataPtr = data;
         /* Read the lesser of the number of entries requested or the number of
          * unread entries available in the queue, starting from the tail ptr */
         bufferDataPtr = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_INT_QUEUE_START + bufferTail);
   
         if (device[board].multiThreadProtect) {
            status = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
            for (wordsRead = 0; (wordsRead < *numberOfWords) && (wordsRead < unreadBufferSize); wordsRead++) {
               status = cei_read_32bit ((CEI_INT16)board, bufferDataPtr, appDataPtr++);
               /* If we reach the end of the h/w interrupt queue, wrap around to the start */
               if (bufferTail == CEIDEV_INT_QUEUE_SIZE) {
                  bufferTail = 0;
                  bufferDataPtr = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_INT_QUEUE_START);
               }
               else {
                  bufferTail++;
                  bufferDataPtr++;
               }
            }
            if (device[board].multiThreadProtect) {
               status = cei_release_thread_lock(LOCK_IDENT);
            }
      
            /* Update the API-tracked interrupt queue tail ptr to upon completion */
            lastApplicationReadIntQueueLocation[board] = bufferTail;
         }
      }
   
      *numberOfWords = wordsRead;
   
#ifndef TARGET_EMBEDDED
   }
#endif
   if (status != ARS_NORMAL)
      rxStatus = status;

   return rxStatus;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ L A B E L _ F I L T E R
 *===========================================================================*
 *
 * FUNCTION:    Function to enable/disable standard label filtering on device.
 * 
 * PARAMETERS:  CEI_INT16 board   -- (input) Board number (0 to 15)  
 *              CEI_INT16 channel -- (input) 429 Rx channel selection,
 *                                    valid range is 0 to N, where N refers
 *                                    to the ARINC 561 channel, and N-1
 *                                    refers to the last 429 channel installed
 *              CEI_UINT16 label  -- (input) Label of interest.  Valid range
 *                                    is 0-255.  Also valid is ARU_ALL_LABELS
 *                                    which takes the action for all labels. 
 *              CEI_INT16 action  -- (input) Enable or disable filtering for
 *                                    this board/channel/label.  Valid values
 *                                    are: ARU_FILTER_ON/ARU_FILTER_OFF
 * 
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL      Success
 *                           ARS_INVBOARD    Invalid board argument
 *                           ARS_INVHARVAL   Invalid receive channel argument
 *                           ARS_INVARG      Invalid label or action argument
 *                           ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: The device supports hardware label filtering based on the
 *              label value for standard label filtering, or the combination
 *              of the label, SDI, and ESSM field values for enhanced label
 *              filtering.  This routine is designed to support standard
 *              label filtering for channel types 1 and 3 (basic ARINC 429).
 *              This routine assigns the specified filter action to the
 *              respective label filter table location for the device receive
 *              channel.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_label_filter 
( 
   CEI_INT16 board,                  /* (input) Board number (0 to 15)       */
   CEI_INT16 channel,                /* (input) 429 Rx channel selection     */
   CEI_UINT16 label,                 /* (input) Label of interest.           */
   CEI_INT16 action                  /* (input) Enable or disable filtering  */
)
{
   CEI_INT16 status;                              /* function return status  */
   /* the routine ar_enh_label_filter handles label filter table processing  */
   status = (CEI_INT16)ar_enh_label_filter (board, channel, label, ARU_ALL_SDI, ARU_ALL_ESSM, action);
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:                A R _ L O A D S L V
 *===========================================================================*
 *
 * FUNCTION:    This routine initializes the API and device.
 *
 * PARAMETERS:  CEI_INT16 board     -- (input) device number (0-15)
 *              CEI_UINT32 base_seg -- (input) not used       (present for
 *              CEI_INT32 base_port -- (input) not used  API compatibility
 *              CEI_UINT16 ram_size -- (input) not used              only)
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL    Success!!!
 *                              ARS_INVBOARD  board parm value not supported
 *                              ARS_BADLOAD   FPGA load process failed
 *                              ARS_BOARD_MUTEX Access to or creation of the
 *                                              Board Lock timed-out/failed
 *                              ARS_WINRTFAIL Board session open failed
 *                              ARS_HW_DETECT Supported device not detected
 *                              ARS_MEMWRERR  Device memory test failure
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine fulfills the standard Condor ARINC API
 *              initialization functionality.  It invokes the following:
 *              ar_initialize_api()    - acquires the device resources & loads FPGA
 *              ar_initialize_device() - initializes the device to the default settings
 *              ar_board_test()        - performs a simple internal wrap test
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_loadslv 
(
   CEI_INT16 board,                  /* (input) device number (0-15)         */
   CEI_UINT32 base_seg,              /* (input) legacy parameter, not used   */
   CEI_INT32 base_port,              /* (input) legacy parameter, not used   */
   CEI_UINT16 ram_size               /* (input) legacy parameter, not used   */
)
{
   CEI_INT16 status = ARS_NORMAL;        /* function return status           */
   CEI_INT32 accessStatus = ARS_NORMAL;  /* status of device lock access     */
   CEI_INT32 mapStatus;                  /* function return status           */
   CEI_INT32  index;                     /* utility index                    */
   CEI_UINT32 alreadyExists = FALSE;  /* indicates if shared board mem block exists  */
   CEI_UINT32 isExclusive = FALSE;    /* indicates if this is only mapping process   */
   CEI_UINT32 forceInit = FALSE;      /* indicates if shared mem init should be done */

#ifndef TARGET_EMBEDDED
   if (board < MAX_BOARDS) {
#endif
      /* for a multi-process application, either create the shared block or just map it*/

	   if (deviceGlobal[board] == NULL) {
         if (device[board].multiThreadProtect == AR_CONC_MULTIPROC) {

            /* assure we are alone in creating the shared memory block */
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);

            if (accessStatus == ARS_NORMAL) {

               /* attempt to map global shared board memory block */
               mapStatus = cei_map_shared_memory( (CEI_UCHAR)board,
                                                  sizeof(ARINC_GLOBAL_TYPE),
                                                  (CEI_VOID **)&deviceGlobal[board],
                                                  &device[board].sharedMemoryHandle,
                                                  &alreadyExists,
                                                  &isExclusive );
                               
               if (mapStatus != ARS_NORMAL) {
                  /* flag shared memory as invalid and bail out */
                  deviceGlobal[board] = NULL;
                  accessStatus = cei_release_thread_lock(LOCK_IDENT);
                  return (CEI_INT16)mapStatus;
               }

               /* force initialization of shared memory if all attached processes previously terminated but the board is still open */
               forceInit = ((alreadyExists) && (isExclusive) && (deviceGlobal[board]->attachedProcessCount == 0));
   
               /* check if this is the first process to map the block or init required */
               if ( (alreadyExists == FALSE) || (forceInit) ) {
                  deviceGlobal[board]->attachedProcessCount = 0;
                  deviceGlobal[board]->configuration = 0;
                  deviceGlobal[board]->ceiPciInterruptEnabled = FALSE;
                  for (index = 0; index < NUMBER_OF_CHANNEL_TYPES; index++) {
                     deviceGlobal[board]->receiveChannelMessageCount[index] = 0;
                  }
               }
               /* first process in a multiprocess initialization is done, safe to release the lock */
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
            else { /* failed to acquire the lock, this should never happen, but if it does fail initialization */
               return status;
            }
         }
         else {
            if (device[board].multiThreadProtect != AR_CONC_MULTITHRD) {
               device[board].multiThreadProtect = AR_CONC_NONE;
            }

            /* For a single-process application, just point to local pseudo-shared block for single-process implementation */
            deviceGlobal[board] = &deviceSingleProcessShared[board];
            deviceGlobal[board]->boardInitialized = FALSE;
            deviceGlobal[board]->attachedProcessCount = 0;
            deviceGlobal[board]->configuration = 0;
            deviceGlobal[board]->ceiPciInterruptEnabled = FALSE;
            for (index = 0; index < NUMBER_OF_CHANNEL_TYPES; index++) {
               deviceGlobal[board]->receiveChannelMessageCount[index] = 0;
            }
         }
      }

      /* acquire the device resources and initialize the API */
      status = ar_initialize_api (board);
      
      if (status == ARS_NORMAL) {
         deviceGlobal[board]->attachedProcessCount++;
         a429_utl_initialize_global_data (board);

         if ( (deviceGlobal[board]->attachedProcessCount == 1) || (device[board].multiThreadProtect == AR_CONC_NONE) ) {
            /* initialize the device to a known state,
             * then perform an internal wrap test (unless bypassed).         */
            status = ar_initialize_device (board);

            if (status == ARS_NORMAL) {
               deviceGlobal[board]->boardInitialized = TRUE;
            }
         }
      }

      if ((status == ARS_NORMAL) && (bypassWrapTest[board] == FALSE)) {
         status = ar_board_test (board, INTERNAL_WRAP);
      }

#ifndef TARGET_EMBEDDED
   }
   else   
      status = ARS_INVBOARD;
#endif

   return status;
}
                

/*===========================================================================*
 * ENTRY POINT:            A R _ M O D I F Y _ M S G
 *===========================================================================*
 *
 * FUNCTION:    Modify a message already defined for scheduled transmission.
 *
 * PARAMETERS:  CEI_INT16 board     -- (input) Board number (0 to 15)  
 *              CEI_INT16 channel   -- (input) 429 Tx channel selection,
 *                                         valid range is 0 to N, where N 
 *                                         refers to the ARINC 561 channel, 
 *                                         and N-1 refers to the last 429 
 *                                         Tx channel installed
 *              CEI_INT16 msgNumber -- (input) the message ident to modify
 *              CEI_INT16 rate      -- (input) transmission rate (in msec)
 *              CEI_INT32 data      -- (input) the ARINC word to transmit
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid channel
 *                              ARS_INVHARVAL   Message scheduling is not
 *                                              supported on the specified channel
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_FAILURE     The supplied msg table index
 *                                              exceeds the available number
 *                                              of table entries
 *
 * GLOBALS:     device[board].multiThreadProtect
 *
 * DESCRIPTION: This routine modifies an existing ARINC 429 or 561 message 
 *              already defined for periodic retransmission. 
 *
 *              The message rate is defined in milliseconds.  Based on h/w 
 *              design considerations, it is recommended that messages with  
 *              faster rates be defined before the slower ones. If the rate is
 *              set to zero, the message is considered to be removed and 
 *              eligable to be reused.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_modify_msg 
(
   CEI_INT16 board,                   /* (input) Board number (0 to 15)      */
   CEI_INT16 channel,                 /* (input) 429 Tx channel selection    */
   CEI_INT16 msgNumber,               /* (input) message index (from ar_def) */
   CEI_INT16 rate,                    /* (input) transmission rate (in msec) */
   CEI_INT32 data                     /* (input) the ARINC word to transmit  */
)
{
   CEI_UINT32 localChannel;                     /* transmit channel select   */
   CEI_INT32 status = ARS_NORMAL;               /* function return status    */
   CEI_INT32 tempStatus = ARS_NORMAL;           /* local temp return status  */
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;        /* message content structure */

#ifdef TARGET_EMBEDDED
   localChannel = (CEI_UINT32)channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
   /* verify calling parameter validity for h/w message scheduling */
   status = a429_utl_validate_scheduled_message_parameters ((CEI_UINT32)board, (CEI_UINT32)channel, &localChannel);

   if (status == ARS_NORMAL) {
      if (msgNumber >= CEIDEV_CHAN_MSG_SCH_ENTRY_MAX) {
         status = ARS_INVARG;
      }
   }
   if (status == ARS_NORMAL) {
#endif
      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = cei_acquire_thread_lock(LOCK_IDENT);
      }

      if (status == ARS_NORMAL) {
         msgEntry.messageIndex = (CEI_UINT32)msgNumber;
         msgEntry.board        = (CEI_UINT32)board;
         msgEntry.channel      = localChannel;
         msgEntry.rate         = (CEI_UINT32)rate;
         msgEntry.start        = (CEI_UINT32)0;         /* start offset and transmit count mods */
         msgEntry.txCount      = (CEI_UINT32)0;         /* are not supported by this routine    */
         msgEntry.data         = (CEI_UINT32)data;
         
         /* modify a single scheduled message entry */
         status = a429_utl_store_message_scheduler_entry (TRUE, 1, &msgEntry);
      
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            if (status == ARS_NORMAL)
               status = cei_release_thread_lock(LOCK_IDENT);
            else
               tempStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
   }
#endif
   
   #ifdef INTEGRITY_PCI_PPC
      if (tempStatus) ;
   #endif
   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:            A R _ M O D I F Y _ M S G _ B L O C K
 *===========================================================================*
 *
 * FUNCTION:    Modify multiple messages previously defined for scheduled 
 *              transmission.
 *
 * PARAMETERS:  CEI_INT32 numberOfEntries -- (input) the number of entries to
 *                                                define or modify, used as a
 *                                                reference for the number of 
 *                                                structure elements defined
 *                                                via the messageEntry pointer
 *              pAR_SCHEDULED_MSG_ENTRY_TYPE
 *                        messageEntry    -- (input) array of structures of  
 *                                              message content, each element  
 *                                              defined as follows:
 *                                                struct {
 *                                                  CEI_UINT32 messageIndex;                  
 *                                                  CEI_UINT32 board;                         
 *                                                  CEI_UINT32 channel;                       
 *                                                  CEI_UINT32 rate;                          
 *                                                  CEI_UINT32 start;                         
 *                                                  CEI_UINT32 txCount;               
 *                                                  CEI_UINT32 data; } 
 *                                                    
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVHARVAL   Message scheduling is not
 *                                              supported on the specified channel
 *                              ARS_INVARG      Invalid channel
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                              ARS_FAILURE     The supplied msg table index
 *                                              exceeds the available number
 *                                              of table entries
 *
 * GLOBALS:     device[board].multiThreadProtect
 *
 * DESCRIPTION: This routine modifies multiple existing ARINC 429 messages
 *              already defined for periodic retransmission. 
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_modify_msg_block 
(
   CEI_INT32 numberOfEntries,                /* (input) number of entries    */
   pAR_SCHEDULED_MSG_ENTRY_TYPE messageEntry /* (input) ptr to struct of msgs*/
)
{
   CEI_UINT32 board;                    /* board id extracted from structure */
   CEI_INT32 arrayIndex;                     /* scheduler access array index */
   CEI_UINT32 localChannel;                    /* transmit channel select    */
   CEI_INT16 status = ARS_NORMAL;              /* function return status     */
   CEI_INT32 tempStatus = ARS_NORMAL;          /* local temp return status   */
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry;       /* message content structure  */
   pAR_SCHEDULED_MSG_ENTRY_TYPE msgPtr = messageEntry; /* local ptr to array */

   for (arrayIndex = 0; (arrayIndex < numberOfEntries) && (status == ARS_NORMAL); arrayIndex++) {
      board = msgPtr->board;

      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      }
      
#ifdef TARGET_EMBEDDED
      localChannel = (CEI_UINT32)msgPtr->channel + deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#else
      /* verify calling parameter validity for h/w message scheduling */
      tempStatus = a429_utl_validate_scheduled_message_parameters (board, msgPtr->channel, &localChannel);

      if (tempStatus == ARS_NORMAL) {
         if (msgPtr->messageIndex >= CEIDEV_CHAN_MSG_SCH_ENTRY_MAX) {
            status = ARS_INVARG;
         }

         if (status == ARS_NORMAL) {
#endif
            msgEntry.messageIndex = msgPtr->messageIndex;
            msgEntry.board        = board;
            msgEntry.channel      = localChannel;
            msgEntry.rate         = msgPtr->rate;
            msgEntry.start        = 0;  /* start offset cannot be modified from original def */ 
            msgEntry.txCount      = msgPtr->txCount;
            msgEntry.data         = msgPtr->data;
            
            /* modify a single scheduled message entry */
            status = (CEI_INT16)a429_utl_store_message_scheduler_entry (TRUE, 1, &msgEntry);
         
           /* move to the next sructure array entry */
            msgPtr++;
#ifndef TARGET_EMBEDDED
         }
      }
      else {
         status = (CEI_INT16)tempStatus;
      }
#endif
      /* if multi-thread protection requested, free semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         if (status == ARS_NORMAL)
            status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
         else
            tempStatus = cei_release_thread_lock(LOCK_IDENT);
      }
   }
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ N U M _ R C H A N S
 *===========================================================================*
 *
 * FUNCTION:    Return the number of receive channels on the board.
 * 
 * PARAMETERS:  CEI_INT16 board -- (input) board to reference (0 to 15)
 *
 * RETURN VAL:  CEI_INT16       -- number of receive channels on this board.
 *                              ARS_INVBOARD  board parm value not supported
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Supply the number of receiver channels for this device
 *              NOTE: The board must have been initialized via ar_loadslv().
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_num_rchans (CEI_INT16 board)
{
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT)
#endif
      return (CEI_INT16)(deviceGlobal[board]->numRx);
#ifndef TARGET_EMBEDDED
   else
      return ARS_INVBOARD;
#endif
}               


/*===========================================================================*
 * ENTRY POINT:         A R _ N U M _ X C H A N S
 *===========================================================================*
 *
 * FUNCTION:    Return the number of transmit channels on the board.
 * 
 * PARAMETERS:  CEI_INT16 board -- (input) board to reference (0 to 15)
 *
 * RETURN VAL:  CEI_INT16       -- number of transmit channels on this board.
 *                              ARS_INVBOARD  board parm value not supported
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Supply the number of receiver channels for this device
 *              NOTE: The board must have been initialized via ar_loadslv().
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_num_xchans (CEI_INT16 board)
{
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT)
#endif
      return (CEI_INT16)(deviceGlobal[board]->numTx);
#ifndef TARGET_EMBEDDED
   else
      return ARS_INVBOARD;
#endif
}               


/*===========================================================================*
 * ENTRY POINT:             A R _ P U T _ 5 7 3 _ F R A M E
 *===========================================================================*
 *
 * FUNCTION:    Transmit ARINC 573 frame data on the ARINC 573 channel. 
 *
 * PARAMETERS:  CEI_INT16 board           -- (input) Board number of interest (0-15)
 *              CEI_UINT32 numberWords    -- (input) no. of ARINC 573 words to send
 *                                                   NOTE: must be an even number
 *              pCEI_UINT32 transmitCount -- (input) number of words transmitted,
 *                                                   a return value of 0 indicates 
 *                                                   a failure to transmit data
 *              pCEI_UINT16 arincData     -- (input) Pointer to the frame data
 * 
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL        success
 *                              ARS_INVBOARD      invalid board parameter or  
 *                                                board not initialized
 *                              ARS_INVHARVAL     invalid channel parameter
 *                              
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine will stuff the ARINC 573 frame data into 
 *              the respective transmit FIFO.  Once the FIFO is full or the
 *              specified number of elements have been copied to the FIFO,
 *              the number of elements "transmitted" will be returned.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_put_573_frame 
( 
   CEI_INT16 board,               /* (input) Board number of interest (0-15) */
   CEI_UINT32 numberWords,        /* (input) no. of ARINC 573 words to send  */
   pCEI_UINT32 transmitCount,     /* (input) number of words transmitted     */
   pCEI_UINT16 arincData          /* (input) Pointer to the frame data       */
) 
{
   CEI_UINT32 localChannel = 0;         /* merged/individual channel select  */
   CEI_UINT32 channelStatus;            /* Rx channel FIFO data avail status */
   CEI_UINT32 index;                    /* local # of words written counter  */
   CEI_UINT32 messageCount;             /* local number of words to transmit */
   CEI_UINT16 transmitData;             /* local temp transmit data storage  */
   pCEI_UINT16 localDataPtr;            /* local ptr to applic. data array   */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer;        /* local pointer to channel memory area */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/

   *transmitCount = 0;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      /* future support: translate the channel using the device channel 
       * allocation instead a the fixed single ARINC 573 channel
       */
      localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
      
#ifndef TARGET_EMBEDDED
      status = a429_utl_validate_channel_request (board, localChannel, ARU_PUT_DATA);
   }
   else {
      status = ARS_INVBOARD;
   }

   if (status == ARS_NORMAL) {
#endif
      /* Message count is half the requested word count (2 words/message) */
      messageCount = numberWords;
      localDataPtr = arincData;
      index = 0;

      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                   (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
      bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
      boardPointer += CEIDEV_CHANNEL_STATUS;

      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         while ( (index < messageCount) && (accessStatus == ARS_NORMAL) ) {
            accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
            if ( !(channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) && (accessStatus == ARS_NORMAL) ) {
               transmitData = *localDataPtr;
               accessStatus = cei_write_32bit (board, bufferPointer, (CEI_INT32)transmitData);
               localDataPtr++;
               index++;
            }
         }
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
      
      *transmitCount = index;
      
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
         status = (CEI_INT16)accessStatus;
#ifndef TARGET_EMBEDDED
   }
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:              A R _ P U T _ 4 2 9 _ M E S S A G E
 *===========================================================================*
 *
 * FUNCTION:    Put a message in the specified ARINC 429 transmit FIFO.
 *
 * PARAMETERS:  CEI_INT16 board   -- (input) board to reference (0 to 15)
 *              CEI_INT16 channel -- (input) 429 transmit channel to use
 *              CEI_INT32  data   -- (input) 32-bit ARINC 429 data to transmit 
 *
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL      Success
 *                           ARS_INVBOARD    Invalid board or board not initialized
 *                           ARS_INVHARVAL   Invalid transmit channel argument
 *                           ARS_XMITOVRFLO  Transmit buffer overrun has occurred
 *                           ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * GLOBALS:     none
 * 
 * DESCRIPTION: This routine utilizes the common ar_putword() routine to
 *              stuff a message into the respective ARINC 429 transmit FIFO.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_put_429_message  
( 
   CEI_INT16 board,             /* (input) board to reference (0 to 15)      */
   CEI_INT16 channel,           /* (input) 429 transmit channel to use       */
   CEI_INT32 data               /* (input) 32-bit ARINC 429 data to transmit */
)
{
   CEI_INT16 status = ARS_NORMAL;                 /* function return status  */
   status = ar_putword ( board, channel, data );
   return (status);
}
           

/*===========================================================================*
 * ENTRY POINT:             A R _ P U T W O R D
 *===========================================================================*
 *
 * FUNCTION:    Put a message into the generic device channel FIFO. 
 *
 * PARAMETERS:  CEI_INT16 board     -- (input) board to reference (0 to 15)
 *              CEI_INT16 channel   -- (input) Transmit channel (0 to *)
 *              CEI_INT32 arincdata -- (input) message data to transmit
 * 
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL      Success
 *                           ARS_INVBOARD    Invalid board or board not initialized
 *                           ARS_INVHARVAL   Invalid transmit channel argument
 *                           ARS_XMITOVRFLO  Transmit buffer overrun has occurred
 *                           ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine will place the specified message in the
 *              respective channel FIFO.  
 *              The channel number passed to this routine corresponds to the
 *              ARINC 429 transmit channel index, starting with zero.  If it *
 *              exceeds the 429 transmit channel count and an ARINC 573/717
 *              transmitter exists, it will be used as the designated transmitter.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_putword 
( 
   CEI_INT16 board,                 /* (input) board to reference (0 to 15)  */
   CEI_INT16 channel,               /* (input) transmit channel to use       */
   CEI_INT32 arincdata              /* (input) 32/16-bit data to transmit    */
) 
{
   CEI_INT16 status;                    /* function return status            */
   CEI_UINT32 channelStatus;            /* Rx channel FIFO data avail status */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer;        /* local pointer to channel memory area */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_UINT32 localChannel = (CEI_UINT32)channel; /* merged/indiv ch select  */

   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      if (localChannel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
         localChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      }
      else if (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] != CEIDEV_MAX_CHANNELS) {
         localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
      }
#ifndef TARGET_EMBEDDED
      else {
         status = ARS_INVHARVAL;
      }
      
      status = a429_utl_validate_channel_request (board, localChannel, ARU_PUT_DATA);
   }
   else {
      status = ARS_INVBOARD;
   }

   if (status == ARS_NORMAL) {
#endif

      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                   (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
      bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;

      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         /* check the transmit channel status, if it indicates "NOT FULL"
          * then stuff the data into the queue; otherwise, it's an overflow.
          */
         accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
         if (accessStatus == ARS_NORMAL) {
            if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
               status = ARS_XMITOVRFLO;
            }
            else {
               /* stuff the data into the transmit FIFO */
               accessStatus = cei_write_32bit (board, bufferPointer, arincdata);
               status =  ARS_NORMAL;
            }
         }
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
         status = (CEI_INT16)accessStatus;
   }
#endif
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:              A R _ P U T B L O C K
 *===========================================================================*
 *
 * FUNCTION:    Put multiple words into the specified channel's sequential
 *              transmit buffer.
 *
 * PARAMETERS:  CEI_UINT32 board       -- (input)  board to reference (0 to 15)
 *              CEI_UINT32 channel     -- (input)  recvr channel to use (0 to 255)
 *              CEI_INT32  maxMessages -- (input)  number of messages to transmit
 *              CEI_INT32  offset      -- (input)  unused parameter (legacy support)
 *              pCEI_INT32 data        -- (input)  source of data
 *              pCEI_INT32 actualCount -- (output) number of messages transmitted
 *
 * RETURN VAL:  ARS_NORMAL        Success
 *              ARS_INVBOARD      Invalid board number
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid argument value
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *              ARS_XMITOVRFLO    Transmit buffer overflow - indicates that 
 *                                there was not enough room in the transmit 
 *                                buffer for all of the ARINC words specified
 *                                ('ActualCount' receives the number of words
 *                                actually transfered)
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine puts multiple ARINC words in the transmit queue
 *              for a particular channel. When this routine returns, the data
 *              has not necessarily been sent, it has only been placed in the
 *              transmit buffer. If other data is in the transmit buffer ahead
 *              of it, it will be transmitted in turn.
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_putblock
(
   CEI_UINT32 board,           /* (input)  board to reference (0 to 15)      */
   CEI_UINT32 channel,         /* (input)  recvr channel to use (0 to 255)   */
   CEI_INT32  maxMessages,     /* (input)  number of messages to transmit    */
   CEI_INT32  offset,          /* (input)  unused parameter (legacy support) */
   pCEI_INT32 data,            /* (input)  source of data                    */
   pCEI_INT32 actualCount      /* (output) number of messages transmitted    */
)
{
   CEI_INT16 status;                    /* function return status            */
   CEI_INT32 messageCount = 0;          /* receiver buffer message counter   */
   CEI_UINT32 channelStatus;            /* Rx channel FIFO data avail status */
   CEI_INT32 arincdata;                 /* temp ARINC message container      */
   pCEI_INT32 localDataPtr;             /* temp pointer to application data  */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer;        /* local pointer to channel memory area */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_UINT32 localChannel = (CEI_UINT32)channel; /* merged/indiv ch select  */


#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      if (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
         localChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      }
      else if (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] != CEIDEV_MAX_CHANNELS) {
         localChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
      }
#ifndef TARGET_EMBEDDED
      else {
         status = ARS_INVHARVAL;
      }
      status = a429_utl_validate_channel_request ((CEI_INT16)board, localChannel, ARU_PUT_DATA);
      
      /* verify required parameters */
      if ( (status == ARS_NORMAL) && ((data == NULL) || (actualCount == NULL)) ) {
         status = ARS_INVARG;
      }
   }
   else {
      status = ARS_INVBOARD;
   }

   if (status == ARS_NORMAL) {
#endif
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                   (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
      bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
      boardPointer += CEIDEV_CHANNEL_STATUS;

      /* attempt to transmit as many messages as requested */
      messageCount = 0;
      localDataPtr = data;
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         while ((messageCount < maxMessages) && (accessStatus == ARS_NORMAL)) {
            /* check the transmit channel status, if it indicates "NOT FULL"
             * then stuff the data into the queue; otherwise, it's an overflow.
             */
            accessStatus = cei_read_32bit ((CEI_INT16)board, boardPointer, &channelStatus);
            if (accessStatus == ARS_NORMAL) {
               if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
                  status = ARS_XMITOVRFLO;
                  break;
               }
               else {
                  arincdata = *localDataPtr;
                  localDataPtr++;
         
                  /* stuff the data into the transmit FIFO */
                  accessStatus = cei_write_32bit ((CEI_INT16)board, bufferPointer, arincdata);
                  if (accessStatus == ARS_NORMAL)
                     messageCount++;
               }
            }
         }
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }

      *actualCount = messageCount;
   
#ifndef TARGET_EMBEDDED
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
         status = (CEI_INT16)accessStatus;
   }
#endif
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:              A R _ P U T B L O C K _ M U L T I _ C H A N
 *===========================================================================*
 *
 * FUNCTION:    Place multiple transmit messages into the multiple channel's 
 *              transmit buffers.
 *
 * PARAMETERS:  CEI_UINT32 board       -- (input)  board to reference (0 to 15)
 *              CEI_INT32 maxMessages  -- (input)  number of messages to transmit
 *              pCEI_UINT32 channels   -- (input)  ARINC 429 transmit channels on
 *                                                 which to transmit data (0 to 15)
 *              pCEI_INT32 data        -- (input)  source of data
 *              pCEI_INT32 actualCount -- (output) number of messages transmitted
 *
 * RETURN VAL:  ARS_NORMAL        Success
 *              ARS_INVBOARD      Invalid board number
 *              ARS_INVHARVAL     Invalid receive channel argument
 *              ARS_INVARG        Invalid argument value
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *              ARS_XMITOVRFLO    Transmit buffer overflow - indicates that 
 *                                there was not enough room in the transmit 
 *                                buffer for all of the ARINC words specified
 *                                ('ActualCount' receives the number of words
 *                                actually transfered)
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine puts multiple ARINC words in the transmit queue
 *              for a multiple channels. When this routine returns, the data
 *              has not necessarily been sent, it has only been placed in the
 *              respective transmit buffer. If other data is in the transmit
 *              buffer ahead of it, it will be transmitted in turn.
 *
 *              Each element of the message data array will be placed in the
 *              transmitter buffer designated by the same "channels" array
 *              element, on an element-by-element basis.
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_putblock_multi_chan
(
   CEI_UINT32  board,          /* (input)  board to reference (0 to 15)      */
   CEI_INT32   maxMessages,    /* (input)  number of messages to transmit    */
   pCEI_UINT32 channels,       /* (input)  transmit channels to use (0 to 15)*/
   pCEI_INT32  data,           /* (input)  source of data                    */
   pCEI_INT32  actualCount     /* (output) number of messages transmitted    */
)
{
   CEI_INT32 messageCount = 0;          /* receiver buffer message counter   */
   CEI_UINT32 channelStatus = 0;        /* Rx channel FIFO data avail status */
   CEI_INT32 arincdata;                 /* temp ARINC message data storage   */
   pCEI_UINT32 localChanPtr;            /* pointer to app's channel array    */
   pCEI_INT32 localDataPtr;             /* pointer to app's message array    */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer;        /* local pointer to channel memory area */
   CEI_INT32 localChannel = 0;          /* temp transmit channel storage     */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/


#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      /* verify required parameters */
      if ((channels == NULL) || (data == NULL) || (actualCount == NULL)){
         status = ARS_INVARG;
      }

      if (status == ARS_NORMAL) {
#endif
         /* attempt to transmit as many messages as requested */
         messageCount = 0;
         localDataPtr = data;
         localChanPtr = channels;
         /* if multi-thread protection requested, acquire semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (accessStatus == ARS_NORMAL) {
            while ((messageCount < maxMessages) && (accessStatus == ARS_NORMAL) && (status == ARS_NORMAL)) {
   
               /* offset 0-basis and switch the channel select for each "channels" array entry */
               localChannel = (CEI_INT32)*localChanPtr;
               localChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
#ifndef TARGET_EMBEDDED
               status = a429_utl_validate_channel_request ((CEI_INT16)board, localChannel, ARU_PUT_DATA);
               if (status == ARS_NORMAL) {
#endif
                  boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                               (localChannel * CEIDEV_CHANNEL_REGION_SIZE));
                  bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
                  boardPointer += CEIDEV_CHANNEL_STATUS;
      
                  /* check the transmit channel status, if it indicates "NOT FULL"
                   * then stuff the data into the queue; otherwise, it's an overflow.
                   */
                  if (accessStatus == ARS_NORMAL)
                     accessStatus = cei_read_32bit ((CEI_INT16)board, boardPointer, &channelStatus);
      
                  if (accessStatus == ARS_NORMAL) {
                     if (channelStatus & CEIDEV_CSR_XMIT_FIFO_FULL) {
                        status = ARS_XMITOVRFLO;
                        break;
                     }
                     else {
                        arincdata = *localDataPtr;
                        localDataPtr++;
                        localChanPtr++;
         
                        /* stuff the data into the transmit FIFO */
                        accessStatus = cei_write_32bit ((CEI_INT16)board, bufferPointer, arincdata);
                        if (accessStatus == ARS_NORMAL)
                           messageCount++;
                     }
                  }
#ifndef TARGET_EMBEDDED
               }
#endif
            }
            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
      
         *actualCount = messageCount;

#ifndef TARGET_EMBEDDED
      }
#endif
   
#ifndef TARGET_EMBEDDED
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL)) {
         status = (CEI_INT16)accessStatus;
      }
   }
   else {
      status = ARS_INVBOARD;
   }
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:                 A R _ P U T F I L T E R
 *===========================================================================*
 *
 * FUNCTION:    Copy the label interrupt and filter buffer definition used 
 *              by the device enhanced label filter feature.
 * 
 * PARAMETERS:  CEI_UINT32 board      -- (input) Board number (0 to 15)  
 *              CEI_UINT32 channel    -- (input) 429 Rx channel selection,
 *                                               valid range is 0 to one less
 *                                               than the number of enhanced
 *                                               ARINC 429 channels installed
 *              pCEI_CHAR filterTable -- (input) pointer to the host filter table
 * 
 * RETURN VAL:  ARS_NORMAL        operation completed successfully
 *              ARS_INVBOARD      invalid board number or board not initialized
 *              ARS_INVARG        invalid receive channel or null table ptr
 *              ARS_BOARD_MUTEX   Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This routine copies the specified label filter table entry
 *              from the caller's memory space to the enhanced label filter
 *              table on the device.  Each element of the enhanced label 
 *              filter buffer consists of a bit field defined as follows:
 *
 *                 FILTER_SEQUENTIAL  0x10   if CLEAR add label to sequential
 *                                              receive buffer
 *                 FILTER_SNAPSHOT    0x20   If CLEAR add label to snapshot
 *                                              receive buffer
 *                 FILTER_INTERRUPT   0x40   If SET on reception insert channel #
 *                                              in the interrupt queue and if 
 *                                              enabled generate a PCI interrupt
 *
 *              The filter buffer for a single channel is defined as follows:
 *
 *                 CEI_UCHAR filterTable[MAX_ESSM][MAX_SDI][MAX_LABEL]
 *
 *              where the bits of the ARINC word are split up as follows:
 *
 *              field   ---eSSM---  -SDI-  --------label---------
 *               bits   30  29  28   9 8   7  6  5  4  3  2  1  0
 *                    
 *              On the device each 32-bit entry in the filter table is 
 *              allocated once for each label/SDI combination, divided 
 *              into individual filter activation fields by eSSM:
 *
 *                 Label Filter Table [MAX_LABEL][MAX_SDI]
 *
 *               bits   31 30 29 28 27 ... 8  7  6  5  4  3  2  1  0
 *              field  | eSSM = 7  |   ...   | eSSM = 1  | eSSM = 0 |
 *                    
 *===========================================================================*/ 
EXPORT32 CEI_INT32 DLL_EXPORTED ar_putfilter 
(
   CEI_UINT32 board,             /* (input) Board number (0 to 15)           */
   CEI_UINT32 channel,           /* (input) 429 Rx channel selection         */
   pCEI_CHAR filterTable         /* (input) pointer to the host filter table */
)
{
   CEI_INT32  labelFilterIndex;         /* label filter channel select index */
   CEI_UINT32 filterActivation;         /* local translation of action parm  */
   CEI_UINT32 essmIndex;                /* local temp enh filter table index */
   CEI_UINT32 labelIndex;               /* local temp enh filter table index */
   CEI_UINT32 sdiIndex;                 /* local temp enh filter table index */
   CEI_INT32  localChannel = 0;         /* merged/individual channel select  */
   CEI_CHAR   hostFilterTableData;      /* local copy of host filterTable ent*/
   pCEI_CHAR  hostFilterTablePtr;       /* local pointer to host filterTable */
   CEI_INT16  status = ARS_NORMAL;      /* function return status            */
   CEI_INT32  accessStatus = ARS_NORMAL;/* status of device read/write access*/
   CEI_INT32  tempStatus = ARS_NORMAL;          /* local temp return status  */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      return ARS_INVBOARD;
   
   /* Label Filter Table access requests are limited to the number of ARINC 429 receive channels installed on the device */
   else if ( ((CEI_UINT32)channel >= deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) ||
             (filterTable == NULL) ) {
      status = ARS_INVARG;
   }
   else {
#endif
      localChannel = (CEI_INT32)channel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];

      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (status == ARS_NORMAL) {
         /* assign the specified action from the host table to the device table */
         hostFilterTablePtr = filterTable;
   
         labelFilterIndex = CEIDEV_ENH_LABEL_FILTER_START + (localChannel * CEIDEV_ENH_LABEL_FILTER_CHAN_SIZE);
         for (labelIndex = 0; labelIndex < MAX_LABEL; labelIndex++) {
            for (sdiIndex = 0; sdiIndex < MAX_SDI; sdiIndex++) {
               filterActivation = 0;
               for (essmIndex = 0; essmIndex < MAX_ESSM; essmIndex++) {
                  hostFilterTablePtr = filterTable  + (essmIndex << 10) + (sdiIndex << 8) + labelIndex;
                  hostFilterTableData = *hostFilterTablePtr;
                  filterActivation <<= 4;
                  filterActivation |= (CEI_UINT32)(hostFilterTableData >> 4) & CEIDEV_ENH_FLTR_BIT_MASK;
               }

               /* load the External Memory Address Register with the Label Filter Table SRAM long-word address  */
               accessStatus = cei_write_32bit ((CEI_INT16)board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_ADDR], labelFilterIndex);
               /* write the filter table entry data pattern to this location */
               accessStatus = cei_write_32bit ((CEI_INT16)board, &device[board].devicePtr[CEIDEV_ACCESS_SRAM_DATA], filterActivation);
               labelFilterIndex++;  
            }
         }

         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            if (status == ARS_NORMAL)
               status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
            else
               tempStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
   }
   
   if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
      status = (CEI_INT16)accessStatus;
#endif
      
   #ifdef INTEGRITY_PCI_PPC
      if (tempStatus) ;
   #endif
   return status;
}


/*===========================================================================*
 * ENTRY POINT:            A R _  Q U E R Y _ D E V I C E
 *===========================================================================*
 *
 * FUNCTION:    Open a session to the board and identify the board type
 * 
 * PARAMETERS:  CEI_INT16  board     -- (input)  device number to open (0-15)
 *              pCEI_INT16 boardType -- (output) if supported, the board type
 *
 *                                                  PRODUCT     DEFINE  VALUE
 *                                                  CEI-830   - CEI830   19
 *                                                  CEI-430   - CEI430   21
 *                                                  AMC-830   - AMC830   22
 *                                                  CEI-530   - CEI530   26
 *                                                  R-830RX   - R830RX   27
 *                                                  RAR-CPCI  - RARCPCI  28
 *                                                  RAR-EC    - RAREC    29
 *
 *                                               unsupported by the API but 
 *                                               reported includes the following:
 *
 *                                                  CEI-520   - CEI520    6 
 *                                                  CEI-620   - CEI620   10 
 *                                                  CEI-820   - CEI820   11 
 *                                                  CEI-715   - CEI715   16 
 *                                                  PMC-708   - PMC708   20
 *                                                  CNIC-A2P  - CNIC    100
 *                                                  PMC-D1    - PMCD1   101
 *                                                  PMC-D2    - PMCD2   102
 *                                                  EPMC      - EPMC    103
 *                                                  P-SER     - PSER    104
 *                                                  P-MIO     - PMIO    105
 *                                                  P-DIS     - PDIS    106
 *                                                  P-10SF    - P10SF   107
 *                                                  AMC-1553  - AMC1553 108
 *                                                  PCI-1553  - PCx1553 109
 *                                                  & cPCI-1553
 *                                                  QPCI-1553 - QPCI    110
 *                                                  QCP-1553  - QCP     111
 *                                                  QPCX-1553 - QPCX    112
 *
 *                                               if not, "UNSUPPORTED"
 *
 * RETURN VAL:  CEI_INT16   -- ARS_NORMAL      Success
 *                             ARS_INVBOARD    Invalid board parameter
 *
 * GLOBALS:     None
 *                                    
 * DESCRIPTION: This routine opens a WinRT session to the device and
 *              checks the PCI manufacturer and device ID.  If these IDs
 *              are of supported x30 products, identify the board;
 *              otherwise report it as "unsupported".
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_query_device ( CEI_INT16 board, pCEI_INT16 boardType )
{
   CEI_INT32  vendorIdent;              /* PCI Vendor ID for opened device   */
   CEI_INT32  deviceIdent;              /* PCI Device ID for opened device   */
   CEI_INT16  status = ARS_NORMAL;      /* OS-specific function return status*/

#ifndef TARGET_EMBEDDED
   if ( board < MAX_BOARDS ) {
#endif
      /*-------------------------------------------------------------------*
       * We are now ready to map and identify the device.
       *-------------------------------------------------------------------*/
      status = (CEI_INT16)cei_acquire_device (board, &device[board]);
      
      /* validate the device configuration */
      if (status == ARS_NORMAL) {
         status = (CEI_INT16)cei_acquire_device_identity ((CEI_INT32)board, &vendorIdent, &deviceIdent);
         if ( (status != ARS_FAILURE) && (vendorIdent == PCI_X30_MANUF_ID) ) {
            switch (deviceIdent) {
               case PCI_DEVICE_ID_830:
                  *boardType = CEI830;
                  break;
               case PCI_DEVICE_ID_830RX:
                  *boardType = R830RX;
                  break;
               case PCI_DEVICE_ID_A30:
                  *boardType = AMCA30;
                  break;
               case PCI_DEVICE_ID_430:
                  *boardType = CEI430;
                  break;
               case PCI_DEVICE_ID_430A:
                  *boardType = CEI430A;
                  break;
               case PCI_DEVICE_ID_530:
                  *boardType = CEI530;
                  break;
               case PCI_DEVICE_ID_630:
                  *boardType = RARCPCI;
                  break;
               case PCI_DEVICE_ID_RAREC:
                  *boardType = RAREC;
                  break;
               case PCI_DEVICE_ID_RAR_PCIE:
                  *boardType = RARPCIE;
                  break;
               case PCI_DEVICE_ID_CNIC:
                  *boardType = CNIC;
                  break;
               case PCI_DEVICE_ID_708 :
                  *boardType = PMC708;
                  break;
               case PCI_DEVICE_ID_520 :
                  *boardType = CEI520;
                  break;
               case PCI_DEVICE_ID_620 :
                  *boardType = CEI620;
                  break;
               case PCI_DEVICE_ID_820 :
                  *boardType = CEI820;
                  break;
               case PCI_DEVICE_ID_PMD1:
                  *boardType = PMCD1;
                  break;
               case PCI_DEVICE_ID_PMD2:
                  *boardType = PMCD2;
                  break;
               case PCI_DEVICE_ID_EPMC:
                  *boardType = EPMC;
                  break;
               case PCI_DEVICE_ID_PSER:
                  *boardType = PSER;
                  break;
               case PCI_DEVICE_ID_PMIO:
                  *boardType = PMIO;
                  break;
               case PCI_DEVICE_ID_PDIS:
                  *boardType = PDIS;
                  break;
               case PCI_DEVICE_ID_P10S:
                  *boardType = P10SF;
                  break;
               case PCI_DEVICE_ID_AMC :
                  *boardType = AMC1553;
                  break;
               case PCI_DEVICE_ID_1553:
                  *boardType = PCx1553;
                  break;
               case PCI_DEVICE_ID_QPCI:
                  *boardType = QPCI;
                  break;
               case PCI_DEVICE_ID_QCP :
                  *boardType = QCP;
                  break;
               case PCI_DEVICE_ID_QPCX:
                  *boardType = QPCX;
                  break;
               default:
                  *boardType = UNSUPPORTED;
            }
         }
         else
            *boardType = UNSUPPORTED;

         status = (CEI_INT16)cei_release_device(board, &device[board]);
      }
#ifndef TARGET_EMBEDDED
   }
   else 
      status = ARS_INVBOARD;
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:        AR_READ_MESSAGE_SCHEDULE_TABLE_ENTRY
 *===========================================================================*
 *
 * FUNCTION:    Retrieve an entry from the message scheduler table
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *              pCEI_UINT32 msgSchedTableEntry -- (output) array of 8 uint
 *
 * RETURN VAL:  CEI_INT16  -- ARS_NORMAL    success
 *                            ARS_INVBOARD  board parm value not supported
 *
 * GLOBALS:     deviceGlobal[]
 *              device[]
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_read_message_schedule_table_entry 
(
   CEI_INT16   board,
   CEI_UINT32  messageIndex,
   pCEI_UINT32 msgSchedTableEntry
)
{
   CEI_UINT32 msgIndex;                 /* temp index for table search       */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   pCEI_UINT32 boardPointer;
   pCEI_UINT32 msgPtr = msgSchedTableEntry;
   
   if (VALID_BOARD_AND_INIT) {
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_MSG_SCHED_TABLE_START) + (messageIndex * CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE);
      for (msgIndex = 0; msgIndex < CEIDEV_MSG_SCHED_TBL_ENTRY_SIZE; msgIndex++) {
         accessStatus = cei_read_32bit (board, boardPointer, msgPtr);
         boardPointer++;
         msgPtr++;
      }
   }
   else
      status = ARS_INVBOARD;

   #ifdef INTEGRITY_PCI_PPC
      if (accessStatus) ;
   #endif
   return status;
}

/*===========================================================================*
 * ENTRY POINT:             A R _ R E S E T
 *===========================================================================*
 *
 * FUNCTION:    Reset the device to the initial state
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *
 * RETURN VAL:  CEI_INT16  -- ARS_NORMAL    success
 *                            ARS_INVBOARD  board parm value not supported
 *
 * GLOBALS:     n/a

 * 
 * DESCRIPTION: see description for ar_initialize_device()
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_reset (CEI_INT16 board)
{
   /* a simple re-initialization of the device will do */
   return ( ar_initialize_device(board) );
}    
     

/*===========================================================================*
 * ENTRY POINT:         A R _ R E S E T _ T I M E R C N T
 *===========================================================================*
 *
 * FUNCTION:    This routine resets the 64-bit device timer to zero.
 * 
 * PARAMETERS:  CEI_INT16 board        -- (input) Board number (0 to 15)  
 *
 * RETURN VAL:  none
 *
 * GLOBALS:     device[board]
 *
 *===========================================================================*/
 EXPORT32 CEI_VOID  DLL_EXPORTED ar_reset_timercnt (CEI_INT16 board) 
{
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_INT16 status = ARS_NORMAL;               /* local return status       */
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT)
   {
#endif
      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      }
      
      if (status == ARS_NORMAL) {
         accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_LOW], (CEI_UINT32)0);
         if (accessStatus == ARS_NORMAL)
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_HIGH], (CEI_UINT32)0);

      
         /* if multi-thread protection requested, free semaphore/mutex-locking */
         if (device[board].multiThreadProtect) {
            status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
   }
#endif
}


#if defined (_LINUX_X86_) || defined (_LINUX_PPC_)
/*===========================================================================*
 * ENTRY POINT:     A R _ S E T _ B A S E _ A D D R E S S E S
 *===========================================================================*
 *
 * FUNCTION:    This routine provides a method to supply the BAR0 and BAR2
 *              virtual address pointers to the base address assigned
 *              to the device acquired from a custom embedded host device driver.
 *
 * PARAMETERS:  CEI_INT16  board       -- (input) board to reference (0 to 15)
 *              CEI_UINT32 bar0Address -- (input) virtual address for BAR0
 *              CEI_UINT32 bar2Address -- (input) virtual address for BAR2
 *
 * RETURN VAL:  ARS_INVBOARD  board parm invalid
 * 
 * GLOBALS:     device[board].devicePtr
 * 
 * DESCRIPTION: This routine provides a method for any embedded platform
 *              application to bypass the standard device driver/host interface
 *              to acquire the device memory resources and supply the device
 *              BAR0 and BAR2 virtual base addresses to the API.
 * 
 *===========================================================================*/
extern CEI_VOID cei_assign_device_resources (CEI_INT16 deviceIndex, CEI_UINT32 bar0Addr, CEI_UINT32 bar2Addr);

EXPORT32 CEI_INT32 DLL_EXPORTED ar_set_base_addresses 
(
   CEI_INT16  board,                 /* (input) board to reference (0 to 15) */
   CEI_UINT32 bar0Address,           /* (input) virtual address for BAR0     */
   CEI_UINT32 bar2Address            /* (input) virtual address for BAR2     */
)
{
   CEI_INT32 status = ARS_NORMAL;
   if (board < MAX_BOARDS) {
      cei_assign_device_resources (board, bar0Address, bar2Address);
   }
   else {
      status = ARS_INVBOARD;
   }
   return status;
}
#endif

/*===========================================================================*
 * ENTRY POINT:             A R _ S E T _ C O N F I G
 *===========================================================================*
 *
 * FUNCTION:    Initialize channel configuration, BusTools product specific.
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *              CEI_INT16 item   -- (input) item for which to set the 
 *                                             current state:
 * 
 *                    ARU_RX_CH0*_BIT_RATE          ARU_DISCRETE_OUT1
 *                    ARU_RX_CH1*_BIT_RATE          ARU_DISCRETE_OUT2
 *                    ARU_TX_CH0*_BIT_RATE          ARU_DISCRETE_OUT2
 *                    ARU_TX_CH1*_BIT_RATE          ARU_DISCRETE_OUT3
 *                    ARU_RX_CH0*_PARITY            ARU_DISCRETE_OUT
 *                    ARU_RX_CH1*_PARITY            ARU_DISCRETE_VALUES/ARU_DISCRETE_OUTPUTS
 *                    ARU_TX_CH0*_SHUT_OFF          ARU_INTERNAL_WRAP         
 *                    ARU_TX_CH1*_SHUT_OFF          ARU_IRIG_WRAP_ENABLE      
 *                    ARU_TX_CH0*_PARITY            ARU_IRIG_INPUT_THRESHOLD  
 *                    ARU_TX_CH1*_PARITY            ARU_IRIG_ADJUST_THRESHOLD 
 *                    ARU_TX_CH**_HB_INJ            ARU_IRIG_QUICK_ADJUSTMENT 
 *                    ARU_TX_CH**_LB_INJ            ARU_IRIG_SET_BIAS         
 *                    ARU_TX_CH**_GAP_INJ           ARU_ACCESS_SNAPSHOT_BUFFER
 *                                                  
 *                    ARU_RX_TIMETAG_MODE w/values of:                          
 *                                                     
 *                              AR_TIMETAG_EXT_IRIG_64BIT                         
 *                              AR_TIMETAG_INT_USEC_64BIT                         
 *                              AR_TIMETAG_INT_20USEC_32BIT                       
 *                              AR_TIMETAG_INT_MSEC_32BIT                         
 *                              AR_TIMETAG_HOST_SEC_32BIT                       
 *                                                     
 *              CEI_UINT32 value -- (input) this is the value to which
 *                                      the specified item will be set
 * 
 * RETURN VAL:  CEI_INT16 -- ARS_NORMAL      Normal succesful completion.
 *                           ARS_INVARG      "item" parameter value is not
 *                                           supported by this function
 *                           ARS_INVHARVAL   Item was invalid or channel
 *                                           not supported on this device
 *                           ARS_INVBOARD    Board parameter value invalid or
 *                                           board has not been initialized
 *                           ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * 
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Calls the function ar_set_device_config () with the
 *      channel(s) of interest based on the expected item definition.
 *      
 *      ar_set_device_config() writes a 16-bit data word to a transmit or 
 *      receive channel configuration register, and should be used by any 
 *      application other than BusTools.
 *
 *      NOTE: The value passed into ar_set_config is UNSIGNED SHORT, the 
 *            value passed here is CEI_UINT32, yet only the least
 *            signifciant 16 bits are considered.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_config 
( 
   CEI_INT16 board,                 /* (input)  board to reference (0 to 15) */
   CEI_INT16 item,                  /* (input)  item to process              */
   CEI_UINT32 value                 /* (input)  state of the item            */
)
{
   CEI_INT32 status;                    /* function return status            */
   CEI_INT16 channel;                   /* local item-derived channel index  */
   CEI_INT16 channelIndex;              /* local item-derived channel index  */
   CEI_INT16 total429TxChannels;        /* local Tx total channel count      */
   CEI_INT16 total429RxChannels;        /* local Rx total channel count      */
   CEI_UINT16 locValue;                 /* local 16bit version of value parm */
   CEI_INT32 locLongValue;              /* temp 32bit variable               */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      return ARS_INVBOARD;
#endif
   
   locValue = (CEI_UINT16)value;
   locLongValue = (CEI_INT32)value;
   status = ARS_NORMAL;

   /*-----------------------------------------------------------------*
    *  CEI-x20 ARINC 429 Set/Get Config Support
    *-----------------------------------------------------------------*/
   /* Configure receive channel bit rate in Hertz */
   if ( ((item >= ARU_RX_CH01_BIT_RATE) && (item <= ARU_RX_CH32_BIT_RATE)) || 
        ((item >= ARU_OLD_RX_CH01_BIT_RATE) && (item <= ARU_OLD_RX_CH16_BIT_RATE)) ) {

      if (item >= ARU_RX_CH01_BIT_RATE)
         channel = (CEI_INT16)(item - ARU_RX_CH01_BIT_RATE);
      else
         channel = (CEI_INT16)(item - ARU_OLD_RX_CH01_BIT_RATE);

      if (value > ARU_SPEED_LOW) 
         locLongValue = (AR_SIXTEEN_MHZ / (CEI_INT32)value) - 2;
      
      status = ar_set_device_config (board, channel, ARU_RX_BITRATE, (CEI_INT16)locLongValue);
   }

   /* Configure transmit channel bit rate in Hertz */
   else if ( ((item >= ARU_TX_CH01_BIT_RATE) && (item <= ARU_TX_CH32_BIT_RATE)) || 
             ((item >= ARU_OLD_TX_CH01_BIT_RATE) && (item <= ARU_OLD_TX_CH16_BIT_RATE)) ) {

      if (item >= ARU_TX_CH01_BIT_RATE)
         channel = (CEI_INT16)(item - ARU_TX_CH01_BIT_RATE);
      else
         channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_BIT_RATE);

      if (value > ARU_SPEED_LOW) 
         locLongValue = (AR_SIXTEEN_MHZ / (CEI_INT32)value) - 2;
      
      status = ar_set_device_config (board, channel, ARU_TX_BITRATE, (CEI_INT16)locLongValue);
   }

   /* Configure receive channel parity */
   else if ( ((item >= ARU_RX_CH01_PARITY) && (item <= ARU_RX_CH32_PARITY)) ||
             ((item >= ARU_OLD_RX_CH01_PARITY) && (item <= ARU_OLD_RX_CH16_PARITY)) ) {    
      
      if (item >= ARU_RX_CH01_PARITY)
         channel = (CEI_INT16)(item - ARU_RX_CH01_PARITY);
      else
         channel = (CEI_INT16)(item - ARU_OLD_RX_CH01_PARITY);
            
      status = ar_set_device_config (board, channel, ARU_RX_PARITY, locValue);
   }

   /* Configure transmit channel parity */
   else if ( ((item >= ARU_TX_CH01_PARITY) && (item <= ARU_TX_CH32_PARITY)) ||
             ((item >= ARU_OLD_TX_CH01_PARITY) && (item <= ARU_OLD_TX_CH16_PARITY)) ) {    
      
      if (item >= ARU_TX_CH01_PARITY)
         channel = (CEI_INT16)(item - ARU_TX_CH01_PARITY);
      else
         channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_PARITY);

      status = ar_set_device_config (board, channel, ARU_TX_PARITY, locValue);
   }

   /* Configure transmit enable */
   else if ( ((item >= ARU_TX_CH01_SHUT_OFF) && (item <= ARU_TX_CH32_SHUT_OFF)) ||
             ((item >= ARU_OLD_TX_CH01_SHUT_OFF) && (item <= ARU_OLD_TX_CH16_SHUT_OFF)) ) {    
      
      if (item >= ARU_TX_CH01_SHUT_OFF)
         channel = (CEI_INT16)(item - ARU_TX_CH01_SHUT_OFF);
      else
         channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_SHUT_OFF);
      
      status = ar_set_device_config (board, channel, ARU_TX_DISABLE, locValue);
   }

   /* Assign transmit channel bit count error injection */
   else if ( ((item >= ARU_TX_CH01_HB_INJ) && (item <= ARU_TX_CH32_HB_INJ)) ||
             ((item >= ARU_OLD_TX_CH01_HB_INJ) && (item <= ARU_OLD_TX_CH16_HB_INJ)) ) {    
      
      if (item >= ARU_TX_CH01_HB_INJ)
         channel = (CEI_INT16)(item - ARU_TX_CH01_HB_INJ);
      else
         channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_HB_INJ);

      if (value == AR_ON)
         locValue = AR_HI;
      else
         locValue = AR_OFF;

      status = ar_set_device_config (board, channel, ARU_TX_BIT_ERROR, locValue);
   }

   /* Assign transmit channel bit count error injection */
   else if ( ((item >= ARU_TX_CH01_LB_INJ) && (item <= ARU_TX_CH32_LB_INJ)) ||
             ((item >= ARU_OLD_TX_CH01_LB_INJ) && (item <= ARU_OLD_TX_CH16_LB_INJ)) ) {    
      
      if (item >= ARU_TX_CH01_LB_INJ)
         channel = (CEI_INT16)(item - ARU_TX_CH01_LB_INJ);
      else
         channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_LB_INJ);

      if (value == AR_ON)
         locValue = AR_LO;
      else
         locValue = AR_OFF;

      status = ar_set_device_config (board, channel, ARU_TX_BIT_ERROR, locValue);
   }

   /* Assign transmit channel message gap error injection */
   else if ( ((item >= ARU_TX_CH01_GAP_INJ) && (item <= ARU_TX_CH32_GAP_INJ)) ||
             ((item >= ARU_OLD_TX_CH01_GAP_INJ) && (item <= ARU_OLD_TX_CH16_GAP_INJ)) ) {    
      
      if (item >= ARU_TX_CH01_GAP_INJ)
         channel = (CEI_INT16)(item - ARU_TX_CH01_GAP_INJ);
      else
         channel = (CEI_INT16)(item - ARU_OLD_TX_CH01_GAP_INJ);

      status = ar_set_device_config (board, channel, ARU_TX_GAP_ERROR, locValue);
   }

   else {
      /* for the backward compatibility with old x20 options, only support
       * the total number of 429 channels here, omit 717 channels */
      total429TxChannels = ar_num_xchans(board);
      total429RxChannels = ar_num_rchans(board);
      if (deviceGlobal[board]->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717) {
         total429TxChannels--;
         total429RxChannels--;
      }

      switch (item)
      {
         case ARU_XMIT_RATE:
            if (value > ARU_SPEED_LOW) 
               locLongValue = (AR_SIXTEEN_MHZ / (CEI_INT32)value) - 2;

            for (channel = 0; channel < total429TxChannels; channel++) {
               status = ar_set_device_config ( board, channel, ARU_TX_BITRATE, (CEI_INT16)locLongValue );
            }
            break;
   
         case ARU_RECV_RATE:
            if (value > ARU_SPEED_LOW) 
               locLongValue = (AR_SIXTEEN_MHZ / (CEI_INT32)value) - 2;

            for (channel = 0; channel < total429RxChannels; channel++) {
               status = ar_set_device_config ( board, channel, ARU_RX_BITRATE, (CEI_INT16)locLongValue );
            }
            break;
   
         case ARU_PARITY:
            for (channel = 0; channel < total429RxChannels; channel++) {
               status = ar_set_device_config ( board, channel, ARU_RX_PARITY, locValue );
            }
            for (channel = 0; channel < total429TxChannels; channel++) {
               status = ar_set_device_config ( board, channel, ARU_TX_PARITY, locValue );
            }
            break;
   
         case ARU_INTERNAL_WRAP:
            /* include 717 channel with internal wrap */
            if (deviceGlobal[board]->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717) {
               status = ar_set_573_config (board, ARU_RECV_MODE, locValue);
            }
            for (channel = 0; channel < total429RxChannels; channel++) {
               status = ar_set_device_config ( board, channel, ARU_RECV_MODE, locValue );
            }
            break;
   
         /* assign output discrete register "output discrete" values */
         case ARU_DISCRETE_OUT1:
         case ARU_DISCRETE_OUT2:
         case ARU_DISCRETE_OUT3:
         case ARU_DISCRETE_OUT4:
            channel = (CEI_INT16)(item - ARU_DISCRETE_OUT1);  /* determine channel number */
            status = ar_set_device_config (board, channel, ARU_DISCRETE_OUT, locValue);
            break;
   
         case ARU_DIGITAL_OUT1:
         case ARU_DIGITAL_OUT2:
         case ARU_DIGITAL_OUT3:
         case ARU_DIGITAL_OUT4:
         case ARU_DIGITAL_OUT5:
         case ARU_DIGITAL_OUT6:
         case ARU_DIGITAL_OUT7:
         case ARU_DIGITAL_OUT8:
         case ARU_DIGITAL_OUT9:
         case ARU_DIGITAL_OUT10:
         case ARU_DIGITAL_OUT11:
         case ARU_DIGITAL_OUT12:
            channel = (CEI_INT16)(item - ARU_DIGITAL_OUT1);  /* determine channel number */
            status = ar_set_device_config (board, channel, ARU_DISCRETE_OUT, locValue);
            break;
   
         case ARU_IRIG_OUTPUT_ENABLE:
         case ARU_ACCESS_SNAPSHOT_BUFFER:
         case ARU_RX_TIMETAG_MODE:
         case ARU_IRIG_WRAP_ENABLE:
         case ARU_IRIG_INPUT_THRESHOLD:
         case ARU_IRIG_ADJUST_THRESHOLD:
         case ARU_IRIG_QUICK_ADJUSTMENT:
         case ARU_IRIG_SET_BIAS:
            status = ar_set_device_config (board, 0, item, locValue);
            break;
   
         case ARU_DISCRETE_VALUES:
         case ARU_DISCRETE_OUTPUTS:
            for (channelIndex = 0; channelIndex < (CEI_INT16)deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_DISC_OUT]; channelIndex++) {
               status = ar_set_device_config (board, channelIndex, ARU_DISCRETE_OUT, (CEI_INT16)(locValue & 1));
               locValue >>= 1;
            }
            break;

         default : status = ARS_INVARG; 
            break;
      }
   }

   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:        A R _ S E T _ D E V I C E _ C O N F I G          
 *===========================================================================*
 *
 * FUNCTION:    Assign protocol-specific device configuration register attributes 
 * 
 * PARAMETERS:  CEI_INT16 board   -- (input) board to reference (0 to 15)
 *              CEI_INT16 channel -- (input) channel number (0 to # 429 channels)
 *              CEI_INT16 item    -- (input) channel config item to (re)set
 *              CEI_INT16 value   -- (input) value to set item to
 * 
 *              Valid combinations of item / value parameters are:
 * 
 *              for this "item"        valid options for "value" are:
 * 
 *              ARU_RX_BITRATE         ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom Speed
 *              ARU_RX_PARITY          ARU_PARITY_NONE ARU_PARITY_ODD  ARU_PARITY_EVEN
 *                                     representing AR_OFF or AR_ON  
 *              ARU_RECV_MODE          AR_WRAP_OFF     AR_WRAP_ON
 * 
 *              ARU_TX_BITRATE         ARU_SPEED_LOW   ARU_SPEED_HIGH  Custom
 *              ARU_TX_PARITY          ARU_PARITY_NONE ARU_PARITY_ODD  ARU_PARITY_EVEN
 *
 *              ARU_RX_DISABLE         AR_ON           AR_OFF
 *              ARU_RX_FIFO_ENABLE     AR_ON           AR_OFF
 *              ARU_TX_DISABLE         AR_ON           AR_OFF
 *              ARU_TX_FIFO_ENABLE     AR_ON           AR_OFF
 *
 *              ARU_TX_GAP_ERROR       AR_ON           AR_OFF
 *              ARU_TX_BIT_ERROR       AR_HI   AR_LO   AR_OFF
 *
 *              ARU_DISCRETE_OUT       AR_HI           AR_LO
 *              ARU_DIGITAL_OUT        AR_HI           AR_LO
 *              ARU_DIFFERENTIAL_OUT   AR_ON   AR_OFF  AR_HI    AR_LO
 *
 *              ARU_RX_MERGED_MODE     AR_ON           AR_OFF
 *              ARU_FAST_SLEW_RATE     AR_ON           AR_OFF
 * 
 *              ARU_ACCESS_SNAPSHOT_BUFFER ARU_LABEL_ONLY   ARU_LABEL_WITH_SDI
 *              ARU_BYPASS_INIT_WRAP_TEST  AR_ON       AR_OFF
 *              ARU_MULTITHREAD_PROTECT    AR_ON       AR_OFF
 *              ARU_HW_INTERRUPT_ENABLE    AR_ON       AR_OFF       
 * 
 *              ARU_IRIG_WRAP_ENABLE   AR_ON           AR_OFF
 *              ARU_IRIG_OUTPUT_ENABLE AR_ON           AR_OFF
 *              ARU_IRIG_INPUT_THRESHOLD      threashold voltage in millivolts
 *              ARU_IRIG_ADJUST_THRESHOLD     display DAC settings AR_ON/AR_OFF
 *              ARU_IRIG_QUICK_ADJUSTMENT
 *              ARU_IRIG_SET_BIAS             user IRIG offset 0 to +/-32768 milliseconds
 *              ARU_HW_ENHANCE_UPDATE  AR_ON           AR_OFF
 *              ARU_RX_TIMETAG_MODE w/values of:                          
 *                                     AR_TIMETAG_EXT_IRIG_64BIT                         
 *                                     AR_TIMETAG_INT_USEC_64BIT                         
 *                                     AR_TIMETAG_INT_20USEC_32BIT                       
 *                                     AR_TIMETAG_INT_MSEC_32BIT
 *                                                     
 * RETURN VAL:  CEI_INT16 status  -- ARS_NORMAL      Normal succesful completion.
 *                                   ARS_INVARG      "item" parameter value is not
 *                                                      supported by this function
 *                                   ARS_INVHARVAL   "value" was invalid or channel
 *                                                      not supported on this device
 *                                   ARS_INVBOARD    Board parameter value invalid or
 *                                                      board has not been initialized
 *                                   ARS_INT_ISR  &  Interrupt was either not initialized
 *                                   ARS_DRIVERFAIL     or not terminated successfully
 *                                   ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * 
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This function assigns the respective device configuration
 *              register attributes based on the combined item/value
 *              parameter selections.  It is designed to support the 
 *              ARINC 429 and all discrete and digital I/O configuration  
 *              register attributes available to the device.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_device_config 
( 
   CEI_INT16 board,          /* (input) board to reference (0 to 15)         */
   CEI_INT16 channel,        /* (input) channel number (0 to # 429 channels) */
   CEI_INT16 item,           /* (input) channel config item to (re)set       */
   CEI_INT16 value           /* (input) value to set item to                 */
)
{
   CEI_UINT32 deviceChannel;         /* current channel to select            */
   CEI_UINT32 bitFieldValue = 0;     /* temp for device register read content*/
   CEI_INT32 accessStatus = ARS_NORMAL;  /* status of read/write access      */
   CEI_INT16 status = ARS_NORMAL;           /* function return status        */
   CEI_INT32 tempStatus = ARS_NORMAL;           /* local temp return status  */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   CEI_UINT32 offsetIntoChannelWindow = 0;  /* offset into chan mem region   */
   CEI_UINT32 localIntEnableFlag = FALSE;
   pCEI_UINT32 intCtrlRegPtr = 0;       /* ptr to the PLX Interrupt Ctrl Reg */
   CEI_UINT32 pciIntCtrlRegOffset = 0;  /* the PLX Interrupt Ctrl Reg Offset */
   CEI_UINT32 interruptBit;             /* the respective PLX Int Enable bit */
   CEI_INT32 memoryRegion = 0;

   if ((item == ARU_HW_ENHANCE_UPDATE) && (board < MAX_BOARDS)) {
      if (value == AR_OFF) {
         status = a429_utl_enh (board, PLX_ORIGINAL_RESTORE);
      }
      else if (value == AR_ON) {
         status = a429_utl_enh (board, PLX_ENHANCED_UPDATE);
      }
      else {
         status = ARS_INVARG;
      }

      return status;
   }
   else if (item == ARU_WAIT_ON_EVENT) {
      accessStatus = (CEI_INT32)value;
      tempStatus = cei_open_shared_lock (board, &accessStatus);
      return status;
   }
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      return ARS_INVBOARD;
#endif
   
   deviceChannel = (CEI_UINT32)channel;

   /* if multi-thread protection requested, semaphore/mutex-lock device access */
   if (device[board].multiThreadProtect) {
      status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      if (status != ARS_NORMAL)
         return (status);
   }
      
   /* Validate the channel parameter for the specific item parameter,
    * along with predefining the Channel Type, Channel Select, and Channel
    * Offset, for options requiring configuration register operations.
    */
   switch (item) {
      case ARU_RECV_MODE:
      case ARU_RX_BITRATE:
      case ARU_RX_PARITY:
      case ARU_RX_DISABLE:
      case ARU_RX_FIFO_ENABLE:
      case ARU_RX_MERGED_MODE:
      case ARU_RX_RESET:
         offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
         deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX];
         break;
   
      case ARU_TX_BIT_ERROR:
      case ARU_TX_GAP_ERROR:
      case ARU_TX_BITRATE:
      case ARU_TX_PARITY:
      case ARU_TX_FIFO_ENABLE:
      case ARU_TX_DISABLE:
      case ARU_FAST_SLEW_RATE:
      case ARU_TX_RESET:
         offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
         deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
         break;

      /* support legacy CEI-710 Discrete item parameters */
      case ARU_DISCRETE_OUT1:
      case ARU_DISCRETE_OUT2:
         deviceChannel = (item == ARU_DISCRETE_OUT1) ? 0 : 1;
         /* no break here, now process below as a generic ARU_DISCRETE_OUT selection */
      case ARU_DISCRETE_OUT:
         deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DISC_OUT];
         offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
         break;
   
      case ARU_DIGITAL_OUT:
         deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DISC_OUT];
         offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
         break;
   
      case ARU_DIFFERENTIAL_OUT:
         deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_DIFF_OUT];
         offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
         break;

      case ARU_ACCESS_SNAPSHOT_BUFFER:
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
         if (value == ARU_LABEL_WITH_SDI) {
            bitFieldValue &= CEIDEV_SNAPSHOT_STOR_LABEL_MASK;
         }
         else if (value == ARU_LABEL_ONLY) {
            bitFieldValue |= CEIDEV_SNAPSHOT_STORAGE_ENHANCED;
         }
         else {
            status = ARS_INVHARVAL;
         }
         if (status == ARS_NORMAL) {
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
            status = EXIT_ROUTINE;
         }
         break;
 
      case ARU_HW_INTERRUPT_ENABLE:
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &bitFieldValue);
         if ((deviceGlobal[board]->configuration == CEIDEV_CONFIG_CEI430) || 
             (deviceGlobal[board]->configuration == CEIDEV_CONFIG_RAR_EC) ||
             (deviceGlobal[board]->configuration == CEIDEV_CONFIG_CEI430A)) {
            pciIntCtrlRegOffset = PLX9030_CFG_INT_CTRL_STAT_REG;
            intCtrlRegPtr = device[board].baseAddr + PLX9030_CFG_INT_CTRL_STAT_REG;
            interruptBit = PLX9030_LOCAL_INT_INP_ENABLE;
            memoryRegion = 1;
         }
         else if (deviceGlobal[board]->configuration != CEIDEV_CONFIG_RAR_PCIE) {
            pciIntCtrlRegOffset = PLX_CONFIG_INT_CTRL_STAT_REG;
            intCtrlRegPtr = device[board].baseAddr + PLX_CONFIG_INT_CTRL_STAT_REG;
            interruptBit = PLX9056_LOCAL_INT_INP_ENABLE;
            memoryRegion = 1;
         }
         else {
            interruptBit = 0;
            memoryRegion = 0;
            pciIntCtrlRegOffset = 0;
         }
         if (value == AR_ON) {
            /* first disable interrupts before initiating the ISR */
            bitFieldValue &= CEIDEV_INTERRUPTS_DISABLED;
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
            /* assign the ISR to the h/w interrupt, then enable PLX local interrupt */
            accessStatus = cei_init_interrupt ( board, 
                                                (pCEI_VOID)cei_utl_interrupt_handler, 
                                                ARU_OS_THREAD_PRIORITY, 
                                                intCtrlRegPtr,
                                                memoryRegion,
                                                pciIntCtrlRegOffset,
                                                interruptBit, 
                                                &localIntEnableFlag );
            if (accessStatus == ARS_NORMAL) {
               deviceGlobal[board]->ceiPciInterruptEnabled = localIntEnableFlag;
               bitFieldValue |= CEIDEV_INTERRUPTS_ENABLED;
            }
            else {
               status = (CEI_INT16)accessStatus;
            }
         }
         else if (value == AR_OFF) {
            bitFieldValue &= CEIDEV_INTERRUPTS_DISABLED;
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
            if (deviceGlobal[board]->ceiPciInterruptEnabled) {
               accessStatus = cei_terminate_interrupt(board, interruptBit, intCtrlRegPtr);
               deviceGlobal[board]->ceiPciInterruptEnabled = FALSE;
            }
         }
         else {
            status = ARS_INVHARVAL;
         }
         if (status == ARS_NORMAL) {
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
            status = EXIT_ROUTINE;
         }
         break;

      /* The remaining items are handled either in utility routines or they 
       * don't actually require interaction with common h/w channel register
       * selection.  For these options, perform the operation then set the
       * status so as to skip the subsequent processing.
       */
      case ARU_IRIG_WRAP_ENABLE:
      case ARU_IRIG_INPUT_THRESHOLD:
      case ARU_IRIG_ADJUST_THRESHOLD:
      case ARU_IRIG_QUICK_ADJUSTMENT:
      case ARU_IRIG_SET_BIAS:
      case ARU_IRIG_OUTPUT_ENABLE:
         if (deviceGlobal[board]->irigInstalled) {
            switch (item) {
               case ARU_IRIG_WRAP_ENABLE:
                  accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], (pCEI_UINT32)&bitFieldValue);
      
                  if ( (value == AR_ON) || (value == AR_WRAP_ON) )
                     bitFieldValue |= CEIDEV_IRIG_INTERNAL_ENABLE;
                  else
                     bitFieldValue &= CEIDEV_IRIG_INTERNAL_MASK;
                  
                  if (accessStatus == ARS_NORMAL)
                     accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
                  break;
         
               case ARU_IRIG_OUTPUT_ENABLE:
                  accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], (pCEI_UINT32)&bitFieldValue);
      
                  if ( (value == AR_ON) || (value == AR_WRAP_ON) )
                     bitFieldValue |= CEIDEV_IRIG_830RX_IRIG_ENABLE;
                  else
                     bitFieldValue &= CEIDEV_IRIG_830RX_IRIG_MASK;
                  
                  if (accessStatus == ARS_NORMAL)
                     accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
                  break;
         
               case ARU_IRIG_INPUT_THRESHOLD:
                  bitFieldValue = (CEI_UINT32)((128L + (CEI_INT32)(((float)256/(float)3.3)*((float)4.99/(float)22.1) * ((float)value/(float)1000) )) & 0x000000FF);
                  accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_DAC_CONTROL], bitFieldValue);
                  irigDacThreshold[board] = (CEI_INT32)bitFieldValue;
                  break;
         
               case ARU_IRIG_ADJUST_THRESHOLD:
                  accessStatus = a429_utl_adjust_irig_threshold (board, (CEI_INT32)value);
                  break;
         
               case ARU_IRIG_QUICK_ADJUSTMENT:
                  accessStatus = a429_utl_quick_adjust_irig_threshold (board, (CEI_INT32)value);
                  break;
         
               case ARU_IRIG_SET_BIAS:
                  deviceGlobal[board]->userProgrammableIrigOffset = (TIME_TAG_TYPE)value * (TIME_TAG_TYPE)1000;
                  break;
            }
            if (accessStatus == ARS_NORMAL)
               status = EXIT_ROUTINE;
            else
               status = (CEI_INT16)accessStatus;
         }
         else
            status = ARS_INVHARVAL;
         break;

      case ARU_BYPASS_INIT_WRAP_TEST:
         status = ar_bypass_wrap_test (board, value);
         if (status == ARS_NORMAL)
            status = EXIT_ROUTINE;
         break;
      
      case ARU_MULTITHREAD_PROTECT:
         status = ar_set_multithread_protect (board, value);
         if (status == ARS_NORMAL)
            status = EXIT_ROUTINE;
         break;

      case ARU_RX_TIMETAG_MODE:
         if ( (value == AR_TIMETAG_EXT_IRIG_64BIT) && (deviceGlobal[board]->irigInstalled == FALSE) )
            status = ARS_INVHARVAL;
         else if (value < AR_TIMER_X20_COMPAT_32BIT) {
            deviceGlobal[board]->deviceTimeTagReference = (CEI_UINT32)value;
            status = EXIT_ROUTINE;
         }
         else
            status = ARS_INVHARVAL;
         break;

      case ARU_573_RX_ENABLE:
      case ARU_573_RX_AUTO_DETECT:
      case ARU_573_RX_BPRZ_SELECT:
      case ARU_573_RX_SUBFRAME_SIZE:
      case ARU_573_TX_SLEW_RATE:
      case ARU_573_SYNC_WORD1:
      case ARU_573_SYNC_WORD2:
      case ARU_573_SYNC_WORD3:
      case ARU_573_SYNC_WORD4:
         /* if a 573 item was referenced, let the 573 Set Config routine handle it */
         status = ar_set_573_config (board, item, value);
         if (status == ARS_NORMAL)
            status = EXIT_ROUTINE;
         break;

      case ARU_INSERT_INT_Q_ENTRY:
         accessStatus = cei_write_32bit(board, &device[board].devicePtr[CEIDEV_INT_QUEUE_HEAD_PTR], value);
         status = EXIT_ROUTINE;
         break;

      default : status = ARS_INVARG; 
         break;
   }
   
#ifndef TARGET_EMBEDDED
   /* validate the specified channel */
   if (status == ARS_NORMAL) {
      status = a429_utl_validate_channel_request (board, deviceChannel, item);
   }
#endif

   if (status == ARS_NORMAL) {
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                   offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE));

      /* read the device register with respect to the attribute specified */
      accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);

      if (accessStatus == ARS_NORMAL) {

         /* now adjust the respective bit field and output the new configuration */
         switch (item) {
            case ARU_RECV_MODE:
            case ARU_RX_PARITY:
            case ARU_RX_BITRATE:
            case ARU_RX_MERGED_MODE:
               /* When assigning Receive Configuration Register attributes, force
                * an Rx channel FIFO disable prior to modification of attribute, 
                * then restore the FIFO enable state afterwards
                */
               accessStatus = cei_write_32bit (board, boardPointer, (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
               if (accessStatus != ARS_NORMAL)
                  break;
               
               if (item == ARU_RECV_MODE) {
                  if (value == AR_WRAP_ON)
                     bitFieldValue |= CEIDEV_INTERNAL_WRAP_ENABLE;
                  else if (value == AR_WRAP_OFF)
                     bitFieldValue &= CEIDEV_INTERNAL_WRAP_MASK;
                  else
                     status = ARS_INVHARVAL;
               }
               else if (item == ARU_RX_BITRATE) {
                  bitFieldValue &= CEIDEV_BAUD_RATE_MASK;
   
                  if (value == ARU_SPEED_HIGH)
                     bitFieldValue |= CEIDEV_BAUD_RATE_100K << CEIDEV_BAUD_RATE_SHIFT;
                  else if (value == ARU_SPEED_LOW)
                     bitFieldValue |= CEIDEV_BAUD_RATE_12pt5K << CEIDEV_BAUD_RATE_SHIFT;
                  else {  
                     /* non-standard baud rates are allowed, where rate = 16,000,000 / value+2 */
                     bitFieldValue |= (value << CEIDEV_BAUD_RATE_SHIFT) & CEIDEV_BAUD_RATE_FIELD;
                  }
               }
               else if (item == ARU_RX_MERGED_MODE) {
                  if (value == AR_ON)
                     bitFieldValue |= CEIDEV_MERGE_MODE_ENABLE;
                  else if (value == AR_OFF)
                     bitFieldValue &= CEIDEV_MERGE_MODE_ENABLE_MASK;
                  else
                     status = ARS_INVHARVAL;
               }
               else {   /* item == ARU_RX_PARITY */
                  if ( (value == ARU_PARITY_ODD) || (value == ARU_PARITY_EVEN) || (value == AR_ON) )
                     bitFieldValue |= CEIDEV_PARITY_ENABLE;
                  else if ( (value == ARU_PARITY_NONE) || (value == AR_OFF) || (value == AR_RAW) )
                     bitFieldValue &= CEIDEV_PARITY_MASK;
                  else
                     status = ARS_INVHARVAL;
               }
               break;
   
            case ARU_RX_DISABLE:
               if (value == AR_ON)
                  bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
               else if (value == AR_OFF)
                  bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
               else
                  status = ARS_INVHARVAL;
               break;
         
            case ARU_TX_BITRATE:
            case ARU_TX_PARITY:
            case ARU_TX_BIT_ERROR:
            case ARU_TX_GAP_ERROR:
            case ARU_FAST_SLEW_RATE:
               /* When assigning Transmit Configuration Register attributes, force
                * a Tx channel FIFO disable prior to modification of attribute, 
                * then restore the FIFO enable state afterwards
                */
               accessStatus = cei_write_32bit (board, boardPointer, (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
               if (accessStatus != ARS_NORMAL)
                  break;
               
               /* non-standard baud rates are allowed, where rate = 16,000,000 / value+2 */
               if (item == ARU_TX_BITRATE) {
                  bitFieldValue &= CEIDEV_BAUD_RATE_MASK;
   
                  if (value == ARU_SPEED_HIGH) {   /* automatically gets fast slew rate */
                     bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                     bitFieldValue |= CEIDEV_BAUD_RATE_100K << CEIDEV_BAUD_RATE_SHIFT;
                  }
                  else if (value == ARU_SPEED_LOW) {   /* automatically gets slow slew rate */
                     bitFieldValue |= CEIDEV_BAUD_RATE_12pt5K << CEIDEV_BAUD_RATE_SHIFT;
                     bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                  }
                  else {  
                     /* non-standard baud rates are allowed, where rate = 16,000,000 / value+2 */
                     bitFieldValue |= (value << CEIDEV_BAUD_RATE_SHIFT) & CEIDEV_BAUD_RATE_FIELD;
                     if (value < CEIDEV_BAUD_RATE_15K)
                        bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                     else
                        bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                  }
               }
               else if (item == ARU_TX_BIT_ERROR) {
                  if (bitFieldValue & CEIDEV_429_TXCFG_GAP_ERROR)
                     bitFieldValue &= ~(CEIDEV_429_TXCFG_BIT_COUNT_LO | CEIDEV_429_TXCFG_BIT_COUNT_HI);
                  else
                     bitFieldValue &= ~(CEIDEV_429_TXCFG_BIT_COUNT_LO | CEIDEV_429_TXCFG_BIT_COUNT_HI | CEIDEV_429_TXCFG_PARAM_ENABLE);

                  if (value == AR_LO)
                     bitFieldValue |= CEIDEV_429_TXCFG_BIT_COUNT_LO + CEIDEV_429_TXCFG_PARAM_ENABLE;
                  else if (value == AR_HI)
                     bitFieldValue |= CEIDEV_429_TXCFG_BIT_COUNT_HI + CEIDEV_429_TXCFG_PARAM_ENABLE;
                  else if (value != AR_OFF)
                     status = ARS_INVHARVAL;
               }
               else if (item == ARU_TX_GAP_ERROR) {
                  if (bitFieldValue & (CEIDEV_429_TXCFG_BIT_COUNT_LO | CEIDEV_429_TXCFG_BIT_COUNT_HI))
                     bitFieldValue &= ~(CEIDEV_429_TXCFG_GAP_ERROR);
                  else
                     bitFieldValue &= ~(CEIDEV_429_TXCFG_GAP_ERROR | CEIDEV_429_TXCFG_PARAM_ENABLE);

                  if (value == AR_ON) 
                     bitFieldValue |= CEIDEV_429_TXCFG_GAP_ERROR + CEIDEV_429_TXCFG_PARAM_ENABLE;
                  else if (value != AR_OFF)
                     status = ARS_INVHARVAL;
               }
               else if (item == ARU_FAST_SLEW_RATE) {
                  if (value == AR_ON) 
                     bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                  else if (value == AR_OFF)
                     bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                  else
                     status = ARS_INVHARVAL;
               }
               else {   /* item == ARU_TX_PARITY */
                  bitFieldValue &= CEIDEV_PARITY_SELECT_MASK;
   
                  if ( (value == ARU_PARITY_NONE) || (value == AR_OFF) || (value == AR_RAW) )
                     bitFieldValue &= CEIDEV_PARITY_MASK;
                  else if (value == ARU_PARITY_ODD)
                     bitFieldValue |= CEIDEV_PARITY_ENABLE;
                  else if (value == ARU_PARITY_EVEN)
                     bitFieldValue |= CEIDEV_PARITY_SELECT_EVEN + CEIDEV_PARITY_ENABLE;
                  else
                     status = ARS_INVHARVAL;
               }
               break;
   
            case ARU_TX_DISABLE:
               if (value == AR_OFF) {
                  bitFieldValue &= CEIDEV_TRANSMIT_DISABLE_MASK;
                  bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
               }
               else if (value == AR_ON) {
                  bitFieldValue |= CEIDEV_TRANSMIT_DISABLE;
               }
               else
                  status = ARS_INVHARVAL;
               break;
         
            case ARU_RX_FIFO_ENABLE:
            case ARU_TX_FIFO_ENABLE:
               if (value == AR_OFF)
                  bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
               else if (value == AR_ON)
                  bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
               else
                  status = ARS_INVHARVAL;
               break;
         
   
            case ARU_DISCRETE_OUT1:
            case ARU_DISCRETE_OUT2:
            case ARU_DISCRETE_OUT:
                  if ( (value == AR_ON) || (value == AR_HI) )
                     bitFieldValue = 0;
                  else if ( (value == AR_OFF) || (value == AR_LO) )
                     bitFieldValue = 1;
                  else
                     status = ARS_INVHARVAL;
               break;
   
            case ARU_DIFFERENTIAL_OUT:
                  if (value == AR_ON)
                     bitFieldValue = CEIDEV_DIFF_OUT_ENABLE;
                  else if (value == AR_OFF)
                     bitFieldValue = 0;
                  else if (value == AR_LO)
                     bitFieldValue = 1;
                  else if (value == AR_HI)
                     bitFieldValue = 0;
                  else
                     status = ARS_INVHARVAL;
               break;
   
            case ARU_DIGITAL_OUT:
               if ( (value == AR_ON) || (value == AR_HI) )
                  bitFieldValue = 1;
               else if ( (value == AR_OFF) || (value == AR_LO) )
                  bitFieldValue = 0;
               else
                  status = ARS_INVHARVAL;
               break;
   
            case ARU_RX_RESET:
            case ARU_TX_RESET:
               bitFieldValue = (CEI_UINT32)value;
               break;
         }
      }

      /* update the configuration register with the specified attribute setting */
      if (accessStatus == ARS_NORMAL)
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
      
      if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
         status = (CEI_INT16)accessStatus;
   }
   
   if (status == EXIT_ROUTINE)
      status = ARS_NORMAL;

   /* if multi-thread protection requested, free semaphore/mutex-locking */
   if (device[board].multiThreadProtect) {
      if (status == ARS_NORMAL)
            status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
         else
            tempStatus = cei_release_thread_lock(LOCK_IDENT);
   }

   #ifdef INTEGRITY_PCI_PPC
      if (tempStatus) ;
   #endif
   return (status);
}


/*===========================================================================*
 * ENTRY POINT:             A R _ S E T _ 5 7 3 _ C O N F I G
 *===========================================================================*
 *
 * FUNCTION:    Set ARINC 573 channel configuration register attributes
 *
 * PARAMETERS:  CEI_INT16 board -- (input) board to reference (0 to 15)
 *              CEI_INT16 item  -- (input) config register attribute to set
 *              CEI_INT32 value -- (input) value applied to item attribute
 * 
 *                  item                 valid values
 * 
 *              ARU_RECV_MODE            AR_WRAP_OFF, AR_WRAP_ON
 *              ARU_573_RX_AUTO_DETECT   AR_OFF, AR_ON
 *              ARU_573_RX_BPRZ_SELECT   AR_OFF, AR_ON
 *              ARU_573_TX_BPRZ_SELECT   AR_OFF, AR_ON
 *              ARU_573_TX_HBP_SELECT    AR_OFF, AR_ON
 *
 *              ARU_RX_BITRATE &         ARU_573_RATE_SIZE_384_32    
 *              ARU_TX_BITRATE           ARU_573_RATE_SIZE_768_64    
 *                                       ARU_573_RATE_SIZE_1536_128  
 *                                       ARU_573_RATE_SIZE_3072_256  
 *                                       ARU_573_RATE_SIZE_6144_512  
 *                                       ARU_573_RATE_SIZE_12288_1024
 *                                       ARU_573_RATE_SIZE_24576_2048
 *                                       ARU_573_RATE_SIZE_49152_4096
 * 
 *              ARU_573_TX_SLEW_RATE     ARU_573_TX_SLEW_1PT5
 *                                       ARU_573_TX_SLEW_10PT0
 * 
 *              ARU_TX_DISABLE           AR_OFF, AR_ON
 *              ARU_TX_FIFO_ENABLE       AR_OFF, AR_ON
 *              ARU_RX_FIFO_ENABLE       AR_OFF, AR_ON
 *              ARU_573_RX_ENABLE        AR_OFF, AR_ON
 * 
 *              ARU_573_SYNC_WORD1       value from 0x0000 to 0x0fff
 *              ARU_573_SYNC_WORD2       value from 0x0000 to 0x0fff
 *              ARU_573_SYNC_WORD3       value from 0x0000 to 0x0fff
 *              ARU_573_SYNC_WORD4       value from 0x0000 to 0x0fff
 *              ARU_RX_MERGED_MODE       AR_ON, AR_OFF
 * 
 *              (Following supported for legacy CEI-710 API only)
 *              ARU_RX_BITRATE &         ARU_573_SPEED_384  
 *              ARU_TX_BITRATE           ARU_573_SPEED_768  
 *                                       ARU_573_SPEED_1536 
 *                                       ARU_573_SPEED_3072 
 *                                       ARU_573_SPEED_6144 
 *                                       ARU_573_SPEED_12288
 *                                       ARU_573_SPEED_24576
 * 
 *              ARU_573_RX_SUBFRAME_SIZE ARU_573_SUB_FRAME_64  
 *                                       ARU_573_SUB_FRAME_128 
 *                                       ARU_573_SUB_FRAME_256 
 *                                       ARU_573_SUB_FRAME_512 
 *                                       ARU_573_SUB_FRAME_1024
 *                                       ARU_573_SUB_FRAME_2048
 *                                       ARU_573_SUB_FRAME_4096
 *                                       ARU_573_SUB_FRAME_8192
 * 
 * RETURN VAL:  CEI_INT16 status -- ARS_NORMAL      Normal succesful completion.
 *                                  ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                                  ARS_INVARG      "item" parameter value is not supported by this 
 *                                                  function or device configuration. 
 *                                  ARS_INVHARVAL   "value" parameter value is not supported for "item"
 *                                  ARS_INVBOARD    board parameter value invalid or
 *                                                     board not initialized or ARINC 
 *                                                     573 not supported on this device
 * 
 * GLOBALS:     deviceGlobal[board]
 *
 * DESCRIPTION: This function assigns the respective device configuration
 *              register attributes based on the combined item/value
 *              parameter selections.  It is designed to support the ARINC
 *              573/717 protocol configuration register attributes only.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_573_config 
( 
   CEI_INT16 board,                 /* (input)  board to reference (0 to 15) */
   CEI_INT16 item,                  /* (input)  item to process              */
   CEI_INT32 value                  /* (input)  state of the item            */
)
{
   CEI_UINT32 deviceChannel;         /* current channel to select            */
   CEI_UINT32 bitFieldValue;         /* temp for device register read content*/
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_INT16 status = ARS_NORMAL;           /* function return status        */
   CEI_INT32 tempStatus = ARS_NORMAL;           /* local temp return status  */
   CEI_UINT32 offsetIntoChannelWindow = 0;  /* offset into chan mem region   */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      status = ARS_INVBOARD;
   else {
#endif
      deviceChannel = (CEI_UINT32)0;  /* here for future growth w/channel parm */
   
      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         status = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
         if (status != ARS_NORMAL)
            return (status);
      }

      /* Assign the h/w-respective channel number and type, then validate 
       * the channel referenced for the specific config item parameter value */
      switch (item) {
         case ARU_RECV_MODE:
         case ARU_RX_BITRATE:
         case ARU_RX_DISABLE:
         case ARU_RX_FIFO_ENABLE:
         case ARU_RX_MERGED_MODE:
         case ARU_573_RX_ENABLE:
         case ARU_573_RX_AUTO_DETECT:
         case ARU_573_RX_BPRZ_SELECT:
         case ARU_573_RX_SUBFRAME_SIZE:
            deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
      
         case ARU_TX_BITRATE:
         case ARU_TX_DISABLE:
         case ARU_TX_FIFO_ENABLE:
         case ARU_573_TX_SLEW_RATE:
         case ARU_573_TX_BPRZ_SELECT:
         case ARU_573_TX_HBP_SELECT:
            deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX];
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_1;
            break;
   
         case ARU_573_SYNC_WORD1:
         case ARU_573_SYNC_WORD2:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_2;
            deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
            break;

         case ARU_573_SYNC_WORD3:
         case ARU_573_SYNC_WORD4:
            offsetIntoChannelWindow = CEIDEV_CHANNEL_CONFIG_3;
            deviceChannel += deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX];
            break;

         default: 
            #if defined DEBUG_PROGRAMMING
               #include "dbg_pcie.h"
            #else
            status = ARS_INVARG; 
            #endif
            break;
      }
      
#ifndef TARGET_EMBEDDED
      /* validate the specified channel */
      if (status == ARS_NORMAL)
         status = a429_utl_validate_channel_request (board, deviceChannel, item);
   
      if (status == ARS_NORMAL) {
#endif
         /* for all control register attribute assignments, first read the 
          * device register with respect to the attribute specified 
          */
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      offsetIntoChannelWindow + (deviceChannel * CEIDEV_CHANNEL_REGION_SIZE));

         accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
         
         if (accessStatus == ARS_NORMAL) {
            switch (item) {
               case ARU_RECV_MODE:
               case ARU_RX_BITRATE:
               case ARU_RX_MERGED_MODE:
               case ARU_573_RX_SUBFRAME_SIZE:
               case ARU_573_RX_AUTO_DETECT:
               case ARU_573_RX_BPRZ_SELECT:
                  /* When assigning Receive Configuration Register attributes, force
                   * a Rx channel FIFO disable prior to modification of attribute, 
                   * then restore the FIFO enable state afterwards
                   */
                  accessStatus = cei_write_32bit (board, boardPointer, (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                  
                  if (item == ARU_RECV_MODE) {
                     if (value == AR_WRAP_ON)
                        bitFieldValue |= CEIDEV_INTERNAL_WRAP_ENABLE;
                     else if (value == AR_WRAP_OFF)
                        bitFieldValue &= CEIDEV_INTERNAL_WRAP_MASK;
                     else
                        status = ARS_INVHARVAL;
                  }
                  else if (item == ARU_RX_BITRATE) {
                     bitFieldValue &= CEIDEV_573_FRAME_SPEC_MASK;
                     /* assign the ARINC 717 h/w bus speed selection,  
                      * supporting legacy CEI-7x0 API options for bus speed */
                     switch (value) {
                        case ARU_573_RATE_SIZE_384_32:
                        case ARU_573_SPEED_384:
                           bitFieldValue |= CEIDEV_573_FRAME_384_32;
                           break;
                        case ARU_573_RATE_SIZE_768_64:
                        case ARU_573_SPEED_768:
                           bitFieldValue |= CEIDEV_573_FRAME_768_64;
                           break;
                        case ARU_573_RATE_SIZE_1536_128:
                        case ARU_573_SPEED_1536:
                           bitFieldValue |= CEIDEV_573_FRAME_1536_128;
                           break;
                        case ARU_573_RATE_SIZE_3072_256:
                        case ARU_573_SPEED_3072:
                           bitFieldValue |= CEIDEV_573_FRAME_3072_256;
                           break;
                        case ARU_573_RATE_SIZE_6144_512:
                        case ARU_573_SPEED_6144:
                           bitFieldValue |= CEIDEV_573_FRAME_6144_512;
                           break;
                        case ARU_573_RATE_SIZE_12288_1024:
                        case ARU_573_SPEED_12288:
                           bitFieldValue |= CEIDEV_573_FRAME_12288_1024;
                           break;
                        case ARU_573_RATE_SIZE_24576_2048:
                        case ARU_573_SPEED_24576:
                           bitFieldValue |= CEIDEV_573_FRAME_24576_2048;
                           break;
                        case ARU_573_RATE_SIZE_49152_4096:
                           bitFieldValue |= CEIDEV_573_FRAME_49152_4096;
                           break;
                        default:
                           status = ARS_INVHARVAL;
                     }
                  }
                  /* The following values are supported for legacy CEI-710 API only */
                  else if (item == ARU_573_RX_SUBFRAME_SIZE) {
                     bitFieldValue &= CEIDEV_573_FRAME_SPEC_MASK;
                     switch (value) {
                        case ARU_573_SUB_FRAME_64:
                           bitFieldValue |= CEIDEV_573_FRAME_768_64;
                           break;
                        case ARU_573_SUB_FRAME_128:
                           bitFieldValue |= CEIDEV_573_FRAME_1536_128;
                           break;
                        case ARU_573_SUB_FRAME_256:
                           bitFieldValue |= CEIDEV_573_FRAME_3072_256;
                           break;
                        case ARU_573_SUB_FRAME_512:
                           bitFieldValue |= CEIDEV_573_FRAME_6144_512;
                           break;
                        case ARU_573_SUB_FRAME_1024:
                           bitFieldValue |= CEIDEV_573_FRAME_12288_1024;
                           break;
                        case ARU_573_SUB_FRAME_2048:
                           bitFieldValue |= CEIDEV_573_FRAME_24576_2048;
                           break;
                        default:
                           status = ARS_INVHARVAL;
                     }
                  }
                  else if (item == ARU_573_RX_AUTO_DETECT) {
                     if (value == AR_OFF)
                        bitFieldValue |= CEIDEV_573_RX_MODE_RAW;
                     else if (value == AR_ON)
                        bitFieldValue &= CEIDEV_573_RX_MODE_AUTO_MASK;
                     else
                        status = ARS_INVHARVAL;
                  }
                  else if (item == ARU_573_RX_BPRZ_SELECT) {
                     if (value == AR_ON)
                        bitFieldValue |= CEIDEV_573_RX_BPRZ_SELECT;
                     else if (value == AR_OFF)
                        bitFieldValue &= CEIDEV_573_RX_ENCODING_MASK;
                     else
                        status = ARS_INVHARVAL;
                  } 
                  else if (item == ARU_RX_MERGED_MODE) {
                     if (value == AR_ON)
                        bitFieldValue |= CEIDEV_MERGE_MODE_ENABLE;
                     else if (value == AR_OFF)
                        bitFieldValue &= CEIDEV_MERGE_MODE_ENABLE_MASK;
                     else
                        status = ARS_INVHARVAL;
                  }
                  break;
      
               case ARU_RX_DISABLE:
                  if (value == AR_ON)
                     bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  else if (value == AR_OFF)
                     bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  else
                     status = ARS_INVHARVAL;
                  break;
            
               case ARU_TX_BITRATE:
               case ARU_573_TX_BPRZ_SELECT:
               case ARU_573_TX_HBP_SELECT:
               case ARU_573_TX_SLEW_RATE:
                  /* When assigning Transmit Configuration Register attributes, force
                   * a Tx channel FIFO disable prior to modification of attribute, 
                   * then restore the FIFO enable state afterwards
                   */
                  accessStatus = cei_write_32bit (board, boardPointer, (bitFieldValue & CEIDEV_CHANNEL_ENABLE_MASK));
                  if (accessStatus != ARS_NORMAL)
                     break;
                  
                  if (item == ARU_TX_BITRATE) {
                     bitFieldValue &= CEIDEV_573_FRAME_SPEC_MASK;
                     /* assign the ARINC 717 h/w bus speed selection,
                      * supporting legacy CEI-7x0 API options for bus speed */
                     switch (value) {
                        case ARU_573_RATE_SIZE_384_32:
                        case ARU_573_SPEED_384:
                           bitFieldValue |= CEIDEV_573_FRAME_384_32;
                           break;
                        case ARU_573_RATE_SIZE_768_64:
                        case ARU_573_SPEED_768:
                           bitFieldValue |= CEIDEV_573_FRAME_768_64;
                           break;
                        case ARU_573_RATE_SIZE_1536_128:
                        case ARU_573_SPEED_1536:
                           bitFieldValue |= CEIDEV_573_FRAME_1536_128;
                           break;
                        case ARU_573_RATE_SIZE_3072_256:
                        case ARU_573_SPEED_3072:
                           bitFieldValue |= CEIDEV_573_FRAME_3072_256;
                           break;
                        case ARU_573_RATE_SIZE_6144_512:
                        case ARU_573_SPEED_6144:
                           bitFieldValue |= CEIDEV_573_FRAME_6144_512;
                           break;
                        case ARU_573_RATE_SIZE_12288_1024:
                        case ARU_573_SPEED_12288:
                           bitFieldValue |= CEIDEV_573_FRAME_12288_1024;
                           break;
                        case ARU_573_RATE_SIZE_24576_2048:
                        case ARU_573_SPEED_24576:
                           bitFieldValue |= CEIDEV_573_FRAME_24576_2048;
                           break;
                        case ARU_573_RATE_SIZE_49152_4096:
                           bitFieldValue |= CEIDEV_573_FRAME_49152_4096;
                           break;
                        default:
                           status = ARS_INVHARVAL;
                     }
                  }
                  else if (item == ARU_573_TX_SLEW_RATE) {
   
                     if (value == ARU_573_TX_SLEW_1PT5)
                        bitFieldValue |= CEIDEV_FAST_SLEW_RATE;
                     else if (value == ARU_573_TX_SLEW_10PT0)
                        bitFieldValue &= CEIDEV_SLEW_RATE_MASK;
                     else
                        status = ARS_INVHARVAL;
                  }
                  else if (item == ARU_573_TX_HBP_SELECT) {
                     if (value == AR_ON) {
                        deviceGlobal[board]->transmitMode = ARU_XMIT_573_HBP;
                        bitFieldValue |= CEIDEV_573_TX_HBP_SELECT;
                     }
                     else if (value == AR_OFF)
                        bitFieldValue &= CEIDEV_573_TX_HBP_MASK;
                     else
                        status = ARS_INVHARVAL;
                  }
                  else if (item == ARU_573_TX_BPRZ_SELECT) {
                     if (value == AR_ON) {
                        deviceGlobal[board]->transmitMode = ARU_XMIT_573_BPRZ;
                        bitFieldValue |= CEIDEV_573_TX_BPRZ_SELECT;
                     }
                     else if (value == AR_OFF)
                        bitFieldValue &= CEIDEV_573_TX_BPRZ_MASK;
                     else
                        status = ARS_INVHARVAL;
                  }
                  break;
         
               case ARU_TX_DISABLE:
                  if (value == AR_OFF)
                     bitFieldValue &= CEIDEV_TRANSMIT_DISABLE_MASK;
                  else if (value == AR_ON)
                     bitFieldValue |= CEIDEV_TRANSMIT_DISABLE;
                  else
                     status = ARS_INVHARVAL;
                  break;
            
               case ARU_573_RX_ENABLE:
               case ARU_RX_FIFO_ENABLE:
               case ARU_TX_FIFO_ENABLE:
                  if (value == AR_OFF)
                     bitFieldValue &= CEIDEV_CHANNEL_ENABLE_MASK;
                  else if (value == AR_ON)
                     bitFieldValue |= CEIDEV_CHANNEL_ENABLE;
                  else
                     status = ARS_INVHARVAL;
                  break;
         
               case ARU_573_SYNC_WORD1:
               case ARU_573_SYNC_WORD2:
               case ARU_573_SYNC_WORD3:
               case ARU_573_SYNC_WORD4:
                  a717syncWords[board][item-ARU_573_SYNC_WORD1] = (CEI_UINT32)value;
      
                  if ((item == ARU_573_SYNC_WORD2) || (item == ARU_573_SYNC_WORD4)) {
                     bitFieldValue &= CEIDEV_573_DATA_MASK;
                     bitFieldValue += (CEI_UINT32)(value & CEIDEV_573_DATA_MASK) << CEIDEV_573_SYNCH_SHIFT;
                  }
                  else {
                     bitFieldValue &= CEIDEV_573_DATA_MASK << CEIDEV_573_SYNCH_SHIFT;
                     bitFieldValue += (CEI_UINT32)(value & CEIDEV_573_DATA_MASK);
                  }
                  break;
      
               default: 
                  status = ARS_INVHARVAL; 
                  break;
            }
         }

         /* update the configuration register with the specified attribute setting */
         if (accessStatus == ARS_NORMAL)
            accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
         
         if (status == EXIT_ROUTINE)
            status = ARS_NORMAL;
      
         if ((status == ARS_NORMAL) && (accessStatus != ARS_NORMAL))
            status = (CEI_INT16)accessStatus;
#ifndef TARGET_EMBEDDED
      }
   }
#endif

   /* if multi-thread protection requested, free semaphore/mutex-locking */
   if (device[board].multiThreadProtect) {
      if (status == ARS_NORMAL)
            status = (CEI_INT16)cei_release_thread_lock(LOCK_IDENT);
         else
            tempStatus = cei_release_thread_lock(LOCK_IDENT);
   }
   
   #ifdef INTEGRITY_PCI_PPC
      if (tempStatus) ;
   #endif
   return status;
}


/*===========================================================================*
 * ENTRY POINT:    A R _ S E T _ M U L T I T H R E A D _ P R O T E C T
 *===========================================================================*
 *
 * FUNCTION:    Enable/Disable Mutex Protection around device channel access
 * 
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *              CEI_INT16 state  -- (input) AR_ON / AR_OFF 
 *
 * GLOBALS:     none
 * 
 * RETURN VAL:  ARS_NORMAL       - Success
 *              ARS_INVARG       - state value not supported
 * 
 * DESCRIPTION: This function controls the execution of Win32 Mutex 
 *      protection around every Receive and Transmit Buffer access in the API.
 *      ON enables Mutex protection, OFF disables it.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_multithread_protect 
( 
   CEI_INT16 board,                  /* (input) board to reference (0 to 15) */ 
   CEI_INT16 state                   /* (input) AR_ON / AR_OFF               */
)
{
   CEI_INT16 status = ARS_NORMAL;                 /* function return status  */

   if (board < MAX_BOARDS) {
      if (state == AR_OFF) 
         device[board].multiThreadProtect = AR_CONC_NONE;
      else if (state == AR_ON)
         device[board].multiThreadProtect = AR_CONC_MULTITHRD;
      else {
         status = ARS_INVARG;
      }
   }
   else
      status = ARS_INVBOARD;
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:        A R _ S E T _ I S R _ F U N C T I O N
 *===========================================================================*
 *
 * FUNCTION:    Define an application supplied ISR routine
 *
 * PARAMETERS:  deviceIndex   -- (input) device to reference (0 to 15)
 *              function      -- (input) void function to execute from ISR
 *
 * RETURN VAL:  ARS_NORMAL    -- Success
 *              ARS_INVBOARD  -- Invalid device ID parameter
 *              ARS_INVARG    -- State parameter value is invalid
 *
 * GLOBALS:     none
 *
 * DESCRIPTION: This routine assigns the supplied function pointer to an
 *              array of pointers indexed by device id.  This function will
 *              be invoked from cei_utl_interrupt_handler() instead of
 *              executing the "flush the h/w interrupt queue" processing.
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_set_isr_function 
(
   CEI_INT32 deviceIndex,             /* the device identifier               */
   pCEI_VOID function                 /* multithread protect state selection */
)
{
   CEI_INT32 status = ARS_NORMAL;                /* function return status   */

   if (deviceIndex < MAX_BOARDS) {
      if (function != NULL)
         x30InterruptFunctionPtr[deviceIndex] = function;
      else
         status = ARS_INVARG;
   }
   else
      status = ARS_INVBOARD;

   return status;
}


/*===========================================================================*
 * ENTRY POINT:        A R _ S E T _ P R E L O A D _ C O N F I G
 *===========================================================================*
 *
 * FUNCTION:    Update a pre-load configuration setting.
 *
 * PARAMETERS:  CEI_INT16 board   -- (input) board to reference (0 to 15)
 *              CEI_INT16 item    -- (input) the item to define, with
 *                                           valid selections from:
 *                                              ARU_CONCURRENCY_MODE           
 *              CEI_UINT32 value  -- (input) the value to define the item 
 *                                           valid selections from:
 *                                              AR_CONC_NONE
 *                                              AR_CONC_MULTITHRD
 *                                              AR_CONC_MULTIPROC
 *           
 * RETURN VAL:  ARS_NORMAL   -- semaphore was opened successfully
 *              ARS_INVBOARD -- invalid board number 
 *              ARS_INVARG   -- invalid item or value parameter
 *              ARS_NO_OS_SUPPORT -- item not supported on host o/s
 *              ARS_BOARD_MUTEX -- failed to create/open the semaphore
 *
 * DESCRIPTION: Call this routine before calling AR_LOADSLV to update the 
 *              value of a particular load configuration setting.  This 
 *              routine should not be called after the calling process has 
 *              executed AR_LOADSLV.
 *
 *              If 'item' is ARU_CONCURRENCY_MODE, the 'value' parameter 
 *              specifies the board concurrency mode.  One of three modes may
 *              be selected: AR_CONC_NONE, AR_CONC_MULTITHRD, or 
 *              AR_CONC_MULTIPROC.  Note that some modes are only supported
 *              on certain boards and operating systems.  Refer to the
 *              AR_SET_PRELOAD_CONFIG documentation in the CEI-x30 User's 
 *              Manual to determine if a particular concurrency mode is 
 *              supported on a certain board and operating system.
 *
 *              The default concurrency mode, AR_CONC_NONE, provides no multi-
 *              thread protection and no multi-process support.  The user 
 *              application must ensure that at most one thread is calling 
 *              into the API at any given time, and only a single process may 
 *              interface with a particular board.
 *
 *              If AR_CONC_MULTITHRD concurrency mode is selected, thread
 *              protection is provided internally within the API.  The user 
 *              application may call into the API from multiple threads, but
 *              all threads must belong to a single process.  The main user
 *              application thread should initialize the board with a call to
 *              AR_LOADSLV before other threads attempt to call into the API.
 *
 *              If AR_CONC_MULTIPROC concurrency mode is selected, thread
 *              protection is provided internally within the API and multiple
 *              processes may interface with a single board.  If any process 
 *              requests multi-process mode, all other processes must also
 *              request multi-process mode.
 *
 *              Some coordination is required when using multiple processes in
 *              conjunction with a single CEI-x30 device.  A single "board" 
 *              process should be launched to configure board-specific and 
 *              application-wide attributes and invoke data processing using
 *              AR_GO.  After the "board" process launches the board it is
 *              then permissible to launch one or more "channel" processes.
 *              A "channel" process should first attach to the board via 
 *              AR_LOADSLV, then execute operations strictly confined to the 
 *              particular channel(s) that process is associated with (e.g., 
 *              AR_GETWORD, AR_PUTWORD, AR_SET_CONFIG, etc.).  When finished,
 *              the "channel" process should invoke AR_CLOSE and terminate.
 *              The "board" process should remain running until all "channel"
 *              processes have terminated; afterwhich, it should invoke 
 *              AR_CLOSEand terminate.
 *
 *              All processes must exercise caution when modifying board
 *              settings that may impact other concurrently-running processes.
 *
 *              Note that board setup/initialization is only executed in 
 *              AR_LOADSLV if no other processes have the board open.  If
 *              another process has the board open (that is, if another process
 *              has opened the board using AR_LOADSLV but hasn't yet closed the
 *              board using AR_CLOSE), AR_LOADSLV attaches to the device
 *              without re-initializing the board or modifying board settings.
 *              Similarly, AR_CLOSE only shuts down the board if no other 
 *              processes have the board open.  If another process has the 
 *              board open, AR_CLOSE detaches from the board without shutting 
 *              it down.  Thus, board settings are preserved across "channel" 
 *              process invocations of AR_LOADSLV and AR_CLOSE.  Note the 
 *              importance of matching every call to AR_LOADSLV with a call to
 *              AR_CLOSE in multi-process mode.
 *              
 *              Multi-process mode is only required when accessing a single 
 *              board from multiple processes.  If multiple boards are 
 *              installed, AR_CONC_NONE concurrency mode can be used as long
 *              as only one process interfaces with each board.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_preload_config
(
   CEI_INT16  board,                 /* (input) board to reference (0 to 15) */ 
   CEI_INT16  item,                  /* (input) the item to define           */
   CEI_UINT32 value                  /* (input) the value to define the item */
)
{
   CEI_INT16 status = ARS_NORMAL;        /* function return status           */

   /* make sure the board number is valid */
#ifndef TARGET_EMBEDDED
   if (board < MAX_BOARDS) {
#endif
      /* process the configuration command */
      if (item == ARU_CONCURRENCY_MODE) {
         switch (value) {
            case AR_CONC_MULTIPROC:
#if defined (VXWORKS) || defined (INTEGRITY_PCI_PPC)
               status = ARS_NO_OS_SUPPORT;
               break;
#else
               status = (CEI_INT16)cei_open_preload_lock (board);
#endif
            case AR_CONC_MULTITHRD:
            case AR_CONC_NONE:
               device[board].multiThreadProtect = value;
               break;
            default:
               status = ARS_INVARG;
               break;
         }
      }
      else {
         status = ARS_INVARG;
      }
#ifndef TARGET_EMBEDDED
   }
   else   
      status = ARS_INVBOARD;
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:             A R _ S E T _ R A W _ M O D E
 *===========================================================================*
 *
 * FUNCTION:    Enable/Disable ARINC 429 Parity Mode
 * 
 * PARAMETERS:  CEI_INT16 board     -- (input) board to reference (0 to 15)
 *              CEI_INT16 direction -- (input) TRANSMIT_CHANNEL / RECEIVE_CHANNEL
 *                                         alternates: ARU_XMIT / ARU_RECV
 *              CEI_INT16 channel   -- (input) channel number (0 to 7)
 *              CEI_INT16 control   -- (input) AR_ON / AR_OFF
 *
 * GLOBALS:     none
 * 
 * RETURN VAL:  CEI_INT16     -- ARS_NORMAL     Normal succesful completion.
 *                               ARS_INVARG     direction, channel or control  
 *                                              parameter value is invalid
 *                               ARS_INVBOARD   Board parameter value invalid or
 *                                              board has not been initialized
 *                               ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 * 
 * DESCRIPTION: This routine assigns the state of the transmitter and receiver
 *              parity control bits based on the specified state of "Raw Mode"; 
 *              where raw mode ON disables parity for the respective channel 
 *              and OFF enables ODD parity by default.  The OFF mode defaults 
 *              to ODD PARITY.
 * 
 *              This routine is provided for x20 compatibility.  The proper
 *              method of assigning parity is via ar_set_device_config().
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_raw_mode 
( 
   CEI_INT16 board,            /* (input) board to reference (0 to 15)       */
   CEI_INT16 direction,        /* (input) TRANSMIT_CHANNEL / RECEIVE_CHANNEL */
   CEI_INT16 channel,          /* (input) channel number (0 to 7)            */
   CEI_INT16 control           /* (input) AR_ON / AR_OFF                     */
)
{
   CEI_INT16 item = ARU_TX_PARITY;      /* Rx/Tx parity select               */
   CEI_INT16 value = ARU_PARITY_NONE;   /* Odd/None parity enable            */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   
   if ( control == AR_ON )
      value = ARU_PARITY_NONE;
   else if ( control == AR_OFF )
      value = ARU_PARITY_ODD;
   else
      status = ARS_INVARG;

   if ( (direction == TRANSMIT_CHANNEL) || (direction == ARU_XMIT) )
      item = ARU_TX_PARITY;
   else if ( (direction == RECEIVE_CHANNEL) || (direction == ARU_RECV) )
      item = ARU_RX_PARITY;
   else
      status = ARS_INVARG;

   /* Command the respective channel to the specified parity. */
   if (status == ARS_NORMAL)
      status = ar_set_device_config ( board, channel, item, value );
   
   return status;
}


/*===========================================================================*
 * ENTRY POINT:      A R _ S E T _ S T O R A G E _ M O D E
 *===========================================================================*
 *
 * FUNCTION:    Set board-level receive data storage mode
 *
 * PARAMETERS:  CEI_INT16 board -- (input) board index (0 to 15)
 *              CEI_INT16 mode  -- (input) valid values are:
 *                                 ARU_BUFFERED  individual buffer per channel
 *                                 ARU_DEDICATED individual buffer per channel
 *                                 ARU_MERGED    one merged circular buffer
 *
 * RETURN VAL:  ARS_NORMAL      Success
 *              ARS_INVARG      "mode" parameter value is not supported
 *              ARS_INVBOARD    Invalid board parameter or board not initialized
 *              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine assigns the respective storage mode to all
 *              receiver channels, all Merged or all Buffered.
 *
 *              This routine is provided for backward compatibility with
 *              the CEI-710 API.  The proper method to assign the Merged
 *              Receive Mode is provided for individual channels via the 
 *              respective channel set-config routine using the item
 *              parameter value ARU_RX_MERGED_MODE and its ON/OFF selection.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_storage_mode 
(  
   CEI_INT16 board,    /* (input) board index (0 to 15)                      */
   CEI_INT16 mode      /* (input) storage mode (ARU_BUFFERED or ARU_MERGED)  */
) 
{
   CEI_INT32 channelIndex;              /* local channel index               */
   CEI_INT32 total429ChannelCount;      /* total number of installed 429 ch's*/
   CEI_INT32 mergedSelect = AR_ON;      /* channel merged mode selection     */
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif

      if ((mode == ARU_BUFFERED) || (mode == ARU_DEDICATED))
         mergedSelect = AR_OFF;
      else if (mode == ARU_MERGED)
         mergedSelect = AR_ON;
#ifndef TARGET_EMBEDDED
      else
         status = ARS_INVARG;

      if (status == ARS_NORMAL) {
#endif
         deviceGlobal[board]->receiveMode = (CEI_UINT32)mode;
         total429ChannelCount = deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX];

         for (channelIndex = 0; channelIndex < total429ChannelCount; channelIndex++) {
            ar_set_device_config (board, (CEI_INT16)channelIndex, ARU_RX_MERGED_MODE, (CEI_INT16)mergedSelect);
         }
         if ( deviceGlobal[board]->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717 ) {
            ar_set_573_config (board, ARU_RX_MERGED_MODE, mergedSelect);
         }
#ifndef TARGET_EMBEDDED
      }
   }
   else
      status = ARS_INVBOARD;
#endif

   return status;
}


/*===========================================================================*
 * ENTRY POINT:             A R _ S E T _ T I M E
 *===========================================================================*
 *
 * FUNCTION:    This routine sets either the CEI-830 internal 64-bit timer   
 *              or the IRIG output time based on an application-supplied value.
 *
 * PARAMETERS:  CEI_INT16 board -- (input)  Board number (0 to 15)  
 * 
 *              CEI_INT16 item  -- (input)  Format of value supplied and timer
 *                                      to be assigned (1usec timer or IRIG)
 * 
 *              pAR_TIMETAG_TYPE timeTagRef -- (input)
 *                             ->timeTagFormat -- timer to assign and format
 *                                                of the "timeTag" value
 *                                              AR_TIMETAG_EXT_IRIG_64BIT
 *                                              AR_TIMETAG_INT_USEC_64BIT
 * 
 *                             ->timeTag -- timer value to assign in the 
 *                                          format specified via "timeTagFormat"
 *                                          For timeTagFormat of ...IRIG... the
 *                                          format of timeTag is BCD, as defined
 *                                          in the User Manual for the IRIG time
 *                                          register.
 *                                          For timeTagFormat of ...USEC... the
 *                                          format of timeTag is a 64-bit integer 
 *                                          value with a 1 microsecond resolution.
 *
 *                             ->referenceTimeTag -- not used
 * 
 * RETURN VAL:  CEI_INT32    -- ARS_NORMAL      Success
 *                              ARS_INVBOARD    Invalid or uninitialized board
 *                              ARS_INVARG      Invalid timeTagFormat selection
 *                              ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     boardInitialized
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_time 
(
   CEI_INT16 board,                  /* (input) board to reference (0 to 15) */
   pAR_TIMETAG_TYPE timeTag          /* (input) selected timer and value     */
)
{
   CEI_UINT32 timeTagHigh;
   CEI_UINT32 timeTagLow;
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   
#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      accessStatus = ARS_INVBOARD;
   else {
#endif
      if (timeTag->timeTagFormat == AR_TIMETAG_INT_USEC_64BIT) {
      #if defined (NO_INT64)
         timeTagLow = (CEI_UINT32)(timeTag->timeTag);
         timeTagHigh = (CEI_UINT32)(timeTag->referenceTimeTag);
      #else   
         timeTagLow = (CEI_UINT32)(timeTag->timeTag);
         timeTagHigh = (CEI_UINT32)(timeTag->timeTag >> 32);
      #endif

         /* if multi-thread protection requested, semaphore/mutex-lock device access */
         if (device[board].multiThreadProtect) {
            accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         }

         if (accessStatus == ARS_NORMAL) {
            accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_LOW], timeTagLow);
            if (accessStatus == ARS_NORMAL)
               accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_TIMER_HIGH], timeTagHigh);
   
            /* if multi-thread protection requested, free semaphore/mutex-locking */
            if (device[board].multiThreadProtect) {
               accessStatus = cei_release_thread_lock(LOCK_IDENT);
            }
         }
      }
      else if (timeTag->timeTagFormat == AR_TIMETAG_EXT_IRIG_64BIT) {
         accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_IRIG_UPDATE], (CEI_UINT32)(timeTag->timeTag));
      }
      else
         accessStatus = ARS_INVARG;

#ifndef TARGET_EMBEDDED
   }
#endif
   
   return (CEI_INT16)accessStatus;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ S E T _ T I M E R R A T E
 *===========================================================================*
 *
 * FUNCTION:    This routine assigns the CEI-x20 compatible timer resolution
 *
 * PARAMETERS:  CEI_INT16 board -- (input) board index (0 to 15)
 *              CEI_INT16 rate  -- (input) resolution of the CEI-x20-type timer
 *                                         in 250 nanoseconds.  Note the CEI-x30
 *                                         message scheduler rate resolution is
 *                                         1ms, so the resolution of the rate
 *                                         value should be 4000 to match.
 * RETURN VAL:  none
 *
 * GLOBALS:     deviceGlobal[board]->deviceTimeTagReference
 * 
 *===========================================================================*/
EXPORT32 CEI_VOID  DLL_EXPORTED ar_set_timerrate 
(
   CEI_INT16 board,  /* (input) board index (0 to 15)                        */
   CEI_INT16 rate    /* (input) resolution of the CEI-x20-type timer (usec)  */
)
{
   if ((board < MAX_BOARDS) && (deviceGlobal[board] != NULL)) {
      deviceGlobal[board]->deviceTimeTagReference = AR_TIMER_X20_COMPAT_32BIT;
      if (rate > 0) {
         deviceGlobal[board]->deviceX20CompatibleResolution = (CEI_UINT32)rate;
      }
   }
}


/*===========================================================================*
 * ENTRY POINT:      A R _ S E T _ T R A N S M I T T E R _ M O D E
 *===========================================================================*
 *
 * FUNCTION:    Set the ARINC 717 transmitter mode
 *
 * PARAMETERS:  CEI_INT16 board -- (input) board index (0 to 15)
 *              CEI_INT16 mode  -- (input) valid values are:
 *
 * RETURN VAL:  ARS_NORMAL       Success
 *              ARS_INVHARVAL    "mode" parameter value is not supported
 *                               Or selected protocol is not available
 *              ARS_INVBOARD     Invalid board parameter or board not initialized
 *              ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine enables the respective transmitter definition
 *              for the ARINC 573/717 channel. 
 *
 *              This routine is provided for backward compatibility with
 *              the CEI-710 API.  The proper method to assign the ARINC 573
 *              Encoding as BPRZ or HBP is provided for individual channel(s) 
 *              via the ar_set_573_config routine using the item parameter
 *              values ARU_573_TX_HBP_SELECT and ARU_573_TX_BPRZ_SELECT.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_transmitter_mode 
(
   CEI_INT16 board,   /* (input)  board index (0 to 15)                      */
   CEI_INT16 mode     /* (input)  transmitter mode (CSDB 717-HBP or 717-BPRZ */
) 
{
   CEI_UINT32  bitFieldValue;           /* temp var for device register data */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   CEI_INT32   status = ARS_NORMAL;     /* function return status            */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
      if ((mode == ARU_XMIT_573_HBP) || (mode == ARU_XMIT_573_BPRZ)) {
#endif
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_TX] * CEIDEV_CHANNEL_REGION_SIZE));
         if (device[board].multiThreadProtect) {
            status = cei_acquire_thread_lock(LOCK_IDENT);
         }
         if (status == ARS_NORMAL) {
            status = cei_read_32bit (board, boardPointer, &bitFieldValue);
      
            if (status == ARS_NORMAL) {
               if (mode == ARU_XMIT_573_HBP) {
                  bitFieldValue &= CEIDEV_573_TX_BPRZ_MASK;
                  bitFieldValue |= CEIDEV_573_TX_HBP_SELECT + CEIDEV_CHANNEL_ENABLE;
               }
               else if (mode == ARU_XMIT_573_BPRZ) {
                  bitFieldValue &= CEIDEV_573_TX_HBP_MASK;
                  bitFieldValue |= CEIDEV_573_TX_BPRZ_SELECT + CEIDEV_CHANNEL_ENABLE;
               }
               status = cei_write_32bit (board, boardPointer, bitFieldValue);
               deviceGlobal[board]->transmitMode = mode;
            }
            if (device[board].multiThreadProtect) {
               status = cei_release_thread_lock(LOCK_IDENT);
            }
         }
#ifndef TARGET_EMBEDDED
      }
      else {
         status = ARS_INVHARVAL;
      }
   }
   else
      status = ARS_INVBOARD;
#endif

   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:             A R _ S T O P
 *===========================================================================*
 *
 * FUNCTION:    Disable the transmitter and receiver FIFOs on the device
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) board to reference (0 to 15)
 *
 * RETURN VAL:  CEI_INT16  -- ARS_NORMAL       Success
 *                            ARS_INVBOARD     Invalid board parameter
 *                                                or board not initialized
 *                            ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: Sets the global enable register global enable bit to DISABLE
 *              and disables the CEI-830 message scheduler
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_stop (CEI_INT16 board)
{
   CEI_UINT32 globalEnableRegData;      /* contents of the Global Enable reg */
   CEI_INT32  status = ARS_NORMAL;      /* function return status            */

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT) {
#endif
      if (device[board].multiThreadProtect) {
         status = cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (status == ARS_NORMAL) {
         status = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
         
         /* disable "global enable" */
         if (status == ARS_NORMAL) {
            globalEnableRegData &= CEIDEV_GLOBAL_ENABLE_MASK;
            status = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegData);
         }

         if (device[board].multiThreadProtect) {
            status = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
   }
   else
      status = ARS_INVBOARD;
#endif

   return (CEI_INT16)status;
}


/*===========================================================================*
 * ENTRY POINT:         A R _ V E R S I O N
 *===========================================================================*
 *
 * FUNCTION:    Return the software version
 * 
 * PARAMETERS:  pCEI_CHAR verstr -- (output) string containing version.
 * 
 * RETURN VAL:  CEI_VOID 
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: The software version API_VERSION defined in cdev_api.h is 
 *      copied to the string parameter (verstr) passed into this function.
 * 
 *===========================================================================*/ 
EXPORT32 CEI_VOID  DLL_EXPORTED ar_version (pCEI_CHAR verstr) 
{
   strcpy (verstr, API_VERSION);
}                  
    

/*===========================================================================*
 * ENTRY POINT:   A R _ W A I T 
 *===========================================================================*
 *
 * FUNCTION:    Wait for the specified number of seconds
 * 
 * PARAMETERS:  CEI_FLOAT nsecs -- (input) number of seconds to wait for
 * 
 * RETURN VAL:  CEI_VOID
 *
 * GLOBALS:     none
 * 
 * DESCRIPTION: Delay the specified number of seconds based on the OS
 *              timer function clock() and the value of TIMER_CLOCK_TICKS_PER_SECOND
 *              (see <time.h>)
 *                    
 *===========================================================================*/ 
EXPORT32 CEI_VOID  DLL_EXPORTED ar_wait (CEI_FLOAT nsecs) 
{
   cei_delay_seconds(nsecs);
}

/*===========================================================================* 
 * Compiler Directive selection for available integer support 32 vs 64-bit
 * for ar_convert_time_to_string
 *===========================================================================*/
#if defined (NO_INT64)

/* First is the 32-bit only integer support for limited operating systems    */
/*===========================================================================*
 * ENTRY POINT:    A R _ C O N V E R T _ T I M E _ T O _ S T R I N G   
 *===========================================================================*
 *
 * FUNCTION:    Convert standard CEI-830 64-bit time value to CEI_CHAR string
 * 
 * PARAMETERS:  CEI_INT16 board         --  (input)  Board number (0 to 15)  
 *              CEI_INT16 displayFormat --  (input)  Format for returned string:
 *                       AR_TD_REL_MIDNIGHT  Relative to Midnight Format and
 *                       AR_TD_IRIG          Full IRIG Format, both defined 
 *                                           as "(DDD)hh:mm:ss.uuuuuu"           
 *                       AR_TD_DATE          Date Format defined as
 *                                              "(MM/DD)hh:mm:ss.uuuuuu"
 * 
 *              pAR_TIMETAG_TYPE timeIn -- (input)  Source 64-bit time structure      
 *              pCEI_CHAR timeString    -- (output) Pointer to destination string    
 * 
 * RETURN VAL:  none
 *
 * DESCRIPTION: This routine converts the time value provided in the timeIn
 *              structure to a character string representation of date/time,
 *              format based on what is specified via the displayFormat parm.
 *              The supplied time format (LSB resolution) must be specified
 *              in the timeIn structure member "timeTagFormat", representing
 *              the resolution of the respective "timeTag" member data.
 *
 *===========================================================================*/  
EXPORT32 CEI_VOID  DLL_EXPORTED ar_convert_time_to_string 
( 
   CEI_INT16 board,               /* (input) Board number (0 to 15)          */
   CEI_INT16 displayFormat,       /* (input)  format of returned time string */
   pAR_TIMETAG_TYPE timeIn,       /* (input)  64-bit time-tag structure      */
   pCEI_CHAR timeString           /* (output) Pointer to store time string   */ 
)
{
   /***********************************************************************
   *  Local variables
   ***********************************************************************/
   CEI_NATIVE_INT days;                  /* Number of days in time tag, 0 to 365(6) */
   CEI_NATIVE_INT daysSinceJan;          /* Number of days in time tag, 0 to 365(6) */
   CEI_NATIVE_INT leapday;               /* Leap year extra day tracker             */
   CEI_NATIVE_INT hours;                 /* Number of hours, 0 to 32                */
   CEI_NATIVE_INT minutes;               /* Number of minutes, 0 to 59              */
   CEI_NATIVE_UINT seconds;              /* Number of seconds, 0 to 59              */
   CEI_NATIVE_UINT microseconds;         /* Number of microseconds, 0 to 4294967296 */
   time_t         now;                   /* Current time                            */
   struct         tm start;              /* Formatted "now" current time            */
   CEI_NATIVE_INT year;                  /* Current year (since 1900)               */
   CEI_NATIVE_INT month = 1;             /* The first month is January              */
   CEI_NATIVE_INT resulting_month = 1;   /* Just in case the day of the year should */
   CEI_NATIVE_INT resulting_day = 0;     /*   happen to be zero (which is illegal). */
   
   double floatingTimeValue;             /* temp IRIG time conversion vars   */
   double floatingMicroseconds;          /* temp IRIG time conversion vars   */
   double floatingSeconds;               /* temp IRIG time conversion vars   */
   
   if (timeIn->timeTagFormat == AR_TIMETAG_HOST_SEC_32BIT) {
      seconds = (CEI_NATIVE_UINT)timeIn->timeTag;
      microseconds = 0L;
   }
   else {
      /* first convert the incoming time value to have a 1 microsecond resolution */
      floatingTimeValue = (double)timeIn->referenceTimeTag * (double)0x10000000;
      floatingTimeValue = floatingTimeValue * (double)0x10;
      floatingTimeValue = floatingTimeValue + (double)timeIn->timeTag;
   
      if ((timeIn->timeTagFormat == AR_TIMETAG_INT_20USEC_32BIT))
         floatingTimeValue *= (double)20;
      else if ((timeIn->timeTagFormat == AR_TIMETAG_INT_MSEC_32BIT))
         floatingTimeValue *= (double)1000;
      else if ((timeIn->timeTagFormat == AR_TIMER_X20_COMPAT_32BIT))
         floatingTimeValue *= (double)(deviceGlobal[board]->deviceX20CompatibleResolution/AR_X20_SCALE_TIMETAG_VALUE);

      floatingMicroseconds = floatingTimeValue;

      if (floatingMicroseconds <= ((double)(31536000) * (double)1000000) ) {
         floatingMicroseconds = fmod(floatingTimeValue, (double)1000000);
         floatingSeconds = floatingTimeValue / (double)1000000;
         seconds = (CEI_NATIVE_UINT)floatingSeconds;
         microseconds = (CEI_NATIVE_UINT)floatingMicroseconds;
      }
      else {
         seconds = (CEI_NATIVE_UINT)0;
         microseconds = (CEI_NATIVE_UINT)0;
      }
   }
   
   /* Extract Minutes from Seconds */
   minutes = seconds/60;
   seconds = seconds%60;

   /* Extract Hours from Minutes */
   hours = minutes/60;  
   minutes = minutes%60;

   /* Extract Days from Hours */
   days = hours/24;     
   hours = hours%24;
   
   /* no such thing as day 0, days range from 1 to 365 (366 for leap yr) */
   days++;
   daysSinceJan = days;
   resulting_day = days;

   /* Acquire the current year */
   time(&now);               /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now); /* Local time in seconds, min, hour, day, month, year */
   year = (CEI_NATIVE_INT)start.tm_year + 1900;

   /* Determine if this is a leap year for total days in Feb */
   if ( ((year % 4) == 0) && (year != 2100) )
      leapday = 1;
   else
      leapday = 0;

   /* Compute the current month to extract current day-of-month */
   while ( days > 0 ) {
      resulting_day   = days;
      resulting_month = month;

      switch (month) {
         case 1:               /* Jan     */
            days -= 31; 
            break;             
         case 2:               /* Feb     */
            days -= (28 + leapday); 
            break;             
         case 3:               /* March   */
            days -= 31; 
            break;             
         case 4:               /* April   */
            days -= 30; 
            break;             
         case 5:               /* May     */
            days -= 31; 
            break;             
         case 6:               /* June    */
            days -= 30; 
            break;             
         case 7:               /* July    */
            days -= 31; 
            break;             
         case 8:               /* August  */
            days -= 31; 
            break;             
         case 9:               /* Sept    */
            days -= 30; 
            break;             
         case 10:              /* Oct     */
            days -= 31; 
            break;             
         case 11:              /* Nov     */
            days -= 30; 
            break;             
         case 12:              /* Dec     */
            days -= 31; 
            break;             
      }
      month++;  /* Go to next month */
   }

   /* Define the return string using only non-zero components (zero values omitted). */
   /* When all components are defined, format is based on the display format requested */
   switch ( displayFormat ) {
      case AR_TD_REL_MIDNIGHT: /* Relative to Midnight Format "(ddd)hh:mm:ss.uuuuuu" */
      case AR_TD_IRIG:         /* Full IRIG Format "(ddd)hh:mm:ss.uuuuuu" */
         sprintf(timeString,"(%ld)%2.2ld:%2.2ld:%2.2ld.%6.6ld", 
                 daysSinceJan, hours, minutes, seconds, microseconds);
         break;
   
      case AR_TD_DATE :        /* Date Format "(MM/dd)hh:mm:ss.uuuuuu" */
         sprintf(timeString,"(%d/%d)%2.2ld:%2.2ld:%2.2ld.%6.6ld", 
                 resulting_month, resulting_day, hours, minutes, seconds, microseconds);
         break;
   }
}

#else
/* Next is the 64-bit integer support for most operating systems             */

/*===========================================================================*
 * ENTRY POINT:    A R _ C O N V E R T _ T I M E _ T O _ S T R I N G   
 *===========================================================================*
 *
 * FUNCTION:    Convert standard CEI-830 64-bit time value to CEI_CHAR string
 * 
 * PARAMETERS:  CEI_INT16 board             (input)  Board number (0 to 15)  
 *              CEI_INT16 displayFormat     (input)  Format for returned string:
 *                       AR_TD_REL_MIDNIGHT  Relative to Midnight Format and
 *                       AR_TD_IRIG          Full IRIG Format, both defined 
 *                                           as "(DDD)hh:mm:ss.uuuuuu"           
 *                       AR_TD_DATE          Date Format defined as
 *                                              "(MM/DD)hh:mm:ss.uuuuuu"
 * 
 *              pAR_TIMETAG_TYPE timeIn (input)  Source 64-bit time structure      
 *              pCEI_CHAR timeString       (output) Pointer to destination string    
 * 
 * RETURN VAL:  none
 *
 * DESCRIPTION: This routine converts the time value provided in the timeIn
 *              structure to a character string representation of date/time,
 *              format based on what is specified via the displayFormat parm.
 *              The supplied time format (LSB resolution) must be specified
 *              in the timeIn structure member "timeTagFormat", representing
 *              the resolution of the respective "timeTag" member data.
 *
 *===========================================================================*/  
EXPORT32 CEI_VOID  DLL_EXPORTED ar_convert_time_to_string 
( 
   CEI_INT16 board,               /* (input) Board number (0 to 15)          */
   CEI_INT16 displayFormat,       /* (input)  format of returned time string */
   pAR_TIMETAG_TYPE timeIn,       /* (input)  64-bit time-tag structure      */
   pCEI_CHAR timeString           /* (output) Pointer to store time string   */ 
)
{
   /***********************************************************************
   *  Local variables
   ***********************************************************************/
   CEI_NATIVE_INT days;                  /* Number of days in time tag, 0 to 365(6) */
   CEI_NATIVE_INT daysSinceJan;          /* Number of days in time tag, 0 to 365(6) */
   CEI_NATIVE_INT leapday;               /* Leap year extra day tracker             */
   CEI_NATIVE_INT hours;                 /* Number of hours, 0 to 32                */
   CEI_NATIVE_INT minutes;               /* Number of minutes, 0 to 59              */
   CEI_NATIVE_UINT seconds;              /* Number of seconds, 0 to 59              */
   CEI_NATIVE_UINT microseconds;         /* Number of microseconds, 0 to 4294967296 */
   time_t   now;                         /* Current time                            */
   struct   tm start;                    /* Formatted "now" current time            */
   CEI_NATIVE_INT year;                  /* Current year (since 1900)               */
   TIME_TAG_TYPE timeTag64Value;         /* temp 64-bit time value placeholder      */
   CEI_NATIVE_INT month = 1;             /* The first month is January              */
   CEI_NATIVE_INT resulting_month = 1;   /* Just in case the day of the year should */
   CEI_NATIVE_INT resulting_day = 0;     /*   happen to be zero (which is illegal). */
   TIME_TAG_TYPE maxMicroSecondsPerYear; /* Max number of usec in this year         */

   /* first convert the incoming time value to have a 1 microsecond resolution */
   timeTag64Value = timeIn->timeTag;
   switch (timeIn->timeTagFormat) {
      case AR_TIMETAG_INT_20USEC_32BIT: timeTag64Value *= (TIME_TAG_TYPE)20; break;
      case AR_TIMETAG_INT_MSEC_32BIT:   timeTag64Value *= (TIME_TAG_TYPE)1000; break;
      case AR_TIMETAG_HOST_SEC_32BIT:   timeTag64Value *= (TIME_TAG_TYPE)AR_MICROSECONDS_PER_SECOND; break;
      case AR_TIMER_X20_COMPAT_32BIT:   
         timeTag64Value *= (TIME_TAG_TYPE)(deviceGlobal[board]->deviceX20CompatibleResolution/AR_X20_SCALE_TIMETAG_VALUE); 
         break;
      default: ;
   }
   
   /* Acquire the current year */
   time(&now);               /* Time in seconds since midnight Jan 1, 1970 GMT */
   start = *localtime(&now); /* Local time in seconds, min, hour, day, month, year */
   year = (CEI_NATIVE_INT)start.tm_year + AR_LOCAL_TIME_YEAR_BASIS;

   /* Determine if this is a leap year for total days in Feb and max microseconds/year */
   if ( ((year % AR_LEAP_YEAR_DIVISOR) == 0) && (year != AR_LEAP_YEAR_EXCEPTION) ) {
      leapday = 1;
      maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_LEAP_YEAR;
   }
   else {
      leapday = 0;
      maxMicroSecondsPerYear = AR_IRIG_ROLLOVER_NON_LEAP_YEAR;
   }

   /* Only process a valid time-tag value (within the max # microseconds/year) */
   if (timeTag64Value <= maxMicroSecondsPerYear) {
      microseconds = (CEI_NATIVE_UINT)(timeTag64Value % AR_MICROSECONDS_PER_SECOND);
      seconds = (CEI_NATIVE_UINT)(timeTag64Value / AR_MICROSECONDS_PER_SECOND);
   
      /* Extract Minutes from Seconds */
      minutes = seconds/60;
      seconds = seconds%60;
   
      /* Extract Hours from Minutes */
      hours = minutes/60;  
      minutes = minutes%60;
   
      /* Extract Days from Hours */
      days = hours/24;     
      hours = hours%24;    
   
      /* no such thing as day 0, days range from 1 to 365 (366 for leap yr) */
      days++;
      daysSinceJan = days;
      resulting_day = days;

      /* Compute the current month to extract current day-of-month */
      while ( days > 0 ) {
         resulting_day   = days;
         resulting_month = month;
   
         switch (month) {
            case 1:               /* Jan     */
               days -= 31; 
               break;             
            case 2:               /* Feb     */
               days -= (28 + leapday); 
               break;             
            case 3:               /* March   */
               days -= 31; 
               break;             
            case 4:               /* April   */
               days -= 30; 
               break;             
            case 5:               /* May     */
               days -= 31; 
               break;             
            case 6:               /* June    */
               days -= 30; 
               break;             
            case 7:               /* July    */
               days -= 31; 
               break;             
            case 8:               /* August  */
               days -= 31; 
               break;             
            case 9:               /* Sept    */
               days -= 30; 
               break;             
            case 10:              /* Oct     */
               days -= 31; 
               break;             
            case 11:              /* Nov     */
               days -= 30; 
               break;             
            case 12:              /* Dec     */
               days -= 31; 
               break;             
         }
         month++;  /* Go to next month */
      }
   }
   else {
      resulting_day   = 1;
      resulting_month = 1;
      daysSinceJan = 1;
      hours = 0;
      minutes = 0;
      seconds = 0;
      microseconds = 0;
   }

   /* Define the return string using only non-zero components (zero values omitted). */
   /* When all components are defined, format is based on the display format requested */
   switch ( displayFormat ) {
      case AR_TD_REL_MIDNIGHT:      /* Relative to Midnight Format "(ddd)hh:mm:ss.uuuuuu" */
      case AR_TD_IRIG:              /* Full IRIG Format "(ddd)hh:mm:ss.uuuuuu" */
         sprintf(timeString,"(%d)%2.2d:%2.2d:%2.2d.%6.6d", daysSinceJan, hours, minutes, seconds, microseconds);
         break;
   
      case AR_TD_DATE :             /* Date Format "(MM/dd)hh:mm:ss.uuuuuu" */
         sprintf(timeString,"(%d/%d)%2.2d:%2.2d:%2.2d.%6.6d", resulting_month, resulting_day, hours, minutes, seconds, microseconds);
         break;
   }
}

#endif
/* End of the 32-bit vs. 64-bit integer support conditional compilation      */


/*===========================================================================*  
 * CEI-x20 API Functions - Providing Limited Compatibility from x20 to x30
 *===========================================================================*/             
EXPORT32 CEI_INT16 DLL_EXPORTED ar_get_raw_mode (CEI_INT16 board, CEI_INT16 direction, CEI_INT16 channel) 
{
   CEI_INT16 status;
   CEI_INT16 parity;
   if (direction == ARU_RECV) {
      status = ar_get_device_config (board, channel, ARU_RX_PARITY, &parity);
      parity = (CEI_INT16)((parity == (CEI_INT16)AR_ON) ? AR_OFF : AR_ON);
   }
   else {
      status = ar_get_device_config (board, channel, ARU_TX_PARITY, &parity);
      parity = (CEI_INT16)((parity == (CEI_INT16)ARU_PARITY_NONE) ? AR_ON : AR_OFF);
   }
   #ifdef INTEGRITY_PCI_PPC
      if (status) ;
   #endif
   return parity;
}

/*===========================================================================*
 * ENTRY POINT:            A R _ I N I T _ D U A L _ P O R T
 *===========================================================================*
 *
 * FUNCTION:    Initialize the device to a ready state
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) device number to reference (0-15)
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL       Success
 *                              ARS_MEMWRERR     Memory test failure
 *                              ARS_INVBOARD     Invalid board parameter
 *                              ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine initializes the device I/O configuration 
 *              registers to a state ready for use by the host application.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_init_dual_port (CEI_INT16 board) 
{
   return ar_initialize_device(board);
}
/*===========================================================================*
 * ENTRY POINT:            A R _ I N I T _ S L A V E
 *===========================================================================*
 *
 * FUNCTION:    Initialize the device transmit and receive buffers
 *
 * PARAMETERS:  CEI_INT16 board  -- (input) device number to reference (0-15)
 *
 * RETURN VAL:  CEI_INT16    -- ARS_NORMAL       Success
 *                              ARS_INVBOARD     Invalid board parameter
 *                              ARS_BOARD_MUTEX  Access to the Board Lock timed-out/failed
 *
 * GLOBALS:     deviceGlobal[board]
 * 
 * DESCRIPTION: This routine initializes/flushes the device's transmit and 
 *              receive buffers.
 * 
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_init_slave (CEI_INT16 board)
{
   CEI_UINT32 labelIndex;               /* local label filter table index    */
   CEI_UINT32 globalEnableRegValue;     /* temp var for device register data */
   CEI_UINT32 bitFieldValue;            /* temp var for device register data */
   CEI_UINT32 configRegValue[64];       /* temp var for device register data */
   CEI_UINT32 deviceChannel;            /* temp var for device register data */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   pCEI_UINT32 configPointer;        /* local pointer to channel memory area */
   pCEI_UINT32 bufferPointer;        /* local pointer to channel memory area */
   CEI_INT16  status = ARS_NORMAL;      /* local invocation return status    */
   CEI_INT32  accessStatus = ARS_NORMAL;/* status of device read/write access*/

#ifndef TARGET_EMBEDDED
   if ( (board < MAX_BOARDS) && (deviceGlobal[board] != NULL) && (deviceGlobal[board]->attachedProcessCount > 0) ) {
#endif
      /* if multi-thread protection requested, acquire semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_acquire_thread_lock(LOCK_IDENT);
         if (accessStatus != ARS_NORMAL) {
            return (CEI_INT16)accessStatus;
         }
      }

      /*------------------------------------------------------------------------------------------*
       *  Enable data processing 
       *------------------------------------------------------------------------------------------*/
      accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegValue);
      bitFieldValue = globalEnableRegValue | CEIDEV_GLOBAL_ENABLE_ON;
      accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], bitFieldValue);
   
      /*--------------------------------------------------------------------*
       * Flush all ARINC 429 transmit buffers using the following process:
       *   Save the original transmit channel configuration
       *   Disable external transmission
       *   Set the receive rate to 8Mb/s
       *   Delay long enough to flush the entire buffer
       *--------------------------------------------------------------------*/
      if (deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX] > 0) {
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE));
         for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX], labelIndex = 0;  
              deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
              deviceChannel++, labelIndex++) {
            accessStatus = cei_read_32bit  (board, boardPointer, &configRegValue[labelIndex]);
            bitFieldValue = 0x00018041;
            accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
            boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
         }
      }
      ar_sleep (5);
      /* restore the original transmit channel configuration */
      if (deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX] > 0) {
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX] * CEIDEV_CHANNEL_REGION_SIZE));
         for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX], labelIndex = 0;  
              deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_TX]; 
              deviceChannel++, labelIndex++) {
            accessStatus = cei_write_32bit (board, boardPointer, configRegValue[labelIndex]);
            boardPointer += CEIDEV_CHANNEL_REGION_SIZE;
         }
      }
   
      /*--------------------------------------------------------------------*
       * if present, flush the 717 receiver
       *--------------------------------------------------------------------*/
      if ( deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC573_RX] > 0 ) {
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
         boardPointer += CEIDEV_CHANNEL_STATUS;
         /* clear any outstanding error status bits */
         bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_FIFO_INVALID_MSG;
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
         while (bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
            accessStatus = cei_read_32bit (board, bufferPointer, &labelIndex);
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);
         }
         /* now enable auto-sync */
         bitFieldValue = CEIDEV_573_FRAME_768_64 + CEIDEV_573_RX_BPRZ_SELECT;
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_CONFIG_1 +
                                      (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC573_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
      }
   
      /*--------------------------------------------------------------------*
       *  Enable and Flush the ARINC 429 receivers
       *--------------------------------------------------------------------*/
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                   (deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX] * CEIDEV_CHANNEL_REGION_SIZE));
      configPointer = boardPointer + CEIDEV_CHANNEL_CONFIG_1;
      bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
      if (deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX] > 0) {
         for (deviceChannel = deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
              deviceChannel <= deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_ARINC429_RX]; 
              deviceChannel++) {
            /* clear any outstanding error status bits */
            bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_FIFO_INVALID_MSG;
            accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);    /* read the status register - check for presense of data */
            while (bitFieldValue & CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
               accessStatus = cei_read_32bit (board, bufferPointer, &labelIndex);   /* read the data register to reset status for this message */
               accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue); /* read the status register - check for presense of more data */
            }
            boardPointer  += CEIDEV_CHANNEL_REGION_SIZE;                            /* setup for next channel */
            configPointer += CEIDEV_CHANNEL_REGION_SIZE;
            bufferPointer += CEIDEV_CHANNEL_REGION_SIZE;
         }
      
         /*--------------------------------------------------------------------*
          * Flush the merged mode receiver FIFO
          *--------------------------------------------------------------------*/
         boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START +
                                      (deviceGlobal[board]->lastChannel[CEIDEV_CH_TYPE_MERGED_RX] * CEIDEV_CHANNEL_REGION_SIZE));
         bufferPointer = boardPointer + CEIDEV_CHANNEL_BUFFER_3;
         /* clear any outstanding error status bits */
         bitFieldValue = CEIDEV_CSR_RECV_FIFO_NOT_EMPTY + CEIDEV_CSR_FIFO_INVALID_MSG;
         accessStatus = cei_write_32bit (board, boardPointer, bitFieldValue);
         accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue);    /* read the status register - check for presense of data */
         while (bitFieldValue == CEIDEV_CSR_RECV_FIFO_NOT_EMPTY) {
            accessStatus = cei_read_32bit (board, bufferPointer, &labelIndex);   /* read the data register to reset status for this message */
            accessStatus = cei_read_32bit (board, boardPointer, &bitFieldValue); /* read the status register - check for presense of more data */
         }
      }
      /* restore the global enable register */
      accessStatus = cei_write_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], globalEnableRegValue);

      /* if multi-thread protection requested, free semaphore/mutex-locking */
      if (device[board].multiThreadProtect) {
         accessStatus = cei_release_thread_lock(LOCK_IDENT);
      }
#ifndef TARGET_EMBEDDED
   }
   else
      status = ARS_INVBOARD;
#endif

   return status;
}

EXPORT32 CEI_INT16 DLL_EXPORTED ar_int_control (CEI_INT16 board, CEI_INT16 channel, CEI_INT16 flag) 
{
   CEI_INT16 status;
   if (flag == ARU_ENABLE_INT)
      status = ar_set_device_config (board, 0, ARU_HW_INTERRUPT_ENABLE, AR_ON);
   else
      status = ar_set_device_config (board, 0, ARU_HW_INTERRUPT_ENABLE, AR_OFF);
   return status;
}

EXPORT32 CEI_VOID  DLL_EXPORTED ar_sleep (CEI_UINT32 sleep_ms)
{
   cei_delay_milliseconds(sleep_ms);
}

EXPORT32 CEI_VOID  DLL_EXPORTED ar_recreate_parity (pCEI_UCHAR word) 
{
   CEI_NATIVE_INT i;         /* loop index                    */
   CEI_UCHAR parity;         /* stores the word parity        */

   /* calculate parity for the word  */
   for (i = 0,parity = (CEI_UCHAR)0x80; i < 4; i++)
      parity += (CEI_UCHAR)byte_parity[word[i]];

   /* set/clear the parity bit  */
   if (parity != (CEI_UCHAR)0)
      word[3] |= 0x80;
   else
      word[3] &= 0x7F;
}

EXPORT32 CEI_INT16 DLL_EXPORTED ar_int_set (CEI_INT16 board, CEI_INT32 int_num) {return ARS_NORMAL;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_reset_int (CEI_INT16 board)  {return ARS_NORMAL;}
EXPORT32 CEI_INT32 DLL_EXPORTED ar_setinterrupts (CEI_UINT32 cardnum, CEI_UINT32 channel, CEI_INT32 TransRcv, CEI_UINT32 Count, CEI_UINT32 Mask) {return ARS_NORMAL;}

/*===========================================================================*
 * ENTRY POINT:              A R _ S E T C H P A R M S
 *===========================================================================*
 *
 * FUNCTION:    Define the operating parameters and modes for an ARINC 429
 *              receiver and/or transmitter.  This routine provides compatibility
 *              with the CEI-x20 API method for assigning channel configurations.
 *
 * PARAMETERS:  CEI_UINT32 cardnum    (i) The board number of interest
 *              CEI_UINT32 channel    (i) The channel to which to set the data (0-based)
 *              CEI_INT32 TransRcv    (i) Set parameters for ARU_TRANSMITTER or ARU_RECEIVER
 *              pAR_CHANNEL_PARMS def (i) Pointer to channel parameter structure
 *
 * RETURN VAL:  CEI_INT32 status ARS_NORMAL      Normal succesful completion.
 *                               ARS_INVARG      Invalid channel parameter value 
 *                                                  was supplied for the TransRcv value
 *                               ARS_INVHARVAL   Structure member value is invalid
 *                               ARS_INVBOARD    Board parameter value invalid or
 *                                                  board has not been initialized
 *                               ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *
 * DESCRIPTION: This function is used to configure an ARINC 429 transmitter or
 *              receiver.
 *
 * GLOBALS:     deviceGlobal[board]
 *
 *===========================================================================*/
EXPORT32 CEI_INT32 DLL_EXPORTED ar_setchparms 
(
   CEI_UINT32 cardnum, 
   CEI_UINT32 channel, 
   CEI_INT32 TransRcv, 
   pAR_CHANNEL_PARMS def
)
{
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   CEI_INT32 locLongValue;              /* temp 32bit variable               */
   CEI_INT16 board = (CEI_INT16)cardnum;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II)
      return ARS_INVBOARD;
#endif

   if ( (TransRcv == ARU_TRANSMITTER) && (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) )
      def->ChannelType = CHAN_ARINC429;
   else if ( (TransRcv == ARU_RECEIVER) && (channel < deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_RX]) )
      def->ChannelType = CHAN_ARINC429;
   else if (deviceGlobal[board]->alternateJconfiguration == CEIDEV_ALT_CONFIGURATION_717)
      def->ChannelType = CHAN_ARINC717;
   else
      def->ChannelType = CHAN_NOT_IMPLEMENTED;

   if (def->ChannelType == CHAN_ARINC429) {
      if (TransRcv == ARU_TRANSMITTER) {
         /* first assign the specified bus speed */
         if (def->BaudRate > ARU_SPEED_LOW) 
            locLongValue = (AR_SIXTEEN_MHZ / (CEI_INT32)def->BaudRate) - 2;
         else
            locLongValue = def->BaudRate;
   
         status = ar_set_device_config (board, (CEI_INT16)channel, ARU_TX_BITRATE, (CEI_INT16)locLongValue);
         
         /* next assign the channel parity */
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (board, (CEI_INT16)channel, ARU_TX_PARITY, (CEI_INT16)def->Parity);
         }
      }
      else if (TransRcv == ARU_RECEIVER) {
         /* first assign the specified bus speed */
         if (def->BaudRate > ARU_SPEED_LOW) 
            locLongValue = (AR_SIXTEEN_MHZ / (CEI_INT32)def->BaudRate) - 2;
         else
            locLongValue = def->BaudRate;
      
         status = ar_set_device_config (board, (CEI_INT16)channel, ARU_RX_BITRATE, (CEI_INT16)locLongValue);
         
         /* next assign the channel parity */
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (board, (CEI_INT16)channel, ARU_RX_PARITY, (CEI_INT16)def->Parity);
         }
         /* next assign the receiver wrap selection */
         if (status == ARS_NORMAL) {
            status = ar_set_device_config (board, (CEI_INT16)channel, ARU_RECV_MODE, (CEI_INT16)def->InternalWrap);
         }
         /* next assign the receiver storage mode selection */
         if (status == ARS_NORMAL) {
            status = ar_set_storage_mode (board, (CEI_INT16)def->OperateMode);
         }
      }
      else
         status = ARS_INVARG;
   }
   else if (def->ChannelType == CHAN_ARINC717) {
      if (TransRcv == ARU_TRANSMITTER) {
         /* assign the specified bus speed */
         locLongValue = (def->BaudRate >> 8) & 0x0000000F;
         status = ar_set_573_config (board, ARU_TX_BITRATE, locLongValue);
         
         /* next assign the channel encoding */
         if (status == ARS_NORMAL) {
            if (def->HWEncoding == ARU_717_H_BI_PHASE) {
               status = ar_set_573_config (board, ARU_573_TX_BPRZ_SELECT, AR_OFF);
               status = ar_set_573_config (board, ARU_573_TX_HBP_SELECT, AR_ON);
            }
            else {
               status = ar_set_573_config (board, ARU_573_TX_BPRZ_SELECT, AR_ON);
               status = ar_set_573_config (board, ARU_573_TX_HBP_SELECT, AR_OFF);
            }
         }
      }
      else if (TransRcv == ARU_RECEIVER) {
         /* assign the specified bus speed */
         locLongValue = (def->BaudRate >> 8) & 0x0000000F;
         status = ar_set_573_config (board, ARU_RX_BITRATE, locLongValue);
         
         /* next assign autosync selection */
         if (status == ARS_NORMAL) {
            locLongValue = (def->AutoSync == ARU_717_AUTOSYNC) ? AR_ON : AR_OFF;
            status = ar_set_573_config (board, ARU_573_RX_AUTO_DETECT, locLongValue);
         }
         /* next assign the receiver wrap selection */
         if (status == ARS_NORMAL) {
            status = ar_set_573_config (board, ARU_RECV_MODE, (CEI_INT16)def->InternalWrap);
         }
         /* next assign the receiver storage mode selection */
         if (status == ARS_NORMAL) {
            status = ar_set_storage_mode (board, (CEI_INT16)def->OperateMode);
         }

         if (status == ARS_NORMAL) {
            locLongValue = (def->HWEncoding == ARU_717_BIPOLAR) ? AR_ON : AR_OFF;
            status = ar_set_573_config (board, ARU_573_RX_BPRZ_SELECT, locLongValue);
         }

         if (status == ARS_NORMAL) {
            status = ar_set_573_config (board, ARU_573_SYNC_WORD1, (CEI_INT16)def->Sync1);
            status = ar_set_573_config (board, ARU_573_SYNC_WORD2, (CEI_INT16)def->Sync2);
            status = ar_set_573_config (board, ARU_573_SYNC_WORD3, (CEI_INT16)def->Sync3);
            status = ar_set_573_config (board, ARU_573_SYNC_WORD4, (CEI_INT16)def->Sync4);
         }
      }
      else
         status = ARS_INVARG;
   }

   return status;
}

/*===========================================================================*
 * ENTRY POINT:         A R _ X M I T _ S Y N C
 *===========================================================================*
 *
 * FUNCTION:    This routine is used to wait for all of the data in the
 *              transmit queue to be sent.  It is useful for a program
 *              that wants to reset the slave, but needs to wait for it
 *              to complete transmission before halting it.  It will
 *              self-terminate if the transmit FIFO hasn't completed
 *              flushing all messages after 12 seconds.
 *
 * PARAMETERS:  CEI_INT16 cardnum  -- (input) The board number of interest
 *              CEI_INT16 channel  -- (input) The transmit channel
 *
 * RETURN VAL:  CEI_INT16   -- ARS_NORMAL      Success
 *                             ARS_INVBOARD    Board parameter value invalid or
 *                                                board has not been initialized
 *                             ARS_NOSYNC      Twelve seconds elapsed without
 *                                                completing the Tx FIFO flush
 *                             ARS_BOARD_MUTEX Access to the Board Lock timed-out/failed
 *                             ARS_NODATA      The board is not in active processing mode
 *
 * GLOBALS:     deviceGlobal[board].boardInitialized
 *              device[board].arincDevicePtr
 *
 * DESCRIPTION: This is a utility that waits for all the data in the transmit
 *              buffer to be loaded into the ARINC transmitter.  It is useful
 *              in an application that is sending data out but doesn't want
 *              to halt the interface until everything has been sent.
 *
 *===========================================================================*/
EXPORT32 CEI_INT16 DLL_EXPORTED ar_xmit_sync 
(
   CEI_INT16 board,            /* (input)  board to reference (0 to 15)      */
   CEI_INT16 channel           /* (input)  transmit channel to use (0 or 1)  */
) 
{
   CEI_INT32 localChannel;     /* channel offset into the label filter table */
   CEI_UINT32 channelStatus;            /* TX channel status register value  */
   CEI_UINT32 globalEnableRegData;      /* contents of the Global Enable reg */
   CEI_INT32 accessStatus = ARS_NORMAL; /* status of device read/write access*/
   CEI_INT16 status = ARS_NORMAL;       /* function return status            */
   pCEI_UINT32 boardPointer;         /* local pointer to channel memory area */
   AR_TIMETAG_TYPE boardTime;
   AR_TIMETAG_TYPE exitTime;

#ifndef TARGET_EMBEDDED
   if (VALID_BOARD_AND_INIT_II) 
      status = ARS_INVBOARD;
   else if ((CEI_UINT32)channel >= deviceGlobal[board]->numberOfChannels[CEIDEV_CH_TYPE_ARINC429_TX]) {
      status = ARS_INVARG;
   }
   else {
#endif
      status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &exitTime); 
      exitTime.timeTag += 12000000;

      localChannel = (CEI_INT32)channel + (CEI_INT32)deviceGlobal[board]->firstChannel[CEIDEV_CH_TYPE_ARINC429_TX];
      boardPointer = (pCEI_UINT32)(device[board].devicePtr + CEIDEV_CHANNEL_REGION_START + CEIDEV_CHANNEL_STATUS +
                                   (localChannel * CEIDEV_CHANNEL_REGION_SIZE));

      /* if multi-thread protection requested, semaphore/mutex-lock device access */
      if (device[board].multiThreadProtect) {
         accessStatus = (CEI_INT16)cei_acquire_thread_lock(LOCK_IDENT);
      }
      if (accessStatus == ARS_NORMAL) {
         accessStatus = cei_read_32bit (board, &device[board].devicePtr[CEIDEV_GLOBAL_ENABLE], &globalEnableRegData);
         if ((globalEnableRegData & CEIDEV_GLOBAL_ENABLE_MASK) == 0) {
            status = ARS_NODATA;
         }
         else {
            accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
            channelStatus = (channelStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
            
            /* read the channel status continuously until it indicates the FIFO is empty */
            while (channelStatus > 0) {
               accessStatus = cei_read_32bit (board, boardPointer, &channelStatus);
               channelStatus = (channelStatus & CEIDEV_CSR_FIFO_COUNT_MASK) >> CEIDEV_CSR_FIFO_COUNT_SHIFT;
               status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &boardTime);
               if (boardTime.timeTag > exitTime.timeTag) {
                  accessStatus = ARS_NOSYNC;
                  break;
               }
            }
         }
         if (device[board].multiThreadProtect) {
            accessStatus = cei_release_thread_lock(LOCK_IDENT);
         }
      }
#ifndef TARGET_EMBEDDED
      status = (CEI_INT16)accessStatus;
   }
#endif
   
   return status;
}

/*===========================================================================*  
 * CEI-x20 API Functions (NO FUNCTIONALITY) Providing Link-Compatibility ONLY
 *===========================================================================*/             

EXPORT32 CEI_INT16 DLL_EXPORTED ar_msg_control (CEI_INT16 board, CEI_INT16 control) {return ARS_NORMAL;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_timetag_control (CEI_INT16 board, CEI_INT16 flag) {return ARS_NORMAL;}

EXPORT32 CEI_INT32 DLL_EXPORTED ar_dump_dp (CEI_INT16 board, pCEI_CHAR file_name, pCEI_CHAR msg) {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_force_version (CEI_INT16 board, CEI_INT16 number) {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_formatarinclabel (CEI_UINT32 label_data, pCEI_CHAR labelstr) {;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_formatbinarylabel (CEI_UINT32 label_data, CEI_INT32 trunc_label, CEI_CHAR *labelstr) {;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_get_errorLV (pCEI_CHAR string, CEI_INT16 status) {;}
EXPORT32 CEI_INT32 DLL_EXPORTED ar_get_harris (CEI_INT16 board, CEI_INT16 item) {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_int_slave (CEI_INT16 board) {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_loadslv_init_discrete (CEI_INT16 board, CEI_UINT32 base_seg, CEI_INT32 base_port, CEI_UINT16 ram_size, CEI_UINT32 disc_init)  {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_putword2x16 (CEI_INT16 board, CEI_INT16 channel, CEI_INT16 lsw, CEI_INT16 msw) {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_reformat (pCEI_VOID lsw, pCEI_VOID msw) {;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_arinc_config (CEI_INT16 board, CEI_INT16 channel, CEI_INT16 item, CEI_UINT32 value) {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_INT16 DLL_EXPORTED ar_set_harris (CEI_INT16 board, CEI_INT16 item, CEI_UINT32 value) {return ARS_NO_HW_SUPRT;}
EXPORT32 CEI_VOID  DLL_EXPORTED ar_unformat (pCEI_INT16 lsword, pCEI_INT16 msword) {;}

